/*
 FCEU.c

 * MacFCEU 0.9d3
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  fceu.c Copyright (C) 2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "cart.h"
#include "PPU.h"
#include "NSF.h"
#include "VSUni.h"
#include "FDS.h"
#include "iNES.h"
#include "UNIF.h"
#include "fsound.h"
#include "netplay.h"
#include "general.h"
#include "utility/endian.h"
#include "memory.h"

#include "file.h"
#include "cheat.h"
#include "palette.h"
#include "state.h"
#include "movie.h"
#include "video.h"
#include "input.h"

#include "GameGenie.h"

#ifdef env68k
 #pragma segment shell
#endif

uint8		RateFixFlags = 0;
uint32		FEmuSpeedRatio = 1000000000u; /* added; 2009.01.29 */

uint64		timestampbase;

FCEUS		FSettings;

FCEUGI		*FCEUGameInfo = 0;

void (*GameInterface)(int h);
void (*GameStateRestore)(int v);

readfunc	ARead[0x10000];
writefunc	BWrite[0x10000];

DECLFW(BNull)
{
(void)A; (void)V;
}

DECLFR(ANull)
{
	return A>>8;
}

writefunc FASTAPASS(1) GetWriteHandler(int32 a)
{
	if (BWriteG && a>=0x8000)
		return (BWriteG-0x8000)[a];
	else
		return BWrite[a];
}

readfunc FASTAPASS(1) GetReadHandler(int32 a)
{
	if (a>=0x8000 && AReadG)
		return (AReadG-0x8000)[a];
	else
		return ARead[a];
}

void FASTAPASS(3) SetReadHandler(int32 s,
	int32 e,
	readfunc f)
{
	SetReadHandlerI(1,s,e,f);
}

void FASTAPASS(3) SetWriteHandler(int32 s,
	int32 e,
	writefunc f)
{
	SetWriteHandlerI(1,s,e,f);
}

/*
NOTE: these set from end; you have to get the final address right
*/
void FASTAPASS(4) SetReadHandlerI(int d,
	int32 start,
	int32 end,
	readfunc f)
{
	int32		a;

	if (start>(a=end) || (unsigned)start>0xFFFF)
		return;
	if (!f)
		f = ANull;
	OptFAddr(f,f);
	if (AReadG)
		do
			if (a>=0x8000)
				(AReadG-0x8000)[a] = f;
			else
				ARead[a] = f;
		while((a-=d)>=start);
	else
		do
			ARead[a] = f;
		while((a-=d)>=start);
}

void FASTAPASS(4) SetWriteHandlerI(int d,
	int32 start,
	int32 end,
	writefunc f)
{
	int32		a;

	if (start>(a=end) || (unsigned)start>0xFFFF)
		return;
	if (!f)
		f = BNull;
	OptFAddr(f,f);
	if (BWriteG)
		do
			if (a>=0x8000)
				(BWriteG-0x8000)[a] = f;
			else
				BWrite[a] = f;
		while((a-=d)>=start);
	else
		do
			BWrite[a] = f;
		while((a-=d)>=start);
}

uint8		GameMemBlock[0x20000]; /*131072*/
uint8		RAM[0x800];
uint8		PAL = 0;

/*Dendy (not represented) would be zero*/
#define CPU_MODEL	(PAL ? 2:1)

/*does a $"00 00 00 00 FF FF FF FF" pattern*/
void FCEU_MemoryRand(uint8 *p,
	uint32 z)
{
	uint8		*d = p;
	int		i, x = 0, m = CPU_MODEL ? ~0:0;

	if (!d)
		return;
	if ((i=z>>2)!=0)
		do
			*d++ = *d++ = *d++ = *d++ = x;
		while(x^=m,--i);
	if (z&2)
		*d++ = *d++ = x;
	if (z&1)
		*d++ = x;
	if (m && z>16) /* init. signature added; 2010.03.17 */
	{ /* bug fix: *it wrote past the end*; 2011.01.19 */
		i = 8, x = 0277;
		do
			p[7+i] = x;
		while(x+=010,--i);
	}
}

DECLFW(FCEU_BRAML);
DECLFW(FCEU_BRAMH);
DECLFR(FCEU_ARAML);
DECLFR(FCEU_ARAMH);

DECLFW(FCEU_BRAML)
{
	RAM[A] = V;
}

DECLFW(FCEU_BRAMH)
{
	RAM[A&0x7FF] = V;
}

DECLFR(FCEU_ARAML)
{
	return RAM[A];
}

DECLFR(FCEU_ARAMH)
{
	return RAM[A&=0x7FF];
}

/* file system, game loading */

static void CloseGame(void)
{
	if (!FCEUGameInfo)
		return;
#ifdef NETWORK
	if (FCEUnetplay)
		FCEUD_NetworkClose();
#endif
	if (FCEUGameInfo->name)
		{FCEU_free(FCEUGameInfo->name); FCEUGameInfo->name = 0;}
	if (FCEUGameInfo->type!=GIT_NSF)
		FCEU_FlushGameCheats(0,0);
	if (GameInterface)
		GameInterface(GI_CLOSE);
	ResetExState(0,0);
	CloseGenie();
	FCEU_free(FCEUGameInfo);
	FCEUGameInfo = 0;
}

void ResetGameLoaded(void)
{
	if (FCEUGameInfo)
		CloseGame();
	GameStateRestore = 0;
	PPU_hook = 0;
	GameHBIRQHook = 0;
	if (GameExpSound.Kill)
		GameExpSound.Kill();
	dwmemset(&GameExpSound,0,sizeof(GameExpSound));
	MapIRQHook = 0;
	PAL&= 1;
	pale = 0;
}

static uint8	firsttime;

FCEUGI *FCEUI_LoadGame(const char *path)
{
	FCEUFILE	*fp;
	char		*ipsfn, *name = 0;
	int		result;

	ResetGameLoaded();

	if ((FCEUGameInfo=FCEU_malloc(sizeof(FCEUGI)))==0)
		return 0;
	FCEUGameInfo->soundchan = 0;
	FCEUGameInfo->soundrate = 0;
	FCEUGameInfo->name = 0;
	FCEUGameInfo->type = GIT_CART;
	FCEUGameInfo->vidsys = GIV_USER;
	FCEUGameInfo->input[0] = FCEUGameInfo->input[1] = -1;
	FCEUGameInfo->inputfc = -1;
	FCEUGameInfo->cspecial = 0;

	MakeFileBase(path);

	if ((ipsfn=FCEU_MakeFName(FCEUMKF_IPS,0,0))==0
			|| (name=malloc(256))==0)
		fp = NULL, result =-2;
	else
		*name = 0, fp = FCEU_fopen(path,ipsfn,"rb",0,name,&result);
	if (ipsfn)
		FCEU_free(ipsfn);
	if (name)
		if ((FCEUGameInfo->name=realloc(name,strlen(name)+1))==NULL)
			{FCEU_fclose(fp); fp = NULL; free(name); name = NULL;}
	if (!fp)
	{
		if (result!=-128) /* cancel */
			FCEU_PrintError("Error opening '%s'!",
				path);
		return 0;
	}

	FCEU_printf(*name ? "\nLoading '%s',\n File '%s'...\n\n":"\nLoading '%s'...\n\n",
		path,name);

	if ((result=iNESLoad(*name ? name:path,fp))==0
			&& (result=NSFLoad(fp))==0
			&& (result=UNIFLoad(path,fp))==0
			&& (result=FDSLoad(path,fp))==0 || result==-3) /*added -3; 2011.02.08*/
	{
		if (result!=-128) /* cancel */
			FCEU_PrintError("The file or file within is corrupt or unsupported.");
		result = 0;
	}
	if ((unsigned)result+4<=3) /*[-4,0)*/
	{
		FCEU_PrintError("A major error occurred while loading the file.");
		if (result==-2)
			FCEUD_Message(" Not enough memory.\n");
	}

	FCEU_fclose(fp);
	if (result<=0)
	{
		FCEUD_Message("File closed.\n");
		return 0;
	}

	FCEU_ResetVidSys();

	if (FCEUGameInfo->type!=GIT_NSF)
		if (FSettings.GameGenie)
			OpenGenie();

	PowerNES();
	FCEUSS_CheckStates();
	FCEUMOV_CheckMovies();

	if (FCEUGameInfo->type!=GIT_NSF)
	{
		FCEU_LoadGamePalette();
		FCEU_LoadGameCheats(0);
	}
	FCEU_ResetPalette(-1);
	FCEU_ResetMessages(0); /*Save state, status messages, etc.*/

	firsttime = 1;

	return FCEUGameInfo;
}

int FCEUI_Initialize(void)
{
	if (!FCEU_InitVirtualVideo())
		return 0;
	dwmemset(&FSettings,0,sizeof(FSettings));

	FSettings.UsrFirstSLine[0] = 8;
	FSettings.UsrFirstSLine[1] = 0; /* 1 recommended */
	FSettings.UsrLastSLine[0] = 231; /* 239 is recommended */
	FSettings.UsrLastSLine[1] = 239;
	FSettings.UsrVBLSLines[0] = 20;
	FSettings.UsrVBLSLines[1] = 70;
	FCEUI_SetSoundVolume(100);
	FSettings.soundactivech = 0xFFF;
	FSettings.soundrectamp = FSettings.soundtndamp = 256;
	FSettings.soundnscale = 2;
	FSettings.soundtscale = 3;
	FCEUPPU_Init();
	X6502_Init();
	X6502_DebugLevel(-1);

	return 1;
}

void FCEUI_Kill(void)
{
	FCEU_KillVirtualVideo();
	FCEU_KillGenie();
}

int FCEUI_Emulate(uint8 **PixBuf,
	int16 **SoundBuf,
	int32 *SoundBufSize,
	int skip)
{
	int		r;

	FCEU_UpdateInput();
	if (GenieStage!=1)
		FCEU_ApplyPeriodicCheats();
	if (FCEUPPU_Loop(skip))
		*PixBuf = 0, r = 0;
	else
		*PixBuf = XBuf, r = 1;

	*SoundBufSize = FlushEmulateSound();

	(void)uint64addu32(timestampbase,timestamp);
	timestamp = 0;

	GetSoundBuffer(SoundBuf);
	return r;
}

void FCEUI_CloseGame(void)
{
	CloseGame();
}

void ResetNES(void)
{
	FCEUMOV_EncodeCmd(FCEUNPCMD_RESET,0);
	if (FCEUGameInfo==0)
		return;
	GameInterface(GI_RESETM2);
	FCEUSND_Reset();
	FCEUPPU_Reset();
	X6502_Reset();
}

void PowerNES(void)
{
	FCEUMOV_EncodeCmd(FCEUNPCMD_POWER,0);
	if (FCEUGameInfo==0)
		return;

	FCEU_CheatResetRAM();
	FCEU_CheatSetRAM(2,0,RAM);

	GeniePower();

	FCEU_MemoryRand(RAM,0x800);

	SetReadHandler(0,0x7FF,FCEU_ARAML); /* slightly faster "low", no &0x7FF */
	SetWriteHandler(0,0x7FF,FCEU_BRAML);
	SetReadHandler(0x800,0x1FFF,FCEU_ARAMH);
	SetWriteHandler(0x800,0x1FFF,FCEU_BRAMH);
	SetReadHandler(0x2000,0xFFFF,ANull);
	SetWriteHandler(0x2000,0xFFFF,BNull);

	InitializeInput();
	FCEUSND_Power();
	FCEUPPU_Power();

	/* Have the external game hardware "powered" after the internal NES stuff.
	   Needed for the NSF code and VS System code.
	*/
        GameInterface(GI_POWER);
	if (FCEUGameInfo->type==GIT_VSUNI)
		FCEU_VSUniPower();
	/* zero all CYC and start the CPU */
	_int64xset(timestampbase,0);
	X6502_Power();
	FCEU_PowerCheats();
}

void FCEU_ResetVidSys(void)
{
	int		w;

	if (FCEUGameInfo)
	{
		if (FCEUGameInfo->vidsys==GIV_NTSC)
			w = 0;
		else if (FCEUGameInfo->vidsys==GIV_PAL)
			w = 1;
		else /* default (user input) */
			w = FSettings.PAL ? 1:0;

		FCEUPPU_SetVideoSystem(PAL=w);
	}

	FCEU_SetSoundVariables();
}

void FCEU_printf(const char *format, ...)
{
	char		temp[2048];

	va_list		ap;

	va_start(ap,format);
	vsprintf(temp,format,ap);
	FCEUD_Message(temp);

	va_end(ap);
}

void FCEU_PrintError(const char *format, ...)
{
	char		temp[2048];

	va_list		ap;

	va_start(ap,format);
	vsprintf(temp,format,ap);
	FCEUD_PrintError(temp);

	va_end(ap);
}

void FCEUI_SetRenderedLines(int ntscf,
	int ntscl,
	int palf,
	int pall)
{
	FSettings.UsrFirstSLine[0] = ntscf;
	FSettings.UsrLastSLine[0] = ntscl;
	FSettings.UsrFirstSLine[1] = palf;
	FSettings.UsrLastSLine[1] = pall;

	if (PAL)
	{
		FSettings.FirstSLine = FSettings.UsrFirstSLine[1];
		FSettings.LastSLine = FSettings.UsrLastSLine[1];
	} else
	{
		FSettings.FirstSLine = FSettings.UsrFirstSLine[0];
		FSettings.LastSLine = FSettings.UsrLastSLine[0];
	}
}

void FCEUI_SetVidSystem(int a)
{
	FSettings.PAL = a ? 1:0;

	if (FCEUGameInfo)
		FCEU_ResetVidSys();

	FCEU_ResetPalette(1); /* now calculated regardless; 2009.01.29 */
}

void FCEUI_SetVBLScanlines(int ntscslv,
	int palslv)
{
	FSettings.UsrVBLSLines[0] = ntscslv;
	FSettings.UsrVBLSLines[1] = palslv;

	if (FCEUGameInfo)
		FCEU_ResetVidSys();
}

int FCEUI_GetCurrentVidSystem(int *slstart,
	int *slend,
	int *slvbl)
{
	if (slstart)
		*slstart = FSettings.FirstSLine;
	if (slend)
		*slend = FSettings.LastSLine;
	if (slvbl)
		*slvbl = FSettings.UsrVBLSLines[PAL];

	return PAL;
}

void FCEUI_SetGameGenie(int a)
{
	FSettings.GameGenie = a ? 1:0;
}

void FCEUI_SetSnapName(int a)
{
	FSettings.SnapName = a;
}

#include "ftrig.h"

uint32 FCEUI_SetSpeed(uint32 v)
{ /* max sound buffer size determines minimum speed by quality setting */
  /* the actual min. is ~3.85%, but rate calc. overflow starts to occur under 20% */
	uint32		min =
	#if HAVE_HQ
		FSettings.soundq ? 717000000:
	#endif
		200000000;

	if (v<min) /* clip on what is buffered */
		if (v<(uint32).01e9) /* way out of range - default 100% */
			v = 1e9;
		else
			v = min;
	FEmuSpeedRatio = v;
	FCEUI_SetTNDFilterAmp(FSettings.soundtndamp); /* to rebuild the tables */

	return v;
}

/* 24 fractional bits */

/*#define PAL_FPS   838977920 // 50.006980896 (~50.007) */
/*#define NTSC_FPS 1008307711 // 60.099822938 (~60.1) */

/* calculated numbers:
 3.2*1662607.125/(341*(242+70)) = 50.0069817279495
 3*1789772.7272727272727272/(341*(242+20)) = 60.0984775561123
*/
/* result is 1.0 for every 15625000 */
uint32 FCEUI_GetDesiredFPS(void)
{
	uint64		ratio;
	uint16		VBLsl;

	if (FEmuSpeedRatio<(uint32).2e9) /*20% - was 24%*/
		FEmuSpeedRatio = 1e9;

	if (PAL)
	{
		if (RateFixFlags&1)
			_int64xset(ratio,3.2*1600000*(1L<<24)+.9);
		else /* 13300857/8 Hz */
			_int64xset(ratio,3.2*PAL_CPU*(1L<<24)+.9);
		VBLsl = FSettings.UsrVBLSLines[1];
	} else
	{
		if (RateFixFlags&0x80)
			_int64xset(ratio,3.0*1600000*(1L<<24)+.9);
		else /* 315/176 MHz */
			_int64xset(ratio,3.0*NTSC_CPU*(1L<<24)+.9);
		VBLsl = FSettings.UsrVBLSLines[0];
	}
	VBLsl+= 242;

	return ufract30mul(uint64divu32(ratio,VBLsl*LCYCS),FEmuSpeedRatio);
}

/* some code moved here to reduce re-compilation time of non-core stuff; 2009.12.29 */

uint8		fceutileview = 0;
extern uint8	PPUrendis, PPUmaxsprites;

void FCEUI_ToggleTileView(void)
{
	fceutileview^= 1;
}

void FCEUI_SetRenderDisable(int sprites,
	int bg)
{
	if (sprites==2)
		PPUrendis^= 1;
	else if (sprites>=0)
		PPUrendis = (PPUrendis&~1)|(sprites ? 1:0);

	if (bg==4)
		PPUrendis^= 6;
	else if (bg==2)
		PPUrendis^= 2;
	else if (bg>=0)
		PPUrendis = (PPUrendis&~2)|(bg ? 2:0);
}

void FCEUI_DisableSpriteLimitation(int a)
{
	PPUmaxsprites = a ? 64:8; /* must be >=8 */
}
