/*
 FDS.c

 * MacFCEU 0.8a9
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  fds.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <zlib.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "FDS.h"
#include "fsound.h"
#include "general.h"
#include "state.h"
#include "file.h"
#include "memory.h"
#include "cart.h"
#include "utility/MD5.h"
#include "netplay.h"

#ifdef env68k
 #pragma segment Sound
 #define _nint16 int16
#else
 #define _nint16 int32
#endif

/*	TODO:  Add code to put a delay in between the time a disk is inserted
	and the when it can be successfully read/written to.  This should
	prevent writes to wrong places OR add code to prevent disk ejects
	when the virtual motor is on(mmm...virtual motor).
*/

static DECLFR(FDSRead4030);
static DECLFR(FDSRead4031);
static DECLFR(FDSRead4032);
static DECLFR(FDSRead4033);

static DECLFW(FDSWrite);

static DECLFW(FDSWaveWrite);
static DECLFR(FDSWaveRead);

static DECLFR(FDSSRead);
static DECLFW(FDSSWrite);
static DECLFR(FDSBIOSRead);
static DECLFR(FDSRAMRead);
static DECLFW(FDSRAMWrite);
static void FDSInit(void);
static void FP_FASTAPASS(1) FDSFix(int a);

#define FDSRAM		GameMemBlock
#define CHRRAM		(GameMemBlock+32768)

static uint8	FDSRegs[6];
static int32	IRQLatch, IRQCount;
static uint8	IRQa;

static uint8	FDSBIOS[8192];

/* Original disk data backup, to help in creating save states. */
static uint8 *diskdatao[8] = {0,0,0,0,0,0,0,0};

static uint8 *diskdata[8] = {0,0,0,0,0,0,0,0};

static unsigned	TotalSides;
static uint8	DiskWritten = 0;	/* Set to 1 if disk was written to. */
static uint8	writeskip;
static uint32	DiskPtr;
static int32	DiskSeekIRQ;
static uint8	SelectDisk, InDisk;

typedef struct FDSSOUND
{
	int64	cycles;		// Cycles per PCM sample
	int64	count;		// Cycle counter
	int64	envcount;		// Envelope cycle counter
	uint32	b19shiftreg60;
//	uint32	b24adder66;	// really temporary
	uint32	b24latch68;	// 25 bits
	uint32	b17latch76;
	uint8	clockcount;	// Counter to divide frequency by 8.
	uint8	b8shiftreg88;	// Modulation register.
	uint8	amplitude[2];	// Current amplitudes.
	uint8	speedo[2];
	uint8	mwcount;
	uint8	mwstart;
	uint8	mwave[0x20];	// Modulation waveform
	uint8	cwave[0x40];	// Game-defined waveform(carrier)
	uint8	SPSG[0xB];
} FDSSOUND;

static FDSSOUND FDSo;

#define FDSClock	(NTSC_CPU/2)

#define	SPSG		FDSo.SPSG
#define b19shiftreg60	FDSo.b19shiftreg60
#define b24latch68	FDSo.b24latch68
#define b17latch76	FDSo.b17latch76
#define b8shiftreg88	FDSo.b8shiftreg88
#define clockcount	FDSo.clockcount
#define amplitude	FDSo.amplitude
#define speedo		FDSo.speedo

static void FDSClose(void);
void FDSGI(int h);

void FDSGI(int h)
{
	if (h==GI_CLOSE)
		FDSClose();
	else if (h==GI_POWER)
		FDSInit();
}

static void FDSStateRestore(int version)
{
	int		x;

	setmirror(~FDSRegs[5]>>3&1);

	if (version>=9810)
		for(x=0; x<TotalSides; ++x)
		{
			int		b;

			if (diskdatao[x])
				for(b=0; b<65500; ++b)
					diskdata[x][b]^= diskdatao[x][b];
		}
	b24latch68&= 0x1FFFFFF;
}

void FDSSound(int32); /* prototype fix; 2008.06.28 09:01 GMT */
void FDSSoundReset(void);
void FDSSoundStateAdd(void);
static void RenderSound(void);
static void RenderSoundHQ(void);

static void FDSInit(void)
{
	dwmemset(FDSRegs,0,sizeof(FDSRegs));
	writeskip = DiskPtr = DiskSeekIRQ = 0;
	setmirror(1);

	setprg8(0xE000,0);	// BIOS
	setprg32r(1,0x6000,0);	// 32KB RAM
	setchr8(0);		// 8KB CHR RAM

	MapIRQHook = FDSFix;
	GameStateRestore = FDSStateRestore;

	SetReadHandler(0x4030,0x4030,FDSRead4030);
	SetReadHandler(0x4031,0x4031,FDSRead4031);
	SetReadHandler(0x4032,0x4032,FDSRead4032);
	SetReadHandler(0x4033,0x4033,FDSRead4033);

	SetWriteHandler(0x4020,0x4025,FDSWrite);

	SetWriteHandler(0x6000,0xdfff,FDSRAMWrite);
	SetReadHandler(0x6000,0xdfff,FDSRAMRead);
	SetReadHandler(0xE000,0xFFFF,FDSBIOSRead);
	IRQCount = IRQLatch = IRQa = 0;

	FDSSoundReset();
	InDisk = 0;
	SelectDisk = 0;
}

void FCEU_FDSInsert(int oride)
{
(void)oride;
	if (InDisk==255)
	{
		FCEU_DispMessage("Disk %d Side %s Inserted",
			SelectDisk>>1,(SelectDisk&1) ? "B":"A");
		InDisk = SelectDisk;
	} else
	{
		FCEU_DispMessage("Disk %d Side %s Ejected",
			SelectDisk>>1,(SelectDisk&1) ? "B":"A");
		InDisk = 255;
	}
}

void FCEU_FDSEject(void)
{
         InDisk = 255;
}

void FCEU_FDSSelect(void)
{
	if (InDisk!=255)
	{
		FCEU_DispMessage("Eject disk before selecting.");
		return;
	}
	SelectDisk = ((SelectDisk+1)%TotalSides)&3;
	FCEU_DispMessage("Disk %d Side %s Selected",
		SelectDisk>>1,(SelectDisk&1) ? "B":"A");
}

static void FP_FASTAPASS(1) FDSFix(int a)
{
	if (IRQa&2 && IRQCount!=0)
		if ((IRQCount-=a)<=0)
		{
			if (!(IRQa&1))
				IRQCount = IRQLatch = 0, IRQa&= ~2;
			else
				IRQCount = IRQLatch;
			//IRQCount=IRQLatch; //0xFFFF;
			X6502_IRQBegin(FCEU_IQEXT);
			//printf("IRQ: %d\n",timestamp);
			//   printf("IRQ: %d\n",scanline);
		}

	if (DiskSeekIRQ>0)
		if ((DiskSeekIRQ-=a)<=0 && FDSRegs[5]&0x80)
			X6502_IRQBegin(FCEU_IQFDS);
}

static DECLFR(FDSRead4030)
{
	uint8		ret = 0;
(void)A;
	/* Cheap hack. */
	if (XFrame.intlow&FCEU_IQEXT)
		ret|= 1;
	if (XFrame.intlow&FCEU_IQFDS)
		ret|= 2;
	if (!fceuindbg)
		XFrame.intlow&=~(FCEU_IQEXT|FCEU_IQFDS); //X6502_IRQEnd

	return ret;
}

static DECLFR(FDSRead4031)
{
	static uint8	z = 0;
(void)A;
	if (InDisk!=255)
	{
		z = diskdata[InDisk][DiskPtr];
		if (!fceuindbg)
		{
			if (DiskPtr<64999)
				++DiskPtr;
			DiskSeekIRQ = 150;

			XFrame.intlow&=~FCEU_IQFDS;
		}
	}

	return z;
}

static DECLFR(FDSRead4032)
{
	uint8		ret = A>>8&~7; //XFrame.DB&

	if (InDisk==255)
		ret|= 5;
	if (InDisk==255 || !(FDSRegs[5]&1) || FDSRegs[5]&2)
		ret|= 2;

	return ret;
}

static DECLFR(FDSRead4033)
{
(void)A;
	return 0x80; // battery
}

static DECLFW(FDSRAMWrite)
{
	(FDSRAM-0x6000)[A] = V;
}

static DECLFR(FDSBIOSRead)
{
	return (FDSBIOS-0xE000)[A];
}

static DECLFR(FDSRAMRead)
{
	return (FDSRAM-0x6000)[A];
}

/* FDS sound */

void FDSSoundStateAdd(void)
{
	AddExState(FDSo.cwave,64,0,"WAVE");
	AddExState(FDSo.mwave,32,0,"MWAV");
	AddExState(amplitude,2,0,"AMPL");
	AddExState(SPSG,0xB,0,"SPSG");

	AddExState(&b8shiftreg88,1,0,"B88");

	AddExState(&clockcount,1,1,"CLOC"); /* was 32-bit */
	/* RLSB was missing; 2013.05.15 */
	AddExState(&b19shiftreg60,4|FCEUSTATE_RLSB,1,"B60");
//	AddExState(&b24adder66,4|FCEUSTATE_RLSB,1,"B66");
	AddExState(&b24latch68,4|FCEUSTATE_RLSB,1,"B68");
	AddExState(&b17latch76,4|FCEUSTATE_RLSB,1,"B76");
}

static DECLFR(FDSSRead0)
{
	return amplitude[0]|(A>>8&0xC0); //XFrame.DB&
}

static DECLFR(FDSSRead2)
{
	return amplitude[1]|(A>>8&0xC0); //XFrame.DB&
}

// $4080 - Fundamental wave amplitude data register 92
// $4082/3 - Fundamental wave freq. data register 58 - low, high 4 bits.
// $4084 - Modulation amplitude data register 78
// $4086/7 - Modulation freq. data register 72 - low, high 4 bits

static DECLFW(FDSSWrite)
{
	if (FSettings.SndRate && !(~FSettings.soundactivech&(0x20|64u<<SCNFDS)))
	#if HAVE_HQ
		if (FSettings.soundq)
			RenderSoundHQ();
		else
	#endif
			RenderSound();
	switch(A-=0x4080)
	{
	case 0:
	case 4:	if (V&0x80)
			amplitude[A>>2&1] = V&0x3F; //)>0x20?0x20:(V&0x3F);
		break;
	case 5://printf("$%04x:$%02x\n",A,V);
		break;
	case 7: b17latch76 = 0, SPSG[5] = 0;//printf("$%04x:$%02x\n",A,V);
		break;
	case 8:
		b17latch76 = 0;
		//   printf("%d:$%02x, $%02x\n",SPSG[0x5],V,b17latch76);
		FDSo.mwave[SPSG[5]&=31] = V&7, ++SPSG[5];
	}
	//if(A>=0x7 && A!=0x8 && A<=0xF)
	//if(A==0xA || A==0x9)
	//printf("$%04x:$%02x\n",A,V);
	SPSG[A] = V;
}

static void DoEnv(void)
{
	static int8	envc[2];

	int		x = 0;
	int8		f, *c = envc;

	do
		if (!(uint8)((f=SPSG[x<<2])&0x80))
			if (*c<=0)
			{
				*c = f&0x3F;
				/*if (!(uint8)(f&0x80))*/
				if (f&=0x40)
				{
					if (amplitude[x]<0x3F)
						++amplitude[x];
				} else if (amplitude[x]>0)
					--amplitude[x];
			} else
				--*c;
	while(++c,++x<2);
}

static DECLFR(FDSWaveRead)
{
	return FDSo.cwave[A-0x4040]|(A>>8&0xC0); //|(XFrame.DB&0xC0)
}

static DECLFW(FDSWaveWrite)
{
	if (SPSG[9]&0x80)
		FDSo.cwave[A-0x4040] = V&0x3F;
}

/* (n*4*1.5)/(d+2) -- <<8 scale */
static int16		amp9table[4] =
{
	(0x600+1)/2,
	(0x600+2)/3,
	(0x600+3)/4,
	(0x600+4)/5,
};

void FCEU_FDSSetVolume(int v)
{
	int		x = 2, k;
#if HAVE_HQ
	if (FSettings.soundq && v<0)
		v =-v; /* HQ needs a positive value... */
#endif
	k = (v=v*(0x600>>2)/25)>>31;
	do
		amp9table[x-2] = (v+((x-1+k)^k))/x;
	while(++x<5);
}

#ifdef env68k
 #define deccount(c)		((c).hi-=1<<8)
#else
 #define deccount(c)		((c)-=(int64)1<<40)
#endif

/*static INLINE int32 FDSDoSound(void)*/

#define FDSDoSound(v,am0,am1,z60,z76) do{ \
	(void)int64add(FDSo.count,FDSo.cycles); \
	while(int64gehl(FDSo.count,0,32768)) { /*if(fdso.count>=((int64)1<<40))*/ \
		deccount(FDSo.count); \
		if (!(clockcount&=7)) { /* clock rise */ \
			b19shiftreg60 = z60, b17latch76+= z76; \
			if (!(SPSG[7]&0x80)) { \
				_nint16	adj, t = FDSo.mwave[b17latch76>>13&31]; \
				if ((adj=t&3)!=0) \
					if (t&=4,adj-=t,adj*=am1,adj+=adj,adj>0x7F) \
						adj = 0x7F; \
					else if (adj<-0x80) \
						adj =-0x80; \
				adj+= 0x80, b8shiftreg88 = adj; \
			} else \
				b8shiftreg88 = 0x80; \
		} else \
			b19shiftreg60+= b19shiftreg60, b8shiftreg88>>= 1; \
		/*b24adder66 = (b24latch68+b19shiftreg60)&0x1FFFFFF;*/ \
		if(b8shiftreg88&1) /* || clockcount==7) // clock fall */ \
			b24latch68=(b24latch68+b19shiftreg60)&0x1FFFFFF; /*0x3FFFFFF*/ \
		++clockcount; \
		if (int64addhls(FDSo.envcount,-1,-1)<=0) { /*if(--fdso.envcount<=0)*/ \
			(void)int64addu32(FDSo.envcount,SPSG[10]*3); \
			if (!(SPSG[3]&0x40)) \
				DoEnv(); \
		} \
	} \
	v+=(FDSo.cwave[b24latch68>>19]*am0)>>8; \
}while(0)

static int32	FBC = 0;

static void RenderSound(void)
{
	int32		x, i, z60, z76, amp;
	_nint16		mod;

	i = FBC, x = (SOUNDTS<<16)/soundtsinc;
	if (x<=i)
		return;
	FBC = x;
	if (SPSG[9]&0x80)
		return;
	amp = (_nint16)MIN(*amplitude,32)*amp9table[SPSG[9]&3], mod = amplitude[1];
	x-= i, z60 = SPSG[2]|(SPSG[3]<<8&0xF00), z76 = SPSG[6]|(SPSG[7]<<8&0xF00);
	do
		FDSDoSound(FCEUWave[i>>4],amp,mod,z60,z76);
	while(++i,--x);
}

static void RenderSoundHQ(void)
{
	if (!(SPSG[9]&0x80))
	{
		int32		x, z60 = SPSG[2]|(SPSG[3]<<8&0xF00);
		int32		z76 = SPSG[6]|(SPSG[7]<<8&0xF00);
		int32		amp = (_nint16)MIN(*amplitude,32)*amp9table[SPSG[9]&3];
		_nint16		mod = amplitude[1];

		for(x=FBC; x<SOUNDTS; ++x)
			FDSDoSound(FCEUWave[x],amp,mod,z60,z76);
	}
	FBC = SOUNDTS;
}

static void HQSync(int32 ts)
{
	FBC = ts;
}

void FDSSound(int32 c)
{
	RenderSound();
	FBC = c;
}

/*
static DECLFR(FDSBIOSPatch)
{
 if(FDSRegs[5]&0x4)
 {
  XFrame.X=FDSRead4031(0x4031);
  FDSWrite(0x4024,XFrame.A);
  XFrame.A=XFrame.X;
  return(0x60);
 }
 else
 {
  return(0x58);
  //puts("Write");
 }
}
*/

static void FDS_ESI(void)
{
	if (FSettings.SndRate)
	#if HAVE_HQ
		if (FSettings.soundq)
			int64sethl(FDSo.cycles,1L<<7,0); /* =(int64)1<<39 */
		else
	#endif
		{
			/*((int64)1<<40)*FDSClock*/
			int64sethl(FDSo.cycles,(1L<<8)*FDSClock,0);
			(void)int64divu32(FDSo.cycles,FSettings.SndRate);
		}
//fdso.cycles=(int64)32768*FDSClock/(FSettings.SndRate *16);
	SetReadHandler(0x4040,0x407F,FDSWaveRead);
	SetWriteHandler(0x4040,0x407F,FDSWaveWrite);
	SetWriteHandler(0x4080,0x408A,FDSSWrite);
	SetReadHandler(0x4090,0x4090,FDSSRead0);
	SetReadHandler(0x4092,0x4092,FDSSRead2);

//SetReadHandler(0xE7A3,0xE7A3,FDSBIOSPatch);
}

void FDSSoundReset(void)
{
	dwmemset(&FDSo,0,sizeof(FDSo));
	FDS_ESI();
	GameExpSound.HiSync[SCNFDS] = HQSync;
	GameExpSound.HiFill[SCNFDS] = RenderSoundHQ;
	GameExpSound.Fill[SCNFDS] = FDSSound;
	GameExpSound.RChange[SCNFDS] = FDS_ESI;
}

static DECLFW(FDSWrite)
{
//extern int scanline;
	//FCEU_printf("$%04x:$%02x, %d\n",A,V,scanline);
	switch(A-=0x4020)
	{
	case 0:
		XFrame.intlow&=~FCEU_IQEXT; //X6502_IRQEnd(FCEU_IQEXT);
		IRQLatch&= 0xFF00;
		IRQLatch|= V;
	//	printf("$%04x:$%02x\n",A,V);
		break;
	case 1:
		XFrame.intlow&=~FCEU_IQEXT; //X6502_IRQEnd(FCEU_IQEXT);
		IRQLatch&= 0xFF;
		IRQLatch|= V<<8;
	//	printf("$%04x:$%02x\n",A,V);
		break;
	case 2:
		XFrame.intlow&=~FCEU_IQEXT; //X6502_IRQEnd(FCEU_IQEXT);
		IRQCount = IRQLatch;
		IRQa = V&3;
	//	printf("$%04x:$%02x\n",A,V);
	case 3:
		break;
	case 4:
		if (InDisk!=255 && !(FDSRegs[5]&4) && FDSRegs[3]&1)
		{
			if (DiskPtr>=0 && DiskPtr<65500)
				if (writeskip)
					--writeskip;
				else if (DiskPtr>=2)
				{
					DiskWritten = 1;
					diskdata[InDisk][DiskPtr-2] = V;
				}
		}
		break;
	case 5:
		XFrame.intlow&=~FCEU_IQFDS; //X6502_IRQEnd
		if (InDisk!=255)
		{
			if (!(V&0x40))
			{
				if (FDSRegs[5]&0x40 && !(V&0x10))
					DiskSeekIRQ = 200, DiskPtr-= 2;
				if (DiskPtr<0)
					DiskPtr = 0;
			}
			if (!(V&4))
				writeskip = 2;
			if (V&2)
				DiskPtr = 0, DiskSeekIRQ = 200;
			if (V&0x40)
				DiskSeekIRQ = 200;
		}
		setmirror(~V>>3&1);
		break;
	}
	FDSRegs[A] = V;
}

static void FreeFDSMemory(void)
{
	int		x;

	for(x=0; x<TotalSides; ++x)
		if (diskdata[x])
			{free(diskdata[x]); diskdata[x] = 0;}
}

static int SubLoad(FCEUFILE *fp)
{
	struct md5_context	md5;

	uint8		header[16];
	int		x;

	FCEU_fread(header,16,1,fp);

	/*"FDS\x1a"*/
	if (header[0]=='F' && header[1]=='D' && header[2]=='S' && header[3]==26)
		TotalSides = header[4];
	else if ((memcmp(header+1,"*NINTENDO-HVC*",14))==0)
	{
		long	t = FCEU_fgetsize(fp);

		if (t<65500)
			t = 65500;
		TotalSides = t/65500;
		FCEU_fseek(fp,0,SEEK_SET);
	} else
		return -3;

	md5_starts(&md5);

	if (TotalSides>8)
		TotalSides = 8;
	else if (TotalSides<1)
		TotalSides = 1;

	for(x=0; x<TotalSides; ++x)
	{
		if ((diskdata[x]=(uint8*)FCEU_malloc(65500))==0)
		{
			int		zol;

			for(zol=0; zol<x; ++zol)
				free(diskdata[zol]);
			return -2;
		}

		FCEU_fread(diskdata[x],1,65500,fp);
		md5_update(&md5,diskdata[x],65500);
	}

	md5_finish(&md5,FCEUGameInfo->MD5);

	return 1;
}

static void PreSave(void)
{
	int		x;

	//if(DiskWritten)
	for(x=0; x<TotalSides; ++x)
	{
		int	b;

		if (diskdatao[x])
			for(b=0; b<65500; b++)
				diskdata[x][b]^= diskdatao[x][b];
	}
}

static void PostSave(void)
{
	int		x;

	//if(DiskWritten)
	for(x=0; x<TotalSides; ++x)
	{
		int		b;

		if (diskdatao[x])
			for(b=0; b<65500; b++)
				diskdata[x][b]^= diskdatao[x][b];
	}
}

static int FDSLoadBIOS(FCEUFILE *fp)
{
	char		*fn, **h = 0;
	FILE		*zp = 0;
	int		r;

	FCEU_fseek(fp,0,SEEK_SET);

	if ((r=SubLoad(fp))<=0)
		return r;

	if ((fn=FCEU_MakeFName(FCEUMKF_FDSROM,0,0))!=0)
	{
		zp = FCEUD_UTF8fopen(fn,"rb");
		free(fn);
	}
	r = -5;
	if (zp)
	{
		if (fread(FDSBIOS,1,8192,zp)!=8192)
		{
			fclose(zp);
			FCEU_PrintError("Error reading FDS BIOS ROM image.");
			r = -1;
		fail:
			FreeFDSMemory();
			return r;
		}
		fclose(zp);
	} else
	{
	#include "FDSbios.h"
	#ifdef fdsbiosdefault
		uLongf	length = 8192;

		if (uncompress(FDSBIOS,&length,fdsbiosdefault,
				sizeof(fdsbiosdefault))!=Z_OK || length!=8192)
	#endif
		{
			FCEU_PrintError("Error getting FDS BIOS ROM.");
			goto fail;
		}
	}

	return 1;
}

static int FDSLoadAuxilary(void)
{
	FCEUFILE	*tp;
	char		*fn;
	int		x;

	if ((fn=FCEU_MakeFName(FCEUMKF_FDS,0,0))==0)
		return 0;
	for(x=0; x<TotalSides; ++x)
	{
		if ((diskdatao[x]=(uint8*)FCEU_malloc(65500))==0)
			return -2;
		else
			FCEU_memcpy(diskdatao[x],diskdata[x],65500);
	}

	if ((tp=FCEU_fopen(fn,0,"rb",0,0,&x))!=0)
	{
		FreeFDSMemory();
		if ((x=SubLoad(tp))<=0)
		{
			FCEU_PrintError("Error reading auxillary FDS file.");
			free(fn);
			return x;
		}
		FCEU_fclose(tp);
		DiskWritten = x = 1;	/* For save state handling. */
	}

	free(fn);

	return x;
}

static void FDSDoExStates(void)
{
	int		x;

	ResetExState(PreSave,PostSave);
	FDSSoundStateAdd();

	for(x=0; x<TotalSides; ++x)
	{
		char		temp[5];

		sprintf(temp,"DDT%d",
			x);
		AddExState(diskdata[x],65500,0,temp);
	}

	AddExState(FDSRAM,32768,0,"FDSR");
	AddExState(FDSRegs,sizeof(FDSRegs),0,"FREG");
	AddExState(CHRRAM,8192,0,"CHRR");
	/* RLSB was missing for four of these; 2013.05.15 */
	AddExState(&IRQCount,4|FCEUSTATE_RLSB,1,"IRQC");
	AddExState(&IRQLatch,4|FCEUSTATE_RLSB,1,"IQL1");
	AddExState(&IRQa,1,0,"IRQA");
	AddExState(&writeskip,1,0,"WSKI");
	AddExState(&DiskPtr,4|FCEUSTATE_RLSB,1,"DPTR");
	AddExState(&DiskSeekIRQ,4|FCEUSTATE_RLSB,1,"DSIR");
	AddExState(&SelectDisk,1,0,"SELD");
	AddExState(&InDisk,1,0,"INDI");
	AddExState(&DiskWritten,1,0,"DSKW");

	ResetCartMapping();
	SetupCartCHRMapping(0,CHRRAM,8192,1);
	SetupCartMirroring(0,0,0);
	dwmemset(CHRRAM,0,8192);
	dwmemset(FDSRAM,0,32768);
}

int FDSLoad(const char *name,
	FCEUFILE *fp)
{
	int		r;
(void)name;
	if ((r=FDSLoadBIOS(fp))<=0 || (r=FDSLoadAuxilary())<0)
		return r;
	FCEUGameInfo->type = GIT_FDS;
	GameInterface = FDSGI;

	SelectDisk = 0;
	InDisk = 255;

	FDSDoExStates();

	FCEU_printf(" Sides: %d\n\n",
		TotalSides);
	return 1;
}

void FDSClose(void)
{
	FILE		*fp;
	int		x;
	char		*fn = FCEU_MakeFName(FCEUMKF_FDS,0,0);

	if (!DiskWritten)
		return;

	if ((fp=FCEUD_UTF8fopen(fn,"wb"))==0)
	{
		free(fn);
		return;
	}
	free(fn);

	for(x=0; x<TotalSides; ++x)
		if (fwrite(diskdata[x],1,65500,fp)!=65500)
		{
			FCEU_PrintError("Error saving FDS image!");
			fclose(fp);
			return;
		}

	FreeFDSMemory();
	fclose(fp);
}
