/*
 NSF.c
  Nintendo Sound File playback and view

 * MacFCEU 0.9d3
 *
 * Copyright notice for this file:
 *  nsf.c (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "fsound.h"
#include "NSF.h"
#include "general.h"
#include "memory.h"
#include "file.h"
#include "FDS.h"
#include "cart.h"
#include "input.h"
#include "video.h"
#include "PPU.h"
#include "boards/MMC5.h"
#include "boards/Namcot.h"
#include "iNES.h"	//for prototypes only

#include "ftrig.h"

#define cosmul(r,t)	(ifract30mul(r+r,fastcos(t))/2)
#define sinmul(r,t)	(ifract30mul(r+r,fastsin(t))/2)

#ifdef env68k
 #pragma segment shell
#endif

/* ---- the MacFCEU NSF ROM */

/* revised: 2013.05.25 */
const uint8	NSFROM[128] =
{ /* see NSFROM.asm */
	0x48,0x2C,0xFF,0x38,0x70,0x10,0x8A,0x48,
	0x98,0x48,0x20,0xD6,0x38,0x68,0xA8,0x68,
	0xAA,0x68,0x8D,0xFF,0x38,0x40,0xA2,0xFF,
	0x9A,0xE8,0x8A,0x9D,0x00,0x40,0xE8,0xE0,
	0x14,0xD0,0xF8,0x8D,0x15,0x40,0xAD,0xFC,
	0x38,0x29,0x04,0xF0,0x0A,0xA9,0x80,0x8D,
	0x89,0x40,0xA9,0xE8,0x8D,0x8A,0x40,0xA9,
	0x0F,0x8D,0x15,0x40,0xA9,0xC0,0x8D,0x17,
	0x40,0xAD,0xFE,0x38,0xAE,0xFD,0x38,0x20,
	0xD3,0x38,0x20,0xD6,0x38,0x8D,0xFF,0x38,
	0x18,0x90,0xFE,0x6C,0xF8,0x38,0xA9,0x00,
	0xAA,0xA8,0x6C,0xFA,0x38 
};

/* ---- code */

#define MaxNamesTotal	(256*32)

static int GetIndName(const uint8 *l,
	uint8 *b,
	int i)
{
	const uint8	*s;

	while(1)
	{
		if ((s=memchr(l,0,MaxNamesTotal))==0)
			return 0;
		if (--i<0)
			break;
		l = s+1;
	}
	if ((i=s-l)>31)
		i = 31;
	memcpy(b,l,i);
	b[i] = 0;

	return 1;
}

static uint8	vismode = 1;

static uint8	*NSFDATA = 0;
static uint8	*ExWRAM = 0;

static uint8	*SongNames = 0;
static uint8	CurrentSongName[32], CaptionStr[32];

static int	NSFSize, NSFMaxBank, ExWRAMSize;
static uint16	LoadAddr, NTSCspeed, PALspeed;

int		NSFCycles; /* PPU cycle time per NSF frame - always 341*240 by default */
int8		NSFMaxInitFrames;
static int8	initframes;

static NSF_HEADER	NSFHeader;
//cheap on memory? (yes); 2010.03.18
#define NSFNMIFlags	NSFHeader.reserved[0]
#define NSFCurrentSong	NSFHeader.reserved[1]
#define NSFSpecial	NSFHeader.reserved[2]
#define NSFBankSW	NSFHeader.reserved[3]
#define NSFSizeLo	NSFHeader.reserved[4]
#define NSFSizeHi	NSFHeader.reserved[5]
#define NSFLastSong	NSFHeader.reserved[6] //maximum
#define NSFMaxBankHi	NSFHeader.reserved[7]

#define NMIReset	0x20
#define NMISongReload	0x40
#define NMIEnabled	0x80

void NSFGI(int h)
{
	switch(h)
	{
	case GI_CLOSE:
		if (NSFDATA)
			{FCEU_free(NSFDATA); NSFDATA = 0;}
		if (ExWRAM)
			{FCEU_free(ExWRAM); ExWRAM = 0;}
		if (SongNames)
			{FCEU_free(SongNames); SongNames = 0;}
		//if (NSFHeader.SoundChip&1<<SCVRC6)
		//	NSFVRC6_Init();
		//if (NSFHeader.SoundChip&1<<SCVRC7)
		//	NSFVRC7_Init();
		//if (NSFHeader.SoundChip&1<<SCNFDS)
		//	FDSSoundReset();
		if (NSFHeader.SoundChip&1<<SCMMC5)
			GenMMC5Close();
		if (NSFHeader.SoundChip&1<<SCN106)
			GenN1xxClose();
		//if (NSFHeader.SoundChip&1<<SCSFME)
		//	NSFAY_Init();
		break;
	case GI_RESETM2:
	case GI_POWER:
		NSF_init();
	}
}

/* this expects NSFSize and NSFHeader.LoadAddressLow/High to already be set */
static int NSFLoadData(FCEUFILE *fp)
{
	if (NSFDATA)
		return 2;
	LoadAddr = NSFHeader.LoadAddressLow;
	LoadAddr|= NSFHeader.LoadAddressHigh<<8;

	if (LoadAddr<0x6000)
	{
		FCEUD_PrintError(" NSF Error: Invalid load address.");
		return -5;
	}
	NSFMaxBank = (NSFSize+(LoadAddr&0xFFF)+4095)/4096;
	NSFMaxBank = uppow2(NSFMaxBank);
	NSFMaxBankHi = NSFMaxBank>>8;
	if ((NSFDATA=(uint8*)FCEU_malloc(NSFMaxBank<<12))==0)
		return -2;
	return FCEU_fread(NSFDATA+(LoadAddr&0xFFF),NSFSize,1,fp);
}

// First 32KB is reserved for sound chip emulation in the iNES mapper code.

//static INLINE void BANKSET(uint32 A,
//	uint32 b)
#define BANKSET(A,b)	do{ uint32 _b=(b)&NSFMaxBank; \
	if (NSFHeader.SoundChip&1<<SCNFDS) \
		FCEU_memcpy(ExWRAM+((A)-0x6000),NSFDATA+(_b<<12),4096); \
	else \
		setprg4(A,_b); \
} while(0)

/* ---- NSFe */

typedef struct NSFeAtomHeader
{
	int32		size;
	uint32		tag;
} NSFeAtomHeader;

static int NSFeLoad(FCEUFILE *fp)
{
	uint8		buf[250];
	NSFeAtomHeader	*header = (NSFeAtomHeader*)buf;
	const char	*errs = "faile error in reading";
	int32		skip;

	LoadAddr = 0; /* not a good idea to assume a default here */
	memset(&NSFHeader,0,sizeof(NSFHeader));
	while(FCEU_fread(header,sizeof(NSFeAtomHeader),1,fp)==1)
	{
		header->size = le32_native(header->size);
		if ((skip=header->size)<0)
		{
			errs = "invalid atom data size";
			goto error;
		}
		switch(header->tag)
		{
		case MK4CC('I','N','F','O'): /* the specifics escape me right now */
			memset(buf,0,12);
			/* the last two bytes are still unknown to me */
			/* the INFO entry in Dr. Mario.nsfe is only 10 bytes */
			if (FCEU_fread(buf,MIN(skip,12),1,fp)!=1)
				goto error;
			skip-= 12;
			memcpy(&NSFHeader.LoadAddressLow,buf,6);
			NSFHeader.VideoSystem = buf[6];
			NSFHeader.SoundChip = buf[7];
			NSFHeader.TotalSongs = buf[8];
			NSFHeader.StartingSong = buf[9];
			break;
		case MK4CC('B','A','N','K'): /* loads almost directly */
			if ((skip-=8)<0)
				break;
			if (FCEU_fread(&NSFHeader.BankSwitch,8,1,fp)!=1)
				goto error;
			break;
		case MK4CC('t','i','m','e'): /* track times in milliseconds, int32 LSB */
			/* terminated w/ -1s but NotSoFatSo ignores the track times */
			/* music players like MODo use times to automatically advance tracks */
			break;
		case MK4CC('f','a','d','e'): /* ditto; pre/post-track effect; not handled */
			break;
		case MK4CC('p','l','s','t'): /* song playlist (uint8s) */
			if (--skip<0)
				break;
			if (FCEU_fread(&NSFHeader.StartingSong,1,1,fp)!=1)
				goto error;
			break;
		case MK4CC('t','l','b','l'): /* C-str separated song titles */
			if (skip>MaxNamesTotal)
				header->size = MaxNamesTotal;
			if ((SongNames=(uint8*)FCEU_malloc(header->size))==0)
				return -2;
			if (FCEU_fread(SongNames,header->size,1,fp)!=1)
				goto error;
			skip-= MaxNamesTotal;
			break;
		case MK4CC('D','A','T','A'):
			NSFSize = skip;
			if ((skip=NSFLoadData(fp))!=0)
				return skip;
			skip = 0;
			break;
		case MK4CC('N','E','N','D'):
			return 1;
		case MK4CC('a','u','t','h'): /* cross our fingers about the encoding... */
			FCEU_fread(buf,MIN(skip,sizeof(buf)),1,fp);
			skip-= sizeof(buf);
			GetIndName(buf,NSFHeader.SongName,0);
			GetIndName(buf,NSFHeader.Artist,1);
			GetIndName(buf,NSFHeader.Copyright,2);
			GetIndName(buf,CaptionStr,3);
		}
		if (skip>0)
			FCEU_fseek(fp,skip,SEEK_CUR);
	}
	errs = "terminator tag (NEND) missing";
error:	FCEU_PrintError(" NSFe Error: %s.",
		errs);
	return -5;
}

int NSFLoad(FCEUFILE *fp)
{
	int		x;
	uint8		autoBS;

	*CaptionStr = 0;
	FCEU_fseek(fp,0,SEEK_SET);
	FCEU_fread(&NSFHeader,1,0x80,fp);
	if (memcmp(NSFHeader.ID,"NESM\x1A",5))
	{ /* check NSFe */
		if (memcmp(NSFHeader.ID,"NSFE",4))
			return 0;
		FCEU_fseek(fp,4,SEEK_SET);
		if ((x=NSFeLoad(fp))<=0)
			goto fail;
	} else
	{
		NSFSize = FCEU_fgetsize(fp)-0x80;
		if (FCEU_fseek(fp,0x80,SEEK_SET)<0)
			return 0;
	}
	/* does nothing if NSFe code already loaded */
	if ((x=NSFLoadData(fp))<=0) /* shouldn't fail */
		goto fail;
	NSFSizeLo = NSFSize&0xFF;
	NSFSizeHi = NSFSize>>8;

	x = NSFHeader.PALspeed[0]|NSFHeader.PALspeed[1]<<8;
	if (!NSFHeader.VideoSystem || x-8192u>=16384u)
		x = 0;
	PALspeed = x;
	x = NSFHeader.NTSCspeed[0]|NSFHeader.NTSCspeed[1]<<8;
	if (NSFHeader.VideoSystem==1 || x-8192u>=16384u)
		x = 0;
	NTSCspeed = x;

	NSFHeader.SongName[31] = NSFHeader.Artist[31] = NSFHeader.Copyright[31] = 0;

	--NSFMaxBank;

	NSFBankSW = x = autoBS = 0;
	do
		NSFBankSW|= NSFHeader.BankSwitch[x];
	while(++x<8);

	if (!NSFBankSW)
 		if ((NSFHeader.LoadAddressHigh&0x70)>=0x70)
			NSFBankSW = 0xFF; /* for Ice Climber and other F000 based NSFs */
		else
		{
			//autoBS = 1, BankCount = 0;
			//for(x=(NSFHeader.LoadAddressHigh&0x70)>>4; x<8; ++x)
			//	NSFHeader.BankSwitch[x] = BankCount, ++BankCount;
			//x = 0;
			//do
			//	NSFBankSW|= NSFHeader.BankSwitch[x];
			//while(++x<8);
		}

	FCEUGameInfo->type = GIT_NSF;
	FCEUGameInfo->input[0] = FCEUGameInfo->input[1] = SI_GAMEPAD;
	FCEUGameInfo->cspecial = SIS_NSF;

	if (!NSFHeader.VideoSystem)
		FCEUGameInfo->vidsys = GIV_NTSC;
	else if (NSFHeader.VideoSystem==1)
		FCEUGameInfo->vidsys = GIV_PAL;
	//otherwise user-defined

	GameInterface = NSFGI;

	FCEU_printf(
"NSF Loaded.  File information:\n"
//"\n"
" Name:          %s\n"
" Artist:        %s\n"
" Copyright:     %s\n",
		NSFHeader.SongName,NSFHeader.Artist,NSFHeader.Copyright);
	if (*CaptionStr)
		FCEU_printf(
" Rip Credit:    %s\n\n",
			CaptionStr);
	if (NSFHeader.SoundChip)
	{
		static char	*tab[6] = {
			"Konami VRC-6",
			"Konami VRC-7",
			"Nintendo FDS",
			"Nintendo MMC5",
			"Namco 106/163",
			"Sunsoft FME-07"
		};

		int		m;

		x = 0, m =-1;
		do
			if (NSFHeader.SoundChip&1<<x)
			{
				if (!++m)
					FCEU_printf(" Expansion hardware:  %s\n",
						tab[x]);
				else
					FCEU_printf("                    + %s\n",
						tab[x]);
				/* Prevent confusing weirdness if more than one bit is set. */
				// actually, a working combo may be used; 2010.03.18
				//r = 1<<x;
			}
		while(++x<6);
		//if (m>0)
		//{
		//	FCEU_printf("   Note: only the last one is used.\n");
		//	NSFHeader.SoundChip = r;
		//}
				
	}

	if (NSFBankSW)
		FCEU_printf(autoBS ?
" Bank-switched (auto)\n"
:" Bank-switched (%02X %02X %02X %02X %02X %02X %02X %02X).\n",
			NSFHeader.BankSwitch[0],
			NSFHeader.BankSwitch[1],
			NSFHeader.BankSwitch[2],
			NSFHeader.BankSwitch[3],
			NSFHeader.BankSwitch[4],
			NSFHeader.BankSwitch[5],
			NSFHeader.BankSwitch[6],
			NSFHeader.BankSwitch[7]);
	FCEU_printf(
" Load address:  $%04X\n"
" Init address:  $%04X\n"
" Play address:  $%04X\n"
" %4u KB / %u KB  %s\n",
		LoadAddr,
		NSFHeader.InitAddressLow|(NSFHeader.InitAddressHigh<<8),
		NSFHeader.PlayAddressLow|(NSFHeader.PlayAddressHigh<<8),
		NSFSize+1023>>10,NSFMaxBank*4+4,
		NSFHeader.VideoSystem==1 ? "PAL":NSFHeader.VideoSystem ? "NTSC/PAL":"NTSC");

	if (PALspeed|NTSCspeed)
	{
		FCEU_printf(" Speed(s):      ");
		if (NTSCspeed)
			FCEU_printf("%g fps NTSC",
				1e6/NTSCspeed);
		if (PALspeed && NTSCspeed)
			FCEU_printf(" / ");
		if (PALspeed)
			FCEU_printf("%g fps PAL",
				1e6/PALspeed);
		FCEU_printf("\n\n");
	}

	if (!NSFHeader.TotalSongs)
		NSFHeader.TotalSongs = 1; //fix it; 2010.03.18
	NSFLastSong = NSFHeader.TotalSongs-1;

	FCEU_printf(" Starting song:  %d / %d\n\n",
		NSFHeader.StartingSong,NSFHeader.TotalSongs);

	if (!NSFHeader.StartingSong)
		NSFHeader.StartingSong = 1;

	ExWRAMSize = NSFHeader.SoundChip&1<<SCNFDS ? 32768+8192:8192;
	if ((ExWRAM=FCEU_malloc(ExWRAMSize))==NULL)
	{
		x =-2;
	fail:	NSFGI(GI_CLOSE); /* frees everything */
		return x;
	}
	return 1;
}

static void NSFReset(void)
{

	int32		a, x;

	NSFCycles = LCYCS*240;
	if (PAL)
	{
		if (PALspeed)
			NSFCycles = ((ufract32mul(PALspeed*256u,312.*341*4096*65536/19997+.5)+8)>>4)-(LCYCS*72);
	} else if (NTSCspeed)
		NSFCycles = ((ufract32mul(NTSCspeed*256u,262.*LCYCS*4096*65536/16639+.5)+8)>>4)-(LCYCS*22);

	dwmemset(ExWRAM,0,ExWRAMSize);
	dwmemset(RAM,0,0x800);

	a = 0x8000, x = 0;
	if (NSFBankSW)
		do
		{
			if (NSFHeader.SoundChip&1<<SCNFDS && x>=6) //also do 6xxx-7xxx
				BANKSET(x<<12,NSFHeader.BankSwitch[x]);
			BANKSET(a,NSFHeader.BankSwitch[x]);
		} while(++x,(a+=0x1000)<0x10000);
	else //new approach, should work
		do
			BANKSET(a,x);
		while(x+=(a>=(LoadAddr&0xF000)),(a+=0x1000)<0x10000);
	initframes = NSFMaxInitFrames;
}

static DECLFR(NSF_ROMRead)
{
	//if (!fceuindbg)
	//	NSFNMIFlags&=~NMIReset;
	return NSFROM[A-0x3880];
}

static DECLFR(NSF_VectorRead) //FFFA
{
	if (NSFNMIFlags)
	{
		if (A==0xFFFD)
			if (!fceuindbg)
				NSFNMIFlags = NMISongReload;
		return A&1 ? 0x38:0x80;//A==0xFFFC ? *NSFROM:0x81;
	} else
		return CartBRF8(A);
}

static DECLFR(NSFR38F8)
{
	return (((const uint8*)&NSFHeader.InitAddressLow)-0x38F8)[A];
}

static DECLFR(NSFR38FC)
{
	if (A&1) /* 38FD */
		return PAL;
	else
		return NSFHeader.SoundChip;
}

static DECLFR(NSFR38FE)
{
	if (A==0x38FF)
	{
		uint8	r = NSFNMIFlags;

		if (!fceuindbg)
			NSFNMIFlags&=~(NMISongReload|NMIEnabled);
		return r;
	}
	if (!fceuindbg)
		NSFReset();
	return NSFCurrentSong;
}

static DECLFW(NSFW38FF)
{
(void)(A,V);
	NSFNMIFlags|= NMIEnabled;
}

static DECLFW(NSF_BankWrite) //5FF6/8-5FFF standard
{
	if (!NSFBankSW)
		return;
	A&= 0xF;
	BANKSET(A*4096,V);
}

static void RenderSongNames(void)
{
	if (SongNames)
		GetIndName(SongNames,CurrentSongName,NSFCurrentSong);
	else
		*CurrentSongName = 0;
	sprintf((char*)CaptionStr,"Song: <%d/%d>",
		NSFCurrentSong+1,NSFHeader.TotalSongs);
}

void NSF_init(void)
{
	ResetCartMapping();

	if (NSFHeader.SoundChip&1<<SCNFDS)
	{
		SetupCartPRGMapping(0,ExWRAM,ExWRAMSize,1);
		setprg32(0x6000,0);
		setprg8(0xE000,4);
		SetWriteHandler(0x6000,0xDFFF,CartBW);
	} else
	{
		SetupCartPRGMapping(0,NSFDATA,(NSFMaxBank+1)*4096,0);
		SetupCartPRGMapping(1,ExWRAM,ExWRAMSize,1);
		setprg8r(1,0x6000,0);
		SetWriteHandler(0x6000,0x7FFF,CartBW);
	}
	SetCartRead(0x6000,0xFFFF);

	NSFReset();

	SetReadHandler(0x2000,0x3FFF,0);
	SetReadHandler(0x3880,0x38F7,NSF_ROMRead);
	SetReadHandler(0x38F8,0x38FB,NSFR38F8);
	SetReadHandler(0x38FC,0x38FD,NSFR38FC);
	SetReadHandler(0x38FE,0x38FF,NSFR38FE);
	SetReadHandler(0xFFFA,0xFFFD,NSF_VectorRead);
	SetWriteHandler(0x2000,0x3DFF,0);
	SetWriteHandler(0x38FF,0x38FF,NSFW38FF);
	SetWriteHandler(0x5FF6,0x5FF7,NSFHeader.SoundChip&1<<SCNFDS ? NSF_BankWrite:0); //was ? 0:..
	SetWriteHandler(0x5FF8,0x5FFF,NSF_BankWrite);

	if (NSFHeader.SoundChip&1<<SCVRC6)
		NSFVRC6_Init();
	if (NSFHeader.SoundChip&1<<SCVRC7)
		NSFVRC7_Init();
	if (NSFHeader.SoundChip&1<<SCNFDS)
		FDSSoundReset();
	if (NSFHeader.SoundChip&1<<SCMMC5)
		NSFMMC5_Init();
	if (NSFHeader.SoundChip&1<<SCN106)
		NSFN106_Init();
	if (NSFHeader.SoundChip&1<<SCSFME)
		NSFAY_Init();
	NSFCurrentSong = NSFHeader.StartingSong-1;
	NSFNMIFlags = NMISongReload|NMIEnabled|NMIReset;

	RenderSongNames();
}

static void TestJoy(void)
{
	static uint8	last = 0;
	uint8		current = FCEU_GetJoyJoy();

	if ((current&JOY_RIGHT) && !(last&JOY_RIGHT))
	{
		if (NSFCurrentSong<NSFLastSong)
		{
			++NSFCurrentSong;
			NSFNMIFlags|= NMISongReload;
		}
	} else if ((current&JOY_LEFT) && !(last&JOY_LEFT))
	{
		if (NSFCurrentSong>0)
		{
			--NSFCurrentSong;
			NSFNMIFlags|= NMISongReload;
		}
	} else if ((current&JOY_UP) && !(last&JOY_UP))
	{
		if (NSFCurrentSong+10>NSFLastSong)
			NSFCurrentSong = NSFLastSong;
		else
			NSFCurrentSong+= 10;
		NSFNMIFlags|= NMISongReload;
	} else if ((current&JOY_DOWN) && !(last&JOY_DOWN))
	{
		if (NSFCurrentSong<10)
			NSFCurrentSong = 0;
		else
			NSFCurrentSong-= 10;
		NSFNMIFlags|= NMISongReload;
	} else if ((current&JOY_START) && !(last&JOY_START))
		NSFNMIFlags|= NMISongReload;
	else if ((current&JOY_A) && !(last&JOY_A))
		NSFSpecial = (NSFSpecial+1)%3;
	else
		goto skip;
	RenderSongNames();
skip:
	last = current;
}

#define ITERATION_FRACTION	(0x80000000u/256)
#define FULL_CIRCLE		0

void DrawNSF(uint8 *XBuf)
{
	int16		*Bufpl;
	int32		mul;
	int		special;
	int		l, x = 0, i = 0;

	if (!vismode)
		return;
	if (!XBuf)
		goto input;
	/* default is PAL_BLACK (as it used to be) */
	l = PPUvoidcolor, l|= l<<8, l|= l<<16;
	{
		uint8		*d = XBuf;

		i = 240;
		do
			dwmemset(d,l,256);
		while(d+=FCEUSCRN_ROWSIZE,--i);
	}

	l = GetSoundBuffer(&Bufpl);

	if (!FSettings.SoundVolume)
		mul = 0;
	else
		mul = (50*240)/ABS(FSettings.SoundVolume); //added ABS to make inversion clear

	if ((special=NSFSpecial&3)==0)
	{
		do
		{
			uint32		y = 142+((Bufpl[i>>8]*mul)>>14);

			if (y<240)
				XBuf[x+y*FCEUSCRN_ROWSIZE] = PAL_BGRAY;
		} while(i+=l,++x<256);
	} else if (special==1)
	{
		ufract		t = 0;

		do
		{ /* all of the rest of the trig functions used can be replaced */
			int32		r = (Bufpl[i>>8]*mul)>>14;
			uint32		xp = 128+cosmul(r,t);
			uint32		yp = 120+sinmul(r,t);

			xp&= 255;
			yp%= 240;
			XBuf[xp+yp*FCEUSCRN_ROWSIZE] = PAL_BGRAY;
		} while(i+=l,++x,(t+=ITERATION_FRACTION*2)!=FULL_CIRCLE);
	} else if (special==2)
	{
		static ufract	theta = 0;

		x = 0;
		do
		{
			int32		xc = 128-x;
			int32		yc = 0-(int32)((Bufpl[i>>8]*mul)>>14);
			ufract		t = theta+0x80000000u+ifractatan2(yc,xc);
			int32		r = ihypot2(xc,yc);
			uint32		m = 128+cosmul(r,t);
			uint32		n = 120+sinmul(r,t);

			if (m<256 && n<240)
				XBuf[m+n*FCEUSCRN_ROWSIZE] = PAL_BGRAY;
		} while(i+=l,++x<128);

		do
		{
			int32		xc = x-128;
			int32		yc = (Bufpl[i>>8]*mul)>>14;
			ufract		t = theta+ifractatan2(yc,xc);
			int32		r = ihypot2(xc,yc);
			uint32		m = 128+cosmul(r,t);
			uint32		n = 120+sinmul(r,t);

			if (m<256 && n<240)
				XBuf[m+n*FCEUSCRN_ROWSIZE] = PAL_BGRAY;
		} while(i+=l,++x<256);
		theta += ITERATION_FRACTION;
	}

	DrawTextTrans(XBuf+10*FCEUSCRN_ROWSIZE+4+(31-strlen((char*)NSFHeader.SongName)<<2),
		FCEUSCRN_ROWSIZE, NSFHeader.SongName, PAL_VIOLET);
/*26*/	DrawTextTrans(XBuf+25*FCEUSCRN_ROWSIZE+4+(31-strlen((char*)NSFHeader.Artist)<<2),
		FCEUSCRN_ROWSIZE,NSFHeader.Artist, PAL_VIOLET);
/*42*/	DrawTextTrans(XBuf+40*FCEUSCRN_ROWSIZE+4+(31-strlen((char*)NSFHeader.Copyright)<<2),
		FCEUSCRN_ROWSIZE,NSFHeader.Copyright, PAL_VIOLET);
/*70*/	DrawTextTrans(XBuf+55*FCEUSCRN_ROWSIZE+4+(31-strlen((char*)CaptionStr)<<2),
		FCEUSCRN_ROWSIZE,CaptionStr, PAL_VIOLET);
	if (*CurrentSongName)
		DrawTextTrans(XBuf+70*FCEUSCRN_ROWSIZE+4+(31-strlen((char*)CurrentSongName)<<2),
			FCEUSCRN_ROWSIZE,CurrentSongName, PAL_VIOLET);
input:	TestJoy();
}

void DoNSFFrame(void)
{
	if (NSFNMIFlags&NMIEnabled)
		XFrame.intlow|= FCEU_IQNMI; //TriggerNMI
	else if (initframes>=0 && --initframes<0)
		NSFNMIFlags|= NMIEnabled;
}

void FCEUI_NSFSetVis(int mode)
{
	vismode = !!mode;
}

int FCEUI_NSFChange(int amount)
{
	int	current = NSFCurrentSong;

	if ((current+=amount)<0)
		current = 0;
	else if (current>NSFLastSong)
		current = NSFLastSong;
	NSFNMIFlags|= NMISongReload;

	return NSFCurrentSong=current;
}

/* Returns total songs */
int FCEUI_NSFGetInfo(uint8 *name,
	uint8 *artist,
	uint8 *copyright,
	int maxlen)
{
	strncpy((char*)name,(char*)NSFHeader.SongName,maxlen);
	strncpy((char*)artist,(char*)NSFHeader.Artist,maxlen);
	strncpy((char*)copyright,(char*)NSFHeader.Copyright,maxlen);

	return NSFHeader.TotalSongs;
}
