/*
 PPU.c
  Picture Processing Unit 2C02/2C07

 * MacFCEU 0.9d3
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  ppu.c Copyright (C) 1998 BERO
 *  ppu.c Copyright (C) 2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "types.h"
#include "x6502.h"
#include "FCEU.h"
#include "git.h"
#include "PPU.h"
#include "NSF.h"
#include "GameGenie.h"
#include "fsound.h"
#include "general.h"
#include "memory.h"

#include "cart.h"
#include "palette.h"
#include "state.h"
#include "video.h"
#include "input.h"
#include "boards/MMC5.h"

#include "ftrig.h"

#ifdef env68k
 #pragma segment shell
#endif

/* ---- constants */

/* The makeup of a 15 bit daisy chain:
 bits 12 : fv.3	- fine vertical scroll
 bit  11 : v.1	- vertical scroll counter
 bit  10 : h.1	- horizontal scroll counter
 bits  5 : vt.5	- vertical tile position   -- bit #1 of this shifts palette by 4
 bits  0 : ht.5	- horizontal tile position -- bit #1 of this shifts palette by 2
 bits(-3): fh.3	- fine horizontal scroll
*/
#define HV_SCR_SHIFT		10
#define HT_POS_MASK		0x1F
#define H_POS_MASKS		0x41F
#define T_POS_MASKS		0x3FF
#define POS_MASKS		0x3FFF

/* delay cyles after transferring the interrupt pending to phase bit (A2002.7) */
/* this is correct with how long it takes for an NMI to take effect with the CPU, etc. */
#define NMI_DELAY		14

/* ---- globals */

int		fceuindbg = 0;

uint8		PPUmaxsprites = 8;
uint8		PPUrendis = 0;

uint32		PPUVFrameCounter;
static uint8	PPUBoot, OAMLatch;

uint8		VRAMBuffer, PPUGenLatch;
uint8		*vnapage[4];
uint8		PPUNTARAM;
uint8		PPUCHRRAM;

/* Color deemphasis emulation.  Joy... */
uint32		PPUEmphasis = 0;

void (*GameHBIRQHook)(void), (*GameHBIRQHook2)(void);
void FP_FASTAPASS(1) (*PPU_hook)(uint32 A);

uint32		TempAddr, RefreshAddr;

/* "scanline" is the current visible scanline we're on */

int		scanline;
static int32	PPUVBL_SLcc; //has NMI_DELAY already subtracted

uint8		PPU[4];
uint8		PPUSPL, PPUSPO;
uint32		XOffset = 0;
uint8		vtoggle = 0;

//OAM rewrite pending
uint8		PPUvoidcolor = 255;

/* new form: a modified *copy* of PALRAM is used; no more hacks. :) */
uint8		PPUPal[0x20];
uint8		PALRAM[0x20], NTARAM[0x800];
uint8		SPRAM[0x104], SPRBUF[0x104];

uint8		*PPU_Pline, *PPU_Plinef;
static int	linepel;
static int	linestartts;
static uint8	linereset;
extern uint8 *FCEU_video;

uint8		bitreverseLUT[256];
#ifndef env68k
uint32		PPU_LUT1to4[256]; /* 1024B */
uint32		PPU_TileLUT[128]; /* 512B */
#endif

/* ---- macros */

#define VBlankON		(PPU[0]&0x80)	/* Generate VBlank NMI */
#define Sprite16		(PPU[0]&0x20)	/* Sprites 8x16/8x8 */
#define BGAdrHI			(PPU[0]&0x10)	/* PF pattern adr $0000/$1000 */
#define SpAdrHI			(PPU[0]&8)	/* OAM pattern adr $0000/$1000 */
#define IncByV			(PPU[0]&4)	/* by 32, otherwise 1 (horiz.) */

#define RenderON		(PPU[1]&0x18)	/* either */
#define SpriteON		(PPU[1]&0x10)	/* show OAM layer */
#define PlayfdON		(PPU[1]&8)	/* show playfield */
#define SpFirst8		(PPU[1]&4)	/* render first 8 OAM pixels on line */
#define BGFirst8		(PPU[1]&2)	/* render first 8 PF pixels on line */
#define GrayLine		(PPU[1]&1)	/* erase line color */

#define PPU_status		PPU[2]

#define PPUNoFrame		(scanline>239||!RenderON)

#define exp8(t,p)		(t=p,t|=(unsigned)t<<8,t|=(unsigned)t<<16)
#ifdef C80x86
 #define tilespr32(P,R,d)	do{uint8 *S=R; uint32 V=d; \
	 asm volatile( \
	 "movl %%ebx,%%eax\n\t" \
	 "andl $15,%%eax\n\t" \
	 "movb (%%esi,%%eax),%%cl\n\t" \
	 "shrl $4, %%ebx\n\t" \
	 "movb %%cl, (%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 1(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 2(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 3(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 4(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 5(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 6(%%edi)\n\t" \
\
         /*"movl %%ebx,%%eax\n\t"*/ \
         /*"andl $15,%%eax\n\t"*/ \
         /*"movb (%%esi,%%eax),%%cl\n\t"*/ \
         /*"movb %%cl, 7(%%edi)\n\t"*/ \
	 "movb (%%esi, %%ebx),%%cl\n\t" \
	 "movb %%cl, 7(%%edi)\n\t" \
	 : \
	 :"S" (S), "D" (P), "b" (V) \
	 :"eax", "cl"); \
}while(0)
#elif defined(env68k)
 #define blittile(d,p,m,a,x)	_blittile(d,PPUPal,(unsigned)p<<16|(uint16)m,a,x)
 #undef exp8
 #define exp8(t,p)		(t=_exp8((uint8)(p)))
 #pragma parameter __D0 _exp8(__D0)
 extern uint32 _exp8(uint8) = {
/*	MOVE.B		D0,D2	*/ 0x1400,
/*	LSL.W		#8,D0	*/ 0xE148,
/*	MOVE.B		D2,D0	*/ 0x1002,
/*	MOVE.W		D0,D2	*/ 0x3400,
/*	SWAP		D0	*/ 0x4840,
/*	MOVE.W		D2,D0	*/ 0x3002};
 #pragma parameter __A0 _blittile(__A0, __A1, __D2, __D1, __D0)
 extern uint8 *_blittile(uint8 *D,
	const void *T,
	uint32 P,
	uint16 A,
	uint16 X) = {
/*	MOVE.W		D1,-(SP) */ 0x3F01,
/*	MOVE.W		D0,-(SP) */ 0x3F00,
/*	LSL.L		D0,D2	*/ 0xE1AA, // could be optimized internally;
/*	MOVE.L		D2,D1	*/ 0x2202, // only two loops needed, both start
/*	CLR.W		D1	*/ 0x4241, //and end of `for`; a fixed 8 would
/*	SUBQ.B		#8,D0	*/ 0x5100, //be done in center; attr. would
/*@0	MOVE.B		3(A7),D1 */ 0x122F, 3, //need no latch
/*	ADD.W		D2,D2	*/ 0xD442,
/*	ADDX.L		D1,D1	*/ 0xD381,
/*	ADDX.B		D1,D1	*/ 0xD301,
/*	MOVE.B		0(A1,D1.W),(A0)+ */ 0x10F1, 0x1000,
/*	ADDQ.B		#1,D0	*/ 0x5200,
/*	BCC.S		@0	*/ 0x64EE,
/*	MOVE.W		(A7)+,D0 */ 0x301F,
/*	BEQ.S		@2	*/ 0x6710,
/*@1	MOVE.B		(A7),D1	*/ 0x1217,
/*	ADD.W		D2,D2	*/ 0xD442,
/*	ADDX.L		D1,D1	*/ 0xD381,
/*	ADDX.B		D1,D1	*/ 0xD301,
/*	MOVE.B		0(A1,D1.W),(A0)+ */ 0x10F1, 0x1000,
/*	SUBQ.B		#1,D0	*/ 0x5300,
/*	BNE.S		@1	*/ 0x66F0,
/*@2	ADDQ.W		#2,A7	*/ 0x544F};
#else
 #define tilespr32(P,R,d)	do{ \
	uint8 *S=R; uint32 V=d; \
	*P=S[V&15]; \
	P[1]=S[(V>>=4)&15]; \
	P[2]=S[(V>>=4)&15]; \
	P[3]=S[(V>>=4)&15]; \
	P[4]=S[(V>>=4)&15]; \
	P[5]=S[(V>>=4)&15]; \
	P[6]=S[(V>>=4)&15]; \
	P[7]=S[V>>=4]; \
}while(0)
#endif

#define MMC5SPRVRAMADR(V)	&MMC5SPRVPage[(V)>>HV_SCR_SHIFT][V]
#define MMC5BGVRAMADR(V)	(Sprite16||MMC5ABMode ? \
	MMC5SPRVRAMADR(V):&MMC5BGVPage[(V)>>HV_SCR_SHIFT][V])
#define VRAMADR(V)		&VPage[(V)>>HV_SCR_SHIFT][V]

#define VRead(d,a)		do{ \
	if ((a)<0x2000) \
		d=VPage[(a)>>HV_SCR_SHIFT][a]; \
	else \
		d=vnapage[(a)>>HV_SCR_SHIFT&3][(a)&0x3FF]; \
}while(0)

/* ---- code */

static void makeluts(
	uint8 *brt,
#ifndef env68k
	uint32 *l1b,
	uint32 *tlu
#endif
	)
{
	{
		uint8	*d = brt, v;
		uint32	x = 0;

		do
		{
			v = x<<5&0x40;
			v|= x<<3&0x20;
			v|= (x+x)&0x10;
			v|= x>>1&8;
			v|= x>>3&4;
			v|= x>>5&2;
			v|= x>>7;
			*d++ = v, v|= 0x80;
			*d++ = v;
		} while((x+=2)<256);
	}
#ifndef env68k
	{
		uint32	*d = l1b, x = 0, v;

		do
		{ /* unrolled */
			v = x>>7;
			v|= x>>6-4&0x10;
			v|= x<<8-5&0x100;
			v|= x<<12-4&0x1000;
			v|= x<<16-3&0x10000;
			v|= x<<20-2&0x100000;
			v|= x<<24-1&0x1000000;
			*d++ = v,v|= 0x10000000, *d++ = v;
		} while((x+=2)<256);
	}
	{
		uint32	*d = tlu, cc = 0, xo;

		do
		{
			xo = 0;
			do
			{
				uint32	i = 8, k = xo, v = 0, a = 2;

				do
					v|= ((cc>>(k>>3)*2)&3)<<a;
				while(++k,a+=4,--i);
				*d++=v;
			} while(++xo<8);
		} while(++cc<16);
	}
#endif
}

DECLFW(PPUB2004);

#ifdef env68k
DECLFR(PPUA2002);
DECLFR(PPUA200x);
DECLFR(PPUA2007);

DECLFW(PPUB2000);
DECLFW(PPUB2001);
DECLFW(PPUB2002);
DECLFW(PPUB2003);
DECLFW(PPUB2005);
DECLFW(PPUB2006);
DECLFW(PPUB2007);
DECLFW(PPUB4014);

void CheckSpriteHit(int e);
void RefreshSprites(void);
 #define STATIC
#else
 #define STATIC static
#endif

static void FASTAPASS(1) RefreshLine(int lastpixel);

#define LinePixel()	(PAL ? /* /15 multiply optimization; 2011.02.08 */ \
	ufract32mul(timestamp*48-linestartts,(4294967296.0*8+14)/15)>>3 \
	:(timestamp*48-linestartts)>>4)

int FCEUPPU_GetPixel(void)
{
	if (PPU_Pline)
		return LinePixel()%LCYCS;
	return -1;
}

#define LineUpdate()		do{ \
	if (fceuindbg) \
		break; \
	if (PPU_Pline) \
		RefreshLine(LinePixel()); \
} while(0)

STATIC DECLFR(PPUA2002)
{ /* models 1.4, 2 and 3 return the top two bits and a signature */
	uint8		ret;
(void)A;
	LineUpdate();
	ret = (PPU_status&0xE0)|(PPUGenLatch&0x1F);

	if (fceuindbg)
		return ret;
	vtoggle = 0;
	PPU_status&= 0x77;

	return PPUGenLatch=ret;
}

STATIC DECLFR(PPUA200x)
{ /* 3000 updates by a cycle */
(void)A;
	LineUpdate();

	return PPUGenLatch;
}

STATIC DECLFR(PPUA2004)
{
	uint8		ret;
(void)A;
	LineUpdate();
	ret = 0xFF;
	if (PPUNoFrame) /* no rendering */
		if (PPUSPL<8)
			ret = SPRAM[PPUSPL];
		else if (PPU[3]>7)
			ret = SPRAM[PPU[3]];
	else //if (linepix)
		ret = OAMLatch;

	return PPUGenLatch=ret;
}

STATIC DECLFR(PPUA2007)
{ /* updates by 4 cycles, CPU gets feedback on address */
	uint8		ret;

	A = RefreshAddr&POS_MASKS;
	if (!fceuindbg)
	{
		LineUpdate();
		if (IncByV)
			RefreshAddr+= 0x20;
		else
			++RefreshAddr;
		if (PPU_hook)
			PPU_hook(A);
	}

	/* palette read */
	if (!(~A&0x3F00))
	{
		//if (!(last2007pel|(raddr&0x1F)) /* anything set, for index 0? */
		//	ret = PALRAM[16];
		//else
		ret = PALRAM[A&31];
		if (GrayLine)
			ret&= 0x30;
		A-= 0x1000; /* FCEUX does this, Nestopia does not */
	} else
		ret = VRAMBuffer;
	if (fceuindbg)
		return ret;

	VRead(VRAMBuffer,A);

	if (PPU_hook)
		PPU_hook(RefreshAddr&0x3FFF);

	return PPUGenLatch=ret;
}

STATIC DECLFW(PPUB2000) //hv latch
{
	int		sv;
(void)A;
	if (PPUBoot)
		return;
	LineUpdate();
	sv = (uint32)V<<HV_SCR_SHIFT;
	TempAddr&= 0xF3FF, TempAddr|= sv&0xC00;
	PPUGenLatch = V;
	V = PPU[0];
	PPU[0] = PPUGenLatch;
	if ((VBlankON&PPU_status)>V) //!VBlankON && V&PPU_status
		XFrame.intlow|= FCEU_IQNMI2; //TriggerNMI2
	/*else if (scanline==241 && !(V&0x80) && ppu_cycle<3)
		XFrame.intlow&=~FCEU_IQNMI2;*/

}

STATIC DECLFW(PPUB2001) //general latch
{
(void)A;
	LineUpdate();
	PPUGenLatch = PPU[1] = V;
}

STATIC DECLFW(PPUB2002) //status
{
(void)A;
	PPUGenLatch = V;
}

STATIC DECLFW(PPUB2003) //OAM addr
{
(void)A;
	PPUGenLatch = V;
	PPU[3] = V;
	PPUSPL = V&7;
}

DECLFW(PPUB2004) //OAM data
{
(void)A;
	//if (PPUNoFrame) /* 2010.05.11, 2010.05.27 */
	{
		if ((PPU[3]&3)==2)
			V&= 0xE3; /* some bits aren't stored */
	}//else
	//	V = 0xFF;

	PPUGenLatch = V;

	if (PPUSPL<8)
		SPRAM[PPUSPL] = V;
	else if (PPU[3]>7)
		SPRAM[PPU[3]] = V;
	++PPU[3];
	++PPUSPL;
}

STATIC DECLFW(PPUB2005) //scroll-toggle h/v
{
	uint32		tmpa = TempAddr;
(void)A;
	LineUpdate();
	PPUGenLatch = V;
	if (vtoggle^=1)
		tmpa&= 0xFFE0, tmpa|= V>>3, XOffset = V&7;
	else
		tmpa&= 0x8C1F, tmpa|= ((uint32)V&~7)<<2,
			tmpa|= (uint32)V<<12&0x7000;
	TempAddr = tmpa;
}

STATIC DECLFW(PPUB2006) //address latch
{
(void)A;
	LineUpdate();
	PPUGenLatch = V;
	if (vtoggle^=1)
		TempAddr&= 0xFF, TempAddr|= (uint32)V<<8&0x3F00;
	else
	{
		TempAddr&= 0xFF00, RefreshAddr = (TempAddr|=V);
		if (PPU_hook)
			PPU_hook(RefreshAddr);
	}
}

STATIC DECLFW(PPUB2007) //post-incr addr data write
{
	uint32		k, addrcode = RefreshAddr&POS_MASKS;
(void)A;
	PPUGenLatch = V;
	if (addrcode>=0x3F00)
	{
		uint8	pel;

		if (!RenderON)
			LineUpdate(); /* for full_nes_palette.nes; 2010.04.13 */

		k = addrcode&0x1F;
		/* bit #6 set so spr pixel won't be zero */
		PALRAM[k] = V, PPUPal[k] = pel = V&0x7F|PLB_S;
		if (!(k&3)) /* transparent attr entries are shared */
			k^= 0x10, PALRAM[k] = V, PPUPal[k] = pel;
		/* index #0 for transparent (based on trial and error) */
		PPUPal[0] = PPUPal[4] = PPUPal[8] = PPUPal[12] = PALRAM[0]&0xBF|PLB_B;
	} else
	{
		k = addrcode>>HV_SCR_SHIFT;
		if (addrcode<0x2000)
		{
			if ((uint8)(PPUCHRRAM&1u<<k))
				VPage[k][addrcode] = V;
		} else
		{
			k&= 3;
			if ((uint8)(PPUNTARAM&1u<<k))
				vnapage[k][addrcode&0x3FF] = V;
		}
	}

	if (IncByV)
		addrcode+= 0x20;
	else
		++addrcode;
	RefreshAddr = addrcode;
	if (PPU_hook)
		PPU_hook(addrcode&POS_MASKS);
}

STATIC DECLFW(PPUB4014) //copy 2004h line
{
	uint32		t = (uint32)V<<8;
	int		x = 256;
(void)A;
	do
		X6502_DMW(0x2004,X6502_DMR(t));
	while(++t,--x);
}

void FCEUPPU_LineUpdate(void)
{
	LineUpdate();
}

/* (incomplete) OAM emulation */

uint8		PPUsphitdata;
uint8		PPUsprlinebuf[256+8];
int32		PPUsphitx;

STATIC void CheckSpriteHit(int e)
{
	const uint8	*r;
	int		x = PPUsphitx;
	uint8		d;

	/* optimized version using Dan Weiss's suggestion; 2010.04.08 */
	if ((e-=16)>254)
		e = 255; /* end before 255; should fix "Battletoads" */
	if ((x+=8)<e) /* take and pin, 8 max */
		e = x;
	if ((x-=8)>=e)
		return;

	r = PPU_Plinef;
	if ((uint8)((d=PPUsphitdata)&0x80) && !(uint8)(r[x]&PLB_B))
		goto hit;
	if (++x>=e)
		return;
	do
	{
		if (!(uint8)((d+=d)&0x80) || (uint8)(r[x]&PLB_B))
			continue;
	hit:	PPU_status|= 0x40;
		PPUsphitx = 0x100;
		break;
	} while(++x<e);
}

uint8		PPUOAMrefreshed = 0; /* spork the world. */
/*	Any sprites on this line?
	Then this will be set to 1.  Needed for zapper
	emulation and *gasp* sprite emulation.
*/

#define V_FLIP		0x80
#define H_FLIP		0x40
#define SP_BACK		0x20

typedef struct SPR
{
        uint8	y, no, atr, x;
} SPR;

typedef struct SPRB
{
        uint8	ca[2], atr, x;
} SPRB;


uint8		PPUnumsprites, PPUlineHasSp0;

#define SpFetchAddr(A,S,O,V,P) do{A=S->no; \
	if (P) \
		A=(A<<12&0x1000)+(A<<4&0xFE0); \
	else \
		A=(A<<4)+O; \
	if (S->atr&V_FLIP) \
		A+=7,A-=V,A+=P,A-=V&8; \
	else \
		A+=V,A+=V&8; \
}while(0)

static void FetchSpriteData(void)
{
	SPRB		dst;
	SPR		*S = (SPR*)SPRAM;
	uint32		vofs, Pf = Sprite16>>2, H = 8+Pf;
	int		n = 64, ns = 0, ms = PPUmaxsprites;
	uint8		spzonline = 0; /* delay for callbacks */

	vofs = (uint32)((SpAdrHI*2)&~(Pf+=Pf))<<8;
	if (!PPU_hook)
	{
		do
		{
			unsigned	V = scanline-S->y;

			if (V>=H)
				continue;
			if (ns>=ms)
				{PPU_status|= 0x20; break;}
			if (n==64) /*n==63 [n was 0-63 in `for`]*/
				spzonline = 1;

			{
				uint8		*C;
				unsigned	vadr;

				SpFetchAddr(vadr,S,vofs,V,Pf);

				/* Fix this geniestage hack */
				if (MMC5Ex && GenieStage!=1)
					C = MMC5SPRVRAMADR(vadr);
				else
					C = VRAMADR(vadr);

				dst.ca[0] = *C, dst.ca[1] = C[8];
				*(int16*)&dst.atr = *(int16*)&S->atr; /*atr,x*/
				*(int32*)&SPRBUF[ns<<2] = *(int32*)&dst;
			}
			++ns;
		} while(((uint8*)S)+=4,--n);

		if (ns>8) /* handle >8 sprites per scanline */
			PPU_status|= 0x20;
	} else
	{
		do
		{
			unsigned	V = scanline-S->y;

			if (V>=H)
				continue;
			if (ns>=ms)
				{PPU_status|= 0x20; break;}
			if (n==64)
				spzonline = 1;

			{
				uint8		*C;
				unsigned	vadr;

				SpFetchAddr(vadr,S,vofs,V,Pf);

				if (MMC5Ex)
					C = MMC5SPRVRAMADR(vadr);
				else
					C = VRAMADR(vadr);
				dst.ca[0] = *C;

				if (ns<8)
				{
					PPU_hook(0x2000);
					PPU_hook(vadr);
				}

				dst.ca[1] = C[8];
				*(int16*)&dst.atr = *(int16*)&S->atr; /*atr,x*/
				*(int32*)&SPRBUF[ns<<2] = *(int32*)&dst;
			}
			++ns;
		} while(((uint8*)S)+=4,--n);

		if (ns>8)
			PPU_status|= 0x20;
		else if ((n=8-ns)>0)
			do
			{
				PPU_hook(0x2000);
				PPU_hook(vofs);
			} while(--n);
	}

	PPUnumsprites = ns;
	PPUlineHasSp0 = spzonline;
}

#ifndef env68k
STATIC void RefreshSprites(void)
{
	SPRB		*spr;
	int		n;

	PPUOAMrefreshed = 0;
	if ((n=PPUnumsprites)==0)
		return;
	FCEU_dwmemset(PPUsprlinebuf,0,256);
	PPUnumsprites = --n;
	spr = n+(SPRB*)SPRBUF;
	do
	{
		register uint32	pixdata;
		register uint8	J = spr->ca[0], atr;

		uint8	*C, *VB;
		int	x;

		pixdata = PPU_LUT1to4[spr->ca[1]], pixdata+= pixdata;
		pixdata|= PPU_LUT1to4[J];
		if ((J|=spr->ca[1])==0)
			continue;
		x = spr->x, atr = spr->atr;
		if (!n && PPUlineHasSp0 && !(PPU_status&0x40))
		{
			if (!(uint8)(atr&H_FLIP))
				PPUsphitdata = J;
			else
				PPUsphitdata = bitreverseLUT[J];
			PPUsphitx = x;
		}

		C = PPUsprlinebuf+x;
		VB = (PPUPal+0x10)+((atr&3)<<2);

		if ((uint8)(atr&SP_BACK))
			if ((uint8)(atr&H_FLIP))
			{
				if ((uint8)(J&0x80))
					C[7] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[6] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[5] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[4] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[3] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[2] = VB[pixdata&3];
				if ((J+=J)&0x80)
					C[1] = VB[pixdata>>4&3];
				if ((J+=J)&0x80)
					*C = VB[pixdata>>=8];
			} else
			{
				if ((uint8)(J&0x80))
					*C = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[1] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[2] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[3] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[4] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[5] = VB[pixdata&3];
				if ((J+=J)&0x80)
					C[6] = VB[pixdata>>4&3];
				if ((J+=J)&0x80)
					C[7] = VB[pixdata>>=8];
			}
		else
			if ((uint8)(atr&H_FLIP))
			{
				if ((uint8)(J&0x80))
					C[7] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[6] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[5] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[4] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[3] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[2] = VB[pixdata&3]|PLB_B;
				if ((J+=J)&0x80)
					C[1] = VB[pixdata>>4&3]|PLB_B;
				if ((J+=J)&0x80)
					*C = VB[pixdata>>=8]|PLB_B;
			} else
			{
				if ((uint8)(J&0x80))
					*C = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[1] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[2] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[3] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[4] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[5] = VB[pixdata&3]|PLB_B;
				if ((J+=J)&0x80)
					C[6] = VB[pixdata>>4&3]|PLB_B;
				if ((J+=J)&0x80)
					C[7] = VB[pixdata>>=8]|PLB_B;
			}
	} while(--spr,--n>=0);

	PPUlineHasSp0 = 0;
	PPUOAMrefreshed = 1;
}
#endif

static void CopySprites(uint8 *target)
{
	const uint8	*S; /* sprite/OAM */
	uint8		*P; /* playfield */
	uint8		n;

	if (!PPUOAMrefreshed)
		return;
	n = ~PPU[1], n+= n, n&= 8; /* have SpFirst8 determine where to start */
	S = n+PPUsprlinebuf;
	P = n+target;

	if (PPUrendis&1) /* sprites hidden */
		goto end;
	if (PPUrendis&4) /* playfield off (pixel data could be anything) */
	{
		do
		{
			uint8		t;

			if (!*(uint32*)S)
				{S+= 4; continue;}
			if ((t=*S++)!=0)
				t|=PLB_B, *P = t; /* background priority should be set */
			if ((t=*S++)!=0)
				t|=PLB_B, P[1] = t;
			if ((t=*S++)!=0)
				t|=PLB_B, P[2] = t;
			if ((t=*S++)!=0)
				t|=PLB_B, P[3] = t;
		} while(P+=4,n+=4);
		goto end;
	}

	do
	{
		uint8		t;

		if (!*(uint32*)S)
			{S+= 4; continue;}
	/*act on each normal or behind bg sprite pixel*/
		if ((t=*S++)!=0 && (uint8)((t|*P)&PLB_B))
			*P = t;
		if ((t=*S++)!=0 && (uint8)((t|P[1])&PLB_B))
			P[1] = t;
		if ((t=*S++)!=0 && (uint8)((t|P[2])&PLB_B))
			P[2] = t;
		if ((t=*S++)!=0 && (uint8)((t|P[3])&PLB_B))
			P[3] = t;
	} while(P+=4,n+=4);
end:	PPUOAMrefreshed = 0;
}

/* line rendering */

static void DoHBlank(void)
{
	uint32		rad, t;

	if (!RenderON)
		return;
	rad = RefreshAddr;
	if (!(~rad&0x7000)) /*(rad&0x7000)==0x7000*/
		if ((t=~(rad^=0x7000)&0x3E0)==0) /*(rad&0x3E0)==0x3E0*/
			rad^= 0x3E0;
		else if (t==0x40) /*(rad&0x3E0)==0x3A0*/
			rad^= 0x3A0^0x800;
		else
			rad+= 0x20;
	else
		rad+= 0x1000;
	RefreshAddr = rad;
}

static void blankPixels(uint8 c,
	int z)
{
	uint8		*d;
	uint32		t, x = linepel-MIN(linepel,24+1);

	d = PPU_Plinef+x;
	exp8(t,c);
	if (z>256)
		z = 256;
	if ((z-=x)<0)
		return;
	if (z&1)
		*d++ = t;
	if (z&2)
		*((uint16*)d)++ = t;
	if (z&4)
		*((uint32*)d)++ = t;
	if (z&8)
		*((uint32*)d)++ = t,*((uint32*)d)++ = t;
	if (z&16)
		*((uint32*)d)++ = t, *((uint32*)d)++ = t,
			*((uint32*)d)++ = t,*((uint32*)d)++ = t;
	if (z&=-32)
		do
			*((uint32*)d)++ = t, *((uint32*)d)++ = t,
				*((uint32*)d)++ = t, *((uint32*)d)++ = t,
				*((uint32*)d)++ = t, *((uint32*)d)++ = t,
				*((uint32*)d)++ = t, *((uint32*)d)++ = t;
		while(z+=-32);
}

static void FASTAPASS(1) RefreshLine(int lastpixel)
{
	static uint32	pshift0, pshift1;
	static uint32	atlatch;

	register uint8	*P;

	uint32		smorkus = RefreshAddr;
#define RefreshAddr smorkus
	uint32		vofs;
	int		X1;
	int		pfetchtile, pfetchpel;
	int		numtiles, firsttile;
#ifndef env68k
	int		xdownshift = 8-XOffset;
#endif
	int		scantileline = scanline>>3;

	/* Yeah, recursion would be bad.
		PPU_hook() functions can call mirroring/chr bank switching functions,
		which call FCEUPPU_LineUpdate, which calls this function
	*/
	static uint8 norecurse = 0;
	if (norecurse)
		return;
	pfetchpel = lastpixel;
	if (PPUsphitx!=0x100 && !(PPU_status&0x40))
		if (PPUsphitx<lastpixel-16
				&& PPUsphitx>=(pfetchpel&~7)-16)
			pfetchpel+= 8; /*first 8 must be rendered*/
	if (pfetchpel>272)
		pfetchpel = 272;
	pfetchtile = pfetchpel>>3;
	firsttile = linepel>>3;
	if ((numtiles=pfetchtile-firsttile)<=0)
		return;
	scantileline = scanline>>3;
	P = PPU_Pline;

	if (!RenderON)
	{ /* modified for full_nes_palette.nes; 2010.04.13, 2015.08.08 */
		uint32		t = RefreshAddr-0x3F00;

		if (t>0x7F)
			t = 0;
		numtiles = numtiles*8u;
		/* prev. PPUPal (some colors would do #0; 2010.05.09) */
		blankPixels((PALRAM[t&63]&~(PLB_B|PLB_S))|PLB_B,pfetchpel);
		PPU_Pline+= numtiles;

#define TOFIXNUM (272-4)
		if (lastpixel>=TOFIXNUM && linereset)
			{DoHBlank(); linereset = 0;}

		if (InputScanlineHook && lastpixel-16>=0)
			InputScanlineHook(PPU_Plinef,
				PPUOAMrefreshed ? PPUsprlinebuf:0,
				linestartts,
				pfetchtile*8u-16);
		goto end;
	}

	vofs = (uint32)BGAdrHI<<8|(RefreshAddr>>12&7);

	if (PPUrendis&4) /* skip playfield rendering (yes, some code put here) */
	{
		uint32		me32;
		//substitute void color
		exp8(me32,PPUvoidcolor!=255 ? PPUvoidcolor:PPUPal[4]);
		for(X1=firsttile; X1<pfetchtile; ++X1)
		{ /* fake everytin' */
			if (PPU_hook)
			{
				PPU_hook(0x2000|(RefreshAddr&0xFFF));
				PPU_hook(vofs);
			}

			*((uint32*)P)++ = me32, *((uint32*)P)++ = me32;

			if (!(~RefreshAddr&HT_POS_MASK))
				RefreshAddr^= H_POS_MASKS;
			else
				++RefreshAddr;

			if (PPU_hook)
				PPU_hook(0x2000|(RefreshAddr&0xFFF));
		}
		atlatch = 0;
		goto skip;
	}

	/* This high-level graphics MMC5 emulation code was written
		for MMC5 carts in "CL" mode.  It's probably not totally
		correct for carts in "SL" mode.
	*/
#define PPUT_MMC5
	if (MMC5Ex && GenieStage!=1)
	{
		if (!MMC5ExMode && MMC5SplitCtrl&0x80)
		{
			int	tochange = (MMC5SplitCtrl&31)-firsttile;

			for(X1=firsttile; X1<pfetchtile; ++X1)
				if ((--tochange<0 && MMC5SplitCtrl&0x40)
						|| (tochange>=0 && !(MMC5SplitCtrl&0x40)))
				{
				#define PPUT_MMC5SP
				#include "PPUtile.h"
				#undef PPUT_MMC5SP
				} else
				{
				#include "PPUtile.h"
				}
		} else if (MMC5ExMode==1 && MMC5SplitCtrl&0x80)
		{
		//	int tochange = MMC5SplitCtrl&0x1F; /* unused */

		//	tochange-= firsttile;
			#define PPUT_MMC5SP
			#define PPUT_MMC5CHR1
			for(X1=firsttile; X1<pfetchtile; ++X1)
			{
			#include "PPUtile.h"
			}
			#undef PPUT_MMC5CHR1
			#undef PPUT_MMC5SP
		} else if (MMC5ExMode==1)
		{
			#define PPUT_MMC5CHR1
			for(X1=firsttile; X1<pfetchtile; ++X1)
			{
			#include "PPUtile.h"
			}
			#undef PPUT_MMC5CHR1
		} else
			for(X1=firsttile; X1<pfetchtile; ++X1)
			{
			#include "PPUtile.h"
			}
	#undef PPUT_MMC5
	} else if (PPU_hook)
	{
		norecurse = 1;
		#define PPUT_HOOK
		for(X1=firsttile; X1<pfetchtile; ++X1)
		{
		#include "PPUtile.h"
		}
		#undef PPUT_HOOK
		norecurse = 0;
	} else
		for(X1=firsttile; X1<pfetchtile; ++X1)
		{
		#include "PPUtile.h"
		}
#undef RefreshAddr
skip:	RefreshAddr = smorkus;
	if (!BGFirst8 && PPU_Pline==PPU_Plinef) /*firsttile<3 && pfetchtile>2*/
	{
		uint32	t, *d = (uint32*)PPU_Plinef;

		*d = *d++ = exp8(t,PPUPal[0]);
	}

	if (!PlayfdON)
		blankPixels(PPUPal[0],pfetchpel);

	if (lastpixel>=TOFIXNUM && linereset)
	{
		DoHBlank();
		linereset = 0;
	}

	CheckSpriteHit(lastpixel);

	if (InputScanlineHook && lastpixel-16>=0)
		InputScanlineHook(PPU_Plinef,
			PPUOAMrefreshed ? PPUsprlinebuf:0,
			linestartts,
			pfetchtile*8u-16);

	PPU_Pline = P;
end:	linepel = pfetchpel;
}

static void line32_and(uint32 *d,
	uint32 m)
{
	int		x = 8;

	do
		*d++&= m, *d++&= m, *d++&= m, *d++&= m,
		*d++&= m, *d++&= m, *d++&= m, *d++&= m;
	while(--x);
}

#if 0
static void line32_mask(uint32 *d,
	uint32 m,
	uint32 k)
{
	int		x = 8;

	do
		*d&= m, *d++|= k, *d&= m, *d++|= k, *d&= m, *d++|= k, *d&= m, *d++|= k,
		*d&= m, *d++|= k, *d&= m, *d++|= k, *d&= m, *d++|= k, *d&= m, *d++|= k;
	while(--x);
}
#endif

#define DoUpdateV()	do{ \
	if(RenderON) \
		RefreshAddr=(RefreshAddr&0xFBE0)|(TempAddr&0x41F); \
}while(0)

#define DoCYC64_192()	do{ \
/*0*/	PPUEmphasis = PPU[1]<<8-5&0x700; \
/*64*/	OAMLatch=0xFF; X6502_Run(1); \
/*65*/	OAMLatch=SPRAM[0]; X6502_Run(191); \
/*256*/	OAMLatch=0xFF; \
}while(0)

#define DoCYC320_5()	do{ \
/*320*/	OAMLatch=SPRAM[0]; X6502_Run(5); \
}while(0)

static void DoLine(void)
{
	uint8		*target = XBuf+((unsigned)scanline*FCEUSCRN_ROWSIZE);

	if (MMC5Ex && RenderON)
	{ /* fixes alignment - does it break anything?; 2015.08.22 */
	//	X6502_Run(10);
		MMC5_hb(scanline);
	//	X6502_Run(64-10);
	}// else
		X6502_Run(64);
	DoCYC64_192();

	/* EndRL */
	RefreshLine(272);
	if (linereset)
		DoHBlank();
	CheckSpriteHit(272); /* finish */
	PPU_Pline = 0;

	if (PPUrendis&2) /* User asked to not display background data. */
	{
		uint32		t = ((PPUvoidcolor!=255 ? PPUvoidcolor:PPUPal[0])*0x101u)*0x10001u;

		{
			uint32		*d = (uint32*)target, c = 256/32;

			do
				*d++ = *d++ = *d++ = *d++ =
				*d++ = *d++ = *d++ = *d++ = t;
			while(--c);
		}
	}

	if (SpriteON)
		CopySprites(target);

	if (PPUrendis&4)
	{
		/* SPZ hit'll fail without something; all have bit #7 on; 2010.04.26 */
		if (SpriteON && scanline==SPRAM[0]) /* only if SpriteON; 2010.05.27 */
			PPU_status|= 0x40;
	} else if (RenderON && GrayLine) /* Yes, very el-cheapo. */
		line32_and((uint32*)target,0x30303030);
//256
	PPUsphitx = 0x100;

	if (RenderON)
		FetchSpriteData();

	if (GameHBIRQHook && RenderON && (PPU[0]&0x38)!=0x18)
	{
		X6502_Run(6);
		DoUpdateV(); //262
		X6502_Run(4);

		GameHBIRQHook();

		X6502_Run(85-16-10-5); /* 54 */
		DoCYC320_5();
	} else
	{
		X6502_Run(6);
		DoUpdateV(); //262
		X6502_Run(85-6-16-5); /* 58 */
		DoCYC320_5();

		// A semi-hack for Star Trek: 25th Anniversary
		if (GameHBIRQHook && RenderON && (PPU[0]&0x38)!=0x18)
			GameHBIRQHook();
	}
//325
	if (SpriteON)
		RefreshSprites();
	if (GameHBIRQHook2 && RenderON)
		GameHBIRQHook2();
	/* remap the line after all the hooks, given emphasis */
	_memrem4(target,256,epal_inv+PPUEmphasis);
}

void FCEUPPU_SetVideoSystem(int which)
{
	if (which==1)
	{ /* PAL */
		PPUVBL_SLcc = FSettings.UsrVBLSLines[1]*LCYCS; /*312-242*/
		FSettings.FirstSLine = FSettings.UsrFirstSLine[1];
		FSettings.LastSLine = FSettings.UsrLastSLine[1];
	} else if (!which)
	{ /* NTSC */
		PPUVBL_SLcc = FSettings.UsrVBLSLines[0]*LCYCS; /*262-242*/
		FSettings.FirstSLine = FSettings.UsrFirstSLine[0];
		FSettings.LastSLine = FSettings.UsrLastSLine[0];
	}

	if ((PPUVBL_SLcc-=NMI_DELAY)<0)
		PPUVBL_SLcc = 0; /* idiot proofing */
}

void FCEUPPU_Init(void)
{
	makeluts(bitreverseLUT
#ifndef env68k
		,PPU_LUT1to4
		,PPU_TileLUT
#endif
		);
}

void FCEUPPU_Reset(void)
{
	XBuf = FCEU_video; /* odd field when frame begins */
	VRAMBuffer = PPU[0] = PPU[1] = PPU_status = PPU[3] = PPUSPL =
		PPUGenLatch = vtoggle = 0;
	RefreshAddr = TempAddr = PPUVFrameCounter = 0;
	PPUBoot = 2;
}

/* from simple memset of zero...; 2010.04.14, 2010.05.26 */
static void InitializePalette(uint8 *d)
{ /* actual does 0x09 (yellow-green) for 1st */
	d[2] = d[4] = d[12] = d[13] = d[20] = d[22] = d[26] = d[28] = 0;
	d[1] = d[3] = d[17] = 1, d[5] = d[6] = d[27] = 2, d[7] = 13;
	d[8] = d[10] = d[24] = d[31] = 8, d[9] = 16, d[11] = 36,
		d[14] = d[21] = 4, d[15] = d[30] = 44; /* #30 did 52 */
	*d = 63, d[16] = 9, d[18] = 52, d[19] = 3, d[23] = 20;
	d[25] = 58, d[29] = 32;
}

void FCEUPPU_UpdatePalette()
{
	int		i;

	/* Playfield: set transparent bit; index #0 used rendering transparent */
	PPUPal[0] = PPUPal[4] = PPUPal[8] = PPUPal[12] = PALRAM[0]&(~PLB_S)|PLB_B;
	i = 0;
	do
		PPUPal[i+1] = PALRAM[i+1]&~(PLB_B|PLB_S),
		PPUPal[i+2] = PALRAM[i+2]&~(PLB_B|PLB_S),
		PPUPal[i+3] = PALRAM[i+3]&~(PLB_B|PLB_S);
	while((i+=4)<16);
	do /* OAM: make sure it isn't zero */
		PPUPal[i] = PALRAM[i]&(~PLB_B)|PLB_S;
	while(++i<32);
}

void FCEUPPU_Power(void)
{
	readfunc	*rdrt, r0x, r02, r04, r07;
	writefunc	*wrrt, w[8];
	int		x;

	InitializePalette(PALRAM);
	FCEUPPU_UpdatePalette();

	dwmemset(NTARAM,0,0x800);
	dwmemset(SPRAM,~0,0x100); //data was 0
	FCEUPPU_Reset();

	if (FCEUGameInfo->type==GIT_NSF)
		return;
	OptFAddr(r0x,PPUA200x);
	OptFAddr(r02,PPUA2002);
	OptFAddr(r04,PPUA2004);
	OptFAddr(r07,PPUA2007);

	OptFAddr(w[0],PPUB2000);
	OptFAddr(w[1],PPUB2001);
	OptFAddr(w[2],PPUB2002);
	OptFAddr(w[3],PPUB2003);
	OptFAddr(w[4],PPUB2004);
	OptFAddr(w[5],PPUB2005);
	OptFAddr(w[6],PPUB2006);
	OptFAddr(w[7],PPUB2007);

	x = 0x2000, rdrt = x+ARead, wrrt = x+BWrite;
	do
		*rdrt++ = r0x,
		*rdrt++ = r0x,
		*rdrt++ = r02,
		*rdrt++ = r0x,
		*rdrt++ = r04,
		*rdrt++ = r0x,
		*rdrt++ = r0x,
		*rdrt++ = r07,

		*wrrt++ = w[0],
		*wrrt++ = w[1],
		*wrrt++ = w[2],
		*wrrt++ = w[3],
		*wrrt++ = w[4],
		*wrrt++ = w[5],
		*wrrt++ = w[6],
		*wrrt++ = w[7];
	while((x+=8)<0x4000);
	BWrite[0x4014] = PPUB4014;
}

static void ResetRL(uint8 *target)
{
//	FCEU_memsetl(target,-1,256); -- no longer necessary at all; 2011.02.14
	if (InputScanlineHook)
		InputScanlineHook(0,0,0,0);
	PPU_Plinef = target;
	PPU_Pline = target;
	linepel = 0;
	linestartts = timestamp*48+XFrame.count;
	linereset = 0;
	LineUpdate();
	linereset = 1;
}

/* PPU-oriented frame */

int FCEUPPU_Loop(int skip)
{
	if (PPUBoot) //Needed for Knight Rider, possibly others.
	{
		dwmemset(FCEU_video,(PPUvoidcolor*0x101u)*0x10001u,FCEUSCRN_ROWSIZE*240);
		if (--PPUBoot)
		{
			PPU_status|= 0x80;
			scanline = 241;
			X6502_Run(PPUVBL_SLcc+NMI_DELAY);
		} else
		{
			PPU_status = 0;
			scanline = 0;
			X6502_Run(242*LCYCS);
		}
		goto rendered;
	}
//340
	XBuf = (uint8*)(256^(long int)XBuf); /* swap fields; 2015.08.03 */

	/* this code passes Shay Green's VBL/NMI tests (yay) */

	/* a number of the cycle counts here have an edge of one cycle
	   because the emulated CPU only executes whole cycles (no pause/resume in microcode)
	*/
	PPU_status|= 8;
	X6502_Run(1);
	PPU_status = (PPU_status&0x77)|(PPU_status<<4&0x80); /* vertical interrupt phase */
	if (FCEUGameInfo->type==GIT_NSF)
	{
		X6502_Run(NMI_DELAY);
		DoNSFFrame();
	} else
	{
		scanline = 241; /* necessary for 2004 reads; 2009.12.29, 2010.05.29 */

		/* reset sprites */
		PPU[3] = PPUSPL = 0; //Not having this here breaks a Super Donkey Kong game.
		X6502_Run(1+1);
		/* according to the documentation, VBL line is a NAND gate of 2000.7 and 2002.7; */
		/* but the Test Cartridge runs at unpredictable speeds with that... */
		/* ** fixed by adjusting the CPU cycle count spent on an NMI */
		PPU_status|= (VBlankON&PPU_status)>>5;
		X6502_Run(NMI_DELAY-1-1);
		if (PPU_status&4)
			XFrame.intlow|= FCEU_IQNMI; /*TriggerNMI*/
	}
	X6502_Run(PPUVBL_SLcc);

	PPU_status = 0; /* out of VINT; on to no-render evaluation line */
//1
	X6502_Run(64-1+1);
	DoCYC64_192();

	if (RenderON)
	{
		if (GameHBIRQHook && (PPU[0]&0x38)!=0x18)
			GameHBIRQHook();
		if (PPU_hook)
		{
			int	x = 42;

			do
			{
				PPU_hook(0x2000);
				PPU_hook(0);
			} while(--x);
		}
		if (GameHBIRQHook2)
			GameHBIRQHook2();
	}
//256
	X6502_Run(85-16-5); /* 64 CYC */
//320 - the hook should be here
	DoCYC320_5();
//325
	if (RenderON)
	{
		RefreshAddr = TempAddr;
		if (PPU_hook)
			PPU_hook(RefreshAddr&0x3FFF);
	}

	//Clean this stuff up later.
	PPUOAMrefreshed = PPUnumsprites = 0;
	ResetRL(XBuf);

	/* frame toggle assessment is done at CYC 338; 2010.04.15 */
	X6502_Run(13);
//338
	if (!(PPUVFrameCounter&1))
	{
		skip&=~255; /* newer implementation...; 2013.05.15 */
		X6502_Run(3);
	} else /* correct way: one less if NTSC (2C02), odd and rendering */
		X6502_Run(!PAL && RenderON ? 2:3);
	if (FCEUGameInfo->type==GIT_NSF)
		X6502_Run(NSFCycles);
	else if (skip)
	{
		int	y = 1+SPRAM[0];

		PPU_status|= 0x20; //set overflow- Fixes "Bee 52".  Does it break anything?
		scanline = 0, PPU_Pline = 0;
		if (GameHBIRQHook)
		{
			X6502_Run(256);
			do
			{
				if (RenderON)
					GameHBIRQHook();
				if (scanline==y && SpriteON)
					PPU_status|= 0x40;
				X6502_Run(LCYCS);
			} while(++scanline<239);
			if (RenderON)
				GameHBIRQHook();
			if (scanline==y && SpriteON)
				PPU_status|= 0x40;
			X6502_Run(85);
		} else
		{
			if (y>239)
				y = 240;
			else
			{
				for(; scanline<y; ++scanline)
					X6502_Run(LCYCS);
				if (SpriteON)
					PPU_status|= 0x40;
				y =-y+240;
			}
			do
				X6502_Run(LCYCS);
			while(++scanline,--y);
		}
	} else
	{
		scanline = 0;
		do
		{
		#if 0
			++deempcnt[PPUemp];
		#endif
			DoLine();
			ResetRL(XBuf+((unsigned)++scanline*FCEUSCRN_ROWSIZE));
			X6502_Run(16);
		} while(scanline<239);
	#if 0
		++deempcnt[PPUemp];
	#endif
		DoLine();
		X6502_Run(16);
		scanline = 240;

		if (MMC5Ex && RenderON)
			MMC5_hb(scanline);
		#if 0
		{
			int	*deempcntp = 1+deempcnt;
			int	x = 1, max = 0, maxref = 0;

			do
				if (*deempcntp>max)
					max = *deempcntp, maxref = x;
			while(*deempcntp++=0,++x<7);
			SetNESDeemph(maxref,0);
		}
		#endif
	}
	//in 0.98.13, this was called at beginning
	X6502_Run(LCYCS-1);
rendered:
	FCEU_PutImage(PPUVFrameCounter++,skip);

	return skip;
}

static uint16	TempAddrT, RefreshAddrT;

void FCEUPPU_LoadState(int v)
{
	if (v<9817)
		PALRAM[4] = PALRAM[8] = PALRAM[12] = PALRAM[0];
	else
		XBuf = FCEU_video+((PPUVFrameCounter&1)*256); /* 2015.08.04 */
	FCEUPPU_UpdatePalette();
	vtoggle&= 1;
	TempAddr = TempAddrT;
	RefreshAddr = RefreshAddrT;
}

SFORMAT FCEUPPU_STATEINFO[] =
{
	{NTARAM, 0x800, "NTAR"},
	{PALRAM, 0x20, "PRAM"},
	{SPRAM, 0x100, "SPRA"},
	{PPU, 4, "PPUR"},
	{&PPUBoot, 1, "DEAD"},
	{&PPUSPL, 1, "PSPL"},
	{BIG_ENDIAN*3+&XOffset, 1, "XOFF"},
	{&vtoggle, 1, "VTOG"},
	{&RefreshAddrT, 2|FCEUSTATE_RLSB, "RADD"},
	{&TempAddrT, 2|FCEUSTATE_RLSB, "TADD"},
	{&VRAMBuffer, 1, "VBUF"},
	{&PPUGenLatch, 1, "PGEN"},
	{&PPUVFrameCounter, 4|FCEUSTATE_RLSB, "VFMC"}, /* sound.c uses "FCNT" */
	{0}
};

void FCEUPPU_SaveState(void)
{
	TempAddrT = TempAddr;
	RefreshAddrT = RefreshAddr;
}
