/*
 driver.h

  MacFCEU 0.9

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __FCEU_DRIVER_H

#ifdef  __cplusplus
extern "C" {
#endif

#include "types.h"
#include "git.h"
#include "debug.h"

struct FILE *FCEUD_UTF8fopen(const char *fn,
	const char *mode);

uint32 FCEUI_CRC32(uint32 crc,
	const void *buf,
	uint32 len);

/* This makes me feel dirty for some reason. */
void FCEU_printf(const char *format, ...);
#define FCEUI_printf FCEU_printf

/* Displays an error.  Can block or not. */
void FCEUD_PrintError(const char *s);
void FCEUD_Message(const char *s);

uint64 FCEUD_GetTime(void); /* system time elapsed */
uint32 FCEUD_GetTimeFreq(void); /* one second of system time -- was uint64 */

#ifdef NETWORK
/* Network interface */

/* Call only when a game is loaded. */
int FCEUI_NetplayStart(int nlocal,
	int divisor);

/* Call when network play needs to stop. */
void FCEUI_NetplayStop(void);

/* Note:  YOU MUST NOT CALL ANY FCEUI_* FUNCTIONS WHILE IN FCEUD_SendData() or
   FCEUD_RecvData().
*/

/* Display text received over the network. */
void FCEUD_NetplayText(uint8 *text);

/* Encode and send text over the network. */
void FCEUI_NetplayText(uint8 *text);
#endif

/* three prototypes moved here for recognition (not called/empty if no network) */
/* Return 0 on failure, 1 on success. */
int FCEUD_SendData(void *data,
	uint32 len);
int FCEUD_RecvData(void *data,
	uint32 len);

/* Called when a fatal error occurred and network play can't continue.  This function
   should call FCEUI_NetplayStop() after it has deinitialized the network on the driver
   side.
*/
void FCEUD_NetworkClose(void);
int FCEUD_NetworkConnect(void);

int FCEUI_BeginWaveRecord(char *fn,
	int format);
int FCEUI_EndWaveRecord(void);

 /* input */

void FCEUI_SetInput(int port,
	int type,
	void *ptr,
	int attr);
void FCEUI_SetInputFC(int type,
	void *ptr,
	int attr);
void FCEUI_DisableFourScore(int s);

#define SI_NONE    		0
#define SI_GAMEPAD 		1
#define SI_ZAPPER  		2
#define SI_POWERPADA		3
#define SI_POWERPADB		4
#define SI_ARKANOID 		5

#define SIFC_NONE		0
#define SIFC_ARKANOID		1
#define SIFC_SHADOW		2
#define SIFC_4PLAYER		3
#define SIFC_FKB		4
#define SIFC_HYPERSHOT		5
#define SIFC_MAHJONG		6
#define SIFC_QUIZKING		7
#define SIFC_FTRAINERA		8
#define SIFC_FTRAINERB		9
#define SIFC_OEKAKIDS		10
#define SIFC_BWORLD		11
#define SIFC_TOPRIDER		12

#define SIS_NONE		0
#define SIS_DETACH		1	/* was SIS_DATACH */
#define SIS_NWC			2
#define SIS_VSUNISYSTEM		3
#define SIS_NSF			4

int FCEUI_DatachSet(const uint8 *rcode);

/* New interface functions */

/* name=path and file to load.  returns 0 on failure, 1 on success */
FCEUGI *FCEUI_LoadGame(const char *name);

/* allocates memory.  0 on failure, 1 on success. */
int FCEUI_Initialize(void);

/* Emulates a frame.  0 on skip. */
int FCEUI_Emulate(uint8 **pbuf,
	int16 **snd,
	int32 *cnt,
	int skip); /* only skips odd frames; 2010.04.26 */

/* Closes currently loaded game */
void FCEUI_CloseGame(void);

/* Deallocates all allocated memory.  Call after FCEUI_Emulate() returns. */
void FCEUI_Kill(void);

/* Enable/Disable game genie. a=0 disable, a=1 enable */
void FCEUI_SetGameGenie(int a);

#define FCEUIOD_STATE		0
#define FCEUIOD_SNAPS		1
#define FCEUIOD_NV		2
#define FCEUIOD_CHEATS		3
#define FCEUIOD_MISC		4
#define FCEUIOD_MOVIE		5

#define FCEUIOD__COUNT		6

void FCEUI_SetDirOverride(int which,
	char *n);

/* Sets the base directory(save states, snapshots, etc. are saved in directories
	after this directory. */
void FCEUI_SetBaseDirectory(char *dir);

 /* video */

void FCEUD_SetPalette(uint8 ix,
	uint8 r,
	uint8 g,
	uint8 b);
void FCEUD_GetPalette(uint8 ix,
	uint8 *r,
	uint8 *g,
	uint8 *b);

/* Tells FCE Ultra to copy the palette data pointed to by pal and use it.
   Data pointed to by pal needs to be 64*3 bytes in length.
*/
void FCEUI_SetPaletteArray(const uint8 *pal);

void FCEUI_NTSCSelect(uint8 w); /* tint = 0, hue, brightness */
void FCEUI_NTSCDEC(void);
void FCEUI_NTSCINC(void);

/* newer 'saturation' version takes an extra bit (for precision accuracy) */

/* returns saturation */
int FCEUI_GetNTSC_HSB(int *hue,
	int *brt); /* brightness; 2010.05.09 */
void FCEUI_SetNTSC_HSB(int n,
	int sat,
	int hue,
	int brt);
#define FCEUI_SetNTSCTHB(n,t,h,b) FCEUI_SetNTSC_HSB(n,h,(int)(t)*2u,b)
#define FCEUI_GetNTSCTHB(t,h,b) do{int *_t=t; *t=FCEUI_GetNTSC_HSB(h,b)>>1;}while(0)

/* 0 to keep 8-sprites limitation, 1 to remove it */
void FCEUI_DisableSpriteLimitation(int a);

/* -1 = no change, 0 = show, 1 = hide, 2 = internal toggle */
void FCEUI_SetRenderDisable(int sprites,
	int bg);

/* Set video system a=0 NTSC, a=1 PAL */
void FCEUI_SetVidSystem(int a);

/* Convenience function; returns currently emulated video system (0=NTSC, 1=PAL).  */
int FCEUI_GetCurrentVidSystem(int *slstart,
	int *slend,
	int *slvbl);

/* First and last scanlines to render, for ntsc and pal emulation. */
void FCEUI_SetRenderedLines(int ntscf,
	int ntscl,
	int palf,
	int pall);
void FCEUI_SetVBLScanlines(int ntscslv,
	int palslv);

void FCEUI_ToggleTileView(void);

 /* sound */

/* Sets up sound code to render sound at the specified rate, in samples per second. */
/* returns the clipped rate */
int FCEUI_Sound(int Rate); /* [0 (off),111861] */
int32 FCEUI_Sound2(int32 Rate); /* 4-bit fraction version [0 (off),111860.75] */

void FCEUI_SetSoundVolume(int32 volume); /*can now be in [-65536,65535]; 2011.02.08*/
void FCEUI_SetSoundActiveChannels(uint16 activebits);
void FCEUI_SetRectFilterAmp(int16 amplitude); /* norm:256 */
void FCEUI_SetTNDFilterAmp(int16 amplitude); /* norm:256 */
void FCEUI_SetNoiseScale(uint16 scale); /* norm:2 */
void FCEUI_SetTriangleScale(uint16 scale); /* norm:3 */
void FCEUI_SetDCFilterLoss(uint16 loss);

#if HAVE_HQ
void FCEUI_SetSoundQuality(int quality);
void FCEUI_SetLowPass(int q);
#endif

uint32 FCEUI_SetSpeed(uint32); /* in per-billion */

void FCEUI_SetVRC6Volume(int16);
void FCEUI_SetVRC7Volume(int16);
void FCEUI_SetSFMEVolume(int16);

 /* file stored state frames */

/* "fname" overrides the default save state filename code if non-NULL. */
void FCEUI_SaveState(const char *fname);
void FCEUI_LoadState(const char *fname);

 /* file stored input movies */

void FCEUI_SaveMovie(const char *fname);
void FCEUI_LoadMovie(const char *fname);

 /* visual interface */

void FCEUI_SelectState(int);
void FCEUI_SelectMovie(int);

#define FCEUI_DispMessage FCEU_DispMessage

uint32 FCEUI_GetDesiredFPS(void);
void FCEUI_SaveSnapshot(void);

/* 0 to order screen snapshots numerically (0.png),
  1 to order them file base-numerically (smb3-0.png). */
void FCEUI_SetSnapName(int a);

 /* cheat decoding */

int FCEUI_DecodePAR(const char *code,
	uint16 *a,
	uint8 *v,
	int *c,
	int *type);
int FCEUI_DecodeGG(const char *str,
	uint16 *a,
	uint8 *v,
	int *c);

 /* cheats */

int FCEUI_AddCheat(const char *name,
	uint32 addr,
	uint8 val,
	int compare,
	int type);
int FCEUI_DelCheat(uint32 which);
int FCEUI_ToggleCheat(uint32 which);

int32 FCEUI_CheatSearchGetCount(void);
void FCEUI_CheatSearchGetRange(uint32 first,
	uint32 last,
	int (*callb)(uint32 a, uint8 last, uint8 current));
void FCEUI_CheatSearchGet(int (*callb)(uint32 a, uint8 last, uint8 current,
		void *data),
	void *data);
void FCEUI_CheatSearchBegin(void);
void FCEUI_CheatSearchEnd(int type,
	uint8 v1,
	uint8 v2);
void FCEUI_ListCheats(int (*callb)(char *name, uint32 a, uint8 v, int compare,
		int s, int type, void *data),
	void *data);

int FCEUI_GetCheat(uint32 which,
	char **name,
	uint32 *a,
	uint8 *v,
	int *compare,
	int *s,
	int *type);
int FCEUI_SetCheat(uint32 which,
	const char *name,
	int32 a,
	int32 v,
	int compare,
	int s,
	int type);

void FCEUI_CheatSearchShowExcluded(void);
void FCEUI_CheatSearchSetCurrentAsOriginal(void);

/* interrupts */

void FCEUI_ResetNES(void);
void FCEUI_PowerNES(void);

 /* other environment functions */

void FCEUI_NSFSetVis(int mode);
int FCEUI_NSFChange(int amount);
int FCEUI_NSFGetInfo(uint8 *name,
	uint8 *artist,
	uint8 *copyright,
	int maxlen);

void FCEUI_VSUniToggleDIPView(void);
void FCEUI_VSUniToggleDIP(int w);
uint8 FCEUI_VSUniGetDIPs(void);
void FCEUI_VSUniSetDIP(int w,
	int state);
void FCEUI_VSUniCoin(void);

int FCEUI_FDSInsert(int oride);
int FCEUI_FDSEject(void);
void FCEUI_FDSSelect(void);

#ifdef  __cplusplus
}
#endif

#define __FCEU_DRIVER_H
#endif
