/*
 file.c

 * MacFCEU 0.8
 *
 * Copyright notice for this file:
 *  file.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef mac
 #include <fcntl.h>

 #define dup(f)		fcntl(f,F_DUPFD,0) /* recommended */
#else
 #include <unistd.h>
#endif

#include <zlib.h>
#include "unzip.h"

#include "driver.h"
#include "file.h"
#include "utility/endian.h"
#include "memory.h"
#include "general.h"

#ifdef env68k
 #pragma segment files
#endif

typedef struct/*(12+)*/
{
	uint8		*data;
	uint32		size;
	uint32		location;
} MEMWRAP;

void ApplyIPS(FILE *ips,
	MEMWRAP *dest);
void ApplyIPS(FILE *ips,
	MEMWRAP *dest)
{
	uint8		header[5];
	uint32		count = 0;

	FCEU_printf(" Applying IPS...\n");
	if (fread(header,1,5,ips)!=5)
	{
		fclose(ips);
		return;
	}
	if (memcmp(header,"PATCH",5))
	{
		fclose(ips);
		return;
	}

	while(fread(header,1,3,ips)==3)
	{
		uint32	offset = (header[0]<<16)|(header[1]<<8)|header[2];
		uint16	size;

		if (!memcmp(header,"EOF",3))
		{
			FCEU_printf(" IPS EOF:  Did %d patches\n\n",
				count);
			fclose(ips);
			return;
		}

		size = fgetc(ips)<<8;
		if ((size|=fgetc(ips))==0)
		{ /* RLE */
			uint8	*start;
			uint8	b;

			size = fgetc(ips)<<8;
			size|= fgetc(ips);

//FCEU_printf("  Offset: %8d  Size: %5d RLE\n",offset,size);

			if ((offset+size)>dest->size)
			{
				uint8	*tmp;

				/* Probably a little slow. */
				if ((tmp=(uint8*)realloc(dest->data,offset+size))==0)
				{
					FCEU_printf(
"  Oops.  IPS patch %d(type RLE) goes beyond end of file.\n"
"Could not allocate memory.\n",
						count);
					fclose(ips);
					return;
				}

				dest->size = offset+size;
				dest->data = tmp;
				memset(dest->data+dest->size,0,offset+size-dest->size);
			}

			b = fgetc(ips);
			start = dest->data+offset;
			do
				*start++ = b;
			while(--size!=0);
		} else		/* Normal patch */
		{
			//FCEU_printf("  Offset: %8d  Size: %5d\n",offset,size);
			if ((offset+size)>dest->size)
			{
				uint8	*tmp;

			/* Probably a little slow. */
				if ((tmp=(uint8*)realloc(dest->data,offset+size))==0)
				{
					FCEU_printf(
"  Oops.  IPS patch %d(type normal) goes beyond end of file.\n"
"Could not allocate memory.\n",
						count);
					fclose(ips);
					return;
				}
				dest->data = tmp;
				memset(dest->data+dest->size,0,offset+size-dest->size);
			}
			fread(dest->data+offset,1,size,ips);
		}
		++count;
	}

	fclose(ips);
	FCEU_printf(" Hard IPS end!\n");
}

static MEMWRAP *MakeMemWrap(void *tz,
	int type)
{
	MEMWRAP		*tmp;

	if ((tmp=(MEMWRAP*)FCEU_malloc(sizeof(MEMWRAP)))==0)
		goto doret;
	tmp->location = 0;

	if (type==0)
	{
		fseek((FILE*)tz,0,SEEK_END);
		tmp->size = ftell((FILE*)tz);
		fseek((FILE*)tz,0,SEEK_SET);
		if ((tmp->data=(uint8*)FCEU_malloc(tmp->size))==0)
		{
			free(tmp);
			tmp = 0;
			goto doret;
		}
		fread(tmp->data,1,tmp->size,(FILE*)tz);
	} else if (type==1)
	{
	/* Bleck.  The gzip file format has the size of the uncompressed data,
	but I can't get to the info with the zlib interface(?). */
		for(tmp->size=0; gzgetc(tz)!=EOF; ++tmp->size)
			;
		gzseek(tz,0,SEEK_SET);
		if ((tmp->data=(uint8*)FCEU_malloc(tmp->size))==0)
		{
			free(tmp);
			tmp = 0;
			goto doret;
		}
		gzread(tz,tmp->data,tmp->size);
	} else if (type==2)
	{
		unz_file_info	ufo;

		unzGetCurrentFileInfo(tz,&ufo,0,0,0,0,0,0);

		tmp->size = ufo.uncompressed_size;
		if ((tmp->data=(uint8*)FCEU_malloc(ufo.uncompressed_size))==0)
		{
			free(tmp);
			tmp = 0;
			goto doret;
		}
		unzReadCurrentFile(tz,tmp->data,ufo.uncompressed_size);
	}

doret:	if (!type)
		fclose((FILE*)tz);
	else if (type==1)
		gzclose(tz);
	else if (type==2)
	{
		unzCloseCurrentFile(tz);
		unzClose(tz);
	}

	return tmp;
}

#ifndef __GNUC__
 /* strcasecmp may not be defined */
 #ifdef mac
  #include "utility/interpreter.h"
 #else
  #define strcasecmp	strcmp
 #endif
#endif

const char	*validExtList =
	".nes.nez.ines" /* iNES */
	".nsf.nsfe" /* NSF, NSFe */
	".unf.unif" /* UNIF */
	".fds" /* FDS */
	".cart.rom"; /* other */

static int getArchiveFile(unzFile tz,
	const char *onlyext,
	char *outname)
{
	char		tn[512];
	char		*nl = 0;
	long		lsz = !outname ? -1:1;
	int		*il = 0, vc = 0, vx = 0, ix = 0, err = 0;

//	printf("getArchiveFile: onlyext '%s'\n",
//		onlyext ? onlyext:"<none>");
	/* first scan the list */
	/* the zip library can't give us the no. of "valid" files */
	for(err=0; err==UNZ_OK; ++ix,err=unzGoToNextFile(tz))
	{
	/* Longer filenames might be possible, but I don't
		think people would name files that long in zip files...
	*/
		char		*tx;
		long		nz;

		if ((err=unzGetCurrentFileInfo(tz,0,tn,512,0,0,0,0))==UNZ_END_OF_LIST_OF_FILE)
			goto end;
		tn[511] = 0;
		/* correct way of getting the extension - prev. used strlen */
		if ((tx=strrchr(tn,'.'))==0)
			continue; /* no extension */
		nz = strlen(++tx);
		if (onlyext)
		{
			if (!strcasecmp(tx,onlyext))
				goto valid;
		} else 
		{
			const char	*cx, *nx;
			long		lz;

			for(cx=validExtList; cx; cx=nx)
			{
				if ((nx=strchr(++cx,'.'))==0)
					lz = strlen(cx);
				else
					lz = nx-cx;
				if (lz==nz && !strncasecmp(tx,cx,lz))
					goto valid;
			}
		}
		continue;
	valid:	++vc, vx = ix;
		/* append to name list */
		if (lsz>=0) /* if there weren't any errors building */
		{
			long		z = strlen(tn);
			int		*p;
			uint8		c;

			if ((tx=realloc(nl,lsz+=z+1))==0 || (p=realloc(il,sizeof(int)*vc))==0)
				{free(nl); free(il); nl = 0, il = 0, lsz =-1;} 
			else
			{
				nl = tx, il = p;
				il[vc-1] = ix;
				tx+= lsz, *--tx = 0, *--tx = '\n';
				while(--z>=0) /* upper and lower chars replaced */
					*--tx = (uint8)((c=tn[z])-' ')>(uint8)(126-' ') ? c=='\t' ? ' ':'_':c;
			}
		}
	}
	if ((err=unzGoToFirstFile(tz))!=UNZ_OK)
		goto end;
//	printf(
//"valid count: %d\n"
//"first name: %.60s, %d\n",
//		vc,nmls ? nmls->name:"<none>",nmls ? nmls->index:-1);
	if (vc>1 && nl) /* multiple candidates - open dialog */
	{
		if ((unsigned)(ix=FCEUI_SelectName(nl,vc))<vc)
			ix = il[ix];
	} else
		ix = vx;

//	printf("selected: %d\n",
//		ix);
	/* lastly, scan to the selected file */
	if (ix<0)
		{err = -128; goto end;} /* cancel */
	if (ix)
		do
			if ((err=unzGoToNextFile(tz))!=UNZ_OK)
				goto end; /* shouldn't err this time */
		while(--ix);

	if (!outname)
		outname = tn;
	unzGetCurrentFileInfo(tz,0,outname,256,0,0,0,0);
	outname[255] = 0;

end:	if (nl)
		free(nl);
	if (il)
		free(il);
	return err;
}

FCEUFILE *FCEU_fopen(const char *path,
	const char *ipsfn,
	const char *mode,
	const char *onlyext,
	char *outname,
	int *err)
{
	unzFile		tz;
	FILE		*ipsfile = 0;
	FCEUFILE	*fceufp = 0;
	void		*t;
	int		dummy;

	if (!err)
		err = &dummy;
	if (strchr(mode,'r') && ipsfn)
		ipsfile = FCEUD_UTF8fopen(ipsfn,"rb");

	if ((fceufp=(FCEUFILE*)malloc(sizeof(FCEUFILE)))==0)
		goto memory;

	/* If it's not a zip file, use regular file handlers.
		Assuming file type by extension usually works,
		but I don't like it. :)
	*/
	if ((tz=unzOpen(path))!=0)
	{
		fceufp->type = 2;
		if ((*err=unzGoToFirstFile(tz))!=UNZ_OK
				|| (*err=getArchiveFile(tz,onlyext,outname))!=UNZ_OK
				|| (*err=unzOpenCurrentFile(tz))!=UNZ_OK)
		{
			unzClose(tz);
			goto fail;
		}

		if ((fceufp->fp=MakeMemWrap(tz,2))==0)
			goto memory;

		if (ipsfile)
			ApplyIPS(ipsfile,(MEMWRAP*)fceufp->fp);
		return fceufp;
	}

	if ((t=FCEUD_UTF8fopen(path,"rb"))!=0)
	{
		uint32		magic;

		magic = fgetc((FILE*)t);
		magic|= fgetc((FILE*)t)<<8;
		magic|= fgetc((FILE*)t)<<16;

		if (magic!=0x088B1F)   /* Not gzip... */
			fclose((FILE*)t);
		else                  /* Probably gzip */
		{
			int	fd = dup(fileno((FILE*)t));

			fclose((FILE*)t);
			lseek(fd,0,SEEK_SET);

			if ((t=gzdopen(fd,mode))!=0)
			{
				fceufp->type = 1;
				fceufp->fp = t;
				if (ipsfile)
				{
					fceufp->fp = MakeMemWrap(t,1);
					gzclose(t);

					if (!fceufp->fp) /* bug fix; 2012.03.30 */
						goto fail;
					fceufp->type = 3;
					ApplyIPS(ipsfile,(MEMWRAP*)fceufp->fp);
				}
				return fceufp;
			}
			close(fd); //failed
		}
	}

	if ((t=FCEUD_UTF8fopen(path,mode))!=0)
	{
		fseek(t,0,SEEK_SET);
		fceufp->type = 0;
		fceufp->fp = t;
		if (ipsfile)
		{
			if ((fceufp->fp=MakeMemWrap(t,0))==0)
				{memory: *err =-2; goto fail;}
			fceufp->type = 3;
			ApplyIPS(ipsfile,(MEMWRAP*)fceufp->fp);
		}
		return fceufp;
	}

fail:	if (fceufp)
		free(fceufp);
	return 0;
}

int FCEU_fclose(FCEUFILE *fp)
{
	if (fp->type==1)
		gzclose(fp->fp);
	else if (fp->type>=2)
	{
		free(((MEMWRAP*)(fp->fp))->data);
		free(fp->fp);
	} else
		fclose((FILE*)fp->fp);

	free(fp);
	return 1;
}

long int FCEU_fread(void *ptr,
	unsigned long size,
	unsigned long nmemb, FCEUFILE *fp)
{
	if (fp->type==1)
		return gzread(fp->fp,ptr,size*nmemb);
	else if (fp->type>=2)
	{
		MEMWRAP		*wz = (MEMWRAP*)fp->fp;
		uint32		total = size*nmemb;

		if (wz->location>=wz->size)
			return 0;

		if ((wz->location+total)>wz->size)
		{
			int	ak = wz->size-wz->location;

			memcpy((uint8*)ptr,wz->data+wz->location,ak);
			wz->location = wz->size;
			return ak/size;
		} else
		{
			memcpy((uint8*)ptr,wz->data+wz->location,total);
			wz->location+= total;
			return nmemb;
		}
	} else
		return fread(ptr,size,nmemb,(FILE*)fp->fp);
}

long int FCEU_fwrite(void *ptr,
	unsigned long size,
	unsigned long nmemb,
	FCEUFILE *fp)
{
	if (fp->type==1)
		return gzwrite(fp->fp,ptr,size*nmemb);
	else if (fp->type>=2)
		return 0;
	else
		return fwrite(ptr,size,nmemb,(FILE*)fp->fp);
}

int FCEU_fseek(FCEUFILE *fp,
	long int offset,
	int whence)
{
	if (fp->type==1)
		return (gzseek(fp->fp,offset,whence)>0) ? 0:(-1);
	else if (fp->type>=2)
	{
		MEMWRAP		*wz = (MEMWRAP*)fp->fp;

		switch(whence)
		{
		case SEEK_SET:
			if (offset>=wz->size)
				return (-1);
			wz->location = offset;
			break;
		case SEEK_CUR:
			if (offset+wz->location>wz->size)
				return (-1);
			wz->location+= offset;
			break;
		}
		return 0;
	} else
		return fseek((FILE*)fp->fp,offset,whence);
}

long int FCEU_ftell(FCEUFILE *fp)
{
	if (fp->type==1)
  		return gztell(fp->fp);
	else if (fp->type>=2)
		return (((MEMWRAP*)(fp->fp))->location);
	else
		return ftell((FILE*)fp->fp);
}

void FCEU_rewind(FCEUFILE *fp)
{
	if (fp->type==1)
		gzrewind(fp->fp);
	else if (fp->type>=2)
		((MEMWRAP*)(fp->fp))->location = 0;
	else /* Rewind */
		fseek(fp->fp,0,SEEK_SET);
}

int FCEU_read16le(uint16 *val,
	FCEUFILE *fp)
{
	uint8		t[2];

	if (fp->type>=1)
	{
		if (fp->type>=2)
		{
			MEMWRAP		*wz = (MEMWRAP*)fp->fp;

			if (wz->location+2>wz->size)
				return 0;
			*(uint32*)t = *(uint32*)(wz->data+wz->location);
			wz->location+= 2;
		} else if (fp->type==1)
			if (gzread(fp->fp,&t,2)!=2)
				return 0;
		return 1;
	} else if (fread(t,1,2,(FILE*)fp->fp)!=2)
		return 0;

	*val = t[0]|(t[1]<<8);

	return 1;
}

int FCEU_read32le(uint32 *Bufo,
	FCEUFILE *fp)
{
	if (fp->type>=1)
	{
		uint8	t[4];
#ifndef LSB_FIRST
		uint8	x[4];
#endif
		if (fp->type>=2)
		{
			MEMWRAP		*wz = (MEMWRAP*)fp->fp;

			if (wz->location+4>wz->size)
				return 0;
			*(uint32*)t = *(uint32*)(wz->data+wz->location);
			wz->location+= 4;
		} else if (fp->type==1)
			gzread(fp->fp,&t,4);
		else
			*(uint32*)t = 0;
#ifdef LSB_FIRST
		*(uint32*)Bufo = *(uint32*)t;
#else
		x[0] = t[3];
		x[1] = t[2];
		x[2] = t[1];
		x[3] = t[0];
		*(uint32*)Bufo = *(uint32*)x;
#endif
		return 1;
	} else
		return read32le(Bufo,(FILE*)fp->fp);
}

int FCEU_fgetc(FCEUFILE *fp)
{
	if (fp->type==1)
		return gzgetc(fp->fp);
	else if (fp->type>=2)
	{
		MEMWRAP		*wz = (MEMWRAP*)fp->fp;

		if (wz->location<wz->size)
			return wz->data[wz->location++];
		return EOF;
	} else
		return fgetc((FILE*)fp->fp);
}

long int FCEU_fgetsize(FCEUFILE *fp)
{
	if (fp->type==1)
	{
		int		x, t = gztell(fp->fp);

		gzrewind(fp->fp);
		for(x=0; gzgetc(fp->fp)!=EOF; ++x)
			; /* slow method */
		gzseek(fp->fp,t,SEEK_SET);
		return x;
	} else if (fp->type>=2)
		return ((MEMWRAP*)(fp->fp))->size;
	else
	{ /* was long type; 2009.10.19 16:11 EDT */
		long int	r, t = ftell((FILE*)fp->fp);

		fseek((FILE*)fp->fp,0,SEEK_END);
		r = ftell((FILE*)fp->fp);
		fseek((FILE*)fp->fp,t,SEEK_SET);
		return r;
	}
}

int FCEU_fisarchive(FCEUFILE *fp)
{
	return fp->type==2 ? 1:0;
}
