/*
 macos/commandline.c
  user command line interface (yay!)

  Created: 2008.08.11

  MacFCEU 0.8a6

Sunday, December 27, 2009 11:49 AM
  finally implemented the line storage mechanism.
Sunday, March 7, 2010
  next store identical to previous is canceled (don't add blatantly redundant lines).
Sunday, April 4, 2010 6:56 PM EDT
  asterisk at end recognized as *anything* for keywords_getproc.
*/

#include <string.h>

#include <MacWindows.h>
#include <QuickdrawText.h>

#include "shell.h"
#include "macvideo.h"
#include "usercommands.h"

#ifdef env68k
 #pragma segment UserCommands
#endif

/* ---- constants */

#define TEXT_INDENT_H			2
#define TEXT_INDENT_V			2
#define TEXT_LINEHEIGHT			12

/* ---- code */

int commandline_update(WindowRef window,
	const Rect *frame,
	const char *text,
	short length,
	short position,
	unsigned char insertmode)
{
	Rect		bounds;
	RgnHandle	oldclip;
	GrafPtr		oldport;

	if (!getvisiblewindowport(window) || !frame)
		return 0;
	GetPort(&oldport);
	setportwindow(window);

	if ((oldclip=NewRgn())!=NULL)
		GetClip(oldclip);

	/* set clip so drawing remains within the user box */
	bounds = *frame;
	ClipRect(&bounds);
	bounds.bottom = bounds.top+TEXT_LINEHEIGHT;

	/* draw text */
	MoveTo(bounds.left+TEXT_INDENT_H,bounds.bottom-TEXT_INDENT_V);
	if (length>0)
		MacDrawText(text,0,length);

	/* erase edges (especially to get rid of any stale cursor pixels) */
	++bounds.top, ++bounds.left, ++bounds.bottom; /*, --bounds.right; 2010.04.04 */
	PenMode(srcBic);
	MacFrameRect(&bounds);
	PenMode(srcCopy);

	/* erase from end; srcCopy mode should erase previous text (no flickering) */
	{
		Point		pen;

		GetPen(&pen);
		bounds.left = pen.h;
		EraseRect(&bounds);
	}

	/* draw cursor */
	if (position>=0)
	{ /* TEXT_INDENT_H-1 appearance fix; 2009.09.03 */
		short		x;

		if (position>length)
			position = length;
		x = frame->left+(TEXT_INDENT_H-1)+TextWidth(text,0,position);
		MoveTo(x,bounds.top);
		MacLineTo(x,bounds.bottom-TEXT_INDENT_V);

		if (!insertmode)
			Line(3,0);
	}

	if (oldclip)
		{SetClip(oldclip); DisposeRgn(oldclip);}

	MacSetPort(oldport);

	return 1;
}

int commandline_key(char *textbuf,
	short bufsize,
	short *textlength,
	short *textposition,
	unsigned char *insertmode,
	char charcode,
	boolean option)
{
	int		length, position, delta = 1;
	unsigned char	overwrite = 0;

	if (textbuf==NULL || textlength==NULL)
		return -1;

	if ((length=*textlength)<0)
		length = 0;
	if (insertmode)
		overwrite = !*insertmode;
	if (length>=bufsize)
		length = bufsize, overwrite = 1;

	if (textposition==NULL || (position=*textposition)<0)
		position = 0;
	else if (position>length)
		position = length;

	if (charcode==kDeleteCharCode) /* forward delete */
	{
		if (option)
			delta = parse_wordbreak(position+textbuf,
				length+textbuf,NULL)-textbuf-position;
	deletechar:
		if ((length-=delta)<0)
			length = 0;
		else if (length>position)
			memmove(position+textbuf,delta+position+textbuf,length-position);

		if (position>length)
			position = length;
		goto end;
	} else if (charcode=='`') /* swap; next key press will bring debugger to front */
	{
		if (getvisiblewindowport(gamewindow))
			BringToFront(gamewindow);
		goto end;
	}

	switch(charcode)
	{
	case kEnterCharCode: /* may be overridden by handleevent */
	case kReturnCharCode:
		return 1;
	case kClearCharCode: /* clear all */
		textbuf[0] = 0, length = 0;
	case kHomeCharCode:
		position = 0;
		break;
	case kEndCharCode:
		position = length;
		break;
	case kBackspaceCharCode:
		if (position==0)
			break;
		if (option)
			if ((delta=textbuf-parse_rwordbreak(position+textbuf,
					textbuf)+position)==0)
				delta = 1; /* delete something; 2011.02.08 */
		position-= delta;
		goto deletechar;
	case kTabCharCode: /* set element / pass field (not implemented) */
		break;
	case kHelpCharCode: /* insert (see 'ins' on the key); enter "help" for help */
		if (insertmode)
			*insertmode^= 1;
		break;
	case kLeftArrowCharCode:
		if (option)
			if ((position=parse_rwordbreak(position+textbuf,
					textbuf)-textbuf)!=*textposition)
				break; /* only if it moves; 2010.03.07 */
		if (position>0)
			--position;
		break;

	default:
		if (position>=bufsize)
		{
			if (--position<0) /* overwrite last character */
				break; /* no writing out of bounds! (this time) */
		} else if (overwrite)
		{
			if (position==length)
				++length;
		} else if (++length-position>0)
			memmove(1+position+textbuf,position+textbuf,length-position-1);

		if (length>bufsize)
			length = bufsize;
		textbuf[position] = charcode;
	case kRightArrowCharCode:
		if (option)
			if ((position=parse_wordbreak(position+textbuf,
					length+textbuf,NULL)-textbuf)!=*textposition)
				break; /* only if it moves; 2010.03.07 */
		if (++position>length)
			position = length;
	}
end:	if (textposition)
		*textposition = position;
	if (textlength)
		*textlength = length;

	return 0;
}

/* ---- execution */

#include <stdlib.h>

keyword_proc keywords_getproc(const char *name,
	long length,
	const commandproc_keyword *keywords)
{
	const char	*s;
	long		z;

	if (!keywords)
		return NULL;
	for(; (s=keywords->name)!=NULL && (z=strlen(s))>0
			&& keywords->proc!=NULL; ++keywords)
	{
		if (s[z-1]=='*')
		{
			if (length<--z)
				continue;
		} else if (length!=z)
			continue;
		if (strncasecmp(s,name,z)==0)
			return keywords->proc;
	}
	return NULL;
}

static char *salloc(const char *s,
	long z)
{
	char		*r;

	if (++z<0 || (r=realloc(NULL,z))==NULL)
		return NULL;
	memcpy(r,s,--z);
	r[z] = '\0';

	return r;
}

static int execute_next_command(const char **p,
	const char *e,
	const commandproc_keyword *keywords,
	keyword_proc defaultproc,
	const char *(*getvalue)(const char*,
		const char*,
		int32*,
		void*),
	unsigned char *userdata,
	evalvalue (*evalindir)(evalvalue),
	keyword_proc *last)
{
	static char	buf[24];

	keyword_proc	proc = NULL;
	char		**argv = NULL;
	const char	*s = *p, *n, *t;
	long		length;
	int		argc = 0;
	int		result = 0;

	if ((s=memchr(s,';',e-s))!=NULL)
		e = s;
	if ((s=*p)>=e)
		{++(*p); return 0;}
	s = parse_skipwsp(s,e);

	if ((n=parse_wordbreak(t=s,e,NULL))==NULL || (length=n-s)<=0)
	{
	dupo:	if ((argv=realloc(argv,(argc+2)*sizeof(*argv)))==NULL
				|| (argv[argc++]=salloc(t,length=e-t))==NULL)
			{result =~argc; goto end;}
		s = e;
		goto expr;
	}
	if ((proc=keywords_getproc(s,length,keywords))==NULL)
		goto dupo;
	for(;;)
	{
		if ((length<=0 && (length=n-(t=s))<=0)
				|| (argv=realloc(argv,(argc+2)*sizeof(*argv)))==NULL
				|| (argv[argc++]=salloc(t,length))==NULL)
			{result =~argc; goto end;}
		s = parse_skipwsp(n,e);
		if (n>=e || (n=parse_parameter(s,e-s,buf,NULL,getvalue,NULL,userdata,
				evalindir))==NULL || (length=strlen(t=buf))<0)
			break;
		if (n==s)
		{ /* make sure unrecognized but valid params are kept; 2011.02.24 */
			if (*s=='#') /* comment break */
				break;
			n = parse_wordbreak(s,e,NULL);
		} else if (*n==',')
			++n;
	}

	argv[argc] = NULL; /* so default proc can process a bare expression */
	result = proc(argc,argv);

	if (last)
		*last = proc;
end:	if (result)
	{
	expr:	if (defaultproc && (result=defaultproc(result,argv))==0
				&& last)
			*last = defaultproc; //succeeded this way; 2010.03.24
	}
	*p = s;
	while(argc>0)
		realloc(argv[--argc],0);
	if (argv)
		realloc(argv,0);

	return result;
}

keyword_proc commandline_execute(const char *text,
	int length,
	const commandproc_keyword *keywords,
	keyword_proc defaultproc,
	const char *(*getvalue)(const char*, const char*, int32*, void*),
	unsigned char *userdata,
	evalvalue (*evalindir)(evalvalue))
{
	const char	*s, *e;
	keyword_proc	prev = NULL;

	if ((s=text)==NULL || length<=0)
		return NULL;

	e = length+text;
	while(execute_next_command(&s,e,keywords,defaultproc,
			getvalue,userdata,
			evalindir,&prev)==0)
		if (s>=e || s<=text)
			break;
		else
			text = s, prev = NULL;

	return prev;
}

/* ---- scroll */

/* previously 20; 2010.03.30 */
#define MAXIMUM_LINES_STORED		50

typedef struct stored_line
{
	short		length;
	char		textbuf[1];
} stored_line;

static stored_line	*linestorage[MAXIMUM_LINES_STORED];
static unsigned short	linecount = 0;
static unsigned short	currentline;

int commandline_scroll(char *textbuf,
	short bufsize,
	short *textlength,
	short direction)
{
	const stored_line	*current;

	if (textbuf==NULL)
		return -1;
	if (linestorage[0]==NULL || linecount==0)
		return 0;

	if ((currentline+=direction)&0x8000)
		currentline = 0;
	else if (currentline>=linecount)
		currentline = linecount-1;

	current = linestorage[currentline];

	if ((*textlength=current->length)>bufsize)
		*textlength = bufsize;
	memcpy(textbuf,current->textbuf,*textlength);

	return 1;
}

static int samelaststring(const char *s,
	short z)
{
	const stored_line	*line;

	if (linecount>MAXIMUM_LINES_STORED
			|| (line=linestorage[linecount-1])==0)
		return EVAL_HUGEVAL;

	return z==line->length
		&& memcmp(s,line->textbuf,z)==0;
}

int commandline_storenext(const char *text,
	short length)
{
	stored_line	*current;

	currentline = linecount;
	if (text==NULL || length<=0 || samelaststring(text,length)>0)
		return 0;

	if (linecount>=(unsigned)MAXIMUM_LINES_STORED)
	{
		linestorage[0] = realloc(linestorage[0],0);
		memmove(linestorage,1+linestorage,
			(MAXIMUM_LINES_STORED-1)*sizeof(linestorage[0]));
		currentline = MAXIMUM_LINES_STORED;
	} else
		currentline = ++linecount;

	if ((current=realloc(0,length+(0+(int)((stored_line*)0)->textbuf)))==NULL)
		return -1;
	linestorage[currentline-1] = current;

	current->length = length;
	memcpy(current->textbuf,text,length);

	return 1;
}

int commandline_free()
{
	unsigned	c;

	if ((c=linecount)-1>=(unsigned)MAXIMUM_LINES_STORED-1)
		return (c==0)-1;
	do
		linestorage[c-1] = realloc(linestorage[c-1],0);
	while(--c!=0);

	c = linecount;
	currentline = linecount = 0;

	return c;
}
