/*
 macos/listwindow.c

  Created: 2013.05.06 (orig. from messagelog.c)
  Revised: 2015.08.16
*/

#include <stdlib.h>
#include <string.h>

#include <Events.h>
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Scrap.h>

#include "listwindow.h"
#include "shell.h"

/* ---- constants */

#define LW_FULL_UPDATE			0x8000
#define LW_MODAL			0x80
#define LW_SLECTABLE			0x1
#define LW_MULTIPLESEL			0x2

#define SCROLL_DOWN_DELAY		20 /* in 60ths ticks - was 30 */

#define DEFAULT_INDENT_H		6
#define DEFAULT_INDENT_V		0

#define MARGIN_HV			1

/* ---- typedefs */

typedef struct lwrec
{ /* 256 bytes buffered; a Classic DialogRecord is 170 bytes */
	int8		priv[180];
	int16		sostart, soend; /* horizontal character range for lines */
	lwClickHook	clickhook;
	lwTextHook	texthook;
	uint32		lastctime;
	int32		lastcsel, selstart, selend; /* selection lines */
	Point		indent;
	int16		maxlsize, flag, lhgt, bottompad;
	uint32		maxlcount;
	int32		tfirstln; /* where to clip when appending */
	int32		tlength, lcount;
	int32		*tlinelist;
	char		*text;
	ControlRef	hscroll, vscroll;
} lwrec, *lwptr;

/* ---- macros */

#define ShowHideControl(c,x)		do{if(x) ShowControl(c); else HideControl(c);}while(0);

#if !TARGET_API_MAC_CARBON
 /* this is actually an Appearance API */
 #define IsControlVisible(h)		(0+(**(h)).contrlVis)
#endif

/* ---- globals */

static ControlActionUPP		lwctlupp = NULL;

/* ---- code */

/* appending text to a clipped end requires that we SCAN IN REVERSE */
/* returns the earliest list index */
int32 rcalclines(int32 *list,
	int32 *count, /* required */
	const char *text,
	const char *e)
{
	int32		*l;
	const char	*s;
	long		t = 0; /* was 1 - the append func will increment when appropriate */
	uint8		p = 0xFF;

	if (!count || text>=(s=e))
		return -1;
	if ((l=list)!=NULL)
		l+= *count;
	do
	{
		uint8		c = *--s;

		if ((c-=10)<=13-10) /* any EOL */
		{ /* p==0 means the following char is a linefeed */
			if (!p && c==(uint8)(13u-10u)) /* DOS line ending (skip) */
				{p = 0xFF; continue;}
			++t;
			if (l)
				if (--l<=list)
					{++s, ++l; break;}
				else
					*l = 1+s-text;
		}
		p = c;
	} while(s>text);

	if (l && --l>=list)
		*l = s-text;
	*count = t;

	return l-list;
}

int listwindow_getbounds(WindowRef window,
	Rect *bounds)
{
	if (!window || !bounds)
		return -1;
	getwindowbounds(window,bounds);
	if (((lwptr)window)->vscroll)
		bounds->right-= 15;
	if (((lwptr)window)->hscroll)
		bounds->bottom-= 15;
	bounds->top+= MARGIN_HV, bounds->left+= MARGIN_HV;
	bounds->bottom-= MARGIN_HV, bounds->right-= MARGIN_HV;

	return bounds->bottom>bounds->top && bounds->right>bounds->left;
}

int listwindow_append(WindowRef window,
	const char *str,
	int z,
	Boolean update)
{
	char		*t;
	int32		max, total;

	if (!window || ((lwptr)window)->texthook)
		return -1;
	if ((total=((lwptr)window)->tlength)<0)
		total = 0;
	max = ((lwptr)window)->maxlcount;
	if (!((lwptr)window)->tlinelist)
		if ((((lwptr)window)->tlinelist=calloc(1,max*sizeof(int32)))==NULL)
			return -2;
	if ((t=((lwptr)window)->text)!=NULL && max)
	{ /* clip */
		int32		d = ((lwptr)window)->lcount;

		if (total>0)
			d+= t[total-1]-10u<=13u-10u;
		if (d>=max)
		{
			if ((d=((lwptr)window)->tfirstln)<0)
				return -1;
			((lwptr)window)->flag|= LW_FULL_UPDATE;
			if ((d=((lwptr)window)->tlinelist[d])>=0)
				if ((total-=d)<0)
					total = 0;
				else
					memmove(t,t+d,total);
		}
	}
	((lwptr)window)->tlength = total;

	if (z<=0)
		return 0;
	if ((t=realloc(t,total+z))==NULL)
		return -2;
	((lwptr)window)->text = t;
	((lwptr)window)->tlength = total+z;

	memcpy(total+t,str,z);

	if (!max)
	{ /* no clipping: gather on what's added instead */
		int32		*l;
		const char	*n = t;//+z, *e = t+total;
		int32		c = 0;
		/* count lines - account for a newly formed CR-LF sequence too */
		/* apparently, MPW is lax on this... (graphical errors pasting disparate eol types) */
		if (total>0 && t[total-1]==13 && *str==10)
			++n; /* skip to discount */
		total+= z;
		rcalclines(NULL,&c,t,t+total);
		/* reallocate */
		max = c+1;/*((lwptr)window)->lcount+c;*/
		if ((l=realloc(((lwptr)window)->tlinelist,max*sizeof(int32)))==NULL)
			return -2;
		((lwptr)window)->tlinelist = l;
		/* dealing with offsets that don't change - only add new ones (faster) */
		//rcalclines(l+((lwptr)window)->lcount,&c,n,e);
		/* now we can update the line count */
		//((lwptr)window)->lcount = max;
		//((lwptr)window)->tfirstln = 0;
	}// else
	{ /* initialize fixed maximum for rcalclines (to do the whole thing) */
		((lwptr)window)->lcount = max;
		((lwptr)window)->tfirstln = rcalclines(((lwptr)window)->tlinelist,
			&((lwptr)window)->lcount,t,t+((lwptr)window)->tlength);
		if (((lwptr)window)->lcount<max)
			++((lwptr)window)->lcount;
	}
	if (((lwptr)window)->tlength>0 && t[((lwptr)window)->tlength-1]-10u<=(13u-10u))
		--((lwptr)window)->lcount; /* graphical consideration - omit an empty last line */
	if (update)
	{
		listwindow_updatescrolls(window,false,true);
		listwindow_update(window,!!(((lwptr)window)->flag&LW_FULL_UPDATE));
		((lwptr)window)->flag&=~LW_FULL_UPDATE;
	}

	return 1;
}

static int listwindow_free(WindowRef window) /* should be valid */
{
	int		r = 0;

	if (((lwptr)window)->text)
		((lwptr)window)->text = realloc(((lwptr)window)->text,0), r = 1;
	((lwptr)window)->tlength = ((lwptr)window)->lcount = 0;
	if (((lwptr)window)->tlinelist)
		((lwptr)window)->tlinelist = realloc(((lwptr)window)->tlinelist,0);
	return r;
}

int listwindow_clear(WindowRef window)
{
	if (!window)
		return -1;
	listwindow_free(window);
	listwindow_updatescrolls(window,false,true);

	return listwindow_update(window,true);
}

int listwindow_select(WindowRef window,
	int32 firstLine,
	int32 endLine,
	int32 oStart,
	int32 oEnd,
	Boolean adjust)
{
	int		c = 0, y;

	if (!window)
		return -1;
	if (firstLine<0)
		firstLine = 0;
	else if (firstLine>=((lwptr)window)->lcount)
		firstLine = ((lwptr)window)->lcount-1;
	if (endLine>=((lwptr)window)->lcount)
		endLine = ((lwptr)window)->lcount-1;
	if (((lwptr)window)->selstart!=firstLine)
		((lwptr)window)->selstart = firstLine, c = 1;
	if (((lwptr)window)->selend!=endLine)
		((lwptr)window)->selend = endLine, ++c;
	if (adjust && ((lwptr)window)->vscroll
			&& listwindow_getscrolls(window,NULL,&y))
	{
		Rect		bounds;

		listwindow_getbounds(window,&bounds);
		if (firstLine<y || firstLine>=(y+=(bounds.bottom-bounds.top)
				/listwindow_getlheight(window)))
		{
			SetControlValue(((lwptr)window)->vscroll,firstLine);
			c|= 4;
		}
	}
	if (oStart!=IMIN(oStart))
	{
		((lwptr)window)->sostart = oStart;
		((lwptr)window)->soend = oEnd;
	}
	return c;
}

int listwindow_getsel(WindowRef window,
	int32 *start,
	int32 *end,
	int32 *lstart,
	int32 *lend)
{
	int32		s, e;

	if (!window)
		return -1;
	e = ((lwptr)window)->lcount-1;
	s = PIN(((lwptr)window)->selstart,0,e);
	e = PIN(((lwptr)window)->selend,-1,e);
	if (start)
		*start = s;
	if (end)
		*end = e;
	if (lstart)
		*lstart = MAX(((lwptr)window)->sostart,0);
	if (lend)
		*lend = ((lwptr)window)->sostart<0 ? ((lwptr)window)->maxlsize
			:((lwptr)window)->soend;
	return ((lwptr)window)->selend>=0 ? e-s+1:0;
}

int listwindow_getlheight(WindowRef window)
{
	int		v;

	if ((v=((lwptr)window)->lhgt)==0)
		v = 1;
	return v;
}

int listwindow_getscrolls(WindowRef window,
	int *h,
	int *v)
{
	ControlRef	control;
	int		c;

	if (!window)
		return -1;
	c = 0;
	if ((control=((lwptr)window)->vscroll)!=NULL)
	{
		++c;
		if (v)
			*v = GetControlValue(control);
	}
	if ((control=((lwptr)window)->hscroll)!=NULL)
	{
		++c;
		if (h)
			*h = GetControlValue(control);
	}
	return c;
}

int listwindow_drawtext(WindowRef window,
	int c)
{
	lwTextHook	hook;
	const int32	*a = NULL, *e;
	char		*t = NULL;
	int32		h, y = 0, sl, el, mxlsz;
	int		i, x, lheight;
	Boolean		hl;

	if (!window)
		return -1;
	if ((hook=((lwptr)window)->texthook)==NULL
			&& ((t=((lwptr)window)->text)==NULL
			|| (a=((lwptr)window)->tlinelist)==NULL
			|| (y=((lwptr)window)->tfirstln)<0))
		return 0;
	if (((lwptr)window)->maxlcount)
		a+= y;
	e = a-1+((lwptr)window)->lcount;
	i = x = 0, mxlsz = ((lwptr)window)->maxlsize+1;

	listwindow_getscrolls(window,&x,&i);
	x*=-6;
	x+= ((lwptr)window)->indent.h;
	y = ((lwptr)window)->indent.v+(lheight=listwindow_getlheight(window));

	/* completion to account for scrolling */
	if (i>0)
		--i, y-= lheight, ++c;
	if ((unsigned)(h=((lwptr)window)->lcount-i)>c)
		h = c;
	if (a)
		a+= i;
	hl = false, sl = ((lwptr)window)->selstart, el = ((lwptr)window)->selend;
	TextMode(srcCopy);
	do
	{
		char		*s = NULL;
		const char	*p;
		int32		z;
		Boolean		final = 1;

		MoveTo(x,y);
		if (a)
		{
			if (a>e)
				break;
			final = a>=e;
			if ((z=(final ? ((lwptr)window)->tlength:a[1])-*a)<=0)
				{++a; continue;}
			s = t+*a++;
			if (s[z-1]=='\012') /* dont draw LF; 2009.06.27 */
				--z;
		}
		if (i>=sl && i<=el)
		{
			if (!hl)
				{hl = true; TextMode(notSrcCopy);}
		} else if (hl)
			{hl = false; TextMode(srcCopy);}
		p = s;
		if (hook)
			p = hook(s,&z,i);
		if (z>mxlsz)
			z = mxlsz;
		if (p)
			if (hl && ((lwptr)window)->sostart>=0)
			{
				TextMode(srcCopy);
				DrawText(p,0,MIN(z,((lwptr)window)->sostart));
				TextMode(notSrcCopy);
				DrawText(p,((lwptr)window)->sostart,MIN(z,((lwptr)window)->soend)
					-((lwptr)window)->sostart);
				TextMode(srcCopy);
				if ((z-=((lwptr)window)->soend)>0)
					DrawText(p,((lwptr)window)->soend,z);
				hl = false;
			} else
				DrawText(p,0,z);
		if (!final)
			DrawChar(' ');
	} while(y+=lheight,++i,--h);
	TextMode(srcCopy);

	return 1;
}

static void listwindow_scrollbox(WindowRef window, /* should be valid */
	short sh,
	short sv)
{
	Rect		newfocus;
	RgnHandle	region;
	GrafPtr		oldport;

	GetPort(&oldport);
	setportwindow(window);

	listwindow_getbounds(window,&newfocus);
	region = NewRgn();
	ScrollRect(&newfocus,sh,sv,region);

	if (region) /* update what ScrollRect has given us (the remainder that needs rendering) */
		SetClip(region);
	listwindow_drawtext(window,1+(newfocus.bottom-newfocus.top)
		/listwindow_getlheight(window));

	if (region)
		DisposeRgn(region);
	/* reset clip (anything large enough will do) */
	*(int32*)&newfocus.top = 0;
	*(int32*)&newfocus.bottom = 0x40004000;
	ClipRect(&newfocus);

	MacSetPort(oldport);
}

int listwindow_scroll(WindowRef window,
	int hcount,
	int vcount)
{
	ControlRef	control;
	int		hsep = IMIN(int), vsep = IMIN(int);
	int		c = 0, vshift = vcount&(cmdKey<<16) ? 11:0;

	if (!window)
		return -1;
	/* values above 0xFFFF are page move indicators */
	if (hcount>0xFFFF)
	{
		if ((hcount&=0xFF)==0)
		{ /* command page -> vertical to horizontal scroll */
			if ((hcount=vcount&0xFF)>kPageDownCharCode || !vshift)
				{hcount = 0; goto skip;}
			vcount = 0;
		}
		switch(hcount)
		{
		case kPageUpCharCode: hcount =-8; break;
		case kPageDownCharCode: hcount = 8; break;
		case kUpArrowCharCode: hcount =-1; break;
		case kDownArrowCharCode: hcount = 1; break;
		default: hcount = 0;
		}
	}
	if (vcount>0xFFFF)
		switch(vcount&0xFF)
		{
		case kPageUpCharCode: vcount =-8<<vshift; break;
		case kPageDownCharCode: vcount = 8<<vshift; break;
		case kUpArrowCharCode: vcount =-1<<vshift; break;
		case kDownArrowCharCode: vcount = 1<<vshift; break;
		default: vcount = 0;
		}
skip:	if ((control=((lwptr)window)->vscroll)!=NULL)
	{
		++c, vsep = GetControlValue(control);
		if ((vcount+=vsep)<-32768)
			vcount =-32768;
		else if (vcount>32767)
			vcount = 32767;
		SetControlValue(control,vcount);
		vsep-= GetControlValue(control);
		vsep*= listwindow_getlheight(window);
	}
	if ((control=((lwptr)window)->hscroll)!=NULL)
	{
		++c, hsep = GetControlValue(control);
		if ((hcount+=hsep)<-32768)
			hcount =-32768;
		else if (hcount>32767)
			hcount = 32767;
		SetControlValue(control,hcount);
		hsep-= GetControlValue(control);
		hsep*= 6;
	}

	listwindow_scrollbox(window,hsep,vsep);
	((lwptr)window)->flag|= LW_FULL_UPDATE;

	return c;
}

int listwindow_updatescrolls(WindowRef window,
	Boolean doBounds,
	Boolean advance)
{
	Rect		bounds;
	ControlRef	control;
	int		value, vertc = 64, width, height;

	if (!window)
		return -1;
	getwindowbounds(window,&bounds);
	height = bounds.bottom-bounds.top-((lwptr)window)->bottompad;
	width = bounds.right-bounds.left;
	if ((control=((lwptr)window)->vscroll)!=NULL)
	{ /* vertical */
		int	lheight = listwindow_getlheight(window);

		if (doBounds)
		{
			bounds.top =-1;
			bounds.left = width-15;
			bounds.bottom = height-14;
			bounds.right = width+1;
			SetControlBounds(control,&bounds);
		} else
			GetControlBounds(control,&bounds);
		if ((vertc=((lwptr)window)->lcount-(bounds.bottom-bounds.top-15)/lheight-1
				-!((lwptr)window)->text)<=0)
			SetControlMaximum(control,0);
		else
		{ /* set maximum, update scroll position if at maximum */
			value = GetControlMaximum(control);

			SetControlMaximum(control,vertc);

			if (advance && GetControlValue(control)==value) /*value>0 && */
			{
				SetControlValue(control,vertc);
				value-= vertc;
				listwindow_scrollbox(window,0,value*lheight);
			}
		}
	}
	if ((control=((lwptr)window)->hscroll)!=NULL)
	{ /* horizontal */
		if (doBounds)
		{
			bounds.top = height-15;
			bounds.left =-1;
			bounds.bottom = height+1;
			bounds.right = width-14;
			SetControlBounds(control,&bounds);
		} else
			GetControlBounds(control,&bounds);
		value = ((lwptr)window)->maxlsize-(bounds.right-bounds.left)/6+1;
		SetControlMaximum(control,value<0 || !((lwptr)window)->text ? 0:value);
	}
	return vertc;
}

int listwindow_maxline(WindowRef window,
	Boolean set)
{
	const int32	*a, *e;
	const char	*t;
	int32		z, x;

	if (!window)
		return -1;
	if ((t=((lwptr)window)->text)==NULL
			|| (a=((lwptr)window)->tlinelist)==NULL
			|| (x=((lwptr)window)->tfirstln)<0)
		return 0;
	if (((lwptr)window)->maxlcount)
		a+= x;
	for(x=0,e=a-1+((lwptr)window)->lcount; a<=e; ++a)
	{
		if ((z=(a>=e ? ((lwptr)window)->tlength:a[1])-*a)<=0)
			continue;
		if (z>x)
			x = z;
	}
	if (set)
	{
		Rect		bounds;

		listwindow_getbounds(window,&bounds);
		((lwptr)window)->maxlsize = x;
	}
	return x;
}

static pascal void mycontrolactionproc(ControlRef control,
	ControlPartCode part)
{
	WindowRef	window;
	unsigned long	ticks = TickCount();
	unsigned long	reference = GetControlReference(control);
	SInt16		value;

	if (ticks<reference)
		return;
	value = GetControlValue(control);
	switch(part)
	{
	case kControlUpButtonPart:
		if (value>0)
			SetControlValue(control,value-1);
		break;
	case kControlDownButtonPart:
		if (value<GetControlMaximum(control))
			SetControlValue(control,value+1);
		break;
	case kControlPageUpPart:
		SetControlValue(control,value-8);
		break;
	case kControlPageDownPart:
		SetControlValue(control,value+8);
		break;
	case kControlIndicatorPart:
		break;
	}

	if ((value-=GetControlValue(control))!=0 /* sub-ing new; negative = forward */
			&& (window=GetControlOwner(control))!=NULL)
	{ /* scroll window and update */
		Str255	title;

		GetControlTitle(control,title);
		if (title[1]=='V')
			listwindow_scrollbox(window,0,value*listwindow_getlheight(window));
		else
			listwindow_scrollbox(window,6*value,0);
	}

	if (!reference)
		SetControlReference(control,ticks+SCROLL_DOWN_DELAY);
}

#if 0
static void alignbounds(int lheight,
	long newdimensions,
	Rect *bounding)
{
	*(int32*)bounding = 0;
	bounding->right = (int16)newdimensions;

	/* make height divisible by the line height */
	newdimensions+= 0+(lheight/2)<<16;
	newdimensions/= lheight<<16;
	newdimensions*= lheight;
	bounding->bottom = (newdimensions-=0);
}
#endif

WindowRef listwindow_create(WindowRef behind,
	const Rect *bounds,
	const uint8 *title,
	const uint8 *fontfamily,
	int16 textsize,
	int16 maxlsize,
	int32 maxlcount,
	int16 bottompad,
	int sel,
	long refCon)
{
	FontInfo	fontinfo;
	Rect		ctlbounds;
	WindowRef	window;
	GrafPtr		oldport;
	Ptr		storage;
	int		lh;
	int16		fontid, height, width;

	GetPort(&oldport);

	if ((storage=NewPtrClear(sizeof(lwrec)))==NULL)
		return 0;
	if ((window=NewCWindow(storage,bounds,title,false,
			bottompad<0 ? noGrowDocProc:documentProc,
			behind,true,refCon))==NULL)
	{
		DisposePtr(storage);
		return 0;
	}

	setportwindow(window);
	if (fontfamily)
		GetFNum(fontfamily,&fontid);
	TextFont(fontid);
	TextSize(textsize);
	TextMode(srcCopy); /* white background */
	/* calculate line height */
	GetFontInfo(&fontinfo);
	if ((lh=fontinfo.ascent+fontinfo.descent)==0) //+1)==0)
		lh = 11; //12;
	((lwptr)window)->lhgt = lh;
	if (sel>=0)
		((lwptr)window)->flag = sel ? LW_SLECTABLE|LW_MULTIPLESEL:LW_SLECTABLE;
	if (bottompad<0)
		((lwptr)window)->flag|= LW_MODAL;
	((lwptr)window)->selend =-1;
	/* stuff in the rest of the parameters */
	((lwptr)window)->indent.h = DEFAULT_INDENT_H;
	((lwptr)window)->indent.v = DEFAULT_INDENT_V;
	((lwptr)window)->bottompad = (bottompad>>15)^bottompad;
	((lwptr)window)->sostart =-1;

	//alignbounds(lh,(bounds->bottom-bounds->top<<16)|(uint16)(bounds->right-bounds->left),&ctlbounds);
	//width = ctlbounds.right;
	//SizeWindow(window,width,ctlbounds.bottom,false);
	//height = ctlbounds.bottom-((lwptr)window)->bottompad;
	height = bounds->bottom-bounds->top-((lwptr)window)->bottompad;
	width = bounds->right-bounds->left;
	((lwptr)window)->maxlcount = ABS(maxlcount);
	if (maxlcount>=0)
	{
		ctlbounds.top =-1;
		ctlbounds.left = width-15;
		ctlbounds.bottom = height-14;
		ctlbounds.right = width+1;
		((lwptr)window)->vscroll = NewControl(window,&ctlbounds,"\pV",true,
			0,0,0,scrollBarProc,0);
	}
	if ((((lwptr)window)->maxlsize=maxlsize)!=0)
	{
		ctlbounds.top = height-15;
		ctlbounds.left =-1;
		ctlbounds.bottom = height+1;
		ctlbounds.right = width-14;
		((lwptr)window)->hscroll = NewControl(window,&ctlbounds,"\pH",true,
			0,0,MAX(maxlsize,64)-64,scrollBarProc,0);
	}

	MacSetPort(oldport);

	return window;
}

int listwindow_dispose(WindowRef window)
{
extern WindowRef myfrontwindow;
	if (!window)
		return -1;
	listwindow_free(window);
	KillControls(window);
	DisposeWindow(window);
	return 1;
}

static void drawlistgrow(WindowRef window)
{
	Rect		bounds;

	getwindowbounds(window,&bounds);
	if (!((lwptr)window)->vscroll)
		bounds.top = bounds.bottom-15;
	if (!((lwptr)window)->hscroll)
		bounds.left = bounds.right-15;
	ClipRect(&bounds);
	drawgrowicon(window);

	*(int32*)&bounds = 0;
	*(int32*)&bounds.bottom = 0x44444444;
	ClipRect(&bounds);
}

int listwindow_update(WindowRef window,
	Boolean erase)
{
	Rect		bounds;
	GrafPtr		oldport;
	int		r;

	if (!getvisiblewindowport(window))
		return -1;
	getwindowbounds(window,&bounds);

	GetPort(&oldport);
	setportwindow(window);

	ClipRect(&bounds);
	DrawControls(window);

	if (((lwptr)window)->vscroll)
		bounds.right-= 15;
	if (((lwptr)window)->hscroll)
		bounds.bottom-= 15;
	ClipRect(&bounds);

	if (erase)
		EraseRect(&bounds);

	bounds.top+= MARGIN_HV, bounds.left+= MARGIN_HV;
	bounds.bottom-= MARGIN_HV, bounds.right-= MARGIN_HV;
	ClipRect(&bounds);

	r = listwindow_drawtext(window,1+(bounds.bottom-bounds.top)
		/listwindow_getlheight(window));

	if (erase==true && !(((lwptr)window)->flag&LW_MODAL))
		drawlistgrow(window);
	else
	{
		*(int32*)&bounds = 0;
		*(int32*)&bounds.bottom = 0x44444444;
		ClipRect(&bounds);
	}

	MacSetPort(oldport);

	return r;
}

int listwindow_grow(WindowRef window,
	const EventRecord *event,
	Rect *bounding)
{
	long		dimensions;

	if (!getvisiblewindowport(window))
		return -1;
	/* GrowWindow subtracts one when min=max, can adjust wholey via command key */
	if ((dimensions=GrowWindow(window,event->where,bounding))!=0)
	{
		//alignbounds(listwindow_getlheight(window),dimensions,bounding);
		*(int32*)bounding = 0;
		SizeWindow(window,bounding->right=(int16)dimensions,bounding->bottom=dimensions>>16,false);
		listwindow_updatescrolls(window,true,true); /* this was missing for so long!; 2011.03.24 */
		/* because of bugs in SizeWindow, invalidation should be done manually */
		listwindow_update(window,true);

		return 1;
	}
	return 0;
}

#include <stdio.h>

static int listwindow_clickselect(WindowRef window,
	EventRecord *event,
	Point hitpoint)
{
	FontInfo	fontinfo;
	Rect		bounds, lbounds;
	GrafPtr		oldport;
	int		x, y, t, l, m, lheight = listwindow_getlheight(window);
	int		lt, nt, td, r = 0;
	int16		ltop, left;
	Boolean		ms;

	listwindow_getbounds(window,&lbounds);
	getwindowbounds(window,&bounds);
	GetPort(&oldport);
	setportwindow(window);

	listwindow_getscrolls(window,&x,&y);
	t = ((lwptr)window)->indent.h-6*x, left = MAX(t,0);
	t = ((lwptr)window)->indent.v-lheight*y, ltop = MAX(t,0);
	if (hitpoint.h<left || hitpoint.v<MAX(ltop,lbounds.top)
			|| hitpoint.h>=lbounds.right || hitpoint.v>=lbounds.bottom)
		goto end;

	GetFontInfo(&fontinfo); /* no more FUDGE_V */
	if (!fontinfo.widMax)
		fontinfo.widMax = 1;

	if ((y=(hitpoint.v-t-fontinfo.descent)/lheight)<0 || y>=((lwptr)window)->lcount)
		goto end;
	l =-1, ms = ((lwptr)window)->flag&LW_MULTIPLESEL && event->modifiers&shiftKey;

	if (((lwptr)window)->selstart<0 || ((lwptr)window)->selend<((lwptr)window)->selstart)
		m = y; /* no prev. range */
	else 
	{ /* direction from middle */
		m = ((lwptr)window)->selstart+((lwptr)window)->selend>>1;
		m = y>m ? ((lwptr)window)->selstart:((lwptr)window)->selend;
	}
	/* double-click check; 2013.05.21 */
	lt = TickCount(), td = lt-((lwptr)window)->lastctime;
	((lwptr)window)->lastctime = lt;

	if (((lwptr)window)->lastcsel==y && td<GetDblTime())
	{
		((lwptr)window)->lastcsel =-1; /* no triple-clicking */
		return lwDoubleClick;
	}

	ClipRect(&lbounds);
	if (((lwptr)window)->clickhook)
		goto fsel;
	do
	{
		int		sa = 0, se;

		if (l!=y)
		{
			if (!ms)
				m = y;
			if (listwindow_select(window,MIN(y,m),MAX(y,m),IMIN(y),0,false) || l==IMIN(l))
				listwindow_drawtext(window,1+(lbounds.bottom-lbounds.top)/lheight);
			l = y;
		}
		GetMouse(&hitpoint);
	fsel:	if ((hitpoint.v<ltop || hitpoint.v>lbounds.bottom) && (nt=TickCount())-lt>=0)
		{
			if ((t=hitpoint.v-ltop)<0)
				t =-t, y =-1;
			else
				t = hitpoint.v-lbounds.bottom, y = 1;
			lt = nt+32/PIN(t,2,32);

			ClipRect(&bounds);
			listwindow_scroll(window,0,y);
			ClipRect(&lbounds);

			listwindow_getscrolls(window,&x,&y);
			t = ((lwptr)window)->indent.v-lheight*y;
			//if ((t=((lwptr)window)->indent.h-6*x)<0)
			//	t = 0;
			//mh = t;
			goto upy;
		}
		if (hitpoint.h>=left && hitpoint.h<lbounds.right)
		{
			sa = !!((lwptr)window)->clickhook;
		upy:	y = (hitpoint.v-t-fontinfo.descent)/lheight;
			if (y>=((lwptr)window)->lcount)
				y =-1;
			if (sa)
			{
				sa = ((lwptr)window)->sostart;
				se = ((lwptr)window)->soend;
				((lwptr)window)->clickhook(y,(hitpoint.h-left)/fontinfo.widMax,&sa,&se);
				if (sa!=((lwptr)window)->sostart)
					((lwptr)window)->sostart = sa, l = IMIN(l);
				if (se!=((lwptr)window)->soend)
					((lwptr)window)->soend = se, l = IMIN(l);
			}
		}
	} while(StillDown());

	if (y>=((lwptr)window)->lcount)
		y =-1;
	r = (((lwptr)window)->lastcsel=y)>=0 ? lwSingleClick:lwWithinBounds;

end:	ClipRect(&bounds);
	MacSetPort(oldport);

	return r;
}

int listwindow_click(WindowRef window,
	EventRecord *event,
	short clickpart,
	Rect *growbounding)
{
	Point		hitpoint = event->where;
	ControlRef	control;

	ControlPartCode		part;

	if (!getvisiblewindowport(window))
		return -1;
	localwindowpoint(window,&hitpoint);

	if (clickpart==inGrow && growbounding)
		return listwindow_grow(window,event,growbounding);
	if (FrontWindow()!=window)
	{
		showhide(window,true);
		return 0;
	}
	/* both scrolls show reflect the same 'visible' */
	if ((control=((lwptr)window)->vscroll)!=NULL && !IsControlVisible(control))
		listwindow_setactive(window,true); /* manual update; 2010.05.27 */

	if (!lwctlupp && (lwctlupp=NewControlActionUPP(mycontrolactionproc))==NULL)
		return -2;
	if ((part=FindControl(hitpoint,window,&control))==kControlNoPart || !control)
	{
		if (((lwptr)window)->flag&LW_SLECTABLE)
			return listwindow_clickselect(window,event,hitpoint);
		return 0;
	}
	SetControlReference(control,0); /* initialize timing for control */

	switch(part)
	{
	case kControlUpButtonPart:
	case kControlDownButtonPart:
	case kControlPageUpPart:
	case kControlPageDownPart:
		TrackControl(control,hitpoint,lwctlupp);
		break;
	case kControlIndicatorPart:
		TrackControl(control,hitpoint,NULL);
		listwindow_update(window,2);
	}
	return 1;
}

int listwindow_setactive(WindowRef window,
	Boolean active)
{
	ControlRef	control;

	if (!getvisiblewindowport(window))
		return 0;
	if ((control=((lwptr)window)->vscroll)!=NULL)
		ShowHideControl(control,active);
	if ((control=((lwptr)window)->hscroll)!=NULL)
		ShowHideControl(control,active);
	/* commenting this out may produce missing grow icon under pre-Appearance */
	//if (!(((lwptr)window)->flag&LW_MODAL))
	//	drawlistgrow(window);
	HiliteWindow(window,active);

	return 1;
}

static int32 gettoffset(WindowRef window,
	int32 x)
{
	const int32	*a;
	int32		i;

	if ((a=((lwptr)window)->tlinelist)==NULL || (i=((lwptr)window)->tfirstln)<0)
		return -1;
	if (((lwptr)window)->maxlcount)
		a+= i;
	return x<0 ? 0:x>=((lwptr)window)->lcount ? ((lwptr)window)->tlength:a[x];
}

int listwindow_copy(WindowRef window)
{
	int32		s, e;

	if (!window)
		return -1;
	if (!((lwptr)window)->text
			|| (s=gettoffset(window,((lwptr)window)->selstart))<0
			|| (e=gettoffset(window,((lwptr)window)->selend+1))<0
			|| s>=e)
		return 0;
#if CALL_NOT_IN_CARBON
	ZeroScrap();
	PutScrap(e-s,kScrapFlavorTypeText,s+((lwptr)window)->text);
#else
	ClearCurrentScrap();
	{
		ScrapRef	scrap;

		GetCurrentScrap(ScrapRef *scrap);
		PutScrapFlavor(scrap,kScrapFlavorTypeText,kScrapFlavorMaskNone,
			e-s,s+((lwptr)window)->text);
	}
#endif
	return 1;
}

/* added; 2015.07.18 */
int listwindow_settexthook(WindowRef window,
	lwTextHook hook,
	int lineCount)
{
	if (!window)
		return -1;
	((lwptr)window)->texthook = hook;
	if (lineCount>=0)
	{
		if (((lwptr)window)->maxlsize<=0)
			((lwptr)window)->maxlsize = 16383/6;
		((lwptr)window)->lcount = lineCount;
		listwindow_updatescrolls(window,false,false);
	}
	return 1;
}

/* added; 2015.07.27 */
int listwindow_setclickhook(WindowRef window,
	lwClickHook hook)
{
	if (!window)
		return -1;
	((lwptr)window)->clickhook = hook;
	return !!hook;
}

const char *listwindow_gettext(WindowRef window,
	int32 *z,
	int i)
{
	int32		s, e;

	if (!window)
		return NULL;
	if (!((lwptr)window)->text)
	{
		if (((lwptr)window)->texthook)
			return ((lwptr)window)->texthook(NULL,z,i);
		return NULL;
	}
	if (i>=((lwptr)window)->selend || (s=gettoffset(window,i))<0
			|| (e=gettoffset(window,i<0 ? IMAX(i):i+1))<0
			|| s>=e)
		return NULL;
	if (z)
		*z = e-s;
	return s+((lwptr)window)->text;
}
