/*
 macos/macdialogs.c

  Created: 2008.07.12
  Revised: 2014.04.03
*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <Files.h>
#include <Dialogs.h>
#include <Controls.h>
#include <Menus.h>
#include <StandardFile.h>
#include <QuickdrawText.h>
#include <Sound.h> /* for SysBeep */

#if TARGET_API_MAC_CARBON
 #include <Navigation.h>
#endif

#include "shell.h"

#include "../FCEU.h"

#include "shellmenus.h"
#include "macdialogs.h"
#include "usercommands.h"

#ifdef env68k
 #pragma segment shell
#endif

/* ---- constants */

#define CPROGRAM_NAME		"MacFCEU"

enum
{ /* getfilepath DITL (one-based) indicies */
	gfpitem_select		= 11, /* only for custom get */
	gfpitem_unused,
	gfpitem_filter
};

/* ---- globals */

static ModalFilterUPP	stdfilter;

/* ---- code */

static char *strndup(const char *s,
	long z)
{
	char		*d;

	if (!s)
		return NULL;
	if (z<0)
		z = strlen(s);
	if ((d=realloc(NULL,1+z))==NULL)
		return NULL;
	_bcpy(d,s,z);
	d[z] = 0;

	return d;
}

int setcontrolitem(DialogRef dialog,
	int item,
	int value,
	int minimum,
	int maximum,
	int hilite,
	const uint8 *title)
{
	Handle		handle;

	if (!dialog || item<=0)
		return -1;
	{
		Rect		bounds;
		short		type;

		GetDialogItem(dialog,item,&type,&handle,&bounds);
		if (!handle || !(type&kControlDialogItem))
			return 0;
	}

	if (minimum!=INT_NO_VALUE && maximum!=GetControlMinimum((ControlRef)handle))
		SetControlMinimum((ControlRef)handle,minimum);
	if (maximum!=INT_NO_VALUE && maximum!=GetControlMaximum((ControlRef)handle))
		SetControlMaximum((ControlRef)handle,maximum);
	if (value!=INT_NO_VALUE && value!=GetControlValue((ControlRef)handle))
		SetControlValue((ControlRef)handle,value);
	if (hilite!=INT_NO_VALUE && hilite!=GetControlHilite((ControlRef)handle))
		HiliteControl((ControlRef)handle,hilite);
	if (title)
	{
		Str255		whatevs;

		GetControlTitle((ControlRef)handle,whatevs);
		if (memcmp(whatevs,title,*whatevs)!=0)
			SetControlTitle((ControlRef)handle,title);
	}

	return 1;
}

static int getcontrolitem(DialogRef dialog,
	int item,
	int *minimum,
	int *maximum,
	int *hilite,
	uint8 *title)
{
	Rect		bounds;
	Handle		handle;
	short		type;

	if (!dialog || item<=0)
		return INT_NO_VALUE;
	GetDialogItem(dialog,item,&type,&handle,&bounds);
	if (!handle || !(type&kControlDialogItem))
		return INT_NO_VALUE;

	if (minimum)
		*minimum = GetControlMinimum((ControlRef)handle);
	if (maximum)
		*maximum = GetControlMaximum((ControlRef)handle);
	if (hilite)
		*hilite = GetControlHilite((ControlRef)handle);
	if (title)
		GetControlTitle((ControlRef)handle,title);
	return GetControlValue((ControlRef)handle);
}

static void makedirspec(int32 *voldir,
	const char *directory)
{
	CInfoPBRec	cinfo;
	FSSpec		spec;
	long		length;

	if (((length=strlen(directory))==0 || path2spec(directory,length,&spec)!=noErr)
			&& path2spec(":",1,&spec)!=noErr) /* shouldnt fail */
	{
		spec.vRefNum = -1;
		spec.parID = fsRtDirID;
		*spec.name = 0;
	}

	cinfo.dirInfo.ioNamePtr = spec.name;
	cinfo.dirInfo.ioFDirIndex = 0;
	cinfo.dirInfo.ioFlAttrib = 0;
	cinfo.dirInfo.ioACUser = 0;
	cinfo.dirInfo.ioVRefNum = spec.vRefNum;
	cinfo.dirInfo.ioDrParID = cinfo.dirInfo.ioDrDirID = spec.parID;
	PBGetCatInfoSync(&cinfo);

	*voldir = cinfo.dirInfo.ioVRefNum;
	voldir[1] = !(cinfo.dirInfo.ioFlAttrib&kioFlAttribDirMask) ?
		cinfo.dirInfo.ioDrParID:cinfo.dirInfo.ioDrDirID; /* use parent if file */
}

short showalert(short alertid,
	const uint8 *p1,
	const uint8 *p2,
	const uint8 *p3)
{
	short		index;

	ParamText(p1,p2,p3,"\p");
	index = Alert(alertid,NULL);
	ParamText("\p","\p","\p","\p"); /* reset */

	if (alertid==alertid_fatal)
		ExitToShell(); /* can't exit() here (memcpy is used) */

	return index;
}

int setbinaryitem(DialogRef dialog,
	short item,
	int value)
{
	Rect		bounds;
	Handle		handle;
	short		type;

	GetDialogItem(dialog,item,&type,&handle,&bounds);
	if (!handle)
		return -1;
	if (value==-2)
		HideDialogItem(dialog,item);
	else if (type==kCheckBoxDialogItem)
		SetControlValue((ControlRef)handle,value!=INT_NO_VALUE ?
			value:!GetControlValue((ControlRef)handle));

	return type;
}

int getbinaryitem(DialogRef dialog,
	short item)
{
	Rect		bounds;
	Handle		handle;
	short		type;

	GetDialogItem(dialog,item,&type,&handle,&bounds);
	if (handle)
		if (type==kCheckBoxDialogItem)
			return GetControlValue((ControlRef)handle);
	return 0;
}

/* should be used sparingly */
static int setdialogtextbox(DialogRef dialog,
	int item,
	const void *text,
	long length,
	boolean update)
{
	int		current, last;

	if (item>0)
		last = item;
	else /* otherwise, were scanning to the next workable text box */
		item = -item, last = CountDITL(dialog);

	if (!dialog || item>last)
		return -1;
	do
	{
		Rect		bounds;
		Handle		handle;
		short		type;

		GetDialogItem(dialog,item,&type,&handle,&bounds);
		type&=~kItemDisableBit;
		if (type!=kStaticTextDialogItem && type!=kEditTextDialogItem
				|| !handle)
			continue;
		/* hack; handle returned for text boxes are the handle to the raw text */
		if (!text)
			length = 0;
		SetHandleSize(handle,length);
		if (MemError()==noErr)
			_bcpy(*handle,text,length);
		if (item==(current=GetDialogKeyboardFocusItem(dialog)))
			SelectDialogItemText(dialog,0,0,0);
		if (update)
			InvalWindowRect(GetDialogWindow(dialog),&bounds);
		if (item==current)
			SelectDialogItemText(dialog,item,0,length);
		return item;
	} while(++item<=last);

	return -1;
}

/* requires that `item` be valid */
static char *getdialogtextbox(DialogRef dialog,
	int item)
{
	Rect		bounds;
	Handle		handle;
	long		length;
	short		type;

	if (!dialog || item<=0)
		return NULL;
	GetDialogItem(dialog,item,&type,&handle,&bounds);
	type&=~kItemDisableBit;
	if (type!=kStaticTextDialogItem && type!=kEditTextDialogItem || !handle
			|| (length=GetHandleSize(handle))<=0)
		return NULL;

	return strndup(*handle,length);
}

void setdialogitem_pstr(DialogRef dialog,
	DialogItemIndex item,
	const uint8 *str)
{
	Rect		bounds;
	Handle		handle;
	short		type;

	GetDialogItem(dialog,item,&type,&handle,&bounds);
	if (handle)
		SetDialogItemText(handle,str);
}

int dialogselect(DialogRef dialog,
	EventRecord *event)
{
	GrafPtr		current;

	DialogItemIndex		item;

	switch(GetWRefCon(GetDialogWindow(dialog)))
	{
	case FOUR_CHAR_CODE('stdf'):
	case FOUR_CHAR_CODE('nfdr'):
	case FOUR_CHAR_CODE('rplc'):
	case FOUR_CHAR_CODE('stat'):
	case FOUR_CHAR_CODE('lock'): //some software (i.e., ResEdit) forget this one
	case FOUR_CHAR_CODE('err '):
		return standardmodal(dialog,event);
	}

	GetPort(&current);
	if ((CGrafPtr)current!=GetDialogPort(dialog))
		BringToFront(GetDialogWindow(dialog));

	if (IsDialogEvent(event))
		if (DialogSelect(event,&dialog,&item))
			return item;
	return 0;
}

int dialogupdate(DialogRef dialog)
{
	Rect		bounds;
	Handle		handle;
	GrafPtr		oldport;
	short		index, type;

	if (!dialog)
		return -1;

	GetPort(&oldport);
	SetPortDialogPort(dialog);

	DrawDialog(dialog);
	if ((index=GetDialogDefaultItem(dialog))>0)
	{
		GetDialogItem(dialog,index,&type,&handle,&bounds);
		PenSize(3,3);
		bounds.top-= 4, bounds.left-= 4, bounds.bottom+= 4, bounds.right+= 4;
		FrameRoundRect(&bounds,16,16);
		PenSize(1,1);
	}
	MacSetPort(oldport);

	return 0;
}

int standardmodal(DialogRef dialog,
	EventRecord *event)
{
	if (!stdfilter)
		if (GetStdFilterProc(&stdfilter)!=noErr)
			stdfilter = NULL;
	if (stdfilter)
	{
		DialogItemIndex		item;

		if (InvokeModalFilterUPP(dialog,event,&item,stdfilter))
			return item;
	}
	return 0;
}

int animatebuttonitem(DialogRef dialog,
	int item)
{
	Rect		bounds;
	Handle		handle;
	short		type, last, next;

	GetDialogItem(dialog,item,&type,&handle,&bounds);
	if (!handle || (type&~kItemDisableBit)!=kButtonDialogItem)
		return -1;
	HiliteControl((ControlRef)handle,1);

	/* extend from 16 bits in case of overflow (do 32 bits and wait 3 years?) */
	next = 6+TickCount(); /* 6/60 sec. */
	do
		last = TickCount();
	while((last-=next)<0);
	HiliteControl((ControlRef)handle,0);

	next+= 2; /* 2/60 sec. */
	do
		last = TickCount();
	while((last-=next)<0);

	return 1;
}

static int dialogeventloop(DialogRef dialog,
	int type)
{
	EventRecord	event;
	int		hit = -1;

	do
	{
		WaitNextEvent(everyEvent,&event,1,NULL);
		if ((hit=handleevent(&event,dialog))!=0)
			continue;
		/* probably a key hit (return -> default to modal) */
		if (event.what==nullEvent) /* for text cursor to flash */
			goto normal;
		if (event.what==keyDown || event.what==autoKey)
		{
			char		charcode = event.message&charCodeMask;

			switch(type)
			{
			case dialogid_accept:
				if (charcode=='a')
					{hit = std_ok; goto anim;}
				break;
			case dialogid_confirm:
				if (charcode=='y')
					{hit = confirm_yes; goto anim;}
				else if (charcode=='n')
					{hit = confirm_no; goto anim;}
				break;
			anim:	animatebuttonitem(dialog,hit);
				goto end;

			default:
				if (charcode!=kEscapeCharCode
						&& charcode!=kReturnCharCode
						&& charcode!=kEnterCharCode)
				normal:	if ((unsigned)(hit=dialogselect(dialog,&event))+1>1)
						return hit;
			}
		}
		return standardmodal(dialog,&event);
	} while(hit<0 && hit!=-2);
end:	return hit;
}

static int valuedialog_parse(DialogRef dialog,
	int item,
	double *valueitems)
{
	double		current;
	char		*str;
	const char	*p;

	if ((str=getdialogtextbox(dialog,item))==NULL)
		return -1;
	p = parse_number(str,strlen(str)+str,&current,true);
	free(str);

	if (!p || (valueitems && (current<valueitems[1] || current>valueitems[2])))
		return 0;
	*valueitems = current;

	return 1;
}

double valuedialog(const uint8 *title,
	double current,
	double lower,
	double higher,
	boolean real)
{
	char		str[128], *strp[2];
	double		items[2];

	/* pin */
	if (current<lower)
		current = lower;
	else if (current>higher)
		current = higher;

	sprintf(str,real ? (current==(int)current ? "%.1f":"%.15g"):"%.f", current);

	*items = current;
	items[1] = lower;
	items[2] = higher;
	*strp = str;
	*(double**)&strp[1] = items;

	if (showdialog(dialogid_value,std_ok,title,(char*)items,NULL,NULL,strp)==std_ok)
	{
		if (*strp && *strp!=str)
			free(*strp);
		return *items;
	}

	return INT_NO_VALUE;
}

DialogRef getnewdialog(int dialogid,
	int defaultitem,
	WindowRef frontwindow)
{
	DialogRef	dialog;
	Ptr		storage;

	if ((storage=NewPtr(256))==NULL)
		return NULL;

	((int32*)storage)[56] = ((int32*)storage)[57] = ((int32*)storage)[58] =
		((int32*)storage)[59] = ((int32*)storage)[60] = ((int32*)storage)[61] =
		((int32*)storage)[62] = 0;
	((int32*)storage)[63] = defaultitem;

	if ((dialog=GetNewDialog(dialogid,storage,frontwindow))==NULL)
		DisposePtr(storage);

	return dialog;
}

short showdialog(short dialogid,
	short defaultitem,
	const uint8 *title,
	const char *caption,
	const uint8 *parameter,
	uint32 *controlbits,
	char **defaulttext)
{
	DialogRef	dialog = NULL;
	GrafPtr		oldport;
	const uint8	*p1 = !parameter ? "\p":parameter;
	int		textbox[2];
	short		index = -1;

	*textbox = textbox[1] = 2;

	GetPort(&oldport);

	ParamText(p1,"\p","\p","\p");

	if ((dialog=getnewdialog(dialogid,defaultitem,(WindowRef)-1))==NULL)
	{
		if (defaulttext)
			*defaulttext = NULL;
		goto close;
	}

	if (caption && dialogid!=dialogid_value)
		textbox[1] = 1+(*textbox=setdialogtextbox(dialog,-*textbox,
			caption,strlen(caption),false));
	if (defaulttext)
	{
		textbox[1] = setdialogtextbox(dialog,-textbox[1],
			*defaulttext,strlen(*defaulttext),false);
		*defaulttext = NULL;
	}
	if (controlbits)
	{
		uint32		bits = *controlbits;
		short		selected;

		for(selected=2; selected<=16+1; ++selected,bits>>=2)
			setbinaryitem(dialog,selected,bits&3);
	}

	ParamText(p1,"\p","\p","\p"); /* make sure ^x is filled */

	if (title)
		SetWTitle(GetDialogWindow(dialog),title);

	MacShowWindow(GetDialogWindow(dialog));
	SetPortDialogPort(dialog);

	if (defaultitem>0)
		SetDialogDefaultItem(dialog,defaultitem);
	if (getcontrolitem(dialog,std_cancel,NULL,NULL,NULL,NULL)>=0)
		SetDialogCancelItem(dialog,std_cancel);
	SetDialogTracksCursor(dialog,true);

/*	if (!modal)*/
		reenablemenus(false);

	do
	{
		if (index>0) /* handle checkboxes, etc. */
			if (setbinaryitem(dialog,index,INT_NO_VALUE)==kButtonDialogItem
					&& index==2)
				break;
		if ((index=dialogeventloop(dialog,dialogid))==std_ok
				&& dialogid==dialogid_value)
			if (!valuedialog_parse(dialog,textbox[1],((double**)defaulttext)[1]))
				{SysBeep(32); index = 0;}
	} while(index!=-2 && (uint16)(index-1)>1); /* up to three buttons */

/*	if (!modal)*/
		reenablemenus(true);

	SetDialogTracksCursor(dialog,false);

	if (index<-1)
		index = -1;
	else
	{
		if (defaulttext && textbox[1]!=-1)
			*defaulttext = getdialogtextbox(dialog,textbox[1]);

		if (controlbits)
		{
			uint32		bits = 0;
			short		selected;

			for(selected=16+1; selected>=2; --selected,bits<<=2)
				bits|= getbinaryitem(dialog,selected)&3;
			*controlbits = bits;
		}
	}
close:	if (dialog)
	{
		if (*(MenuRef*)&((int32*)dialog)[56])
		{
			MacDeleteMenu(popupmenuid_myfilter);
			DisposeMenu(*(MenuRef*)&((int32*)dialog)[56]);
		}
		DisposeDialog(dialog);
		InitCursor();
	}

	ParamText("\p","\p","\p","\p"); /* reset */
	SetPort(oldport);

	return index;
}

/* ---- getfilepath main code */

typedef struct gfpdata/*(32+)*/
{
	const uint8	*title; /* whenever set, the dialog window title is set */
	const uint8	*select; /* whenever set, the default button title is set */
	const char	*filter; /* string of the selected; NULL at initialization */
	const char	*filterlist;
	uint16		*filterx;
	FSSpec		*selected;
	uint32		flags; /* gfpf flags passed to getfilepath */
	gfptype		type;
	int		hilite;
} gfpdata;

static int setgfpfilter(gfpdata *data,
	int index)
{
	const char	*filter;

	if ((filter=data->filterlist)==NULL || !*(uint16*)filter || index<=0)
		return -1;
	do
	{
		filter+= 1+strlen(filter);
		if (!--index)
			if (data->filter==filter)
				return 1;
			else
				{data->filter = filter; return 2;}
	} while(*(filter+=1+strlen(filter))!=0);

	return 0;
}

static int buildgfpfiltermenu(DialogRef dialog,
	int item,
	const char *filterlist)
{
	Str255		name;
	MenuRef		menu;
	const char	*filter;
	int		maxwidth, count = 0;

	if ((filter=filterlist)==NULL || !dialog || !*(uint16*)filter || item<=0)
		return 0;
	else
	{
		Rect		bounds;
		ControlRef	control;
		short		type;

		GetDialogItem(dialog,item,&type,(Handle*)&control,&bounds);
		if (!control || !(type&kControlDialogItem)
				|| (menu=GetControlPopupMenuHandle(control))==NULL)
			return 0;
		if ((maxwidth=bounds.right-bounds.left)<=0)
			maxwidth = 160;
		/* add excess and indent */
		maxwidth+= 64+40; /* was 64+32; 2010.04.08*/
	}

	DeleteMenuItems(menu,1,-1UL);
	do
	{
		long	length = strlen(filter);

		memcpy(1+name,filter,*name=(length>255 ? 255:length));
		if (length>255)
			name[255] = ''; /* elipsis (technically truncated) */
		if (name[1]=='-') /* system takes starting hypen as divider... */
			*(uint16*)name = ''; /* dot hyphen as substitute */
		TruncString(maxwidth,name,truncEnd);
		AppendMenu(menu,"\p "); //add new via 'space' template, then set name
		SetMenuItemText(menu,++count,name);

		filter+= length;
		filter+= 1+strlen(filter); //skip filter data
	} while(*(filter+=1+strlen(filter))!=0); //empty title indicates end

	return count;
}

static int patternlistmatch(const char *pattern,
	const void *match,
	int length)
{
	const char	*matchend = length+(const char*)match;
	const char	*pe;
	uint8		scan;

	if (!pattern)
		return -1;
	if (length<=0)
		return 0;
	do
	{
		const char	*ps = pe = pattern;

		if ((scan=*pe++)==0 || scan==';')
			continue; /* zero length; skip */
		while((scan=*pe++)!=0 && scan!=';') ;
		pattern = pe;
		--pe;

		if (*(uint32*)ps!=FOUR_CHAR_CODE('*.*\0')
				&& *(uint32*)ps!=FOUR_CHAR_CODE('*.*;'))
		{ /* match contained pattern */
			const char	*me = matchend;

			do
			{
				uint8		c = *--pe;

				if ((c+=-'*')==0)
				{ /* any string of characters; take next char, reverse scan match */
					if (pe<=ps)
						return 1; /* any string to start; nothing left */
					c = *--pe;
					do
						if (me<=(const char*)match)
							{length = 0; goto next;} /* not found */
					while(*--((uint8*)me)!=c);
				} else if ((c+='*'-'?')==0) /* single skip */
				{
					if (--me<(const char*)match)
						{length = 0; goto next;}
				} else
				{ /* otherwise, literal (any other special characters?) */
					if (--me<(const char*)match
							|| tolower(c+='?')!=tolower(*(uint8*)me))
						{length = 0; goto next;}
				}
			} while(pe>ps);
		}
		return 1;
	next:	;
	} while(scan!=0);

	return 0;
}

static pascal Boolean myobjfilteryd(CInfoPBPtr pb,
	void *yd)
{
	gfpdata		*mydata;

	if ((mydata=yd)==NULL || !pb || !pb->dirInfo.ioNamePtr)
		return false;
	/* handle directories, show invisible directories */
	if (pb->dirInfo.ioFlAttrib&kioFlAttribDirMask)
		return mydata->flags&gfpf_showinvisibledirs ?
			false:(pb->dirInfo.ioDrUsrWds.frFlags&kIsInvisible)!=0;
	if (mydata->type==gfp_getdirectory)
		return true;
	/* handle show invisible files */
	if (!(mydata->flags&gfpf_showinvisiblefiles)
			&& pb->hFileInfo.ioFlFndrInfo.fdFlags&kIsInvisible)
		return true;
	/* handle hide read-only */
	if ((mydata->flags&gfpf_hidereadonly)
			&& (pb->hFileInfo.ioFlAttrib&(kioFlAttribLockedMask
			|kioFlAttribFileOpenMask)))
		return true;
	/* handle file must exist */
	if ((mydata->flags&gfpf_filemustexist)
			&& (pb->hFileInfo.ioFlFndrInfo.fdFlags&kIsStationery))
		return true;
	/* finally, match name with pattern */
	return patternlistmatch(mydata->filter,
		1+pb->dirInfo.ioNamePtr,*pb->dirInfo.ioNamePtr)<=0;
}

//static gfpdata *mynonyddata;

//static pascal Boolean myobjfilter(CInfoPBPtr pb)
//{
//	return myobjfilteryd(pb,mynonyddata);
//}

static boolean isfolder(const FSSpec *spec)
{
	CInfoPBRec	cinfo;

	cinfo.dirInfo.ioNamePtr = (UInt8*)spec->name;
	cinfo.dirInfo.ioFDirIndex = 0;
	cinfo.dirInfo.ioFlAttrib = 0;
	cinfo.dirInfo.ioACUser = 0;
	cinfo.dirInfo.ioVRefNum = spec->vRefNum;
	cinfo.dirInfo.ioDrParID = cinfo.dirInfo.ioDrDirID = spec->parID;

	if (PBGetCatInfoSync(&cinfo)!=noErr)
		return false;
	return (cinfo.dirInfo.ioFlAttrib&kioFlAttribDirMask)!=0;
}

static void customgfpupdate(DialogRef dialog,
	gfpdata *mydata)
{
	if (mydata->title)
	{
		SetWTitle(GetDialogWindow(dialog),mydata->title);
		mydata->title = NULL;
	}
	if (mydata->select)
	{
		setcontrolitem(dialog,sfItemOpenButton,INT_NO_VALUE,INT_NO_VALUE,
			INT_NO_VALUE,INT_NO_VALUE,mydata->select);
		mydata->select = NULL;
	}
	if (!mydata->filter) /*&& mydata->type!=gfp_putfile*/
	{
		int	index = !mydata->filterx ? 1:*mydata->filterx+1;

		setcontrolitem(dialog,gfpitem_filter,
			index,
			1,
			buildgfpfiltermenu(dialog,gfpitem_filter,mydata->filterlist),
			INT_NO_VALUE,
			NULL);
		setgfpfilter(mydata,index);
	}
}

static pascal Boolean mymodalfilteryd(DialogRef dialog,
	EventRecord *event,
	DialogItemIndex *hit,
	void *mydata)
{
	if (mydata)
		customgfpupdate(dialog,(gfpdata*)mydata);
	if (!stdfilter)
		if (GetStdFilterProc(&stdfilter)!=noErr)
			stdfilter = NULL;
	if ((*hit=handleevent(event,dialog))==0)
	{
		if (stdfilter)
			return InvokeModalFilterUPP(dialog,event,hit,stdfilter);
	} else if (*hit<0)
	{
		if (*hit==-2) /* map -1,-2 to false,true */
			{*hit = sfItemCancelButton; return true;}
		if (*hit<-2) /* likely dragged window bar; do nothing */
			{*hit = sfHookNullEvent; return true;} /*sfItemPromptStaticText*/
	} else
		return true;

	return false;
}

static pascal short myhookyd(short hit,
	DialogRef dialog,
	void *yd)
{
	gfpdata		*mydata;

	if ((mydata=yd)==NULL)
		return hit;
	if (hit!=sfHookNullEvent && hit>0)
		mydata->flags|= 0x80000000UL;
	else
	{
		int	hilite = mydata->hilite;

		getcontrolitem(dialog,sfItemOpenButton,NULL,NULL,&mydata->hilite,NULL);
		if (mydata->hilite!=hilite)
			mydata->flags|= 0x80000000UL;
	}

	if (hit==sfHookFirstCall)
	{
		if (mydata->type<=gfp_getfile)
			HideDialogItem(dialog,gfpitem_select);
		if (mydata->type==gfp_getdirectory) /*|| mydata->type==gfp_putfile*/
		{
			if (mydata->filterlist)
				HideDialogItem(dialog,gfpitem_filter);
			mydata->filterlist = NULL;
		}
	} else if (hit==gfpitem_filter)
	{
		int	index = getcontrolitem(dialog,gfpitem_filter,NULL,NULL,NULL,NULL);

		if (mydata->filterx)
			*mydata->filterx = index-1;
		if (mydata->type!=gfp_putfile && setgfpfilter(mydata,index)>1)
			hit = sfHookRebuildList;
	} else if (hit==gfpitem_select)
		return sfItemOpenButton;

	if (mydata->type>gfp_getfile && mydata->flags&0x80000000UL)
	{
		mydata->flags&= 0x7FFFFFFF;
		setcontrolitem(dialog,gfpitem_select,INT_NO_VALUE,INT_NO_VALUE,
			INT_NO_VALUE,isfolder(mydata->selected) ? 0:-1,
			*mydata->selected->name!=0 ? "\pSelect":"\pCurrent");
	}

	return hit;
}

#if TARGET_API_MAC_CARBON
static OSErr getfirstaefsspec(AEDescList *list,
	FSSpec *spec)
{
	long		index, count;
	OSErr		err;

	if ((err=AECountItems(list,&count))!=noErr)
		return err;
	for(index=1; index<=count; ++index) 
	{
		Size		actual;
		DescType	type;
		AEKeyword	keyword;

		if ((err=AEGetNthPtr(list,index,typeFSS,&keyword,&type,
				(Ptr)spec,sizeof(FSSpec),&actual))==noErr)
			break;
	}

	return err;
}
#endif

static StandardFileReply	lastreply;

char		*lastpath = NULL;

const char *getlastfilepath(short *vrefnum,
	long *parid,
	uint8 *name)
{
	if (vrefnum)
		*vrefnum = lastreply.sfFile.vRefNum;
	if (parid)
		*parid = lastreply.sfFile.parID;
	if (name)
	{
		unsigned	z = 1+*lastreply.sfFile.name;

		memcpy(name,lastreply.sfFile.name,z>64 ? 64:z);
	}

	return lastpath;
}

/* added; 2010.05.21 */
const char *setlastfilepath(short vrefnum,
	long parid,
	const uint8 *name)
{
	char		*spath;
	int		pathlength = 0;

	if (!name)
		name = "\p";
	if (FSMakeFSSpec(vrefnum,parid,name,&lastreply.sfFile)!=noErr)
		return NULL;
	if ((spath=makepathforfsspec(&lastreply.sfFile,&pathlength))==NULL)
		return NULL;
	if (lastpath) /* pass all zeros for parameters to clear w/o dialog */
		lastpath = realloc(lastpath,0);
	return lastpath=spath;
}

char *getfilepath(gfptype type,
	const uint8 *title,
	const uint8 *select,
	const char *initdirstr,
	const char *initnamestr,
	const char *filterlist,
	uint16 *nameoffset,
	uint16 *filterindex, /* 2010.04.19 */
	uint32 flags)
{
#if TARGET_API_MAC_CARBON
 #pragma unused(nameoffset,pathlength)
	NavReplyRecord		reply;
	NavDialogOptions	dlgoptions;
	NavObjectFilterUPP	objfilterupp = NULL;
	NavEventUPP		eventupp = NULL;
	AEDesc			defaultdir;
#else
	uint8		name[256];
	Point		where;

	ModalFilterYDUPP	modalfilteryd = NULL;
	FileFilterYDUPP		objfilteryd = NULL;
	DlgHookYDUPP		hookyd = NULL;
#endif
	gfpdata		mydata;
	int32		initialdir[3];
	int		pathlength = 0;

	if (lastpath) /* pass all zeros for parameters to clear w/o dialog */
		lastpath = realloc(lastpath,0);

	mydata.type = type;
	mydata.flags = flags|0x80000000UL;
	mydata.title = title;
	mydata.select = select;
	mydata.filterlist = filterlist;
	mydata.filter = NULL;
	mydata.filterx = filterindex;
	mydata.hilite = 0;

	makedirspec(initialdir,initdirstr);

	ShowMenuBar(); /* StandardFile sometimes crashes for some reason when hidden */

#if TARGET_API_MAC_CARBON
	/*objfilterupp = NewNavObjectFilterUPP(myobjfilter);*/

	memset(&reply,0,sizeof(dlgoptions));
	memset(&dlgoptions,0,sizeof(dlgoptions));

	*(SInt32*)&dlgoptions.location = -1;
	cpstrcpy(dlgoptions.clientName,CPROGRAM_NAME,sizeof(dlgoptions.clientName));
	memcpy(dlgoptions.windowTitle,title,*title+1);
	cpstrcpy(dlgoptions.savedFileName,initnamestr,sizeof(dlgoptions.savedFileName));
	dlgoptions.dialogOptionFlags = kNavDontAutoTranslate|kNavDontAddTranslateItems
		|kNavAllFilesInPopup;

	initialdir[2] = 0; /*was [3] -- tsk, tsk.; 2010.04.08 */
#if !BIG_ENDIAN
	((int16*)initialdir)[1] = (int16)*initialdir; //for the BE +2 displacement
#endif
	AECreateDesc(typeFSS,2+(const uint8*)initialdir,sizeof(FSSpec),&defaultdir);

	switch(type)
	{
	case gfp_putfile:
		NavPutFile(&defaultdir,&reply,&dlgoptions,eventupp,
			'****','****',&mydata);
		break;
	case gfp_getfile:
		NavGetFile(&defaultdir,&reply,&dlgoptions,eventupp,NULL,
			objfilterupp,NULL,&mydata);
		break;
	case gfp_getdirectory:
		NavChooseFolder(&defaultdir,&reply,&dlgoptions,eventupp,
			objfilterupp,&mydata);
		break;
	case gfp_getany:
		NavChooseObject(&defaultdir,&reply,&dlgoptions,eventupp,
			objfilterupp,&mydata);
		break;
	}

	/*DisposeNavObjectFilterUPP(objfilterupp);*/

	AEDisposeDesc(&defaultdir);

	if ((lastreply.sfGood=reply.validRecord)==true)
		if (getfirstaefsspec(&reply.selection,&lastreply.sfFile)!=noErr)
		{
			FCEU_PrintError("Getting FSSpec for Navigation failed.");
			lastreply.sfGood = false;
		}
#else
	cpstrcpy(name,initnamestr,sizeof(name));

	objfilteryd = NewFileFilterYDUPP(myobjfilteryd);
	modalfilteryd = NewModalFilterYDUPP(mymodalfilteryd);
	hookyd = NewDlgHookYDUPP(myhookyd);

	*(SInt32*)&where = -1;
	mydata.selected = &lastreply.sfFile;
	*((volatile SInt16*)0x214) = -(lastreply.sfFile.vRefNum=*initialdir); /* SFSaveDisk */
	*((volatile SInt32*)0x398) = lastreply.sfFile.parID = initialdir[1]; /* CurDirStore */
	*lastreply.sfFile.name = 0;

	switch(type)
	{
	case gfp_putfile:
		//if (sysVer<0x71)
		//	StandardPutFile(NULL,name,&lastreply);
		//else
		CustomPutFile(NULL,name,&lastreply,dialogid_customput,
			where,hookyd,modalfilteryd,NULL,NULL,&mydata);
		break;
	case gfp_getfile:
	case gfp_getdirectory:
	case gfp_getany:
		//if (sysVer<0x71)
		//{
		//	if (!nonyd_objfilter)
		//		nonyd_objfilter = NewFileFilterUPP(myobjfilter);
		//	mydata.filter = mydata.filterlist; /* first item */
		//	mynonyddata = &mydata;
		//	StandardGetFile(nonyd_objfilter,-1,NULL,&lastreply);
		//} else
		CustomGetFile(objfilteryd,-1,NULL,&lastreply,dialogid_customget,
			where,hookyd,modalfilteryd,NULL,NULL,&mydata);
		break;

	default:
		lastreply.sfGood = false; /* fake fail */
	}

	DisposeModalFilterYDUPP(modalfilteryd);
	DisposeFileFilterYDUPP(objfilteryd);
	DisposeDlgHookYDUPP(hookyd);
#endif
	lastpath = lastreply.sfGood ?
		makepathforfsspec(&lastreply.sfFile,&pathlength):NULL;

	if (nameoffset)
		*nameoffset = pathlength-*lastreply.sfFile.name;
#if TARGET_API_MAC_CARBON
	NavDisposeReply(&reply);
#endif
	return strndup(lastpath,-1);
}
