/*
 movie.c

 * MacFCEU 0.7a3
 *  This still does _not_ support the newer 0.98.15+ extended formats.
 *  Just the state file format, directly followed by input data is written.
 *  (Test against the version number in the state header.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "driver.h"
#include "FCEU.h"
#include "input.h"
#include "movie.h"
#include "state.h"
#include "general.h"
#include "video.h"

#ifdef env68k
 #pragma segment files
#endif

#define ENCODE_DUMMY		0x80

static int	current = 0;		/* <0 for playback; slot index is `abs(current)-1` */
static FILE	*slots[11] = {0};	/* #10 is the external one; 2010.04.18 */
static uint8	joop[4];		/* previous joystick data */
static uint32	framets;		/* our counter (regardless of Power, Reset...) */

/* Cache variables used for playback. */
static uint32	nextts, nextxd;		/* timestamp, extended data */
static int	nextcd;			/* command or input data (8 bit); <0 for none */

static int	CurrentMovie = 0;
static uint8	MovieStatus[11];
static uint8	MovieShow = 0;

static void DoEncode(uint8 cmd,
	uint32 xdata)
{
	FILE		*fp;
	uint8		data = cmd;

	if ((fp=slots[current-1])==0)
		return;
	if (framets>65535)
		data|= 3<<5; /* 24 bit */
	else if (framets>255)
		data|= 2<<5; /* 16 bit */
	else if (framets>0)
		data|= 1<<5; /* 8 bit */
	fputc(data,fp);
	if (framets)
		do
			fputc(framets&0xFF,fp);
		while((framets>>=8)!=0);
	if (cmd>(0x1B|ENCODE_DUMMY))
	{ /* extended data (>=16 bits); 2010.04.16 */
		fputc(xdata&0xFF,fp);
		fputc((xdata>>=8)&0xFF,fp);
		if (cmd&2) /* 32 bits */
			fputc((xdata>>=8)&0xFF,fp),
			fputc(xdata>>=8,fp);
	}
}

static void DoExtended(uint8 w,
	uint32 d)
{
(void)w; (void)d;
}

int FCEUMOV_Status(void)
{
	return current;
}

static void DisplayMovieSelect(void)
{
	int		x;

	if ((x=current)<0)
		FCEU_DispMessage(x<-10 ? "-playing movie-":"-playing movie %d-",
			~x);
	else if (x>0)
		FCEU_DispMessage(x>10 ? "-recording movie-":"-recording movie %d-",
			x-1);
	else
		FCEU_DispMessage("-select movie-");
}

static void StopPlayback(void)
{
	FILE		*fp;

	if ((fp=slots[~current])!=0)
	{
		fclose(fp);
		FCEU_DispMessage("Movie playback stopped.");
	}
	current = 0;
}

static void StopRecording(void)
{
//	if (current<=0)
//		return;
	/* Write a dummy timestamp value so that the movie will keep
		"playing" after user input has stopped.
	*/
	DoEncode(ENCODE_DUMMY,0);
	--current;
	fclose(slots[current]), slots[current] = 0;
	MovieStatus[current] = 1;
	current = 0;
	FCEU_DispMessage("Movie recording stopped.");
}

int FCEUMOV_Stop(void)
{
	int		v;

	if ((v=current)<0)
		StopPlayback();
	else if (v>0)
		StopRecording();

	return v;
}

void FCEUI_SaveMovie(const char *fname)
{
	FILE		*fp;
	char		*fn;

	if (FCEUMOV_Stop() && !fname)
		return;

	if (fname)
		fp = FCEUD_UTF8fopen(fname,"wb"), current = 10;
	else if ((fn=FCEU_MakeFName(FCEUMKF_MOVIE,current=CurrentMovie,0))!=0)
	{
		fp = FCEUD_UTF8fopen(fn,"wb");
		free(fn);
	} else
		return;

	if (!fp)
		return;

	/* prev. passed 1 for movie -- not ready!; 2010.05.16 */
	if (FCEUSS_SaveFP(fp,0)<0)
		return;
	if (fseek(fp,0,SEEK_END)<0)
	{
		FCEU_DispMessage("Movie recording failed.");
		return;
	}

	slots[current++] = fp;
	memset(joop,0,sizeof(joop));

	framets = 0;
	DisplayMovieSelect();
}

void FCEUI_LoadMovie(const char *fname)
{
	FILE		*fp;
	char		*fn;

	if (FCEUMOV_Stop() && !fname)
		return;

	if (fname)
		fp = FCEUD_UTF8fopen(fname,"rb"), current = 10;
	else if ((fn=FCEU_MakeFName(FCEUMKF_MOVIE,current=CurrentMovie,0))!=0)
	{
		fp = FCEUD_UTF8fopen(fn,"rb");
		free(fn);
	} else
		return;

	if (!fp || FCEUSS_LoadFP(fp)<=0) /* load the starting state */
	{ /* none or bad */
		FCEU_DispMessage("Movie loading failed.");
		if (fp)
			fclose(fp);
		/* write (record) after fail bug fix; 2011.03.28 */
		slots[current] = 0;
		MovieStatus[current] = 0;
		current = 0;
		return;
	}

	slots[current] = fp;
	MovieStatus[current] = 1;

	memset(joop,0,sizeof(joop));
	current = ~current; /* playing (negative value) */
	framets = 0;
	nextts = nextxd = 0;
	nextcd = -1;
	DisplayMovieSelect();
}

/* this is the function to read/write from the movie file */
void FCEUMOV_AddJoy(uint8 *js)
{
	int		x, t;

	if (!current)
		return;	/* Neither playing nor recording. */

	if (current<0) /* Playback */
	{
		while(nextts==framets)
		{
			FILE	*fp = slots[~current];
			int	sz;
			uint8	d;

			if (nextcd>=0)
				if (nextcd&ENCODE_DUMMY)
					if (nextcd>(0x1B|ENCODE_DUMMY))
						DoExtended(nextcd&1,nextxd);
					else
						FCEU_DoSimpleCommand(nextcd&0x1F);
				else
					joop[nextcd>>3&3]^= 1<<(nextcd&7);

			if ((sz=fgetc(fp))<0)
			{ /* file error or EOF */
			error:	StopPlayback();
				return;
			}
			d = sz;

			nextts = 0;
			sz>>= 5;
			if (sz&=3)
			{
				x = sz, t = 0;
				do
					nextts|= fgetc(fp)<<t;
				while(t+=8,--x!=0);

	// This fixes a bug in movies recorded before version 0.98.11
	// It's probably not necessary, but it may keep away CRAZY PEOPLE who recorded
	// movies on <= 0.98.10 and don't work on playback.
				if (!nextts)
					if (sz==1)
						nextts|= fgetc(fp)<<8;
					else if (sz==2)
						nextts|= fgetc(fp)<<16;
			}

			framets = 0;
			if ((nextcd=d)>(0x1B|ENCODE_DUMMY))
			{ /* future (stupid) custom to MacFCEU 0.6.2+ */
				nextxd = t = fgetc(fp);
				nextxd|= (t|=fgetc(fp))<<8;
				if (d&2) /* 32 bits, otherwise 16 bits */
					nextxd|= (t|=fgetc(fp))<<16,
					nextxd|= (t|=fgetc(fp))<<24;
				if (t<0) /* any errors? */
					goto error;
			}
		}
		/* copy to emulated */
		js[0] = joop[0], js[1] = joop[1], js[2] = joop[2], js[3] = joop[3];
	} else	/* Recording */
	{
		for(t=x=0; x<4; ++x,t+=8)
		{
			if (js[x]==joop[x])
			{
				if (framets==(1<<24)-1) /* if inevitable overflow... */
					DoEncode(ENCODE_DUMMY,0); /* dummy update. */
				continue;
			}
			/* unrolled */
		#define _joop(x,y) {if((js[x]^joop[x])&(1<<y)) DoEncode(t|y,0);}
			_joop(x,0);
			_joop(x,1);
			_joop(x,2);
			_joop(x,3);
			_joop(x,4);
			_joop(x,5);
			_joop(x,6);
			_joop(x,7);
		#undef _joop
			joop[x] = js[x];
		}
		/* and here's where we encode the future stupid custom (FSC) stuff */
	}
	++framets;
}

/* was `void FCEUMOV_AddCommand(int cmd)` */
void FCEUMOV_EncodeCmd(uint8 cmd,
	uint32 xdata)
{
	if (current<=0)
		return;
	DoEncode((cmd&=0x1F)|ENCODE_DUMMY,xdata);
}

void FCEUMOV_CheckMovies(void)
{
	FILE		*st = 0;
	char		*fn;
	int		ssel = 0;

	do
	{
		if ((fn=FCEU_MakeFName(FCEUMKF_MOVIE,ssel,0))==0)
			continue;
		st = FCEUD_UTF8fopen(fn,"rb");
		free(fn);
		if (st)
			MovieStatus[ssel] = 1, fclose(st);
		else
			MovieStatus[ssel] = 0;
	} while(++ssel<10);
}

void FCEUI_SelectMovie(int w)
{
	if (w<0 || FCEUGameInfo->type==GIT_NSF) /* added NSF check; 2010.04.25 */
		{MovieShow = 0; return;}
	FCEUI_SelectState(-1);

	CurrentMovie = w;
	MovieShow = 180; /* FCEU_DrawMovies is called anyway, so this signals to render */
	DisplayMovieSelect();
}

void FCEU_DrawMovies(uint8 *XBuf)
{
	if (!XBuf)
		MovieShow = 0;
	if (!MovieShow || (unsigned)CurrentMovie>9)
		return;
	FCEU_DrawNumberRow(XBuf,MovieStatus,CurrentMovie);
	--MovieShow;
}
