/*
 palette.c

 * MacFCEU 0.9
 *
 * Copyright notice for the original file:
 *  palette.c Copyright (C) 2002,2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>

#include "driver.h"
#include "FCEU.h"
#include "general.h"
#include "palette.h"
#include "palettes/palettes.h" /* keep this here as to not intefere w/ <Palettes.h> */

#include "ftrig.h"
#include "video.h" //for rowsize

#ifdef env68k
 #pragma segment shell
#endif

/* ---- constants */

/* product and end-curve - not true emphasis, but will do for the development phase */
static const uint16	emphasis_lut[3*7] =
{ /* adjusted table, based on some actual results, below */
  /* 8 bits in, 8 bits out, it's kind of pointless to go >10 bits fractional precision */
	1.109*1024,0.837*1024,0.911*1024,
	0.808*1024,1.028*1024,0.653*1024,
	0.878*1024,0.837*1024,0.658*1024,
	0.843*1024,0.912*1024,1.219*1024,
	0.889*1024,0.769*1024,0.819*1024,
	0.723*1024,0.809*1024,0.843*1024,
	0.743*1024,0.743*1024,0.743*1024
};
/*
 Interpolated 1/16 against next row down.
 Rounding to nearest, some down (tolerance <0.7%).

1.000,1.000,1.000
1.116,0.853,0.914
0.839,1.035,0.688
0.888,0.853,0.692
0.853,0.920,1.225
0.900,0.794,0.990
0.747,0.895,0.853
0.808,0.808,0.808

1.002,1.002,1.002
1.107,0.835,0.911
0.806,1.026,0.650
0.876,0.835,0.656
0.841,0.910,1.217
0.887,0.766,0.806
0.720,0.802,0.841
0.771,0.771,0.771

 Only 9 non-black/white grays in the entire 2C07 PPU palette.

.000,.188,.271,.310,.400,.522,.557,.678,.722,.808,1.00
0x00,0x30,0x45,0x4F,0x66,0x85,0x8E,0xAD,0xB8,0xCE,0xFF
*/

/* ---- globals */

uint8		pale = 0;
static uint8	pali = 0;
static uint8	prevemph = 0;

static uint8	ntsccol = 0;
/* these corrected via trial and error */
static int	ntschue = 66; //72
static int	ntscsat = 51.5*2; //(46+10)*2
static int	ntscbrt = 64; /* added; 2010.05.09 */

/* These are dynamically filled/generated palettes: */
pal		palettei[64]; // Loaded palette for an individual game (based on its name).
pal		palettec[64]; // Custom "global" palette.
pal		paletten[64]; // Mathematically generated NTSC palette.

pal		aggr_pal[255]; // aggregate screen colors
uint8		aggr_cc;	// aggregate screen color count
uint8		epal_inv[8*256]; // emphasis-adjusted inverse table to aggregate colors
static uint16	palb[3*64];	// cached intensified colors

const pal	*palo;

static pal	*palpoint[8] =
{
	palette,
	rp2c04001,
	rp2c04002,
	rp2c04003,
	rp2c05004,
};

/* ---- code */

void FCEUI_SetPaletteArray(const uint8 *p)
{
	if (!p)
		palpoint[0] = palette;
	else
	{
		pal	*ent = palettec;
		int	c = 64;

		palpoint[0] = ent;
		do
			ent->r = *p++,
			ent->g = *p++,
			ent->b = *p++;
		while(++ent,--c);
	}
	FCEU_ResetPalette(-1);
}

void FCEUI_SetNTSC_HSB(int n,
	int hue,
	int sat,
	int brt)
{
	ntsccol = !!n;
	ntscsat = sat;
	ntschue = hue;
	ntscbrt = brt;
	FCEU_ResetPalette(1); //always recalculate here; 2009.01.29
}

int FCEUI_GetNTSC_HSB(int *hue,
	int *brt)
{
	*hue = ntschue;
	*brt = ntscbrt;
	return ntscsat;
}

#if 0
/* classic method */
void SetNESDeemph(uint8 d,
	int force)
{ /* implementations suggest 2C07/2C02 treat emphasis as a 1.0:0.75|0.85 form */
  /* the other 2C0x's apparently render a white-out... */
	static uint16	rtmul[7] =
		{32768*1.239,32768*0.794,32768*1.019,32768*0.905,32768*1.023,32768*0.741,32768*.75};
	static uint16	gtmul[7] =
		{32768*0.915,32768*1.086,32768*0.980,32768*1.026,32768*0.908,32768*0.987,32768*.75};
	static uint16	btmul[7] =
		{32768*0.743,32768*0.882,32768*0.653,32768*1.277,32768*0.979,32768*1.001,32768*.75};

	const pal	*s;
	uint32		r, g, b;
	int		x;

	if (palo==0)
		return;

	/* If it's not forced (only forced when the palette changes),
		don't waste cpu time if the same deemphasis bits are set as the last call.
	*/
	if (!force)
	{
		if (d==prevemph)
			return;
	} else
	{ /* Only set this when palette has changed. */
		r = rtmul[6];
		g = rtmul[6];
		b = rtmul[6];
		x = 0xC0, s = palo;
		do
		{
			uint32		m = s->r, n = s->g, o = s->b;

			if ((m=(m*r)>>15)>255)
				m = 255;
			if ((n=(n*g)>>15)>255)
				n = 255;
			if ((o=(o*b)>>15)>255)
				o = 255;
			FCEUD_SetPalette(x,m,n,o);
		} while(++s,++x<0x100);
	}

	if (d==0)
		return; //No emphasis

	r = rtmul[d-1];
	g = gtmul[d-1];
	b = btmul[d-1];
	x = 0x40, s = palo; //colors dependent on 0x40-0x7F will screw up
	do
	{
		uint32		m = s->r, n = s->g, o = s->b;

		if ((m=(m*r)>>15)>255)
			m = 255;
		if ((n=(n*g)>>15)>255)
			n = 255;
		if ((o=(o*b)>>15)>255)
			o = 255;
		FCEUD_SetPalette(x,m,n,o);
	} while(++s,++x<0x80);

	prevemph = d;
}
#endif

/* new method */
static int FindOrAddRGB(int x,
	int cr,
	int cg,
	int cb)
{
	pal		*cs;
	int		i, th, mmin, imin = PAL_BLACK, c, m, t, ci, gi, h, r, y;

	if (cr>255u)
		cr =~cr>>31&255;
	if (cg>255u)
		cg =~cg>>31&255;
	if (cb>255u)
		cb =~cb>>31&255;
	if (!(cr|cg|cb))
		return imin;
	cs = aggr_pal;
	if ((c=aggr_cc)<sizeof(unvpalette)/sizeof(unvpalette[0]))
		{i = c, cs+= i; goto set;}
	if (x<0)
	{
		if ((i=~x)>254)
			return imin;
		cs+= i;
		goto set;
	}
	/* threshold increases the fewer colors are available */
	if ((th=(c-96)>>4)<0)
		th = 0;
	i = 0, mmin = IMAX(mmin), ci = ((cg+cg+cr)*2u+cb)*2u+cr+cg; /* 5r/9g/2b */
	y = cr-cg>=0 && cg-cb>3;
	do
	{
		if (x<2 && i-PAL_YELLOW+0u<sizeof(unvpalette)/sizeof(unvpalette[0])-PAL_YELLOW)
			continue; /* not dark-adjusted */
		/* compute total difference for closest match */
		t = cs->r, gi = t*5u;
		if ((m=cr-t)<0)
			m =-m, m+= m>>2;
		h = r = m, t = cs->g, gi+= t*9u;
		if ((t=cg-t)<0)
			t =-t, t+= t>>2;
		if (t>h)
			h = t;
		m+= t, m+= m>>1;
		t = cs->b, gi+= t*2u;
		if ((t=cb-t)<0)
			t =-t;
		m+= t, t-= t>>2;
		if (t>h)
			h = t;
		if ((t=(gi-ci)>>1)<0)
			t =-t;
		if (c>136 && h<3+!r)
			m = 0, t>>= 2;
		t+= ((cs->r-cs->g>=0 && cs->g-cs->b>3)!=y)*32u;
		m+= t;
		if (m<mmin)
		{
			imin = i, mmin = m;
			if (mmin<=th && (x&1)!=i)
				return imin;
		}
	} while(++cs,++i<c);

	if (c>=255) /* out of colors */
		return imin;
set:	if (i>=c)
		aggr_cc = i+1;
	cs->r = cr; /* aggregate */
	cs->g = cg;
	cs->b = cb;
	FCEUD_SetPalette(i,cr,cg,cb);
	return i;
}

static void AddEmphRGB(int r,
	int i,
	int cr,
	int cg,
	int cb)
{
	static int16	last =-1;
	uint32		br, fl, sb;
	/* add the base color */
	sb = FindOrAddRGB(1,cr,cg,cb);
	if (r)
	{
		if (last==sb && sb!=PAL_WHITE && sb!=PAL_BLACK)
			sb+= 2;
		last = sb;
	} else
		last =-1;
	epal_inv[i] = epal_inv[i+64] = epal_inv[i+128] = epal_inv[i+192] = epal_inv[i] = sb;

	cr+= cr<<8, cg+= cg<<8, cb+= cb<<8;
	if (cr>cg)
		br = cr, fl = cg;
	else
		br = cg, fl = cr;
	if (cb>br)
		br = cb;
	/* there's a slight exponential scale to this... */
	sb = (br*(br+1)>>16)*br>>16;
	cr+= (cr*sb>>8)*5>>14, cg+= (cg*sb>>8)*5>>14, cb+= (cb*sb>>8)*5>>14;
	/* cache the intensified color */
	palb[i*3] = cr>>2;
	palb[i*3+1] = cg>>2;
	palb[i*3+2] = cb>>2;
	fl = br-fl;
	/* dark emphasis has a larger dark curve, especially on grays */
	if (br<0x9C00)
		cr-= cr>>6, cg-= cg>>6, cb-= cb>>6;
	if (br<0x8E80)
		cr-= cr>>5, cg-= cg>>5, cb-= cb>>5;
	if (fl<0x8000 && br<0x8100)
		cr-= cr>>5, cg-= cg>>5, cb-= cb>>5;
	if (fl<0x4000 && br<0x7380)
		cr-= cr>>5, cg-= cg>>5, cb-= cb>>5;
	if (fl<0x2000 && br<0x6600)
		cr-= cr>>5, cg-= cg>>5, cb-= cb>>5;
	if (fl<0x1000 && br<0x5880)
		cr-= cr>>5, cg-= cg>>5, cb-= cb>>5;
	if (fl<0x800 && br<0x5100)
		cr-= cr>>5, cg-= cg>>5, cb-= cb>>5;
	if (fl<0x400 && br<0x4980)
		cr-= cr>>5, cg-= cg>>5, cb-= cb>>5;

	/* dark emphasis added second */
	r = epal_inv[i]|1;
	epal_inv[i+7*256+0] = epal_inv[i+7*256+64] = epal_inv[i+7*256+128] = epal_inv[i+7*256+192] = r;
	FindOrAddRGB(~r,emphasis_lut[3*6]*cr>>18,
			emphasis_lut[3*6+1]*cg>>18,
			emphasis_lut[3*6+2]*cb>>18);
}

static void WritePalette(int redundant)
{
	const uint16	*cb;
	int		x;

	aggr_cc = 0;
	for(x=0; x<sizeof(unvpalette)/sizeof(unvpalette[0]); ++x)
		FindOrAddRGB(2,unvpalette[x].r,unvpalette[x].g,unvpalette[x].b);
	if (FCEUGameInfo->type==GIT_NSF)
		return;
	x = 0;
	do
		AddEmphRGB(redundant,x,palo[x].r,palo[x].g,palo[x].b);
	while(++x<64);
	/* fill in the alternate emphasis colors last */
	/* interleave hues to get more color variations in */
	x = 0, cb = palb;
	do
	{
		const uint16	*em = emphasis_lut;
		int		c = 6; /* base and dark already set */

		do
		{
			x+= 256;
			epal_inv[x] = epal_inv[x+64] = epal_inv[x+128] = epal_inv[x+192] =
				FindOrAddRGB(2,*em*cb[0]>>16,
					em[1]*cb[1]>>16,
					em[2]*cb[2]>>16);
		} while(em+=3,--c);
	} while(cb+=6,x+=2,x&=63);
	x = 0, cb = palb+3;
	do
	{
		const uint16	*em = emphasis_lut;
		int		c = 6;

		do
		{
			x+= 256;
			epal_inv[x+1] = epal_inv[x+65] = epal_inv[x+129] = epal_inv[x+193] =
				FindOrAddRGB(2,*em*cb[0]>>16,
					em[1]*cb[1]>>16,
					em[2]*cb[2]>>16);
		} while(em+=3,--c);
	} while(cb+=6,x+=2,x&=63);
	//fwrite(epal_inv,1,sizeof(epal_inv),stdout);
}

// nes_ntsc (by Shay Green) is the one to look up to,
// but this routine's based on what's found in Nintendulator/GFX.cpp
// produces very similar results

#define W		255		//chromatic white/full (8bpc)

#define YS		7		//luminescence sub-fractional bits
#define YW		(W<<YS)		//scaled white
#define DS		9		//decode fractional bits
#define DM		((1<<DS)*5/4)

#define TM		(1<<32-YS-DS)	//output scaler

ifixed		CalcPOutscales[3] = {1.*TM,1.*TM,1.*TM}; /*can be adjusted externally*/

int16		CalcPDecodeIQ[3][2] =
{ /* these are supposed to be correct */
	{ 0.956*DM, 0.621*DM},
	{-0.272*DM,-0.647*DM},
	{-1.105*DM, 1.702*DM}
};

/* the entries can only barely exceed +/-1.0 */
int16		CalcPHiY[4] = {0.40*YW,0.68*YW,1.00*YW,1.00*YW};
int16		CalcPLoY[4] = {-.12*YW,0.00*YW,0.31*YW,0.72*YW};
/* the third here makes for the difference of the above two */
int16		CalcPDfY[4] = {0.52*YW,0.68*YW,0.69*YW,0.28*YW};

static void CalculatePalette(void)
{

	pal		*ent;
	uint32		S, sat;
	ufract		H, hue;
	int32		sr, sg, sb;
	int32		brt = ntscbrt-64<<YS+DS+1;
	int		x, z, v;

	if ((sat=ntscsat)>1999u)
		sat =~sat>>31&2000;
	hue = (long)(4294967296./5)+umuldiv(ntschue%360+360,4294967296./2,360);
	if ((sr=CalcPOutscales[0])<0)
		sr = 0;
	sr+= sr;
	if ((sg=CalcPOutscales[1])<0)
		sg = 0;
	sg+= sg;
	if ((sb=CalcPOutscales[2])<0)
		sb = 0;
	sb+= sb;

	for(x=0,ent=paletten; x<4; ++x)
		for(z=0,H=hue; z<16; ++z,++ent)
		{
			long	Y, I = 0, Q = 0;

			if (z==0)
				Y = CalcPHiY[x]<<DS;
			else if (z<13)
			{
				Y = (CalcPLoY[x]<<DS)+(CalcPDfY[x]<<DS-1);
				S = CalcPDfY[x]*sat>>8;
				I = ifract30mul(S,fastsin(H));
				Q = ifract30mul(S,fastcos(H));
				H+= (ufract)(30.*(double)4294967296./360);
			} else if (z==13)
				Y = CalcPLoY[x]<<DS;
			else
				Y = 0;
			Y+= brt;
			if ((v=Y+I*CalcPDecodeIQ[0][0]+Q*CalcPDecodeIQ[0][1])<0)
				v = 0;
			else if ((v=(ufract32mul(v,sr)+1)>>1)>W)
				v = W;
			ent->r = v;
			if ((v=Y+I*CalcPDecodeIQ[1][0]+Q*CalcPDecodeIQ[1][1])<0)
				v = 0;
			else if ((v=(ufract32mul(v,sg)+1)>>1)>W)
				v = W;
			ent->g = v;
			if ((v=Y+I*CalcPDecodeIQ[2][0]+Q*CalcPDecodeIQ[2][1])<0)
				v = 0;
			else if ((v=(ufract32mul(v,sb)+1)>>1)>W)
				v = W;
			ent->b = v;
		}
}

void FCEU_LoadGamePalette(void)
{
	uint8	ptmp[192];
	FILE	*fp;
	char	*fn;

	pali = 0;

	if ((fn=FCEU_MakeFName(FCEUMKF_PALETTE,0,0))==0)
		return;
	if ((fp=FCEUD_UTF8fopen(fn,"rb"))!=0)
	{
		const uint8	*p;
		int		x, z;

		if ((z=fread(ptmp,3,64,fp))>64)
			z = 64; //use the amount read; clip to 64 color entries
		fclose(fp);

		for(x=0,p=ptmp; x<z; ++x)
			palettei[x].r = *p++,
			palettei[x].g = *p++,
			palettei[x].b = *p++;
		for(; x<64; ++x) //clear out the rest if incomplete
			palettei[x].r =
			palettei[x].g =
			palettei[x].b = 0;

		pali = 1;
	}
	free(fn);
}

void FCEU_ResetPalette(int set)
{
	if ((FCEUGameInfo || set>0) && set)
	{	
		if (FCEUGameInfo->type==GIT_NSF)
			palo = 0;
		else if (pali)
			palo = palettei;
		else if (ntsccol && !PAL && FCEUGameInfo->type!=GIT_VSUNI)
		{
			palo = paletten;
			CalculatePalette();
		} else
			palo = palpoint[pale];
		WritePalette(ntsccol);
	}
}

static int	*ctrlselect = 0;
static uint16	ctrltime = 0;

void FCEUI_NTSCDEC(void)
{
	if (ntsccol && !PAL && FCEUGameInfo->type!=GIT_VSUNI
			&& FCEUGameInfo->type!=GIT_NSF && ctrlselect)
	{
		if (ctrltime)
		{
			if (--(*ctrlselect)<0)
				*ctrlselect = 0;
			CalculatePalette();
			WritePalette(0);
		}
		ctrltime = 360;
	}
}

void FCEUI_NTSCINC(void)
{
	if (ntsccol && !PAL && FCEUGameInfo->type!=GIT_VSUNI
			&& FCEUGameInfo->type!=GIT_NSF && ctrlselect)
	{
		if (ctrltime)
		{
			if (ctrlselect==&ntscsat)
				ntscsat>>= 1;
			if (++(*ctrlselect)>128u)
				*ctrlselect =~*ctrlselect>>31&128;
			if (ctrlselect==&ntscsat)
				ntscsat<<= 1;
			CalculatePalette();
			WritePalette(0);
		}
		ctrltime = 360;
	}
}

void FCEUI_NTSCSelect(uint8 w)
{
	if (ntsccol && !PAL && FCEUGameInfo->type!=GIT_VSUNI
			&& FCEUGameInfo->type!=GIT_NSF)
	{
		switch(w)
		{
		case 0: ctrlselect = &ntscsat; break;
		case 1: ctrlselect = &ntschue; break;
		case 2: ctrlselect = &ntscbrt; break;

		default: return;
		}
		ctrltime = 360;
	}
}

#define BAR_COLOR	PAL_DRED // 0x85

void FCEU_DrawNTSCControlBars(uint8 *XBuf)
{
	int		x, y, which = 0;

	if (!XBuf)
		ctrltime = 0;
	if (!ctrltime)
		{ctrlselect = 0; return;}
	--ctrltime;

	which = *ctrlselect;
	if (ctrlselect==&ntschue)
		DrawTextTrans(XBuf+128-12+180*FCEUSCRN_ROWSIZE,FCEUSCRN_ROWSIZE,(uint8*)"Hue",BAR_COLOR);
	else if (ctrlselect==&ntscsat)
	{
		DrawTextTrans(XBuf+128-16+180*FCEUSCRN_ROWSIZE,FCEUSCRN_ROWSIZE,(uint8*)"Saturation",BAR_COLOR);
		goto nodoub;
	} else if (ctrlselect==&ntscbrt)
		DrawTextTrans(XBuf+128-40+180*FCEUSCRN_ROWSIZE,FCEUSCRN_ROWSIZE,(uint8*)"Brightness",BAR_COLOR);

	which+= which;
nodoub:	XBuf+= 200*FCEUSCRN_ROWSIZE;
	for(x=0; x<which; x+=2)
		for(y=6; y>=-6; --y)
			XBuf[x-FCEUSCRN_ROWSIZE*y] = BAR_COLOR,
				XBuf[x-FCEUSCRN_ROWSIZE*y+(1+FCEUSCRN_ROWSIZE)]|= 1;
	for(; x<256; x+=2)
		for(y=2; y>=-2; --y)
			XBuf[x-FCEUSCRN_ROWSIZE*y] = BAR_COLOR,
				XBuf[x-FCEUSCRN_ROWSIZE*y+(1+FCEUSCRN_ROWSIZE)]|= 1;
}
