/*
 state.c

 * MacFCEU 0.8a7
 *
 * Copyright notice for this file:
 *  state.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef mac
 #include <fcntl.h>
#else
 #include <unistd.h>
#endif

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "PPU.h"
#include "FDS.h"
#include "GameGenie.h"
#include "fsound.h"
#include "general.h"
#include "state.h"
#include "movie.h"
#include "memory.h"
#include "netplay.h"
#include "video.h"
#include "utility/endian.h"

#ifdef env68k
 #pragma segment files
#endif

static void (*SPreSave)(void);
static void (*SPostSave)(void);

static uint8	SaveStateStatus[10];
static uint8	StateShow;

static uint32	ILB;

static uint8	SFEXINDEX;
static SFORMAT	SFMDATA[64];

#define RLSB 		FCEUSTATE_RLSB	//0x80000000

extern SFORMAT FCEUPPU_STATEINFO[];
extern SFORMAT FCEUSND_STATEINFO[];
extern SFORMAT FCEUCTRL_STATEINFO[];

SFORMAT SFCPU[] =
{
	{&XFrame.PC,2|RLSB,"PC\0"},
	{&XFrame.A,1,"A\0\0"},
	{&XFrame.P,1,"P\0\0"},
	{&XFrame.X,1,"X\0\0"},
	{&XFrame.Y,1,"Y\0\0"},
	{&XFrame.S,1,"S\0\0"},
	{RAM,0x800,"RAM"},
	{0}
};

SFORMAT SFCPUC[] =
{
//	{&XFrame.jammed,1,"JAMM"},
	{&ILB,4|RLSB,"IQLB"},
	{&XFrame.tcount,4|RLSB,"ICoa"},
	{&XFrame.count,4|RLSB,"ICou"},
	{&timestampbase,sizeof(timestampbase)|RLSB,"TSBS"},
	{&XFrame.BP,1,"PI\0\0"}, /* 2010.03.08, 2013.05.27 */
	{0}
};

/* info state added; 2010.04.25 */
uint8		SFMD5[16];
char		SFName[100];

SFORMAT SFINFO[] =
{
	{&SFMD5,sizeof(SFMD5),"*MD5"},
	{&SFName,sizeof(SFName),"*NAM"},
	{0}
};

/* ---- code */

static int SameOrNoMD5(const uint8 *a,
	const uint8 *b)
{
	uint8		ao, xo, c, i;

	i = 16, ao = xo = 0;
	do
		ao|= c=*a++, xo|= c^*b++;
	while(--i!=0);

	return !ao || !xo;
}

static int SubWrite(FILE *st,
	SFORMAT *sf)
{
	uint32	acc = 0;

	for(; sf->v; ++sf)
	{
		int32	z;

		if (sf->s==~0)	//Link to another struct.
		{
			uint32	tmp;

			if ((tmp=SubWrite(st,(SFORMAT*)sf->v))==0)
				return 0;
			acc+= tmp;
			continue;
		}

		acc+= 8;	/* Description + size */
		acc+= sf->s&~RLSB;

		if (!st)
			continue; //only calculating size of block
		//writing block
		z = sf->s&~RLSB;

		fwrite(sf->desc,1,4,st);
		if (!write32le(z,st))
			return 0;
	//temporarily reverse byte order for big endian systems
	#ifndef LSB_FIRST
		if (sf->s&RLSB)
			FlipByteOrder(sf->v,z);
	#endif
		if (fwrite((uint8*)sf->v,1,z,st)!=z)
			return 0;
	//restore byte order
	#ifndef LSB_FIRST
		if (sf->s&RLSB)
			FlipByteOrder(sf->v,z);
	#endif
	}

	return acc;
}

static int WriteStateChunk(FILE *st,
	int type,
	SFORMAT *sf)
{
	int		bsize;

	fputc(type,st);

	bsize = SubWrite(0,sf);
	if (!write32le(bsize,st) || !SubWrite(st,sf))
		return 0;
	return bsize+5;
}

static SFORMAT *CheckS(SFORMAT *sf,
	uint32 tsize,
	const char *desc)
{
	for(; sf->v; ++sf)
	{
		int	z;

		if (sf->s==~0)	/* Link to another SFORMAT structure. */
		{
			SFORMAT	*tmp;

			if ((tmp=CheckS((SFORMAT*)sf->v,tsize,desc))!=0)
				return tmp;
			continue;
		}
		z = strlen(sf->desc)+1;
		if (!memcmp(desc,sf->desc,z>3 ? 4:z)) //z>7 ? 8:z
		{
			if (tsize!=(sf->s&~RLSB))
				return 0;
			return sf;
		}
	}
	return 0;
}

static int ReadStateChunk(FILE *st,
	SFORMAT *sf,
	int size)
{
	SFORMAT		*tmp;
	int		blockend = size+ftell(st);

	while(ftell(st)<blockend)
	{
		uint32	tsize, z;
		char	toa[4];

		if (fread(toa,1,4,st)<=0 || !read32le(&tsize,st))
			return 0;
		if ((tmp=CheckS(sf,tsize,toa))==0)
		{
			fseek(st,tsize,SEEK_CUR);
			continue;
		}
		z = tmp->s&~RLSB;

		fread((uint8*)tmp->v,1,z,st);
	#ifndef LSB_FIRST
		if (tmp->s&RLSB)
			FlipByteOrder(tmp->v,z);
	#endif
	}
	return 1;
}

static int ReadStateChunks(FILE *st,
	int32 totalsize)
{
	uint32		size;
	int		t, ret = 1, warn = 0;

	while(totalsize>0)
	{
		if ((t=fgetc(st))==EOF //type byte
				|| !read32le(&size,st))
			break;
		totalsize-= size+5;

		if (!--t) //1
		{
			ret = ReadStateChunk(st,SFCPU,size);
			/* prevent possible hang-time */
			if ((unsigned)(XFrame.count+0x800000)>0x1000000)
				XFrame.count = 0, warn = 1;
			if ((unsigned)(XFrame.tcount+0x800000)>0x1000000)
				XFrame.tcount = 0, warn = 1;
		} else if (!--t) //2
		{
			XFrame.BP = XFrame.P; //in case it's not set (old state)
			ret = ReadStateChunk(st,SFCPUC,size);
		} else if (!--t) //3
			ret = ReadStateChunk(st,FCEUPPU_STATEINFO,size);
		else if (!--t) //4
			ret = ReadStateChunk(st,FCEUCTRL_STATEINFO,size);
		else if (!--t) //5
			ret = ReadStateChunk(st,FCEUSND_STATEINFO,size);
		else if (!--t) //6 -- MacFCEU 0.6.3+
		{
			if ((ret=ReadStateChunk(st,SFINFO,size))>=0
					&& FCEUGameInfo
					&& !SameOrNoMD5(SFMD5,FCEUGameInfo->MD5))
			{
				ret = -5;
				break;
			}
		} else if ((t-=0x10-6)==0)
			ret = ReadStateChunk(st,SFMDATA,size);
		else if (fseek(st,size,SEEK_CUR)<0) /* unknown, skip */
			break;
	}

	if (warn)
		FCEUD_Message("##Loaded state file appears to be corrupt...");
	return ret;
}

int		CurrentState = 0;

static void MakeSFInfo(void)
{
	char		*n;

	if (!FCEUGameInfo)
		return;
	memcpy(SFMD5,FCEUGameInfo->MD5,sizeof(SFMD5));
	if ((n=FCEU_MakeFName(FCEUMKF_DEFAULT,0,0))==0)
		memset(SFName,0,sizeof(SFName));
	else
	{
		strncpy(SFName,n,sizeof(SFName));
		free(n);
	}

}

int FCEUSS_SaveFP(FILE *st,
	int movie)
{
	static uint32	totalsize;
	static uint8	header[16] = {'F','C','S',~0,
		0,0,0,0, /* length */
		FCEU_VERSION_NUMERIC,FCEU_VERSION_NUMERIC>>8, /* version.. */
		FCEU_VERSION_NUMERIC>>16,FCEU_VERSION_NUMERIC>>24,
		~0,~0,~0,~0}; /* compressed length (not implemented) */

	header[2] = movie ? 'M':'S';

	if (!st)
		return 0;
	fwrite(header,1,16,st);
	FCEUPPU_SaveState();
	FCEUSND_SaveState();

	/*2010.04.25*/
	MakeSFInfo();
	totalsize = WriteStateChunk(st,6,SFINFO); /* done first for checksum */

	totalsize+= WriteStateChunk(st,1,SFCPU);
	ILB = XFrame.intlow;
	totalsize+= WriteStateChunk(st,2,SFCPUC);
	totalsize+= WriteStateChunk(st,3,FCEUPPU_STATEINFO);
	totalsize+= WriteStateChunk(st,4,FCEUCTRL_STATEINFO);
	totalsize+= WriteStateChunk(st,5,FCEUSND_STATEINFO);

	if (SPreSave)
		SPreSave();
	totalsize+= WriteStateChunk(st,0x10,SFMDATA);
	if (SPreSave)
		SPostSave();

	fseek(st,4,SEEK_SET);

	return write32le(totalsize,st) ? 1:0;
}

void FCEUSS_Save(const char *fname)
{
	FILE		*st = 0;
	char		*fn;

	if (!FCEUGameInfo)
		return; /* inalid call; ignore */
	if (GenieStage==1)
	{
		FCEU_DispMessage("Cannot do FCS/MOV in GG screen");
		return;
	}

	if (fname)
		st = FCEUD_UTF8fopen(fname,"wb");
	else if ((fn=FCEU_MakeFName(FCEUMKF_STATE,CurrentState,0))!=0)
	{
		st = FCEUD_UTF8fopen(fn,"wb");
		free(fn);
	}
	if (!st)
	{
		FCEU_DispMessage(fname ? "State save error.":"State %d save error.",
			CurrentState);
		return;
	}

	FCEUSS_SaveFP(st,0);

	if (fname)
		SaveStateStatus[CurrentState] = 1;
	fclose(st);
	FCEU_DispMessage(fname ? "State saved.":"State %d saved.",
		CurrentState);
}

int FCEUSS_LoadFP(FILE *st)
{
	uint8		header[16];
	long		compressedLen = -1;
	int		x, version;

	/* clear data that might not be stored (zero = n/a); 2010.04.25, 05.15 */
	memset(SFMD5,0,sizeof(SFMD5));
	memset(SFName,0,sizeof(SFName));

	if (!st)
		return 0;

	if (fread(&header,1,16,st)!=16
			|| header[0]!='F' || header[1]!='C')
	{
		if (!memcmp(header,"version 3",9))
			return -3; /* FCM version 3 */
		if (!memcmp(header,"FNSS",4))
			return -6;
		if (!memcmp(header,"NST\x1A",4))
			return -7;
		if (!memcmp(header,"ZSNES Save State",16))
			return -8;
		return 0;
	}
	if (header[2]=='M')
	{
	//	if (header[3]==0xFF) /* version 0.6.4 */
	//		goto shortlived;
		if (header[3]!=0x1A) /* FCM version 1 and 2 */
			return -3;
		version = FCEU_de32lsb(header+4);
		if (version==1)
			x = 0; /* flags @ 8 has no power/reset flags */
		else if (version==2)
			x = header[8]; /* flags */
		else
			return -3;
		fseek(st,8,SEEK_CUR);
		if (fread(&header,1,4,st)!=4
				|| fseek(st,FCEU_de32lsb(header),SEEK_SET)<0
				|| fread(&header,1,16,st)!=16)
			return -1;
		if (header[0]!='F' || header[1]!='C' || header[2]!='S')
			return 0;
		//if (x&(8|2))
		//	return x; /* no state data? */
	} else if (header[2]!='S')
		return 0;
	if (header[3]=='X')
		compressedLen = FCEU_de32lsb(header+12);
	else if (header[3]!=0xFF)
		return -3; /* ancient version */

	version = FCEU_de32lsb(8+header);

	/* this should've been here since day one! */
	if (version>FCEU_VERSION_NUMERIC)
		return -3;
	if (compressedLen!=-1)
		return -4; /* can't yet handle compressed state files */

	//MacFCEU 0.6a3 and earlier bug fix (oh, boy...)
#ifdef mac
	if ((version-9813)<=1)
		return -3;
#endif
	//FCE Ultra bug fix - stored LE, read without endian conversion; 2008.09.11
	if ((x=ReadStateChunks(st,FCEU_de32lsb(header+4)))<0)
		goto end;
	XFrame.intlow = version<9500 ? 0:ILB;
	if (GameStateRestore)
		GameStateRestore(version);
	if (x)
	{
		FCEUPPU_LoadState(version);
		FCEUSND_LoadState(version);
	}
end:
	return x;
}

int FCEUSS_Load(const char *fname)
{
	FILE		*st;
	char		*fn;
	int		r;

	if (!FCEUGameInfo)
		goto endo; /* inalid call; ignore */
	if (GenieStage==1)
	{
		FCEU_DispMessage("Cannot do FCS/MOV in GG screen");
		goto endo;
	}

	if (fname)
		st = FCEUD_UTF8fopen(fname,"rb");
	else if ((fn=FCEU_MakeFName(FCEUMKF_STATE,CurrentState,fname))!=0)
	{
		st = FCEUD_UTF8fopen(fn,"rb");
		free(fn);
	} else
		st = 0;

	if (!st)
	{
		if (fname)
			FCEU_DispMessage("State load error.");
		else
		{
			FCEU_DispMessage("State %d load error.",
				CurrentState);
			SaveStateStatus[CurrentState] = 0;
		}
		goto endo;
	}

	if ((r=FCEUSS_LoadFP(st))>0)
	{
		if (r&8)
			FCEU_DispMessage("File only Powers-on.");
		else if (r&2)
			FCEU_DispMessage("File only Resets.");
		else if (fname)
			FCEU_DispMessage("State loaded.");
		else
		{
			FCEU_DispMessage("State %d loaded.",
				CurrentState);
			SaveStateStatus[CurrentState] = 1;
		}
		fclose(st);
		return 1;
	} else
	{
	//	SaveStateStatus[CurrentState] = 0; //not reliable...
		fclose(st);

		if (r==-3)
			FCEU_DispMessage("Incompatible FCS/MOV file");
		else if (r==-4)
			FCEU_DispMessage("Can't uncompress FCS/MOV");
		else if (r==-5)
		{
			FCEU_DispMessage("State not made w/ this cart");
			if (SFName[0]) /* print its name; 2010.05.15 */
				FCEU_PrintError(
"State file appears to have been made with:\n %s",
					SFName);
		} else if (r==-6)
			FCEU_DispMessage("This isn't FakeNES.");
		else if (r==-7)
			FCEU_DispMessage("This isn't Nestopia.");
		else if (r==-8)
			FCEU_DispMessage("An SNES state- really?");
		else if (!r)
			FCEU_DispMessage("Not an FCS file.");
		else
			FCEU_DispMessage(fname ?
				"Error(s) reading state file":
				"Error(s) reading state %d!",
				CurrentState);
	}
endo:
	return 0;
}

void FCEUSS_CheckStates(void)
{
	char	*fn;
	int	ssel;

	for(ssel=0; ssel<10; ++ssel)
	{
		if ((fn=FCEU_MakeFName(FCEUMKF_STATE,ssel,0))!=0
				&& CheckFileOrDirExists(fn)==1)
			SaveStateStatus[ssel] = 1;
		else
			SaveStateStatus[ssel] = 0;
		if (fn)
			free(fn);
	}
	CurrentState = 0;
	StateShow = 0;
}

void ResetExState(void (*PreSave)(void),
	void (*PostSave)(void))
{
	int		x;

	for(x=0; x<SFEXINDEX; ++x)
		SFMDATA[x].desc[0] = 0;
	SFMDATA[0].s = 0; //initialize

	SPreSave = PreSave;
	SPostSave = PostSave;
	SFEXINDEX = 0;
}

void AddExState(void *v,
	uint32 s,
	int type,
	const char *desc)
{
	int		x = SFEXINDEX;

	if (desc)
		memcpy(SFMDATA[x].desc,desc,8);
	else
		SFMDATA[x].desc[0] = 0;
	SFMDATA[x].v = v;
	SFMDATA[x].s = s;
	if (type)
		SFMDATA[x].s|= RLSB;
	if (x<sizeof(SFMDATA)/sizeof(SFMDATA[0])-1)
		++x;
	SFMDATA[SFEXINDEX=x].v = 0;	// End marker.
}

void FCEUI_SelectState(int w)
{
	/* checks; 2010.04.25, 2010.05.21 */
	if (w<0 || !FCEUGameInfo || FCEUGameInfo->type==GIT_NSF)
		{StateShow = 0; return;}
	FCEUI_SelectMovie(-1);

	CurrentState = w;
	StateShow = 180;
	FCEU_DispMessage("-select state-");
}

void FCEUI_SaveState(const char *fname)
{
	StateShow = 0;
	FCEUSS_Save(fname);
}

void FCEUI_LoadState(const char *fname)
{
	StateShow = 0;

	FCEUMOV_Stop();

 /* For network play, be the state loaded locally, and then save the state to a temporary
  file, and send that.  This insures that if an older state is loaded that is missing
  some information expected in newer save states, desynchronization won't occur
  (at least not from this ;) ).
 */
	if (FCEUSS_Load(fname) && FCEUnetplay)
	{
		char	*fn = FCEU_MakeFName(FCEUMKF_NPTEMP,0,0);
		FILE	*fp;

		if ((fp=fopen(fn,"wb"))!=0)
		{
			int	r = FCEUSS_SaveFP(fp,0);

			fclose(fp);
			if (r)
				FCEUNET_SendFile(FCEUNPCMD_LOADSTATE,fn);
			unlink(fn);
		}
		free(fn);
	}
}

void FCEU_DrawSaveStates(uint8 *XBuf)
{
	if (!XBuf) /* blah; 2010.04.25 */
		StateShow = 0;
	if (!StateShow)
		return;
	--StateShow;
	FCEU_DrawNumberRow(XBuf,SaveStateStatus,CurrentState);
}
