/*
 asm6502.h

  Revised: 2013.07.05

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __ASM6502_H
 #include "disasm6502.h"

/* ---- constants */

#define _LabDataSize	(4+4)

/* opt (option) bits */
#define _Prescan	0x80	/* allow temporary auto-creation */
#define _Optimize	0x40	/* make smaller codes - a file can turn this on */
#define _6502Only	0x20	/* no 6510 instructions, please */
#define _DataOnly	0x10	/* no instructions at all, please */
#define _QuietMode	0x8	/* no printf */
#define _MirrorMask	0x3	/* all bits on = default*/

/* flg (flag) bits */
#define _Preprocessed	0x8000	/* comments have been removed */
#define _IntoProcList	0x4000	/* processing the procedure list */
#define _orgInited	0x80	/* if not set, ORG is set to ->pos on write */
#define _IntoLine	0x40	/* otherwise start of line */
#define _IntoProc	0x20	/* 'proc' to 'endproc' */
#define _PreEstID	0x10	/* pre-established ID on line (don't optimize this) */
#define _EndKeyw	0x8	/* 'end' statement reached (not EOF) */
#define _PrintNL	0x4	/* printt/printv calls for '\n' */
#define _PWrErr		0x2	/* writef should set this on error */
#define _LScanErr	0x1	/* - or + went out of range */

#define _LineFlags	(_IntoLine|_PreEstID)

enum
{ /* "lab" type values - low 6 bits */
  /* bit #7 = pre-established */
  /* bit #6 = pre-scanned */
	_Unestablished,
	_Default,
	_Label,
	_Equate
};

enum
{ /* predefined segment indices */
	_textSegment,
	_dataSegment,
	_bssSegment,
	_zpSegment
};

enum
{ /* assembler errors (other than the interpreter.h standard) */
	_e_include		=-1000, /* high-level operation (handled in AsmFile) */
	_e_include_depth	=-1001,	/* too many include-within-include or recursion */
	_e_include_open		=-1002, /* failed to open */
	_e_off_or_on		=-1003, /* bad parameter for opt */
	_e_write_in_addr_seg	=-1004,	/* attempt to write code/data in ZP/BSS */
	_e_pos_write_error	=-1005, /* writef set _PWrError */
	_e_bank_overflow	=-1006,	/* auto-bank (org $C000) wrapped */
	_e_labelscan_overflow	=-1007,	/* _LScanErr set */
	_e_org_wrap		=-1008,	/* address wrapped in operation */
	_e_already_in_proc	=-1009,	/* no .endproc before another .proc */
	_e_open_proc		=-1010,	/* reached end, no endproc */
	_e_no_proc		=-1011,	/* hit endproc, not in proc */
	_e_name_too_long	=-1012,	/* total name too long */
	_e_data_type		=-1013,	/* data type not supported with operation */
	_e_data_only		=-1014,	/* BSS, etc. - data-only option set */
	_e_6510_instruction	=-1015,	/* hit 6510 op - disable option set */
	_e_unprep_org		=-1016,	/* ORG not init. for segment */
	_e_open_if		=-1017,	/* if/ifdef/ifndef, no endif closer in file */
	_e_no_if		=-1018,	/* else/elif/endif, no if/ifdef/ifndef */
	_e_else_else		=-1019,	/* else after else */
	_e_check_error		=-1020,	/* ORG above parameter value */
	_e_error_directive	=-1021,	/* hit .error or .fail */
	_e_size			=-1022,	/* bad/unsupported size value */
	_e_unsupported_param	=-1023,	/* e.g., not skip or read with incbin */
	_e_dec_or_hex		=-1024,	/* printv error */
	_e_module_in_proc	=-1025,	/* cannot define a module within a proc */
	_e_name_null		=-1026	/* nulls not allowed in names */
};

enum
{ /* corresponding index is +OPN_COUNT */
	_incbin			= OPN_6510_COUNT,
	_include,
	_addr,
	_align,
	_bank,
	_block,
	_byt,
	_byte,
	_db,
	_dcb,
	_dc_b,
	_defchr,
	_dw,
	_dcw,
	_dc_w,
	_dcl,
	_dc_l,
	_end,
	_endproc,
	_equ,
	_fill,
	_ineschr,
	_inesmap,
	_inesmir,
	_inesprg,
	_mem,
	_module,
	_opt,
	_org,
	_orga,
	_p02,
	_p10,
	_pad,
	_processor,
	_res,
	_segment,
	_subroutine,
	_word,
	_text,		/* dot prefix-required start - _text-_zp range should be left alone */
	_code,
	_data,
	_bss,
	_zeropage,
	_zp,
	_default,	/* _default to _space range should be left alone */
	_alias,
	_proc,
	_space,
	_accu,
	_advance,
	_ascii,
	_checkpc,
	_index,
	_pushseg,
	_popseg,
	_printt,
	_printv,
	_export,
	_import
};

enum
{ /* built-in preprocessing IDs */
	_date_def		= 0x40000000,
	_file_def,
	_line_def,
	_time_def,
	_6510_def,
	_define,
	_undef,
	_error,
	_fail,
	_if,
	_ifdef,
	_ifndef,
	_else,
	_elif,
	_endif
};

/* ---- typedefs */

typedef struct preDef
{
	char		*data;	/* C-string name, text */
	long		size;
} preDef;

typedef struct asm6502frame
{
	uint16		flg;
	uint8		opt;
	uint8		bsz;	/* bank size, in bits */
	uint8		seg;	/* segment index */
	uint8		pei;	/* [dis.1] [else.1] [predis.6] */
	uint16		org, pos; /* pos is the actual writing position */
	uint16		flr;
	int16		bnk, maxbnk, maxchr; /* 0xFFF maximum */
	int16		map;	/* mapper type/index (or negative) */

	uint32		ctr;	/* total low-level line counter */
	uint32		lnx, lnp; /* line index, and in-proc previous, if applicable */
	const char	*lfn;	/* file name, if applicable */

	uint8		*lab;	/* {[hsh.8] [pstr] [len.8] [typ.8] [val.32] [ctr.32]} */
	char		*mod;	/* current module name (or NULL) */
	preDef		*prc;	/* deferred procedure list */
	int32		prx;	/* one-based current proc scan */

	preDef		*pre;	/* preprocessing defines, procs, etc. */
	uint8		*est;	/* preproc. stack - [ofs.32]{pei.8} */

	uint8		sbi[_zpSegment+1][4]; /* see list above */
	uint8		*sgs;	/* null term'd list: [pstr] [[flg.1] [bnk.15] [org.16]] */
	uint8		*sst;	/* segment stack - [ofs.32]{[seg.8]} */

	/* when writing bytes; can be NULL (i.e., check syntax) */
	/* IMPORTANT: this function is expected to use and update `pos` */
	void (*writef)(struct asm6502frame*,
		uint8);
	void		*u;	/* user-defined */
} asm6502frame;

/* ---- prototypes */

/* returns pstr offset (positive), ~end, zero on error */
int32 X6502_AsmGetID(const char *s,
	const char *e,
	uint8 *n, /* buffer capacity >=256 */
	const asm6502frame *f,
	int *z); /* full ID length stored in `n` */
/* returns data offset; negated if already added, zero on error */
int X6502_AsmAddID(const char *s,
	const char *e,
	uint8 *n, /* buffer capacity >=256 */
	asm6502frame *f,
	uint8 t); /* see type values above */
/* returns the ID count */
int X6502_AsmDumpIDs(FILE *f,
	const asm6502frame *af);

/* returns index or ~end */
int X6502_GetPreDef(const asm6502frame *af,
	const char *s,
	const char *e);
/* returns error or name length */
int X6502_AddPreDef(asm6502frame *af,
	const char *str,
	const char *e);
/* returns error or name length */
int X6502_SubPreDef(asm6502frame *af,
	const char *str,
	const char *e);

/* returns index */
int X6502_ParseKey(const char *s,
	const char *e);
/* returns text length parsed, error if negative */
long X6502_ParseName(const char *s,
	const char *e,
	long *z, /* out actual length */
	char **b,
	const char *parent);

/* returns error on negative, warning on positive */
long X6502_Assemble(const char *s,
	const char *e,
	int *err, /* can be NULL */
	asm6502frame*); /* zero this out first thing; can be NULL (one call) */

int X6502_AsmFile(struct FILE*,
	const char *fname, /* can be NULL */
	asm6502frame*,
	uint8 includeDepth);

int X6502_FreeFrame(asm6502frame*);

int X6502_FreeScan(asm6502frame*); /* everything but the labels, for re-scan */
#endif
