/*
 utility/disasm6502.c
  the 6502/10-compatible disassembler, public arrays in one file

  Revised: 2015.07.27

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef SAMPLE
 #include "enviro64.h"
#else
 #include "../types.h"
#endif

#include "../6502struct.h"
#include "disasm6502.h"

#define OPB_RD			(OPB_READ|OPB_FLAG)
#define OPB_WR			(OPB_WRITE|OPB_FLAG)
#define OPB_RW			(OPB_RD|OPB_WR)

#define OPN(n)			(OPN_##n)

/* operator only (condition flags unaffected) */
#define ORL(n)			((OPM_RL<<8)|OPN(n))
#define OAB(n)			((OPM_AB<<8)|OPN(n))
#define OIP(n)			((OPM_IP<<8)|OPN(n))

/* operator, may affect condition flags */
#define FIM(n)			((OPM_IM<<8)|OPB_FLAG|OPN(n))
#define FIP(n)			((OPM_IP<<8)|OPB_FLAG|OPN(n))

/* read, may affect condition flags */
#define RIM(n)			((OPM_IM<<8)|OPB_RD|OPN(n))
#define RZP(n)			((OPM_ZP<<8)|OPB_RD|OPN(n))
#define RZX(n)			((OPM_ZX<<8)|OPB_RD|OPN(n))
#define RZY(n)			((OPM_ZY<<8)|OPB_RD|OPN(n))
#define RAB(n)			((OPM_AB<<8)|OPB_RD|OPN(n))
#define RAX(n)			((OPM_AX<<8)|OPB_RD|OPN(n))
#define RAY(n)			((OPM_AY<<8)|OPB_RD|OPN(n))
#define RIB(n)			((OPM_IB<<8)|OPB_RD|OPN(n))
#define RIX(n)			((OPM_IX<<8)|OPB_RD|OPN(n))
#define RIY(n)			((OPM_IY<<8)|OPB_RD|OPN(n))
#define RIP(n)			((OPM_IP<<8)|OPB_RD|OPN(n))

/* write, may affect condition flags */
#define WZP(n)			((OPM_ZP<<8)|OPB_WR|OPN(n))
#define WZX(n)			((OPM_ZX<<8)|OPB_WR|OPN(n))
#define WZY(n)			((OPM_ZX<<8)|OPB_WR|OPN(n))
#define WAB(n)			((OPM_AB<<8)|OPB_WR|OPN(n))
#define WAX(n)			((OPM_AX<<8)|OPB_WR|OPN(n))
#define WAY(n)			((OPM_AY<<8)|OPB_WR|OPN(n))
#define WIX(n)			((OPM_IX<<8)|OPB_WR|OPN(n))
#define WIY(n)			((OPM_IY<<8)|OPB_WR|OPN(n))
#define WIP(n)			((OPM_IP<<8)|OPB_WRITE|OPN(n)) /* P unaffected */

/* dual read/write, affects condition flags */
#define DIM(n)			((OPM_IM<<8)|OPB_RW|OPN(n))
#define DZP(n)			((OPM_ZP<<8)|OPB_RW|OPN(n))
#define DZX(n)			((OPM_ZX<<8)|OPB_RW|OPN(n))
#define DAB(n)			((OPM_AB<<8)|OPB_RW|OPN(n))
#define DAX(n)			((OPM_AX<<8)|OPB_RW|OPN(n))
#define DAY(n)			((OPM_AY<<8)|OPB_RW|OPN(n))
#define DIX(n)			((OPM_IX<<8)|OPB_RW|OPN(n))
#define DIY(n)			((OPM_IY<<8)|OPB_RW|OPN(n))
#define DIA(n)			((OPM_IA<<8)|OPB_RW|OPN(n))

/* 65816-reserved (2A03 not implemented) instructions */
#define oIM(n)			((OPM_IM<<8)|OPB_HI|OPN(n))
#define oRL(n)			((OPM_RL<<8)|OPB_HI|OPN(n))
#define oIA(n)			((OPM_IA<<8)|OPB_HI|OPN(n))
#define oIP(n)			((OPM_IP<<8)|OPB_HI|OPN(n))

#define rRL(n)			((OPM_RL<<8)|OPB_RD|OPB_HI|OPN(n))
#define rZP(n)			((OPM_ZP<<8)|OPB_RD|OPB_HI|OPN(n))
#define rZX(n)			((OPM_ZX<<8)|OPB_RD|OPB_HI|OPN(n))
#define rAX(n)			((OPM_AX<<8)|OPB_RD|OPB_HI|OPN(n))

#define dIM(n)			((OPM_IM<<8)|OPB_RW|OPB_HI|OPN(n))

/* jams the 6510 */
#define bIC(n)			((OPM_IC<<8)|OPB_BAD|OPB_HI|OPN(n))
#define bRL(n)			((OPM_RL<<8)|OPB_BAD|OPB_HI|OPN(n))
#define bIN(n)			((OPM_IN<<8)|OPB_BAD|OPB_HI|OPN(n))
#define bIP(n)			((OPM_IP<<8)|OPB_BAD|OPB_HI|OPN(n))

#if __POCC__
 /* this is for the following array - no room for ea. C-string null */
 #pragma warn(disable:2237)
#endif

const char	x65xx_opsizes[OPM_COUNT] = "\1\1\1\1\1\1\1\1\2\2\2\2\0\0\0\1\1\1\1\2\2\3";

const char	x6510_opnametable[OPN_65xx_COUNT][3] =
{
	"NOP",
	"BRK",
	"JMP",
	"JSR",
	"RTI",
	"RTS",
	"PHA",
	"PHP",
	"PLA",
	"PLP",
	"BPL",
	"BMI",
	"BVC",
	"BVS",
	"BCC",
	"BCS",
	"BNE",
	"BEQ",
	"STX",
	"STY",
	"STA",
	"TXS",
	"TSX",
	"TXA",
	"TAX",
	"TYA",
	"TAY",
	"CPX",
	"CPY",
	"CMP",
	"BIT",
	"SEC",
	"SEI",
	"SED",
	"CLV",
	"CLC",
	"CLI",
	"CLD",
	"LDX",
	"LDY",
	"LDA",
	"EOR",
	"ORA",
	"AND",
	"ADC",
	"SBC",
	"DEX",
	"DEY",
	"INX",
	"INY",
	"DEC",
	"INC",
	"ASL",
	"LSR",
	"ROL",
	"ROR",
	"SHX",
	"SHY",
	"SHA",
	"SHS",
	"SAX",
	"SBX",
	"ARR",
	"ASR",
	"RLA",
	"RRA",
	"LAX",
	"LAS",
	"SLO",
	"SRE",
	"ANE",
	"ANC",
	"DCP",
	"ISB",
	"BRA",
	"STZ",	/* 64, 74, 9C(ab), 9E */
	"PLX",	/* FA */
	"PLY",	/* 7A */
	"PLD",	/* 2B */
	"PLB",	/* AB */
	"PHX",	/* DA */
	"PHY",	/* 5A */
	"PHD",	/* 0B */
	"PHB",	/* 8B */
	"PHK",	/* 4B */
	"PEA",	/* F4-PEA(ab), D4-PEI(in), 62-PER(rl) */
	"SWA",	/* EB */
	"XCE",	/* FB */
	"TCD",	/* 5B */
	"TDC",	/* 7B */
	"TSB",	/* 04, 0C */
	"TRB",	/* 14(dp), 1C(ab) */
	"REP",	/* C2 */
	"SEP",	/* E2 */
	"MVN",	/* 54 */
	"MVP",	/* 44 */
	"WAI",	/* CB */
	"STP",	/* DB */
	"COP",	/* 02 */
	"WDM"	/* 42 */
};

const uint16	x6510_opn_table[256] =
{ /* some names differ from what's found in 6502_cpu.txt */
  /* some no-ops share addressing modes relative to the line they're on */
  /* some implied operands will show up with this disassembler, such as "ASL A" */
/*00*/	OIP(BRK),RIX(ORA),bIC(COP),DIX(SLO), /* BIP(KIL) now COP(imm) */
/*04*/	rZP(TSB),RZP(ORA),DZP(ASL),DZP(SLO), /* RZP(ZOP) now TSB */
/*08*/	WIP(PHP),RIM(ORA),DIA(ASL),FIM(ANC),
/*0C*/	RAB(TSB),RAB(ORA),DAB(ASL),DAB(SLO),
/*10*/	ORL(BPL),RIY(ORA),bIN(ORA),DIY(SLO), /* BRL(ASL) now ORA */
/*14*/	RZX(TRB),RZX(ORA),DZX(ASL),DZX(SLO),
/*18*/	FIP(CLC),RAY(ORA),oIA(INC),DAY(SLO), /* OIP(ZOP) now INC */
/*1C*/	RAX(TRB),RAX(ORA),DAX(ASL),DAX(SLO),
/*20*/	OAB(JSR),RIX(AND),bIP(JSR),DIX(RLA), /* BIP(ROL) now JSR */
/*24*/	RZP(BIT),RZP(AND),DZP(ROL),DZP(RLA),
/*28*/	RIP(PLP),RIM(AND),DIA(ROL),FIM(ANC),
/*2C*/	RAB(BIT),RAB(AND),DAB(ROL),DAB(RLA),
/*30*/	ORL(BMI),RIY(AND),bIN(AND),DIY(RLA), /* BRL(ROL) now AND */
/*34*/	rZX(BIT),RZX(AND),DZX(ROL),DZX(RLA),
/*38*/	FIP(SEC),RAY(AND),oIA(DEC),DAY(RLA), /* OIP(ZOP) now DEC */
/*3C*/	rAX(BIT),RAX(AND),DAX(ROL),DAX(RLA),
/*40*/	OIP(RTI),RIX(EOR),bIP(WDM),DIX(SRE), /* BIP(LSR) now WDM */
/*44*/	rZP(MVP),RZP(EOR),DZP(LSR),DZP(SRE), /* RZP(ZOP) now MVP-sdk */
/*48*/	WIP(PHA),RIM(EOR),DIA(LSR),rZP(PHK), /* RZP(ASR) now PHK */
/*4C*/	OAB(JMP),RAB(EOR),DAB(LSR),DAB(SRE),
/*50*/	ORL(BVC),RIY(EOR),bIN(EOR),DIY(SRE), /* BRL(LSR) now EOR*/
/*54*/	rZX(MVN),RZX(EOR),DZX(LSR),DZX(SRE), /* RZX(ZOP) now MVN-sdk */
/*58*/	FIP(CLI),RAY(EOR),oIP(PHY),DAY(SRE), /* OIP(ZOP) now PHY */
/*5C*/	rAX(JMP),RAX(EOR),DAX(LSR),DAX(SRE), /* RAX(ZOP) now JMP */
/*60*/	OIP(RTS),RIX(ADC),bRL(PEA),DIX(RRA), /* BIP(ROR) now PEA-rel */
/*64*/	rZP(STZ),RZP(ADC),DZP(ROR),DZP(RRA), /* RZP(ZOP) now STZ */
/*68*/	RIP(PLA),RIM(ADC),DIA(ROR),DIM(ARR), /* 65816 6B=RTL (ret.long) */
/*6C*/	RIB(JMP),RAB(ADC),DAB(ROR),DAB(RRA),
/*70*/	ORL(BVS),RIY(ADC),bIN(ADC),DIY(RRA), /* BRL(ROR) now ADC */
/*74*/	rZX(STZ),RZX(ADC),DZX(ROR),DZX(RRA), /* RZP(ZOP) now STZ */
/*78*/	FIP(SEI),RAY(ADC),oIP(PLY),DAY(RRA),
/*7C*/	rAX(JMP),RAX(ADC),DAX(ROR),DAX(RRA), /* RAX(ZOP) now JMP */
/*80*/	oRL(BRA),WIX(STA),rRL(BRA),DIX(SAX), /* presumed IM-STY/X now 65816 BRA */
/*84*/	WZP(STY),WZP(STA),WZP(STX),DZP(SAX),
/*88*/	FIP(DEY),oIM(BIT),FIP(TXA),RIM(ANE), /* presumed IM-STA now *BIT */
/*8C*/	WAB(STY),WAB(STA),WAB(STX),WAB(SAX),
/*90*/	ORL(BCC),WIY(STA),bIN(STA),WIY(SHA), /* BRL(STX) now STA */
/*94*/	WZX(STY),WZX(STA),WZY(STX),WZY(SAX),
/*98*/	FIP(TYA),WAY(STA),FIP(TXS),WAY(SHS),
/*9C*/	RAX(SHY),WAX(STA),RAY(SHX),WAY(SHA),
/*A0*/	FIM(LDY),RIX(LDA),FIM(LDX),RIX(LAX),
/*A4*/	RZP(LDY),RZP(LDA),RZP(LDX),RZP(LAX),
/*A8*/	FIP(TAY),RIM(LDA),FIP(TAX),RIM(LAX), /* LXA is now just another LAX */
/*AC*/	RAB(LDY),RAB(LDA),RAB(LDX),RAB(LAX),
/*B0*/	ORL(BCS),RIY(LDA),bIN(LDA),RIY(LAX), /* BRL(LDX) now LDA */
/*B4*/	RZX(LDY),RZX(LDA),RZY(LDX),RZY(LAX),
/*B8*/	FIP(CLV),RAY(LDA),OIP(TSX),RAY(LAS),
/*BC*/	RAX(LDY),RAX(LDA),RAY(LDX),RAY(LAX),
/*C0*/	FIM(CPY),RIX(CMP),dIM(REP),DIX(DCP), /* presumed IM-DEC now REP */
/*C4*/	RZP(CPY),RZP(CMP),DZP(DEC),DZP(DCP),
/*C8*/	FIP(INY),RIM(CMP),FIP(DEX),RIM(SBX),
/*CC*/	RAB(CPY),RAB(CMP),DAB(DEC),DAB(DCP),
/*D0*/	ORL(BNE),RIY(CMP),bIN(CMP),DIY(DCP), /* BRL(DEC) now CMP */
/*D4*/	rZX(PEA),RZX(CMP),DZX(DEC),DZX(DCP), /* RZX(ZOP) now PEA(rel) */
/*D8*/	FIP(CLD),RAY(CMP),FIP(DEC),DAY(DCP),
/*DC*/	rAX(JMP),RAX(CMP),DAX(DEC),DAX(DCP), /* RAX(ZOP) now JMP */
/*E0*/	FIM(CPX),RIX(SBC),dIM(SEP),DIX(ISB), /* presumed IM-INC now SEP */
/*E4*/	RZP(CPX),RZP(SBC),DZP(INC),DZP(ISB),
/*E8*/	FIP(INX),RIM(SBC),OIP(NOP),RIM(SBC), /* E9 the offic. SBC */
/*EC*/	RAB(CPX),RAB(SBC),DAB(INC),DAB(ISB),
/*F0*/	ORL(BEQ),RIY(SBC),bIN(SBC),DIY(ISB), /* BRL(INC) now SBC */
/*F4*/	rZX(PEA),RZX(SBC),DZX(INC),DZX(ISB), /* RZX(ZOP) now PEA */
/*F8*/	OIP(SED),RAY(SBC),FIP(INC),DAY(ISB),
/*FC*/	rAX(JSR),RAX(SBC),DAX(INC),DAX(ISB), /* RAX(ZOP) now JSR(iax) */
};

static char *insstr(char *d,
	uint8 x)
{
	const char	*s;

	x&= 0x7F;
	s = x6510_opnametable[MIN(x,OPN_WDM)];
	*d++ = *s++, *d++ = *s++, *d++ = *s;

	return d;
}

char *Disasm_HexStr(char *d,
	unsigned v,
	unsigned c)
{
	const char	*hexchars = "0123456789ABCDEF";
	char		*r;

	if (!c)
		return d;
	r = d+=c;
	do
		*--d = hexchars[v&15], v>>= 4;
	while(--c);

	return r;
}
#define u8hex(x)		(d=Disasm_HexStr(d,x,2))
#define u16hex(x)		(d=Disasm_HexStr(d,x,4))

#define RMem(A)			readers[A>>dnsh](A)

/* "xx xxxx xxx $xxxx,X  @ $xxxx = xx" */
uint32 X6502_Disassemble(
	const X6502struct *XA,
	const readfunc *readers,
//	const uint8 *mem,
	uint32 A,
	uint32 mask,
	int dnsh,
	char *outstr)
{
	register char	*d;
	unsigned	arg, t;
	uint16		x;
	char		sp = dnsh<0 ? (dnsh=-dnsh,'\t'):' ';

	if (!readers || (d=outstr)==0)
		return A;

	t = RMem(A), ++A, A&= mask; u8hex(t);
	*d++ = sp; /* "xx " */

	x = x6510_opn_table[t];

	if ((t=x65xx_opsizes[x>>8&15])==0)
		*d++ = *d++ = *d++ = *d++ = sp, arg = 0;
	else
	{
		arg = RMem(A), ++A, A&= mask; u8hex(arg);
		if (t>1) /* "xxxx" */
			{arg|= (t=RMem(A))<<8, ++A, A&= mask; u8hex(t);}
		else /* "xx  " */
			*d++ = *d++ = sp;
	}
	*d++ = sp, d = insstr(d,x), *d++ = sp;

	if (x&OPB_BAD)
		*d++ = '!'; /* bad addressing mode for instruction */
	else if ((uint8)x>=OPN_6510_COUNT)
		*d++ = '*'; /* 65816-only */
	else
		*d++ = sp;
	switch(x>>8&15)
	{
	case OPM_IM: /* "#$xx"                 - u8 */
		*d++ = '#', *d++ = '$'; u8hex(arg);
		break;
	case OPM_RL: /* "$xxxx"                - *+u8 */
		arg = (A+(char)arg)&mask;
		*d++ = '$'; u16hex(arg);
		if (x&OPB_READ)
		{ /* really *+u16 */
			*d++ = ' ', *d++ = ';', *d++ = ' ';
			*d++ = '$'; u16hex((arg+(RMem(A)<<8))&mask);
		}
		break;
	case OPM_ZP: /* "$xx"                  - *u8 */
		*d++ = '$'; u8hex(arg);
		if (!XA)
			break;
		goto peekb; /* address already printed */
	case OPM_ZX: /* "$xx,X    @ $xxxx"     - u8[X] */
		*d++ = '$'; u8hex(arg); *d++ = ',', *d++ = 'X';
		if (!XA)
			break;
		arg+= XA->X;
		goto peekz;
	case OPM_ZY: /* "$xx,Y    @ $xxxx"     - u8[Y] */
		*d++ = '$'; u8hex(arg); *d++ = ',', *d++ = 'Y';
		if (!XA)
			break;
		arg+= XA->Y;
	peekz:	arg&= 0xFF; /* zero page all the way */
		*d++ = *d++ = sp;
		goto peeka;
	case OPM_IN: /* "($xx)    @ $xxxx"     - **u8 */
		*d++ = '(', *d++ = '$'; u8hex(arg); *d++ = ')';
		if (!XA)
			break;
		t = RMem(arg), ++arg, arg&= 0xFF, arg = t+(RMem(arg)<<8);
		*d++ = *d++ = sp;
		goto peeka;
	case OPM_IB: /* "($xxxx)"              - **u16 */
		*d++ = '(', *d++ = '$'; u16hex(arg); *d++ = ')';
		if (!XA || !(x&OPB_READ))
			break;
		/* bytes were prev. in the wrong order */
		t = RMem(arg), ++arg, arg&= 0xFFFF, arg = t+(RMem(arg)<<8);
		goto peeka;
	case OPM_IX: /* "($xx,X)  @ $xxxx"     - *u8[X] */
		*d++ = '(', *d++ = '$'; u8hex(arg); *d++ = ',', *d++ = 'X', *d++ = ')';
		if (!XA || !(x&OPB_READ))
			break;
		arg+= XA->X, arg&= 0xFF;
		t = RMem(arg), ++arg, arg&= 0xFF, arg = t+(RMem(arg)<<8);
		goto peeka;
	case OPM_IY: /* "($xx),Y  @ $xxxx"     - (*u8)[Y] */
		*d++ = '(', *d++ = '$'; u8hex(arg); *d++ = ')', *d++ = ',', *d++ = 'Y';
		if (!XA)
			break;
		t = RMem(arg), ++arg, arg&= 0xFF, arg = t+(RMem(arg)<<8);
		arg+= XA->Y, arg&= 0xFFFF;
		goto peeka;
	case OPM_AB: /* "$xxxx"                - *u16 */
		*d++ = '$'; u16hex(arg);
		if (!XA)
			break;
		goto peekb; /* address already printed */
	case OPM_AX: /* "$xxxx,X  @ $xxxx"     - u16[X] */
		*d++ = '$'; u16hex(arg); *d++ = ',', *d++ = 'X';
		if (!XA)
			break;
		arg+= XA->X, arg&= 0xFFFF;
		goto peeka;
	case OPM_AY: /* "$xxxx,Y  @ $xxxx"     - u16[Y] */
		*d++ = '$'; u16hex(arg); *d++ = ',', *d++ = 'Y';
		if (!XA)
			break;
		arg+= XA->Y, arg&= 0xFFFF;
		/* peek */
	peeka:	*d++ = *d++ = sp, *d++ = '@', *d++ = sp;
		*d++ = '$'; u16hex(arg);
	peekb:	if (!(x&OPB_READ))
			break;
		*d++ = sp, *d++ = '=', *d++ = sp; u8hex(RMem(arg));
		break;
	case OPM_IA:
		*d++ = 'A';
		break;
	case OPM_IC: /* "#$xx"                 - u8 */
		*d++ = ';', *d++ = '$'; u8hex(RMem(A));
		break;
	case OPM_SR: /* "$xx,S"                - S[u8] */
		*d++ = '$'; u8hex(arg); *d++ = ',', *d++ = 'S';
		break;
	case OPM_SI: /* "($xx,S),Y"            - S[u8][Y] */
		*d++ = '(', *d++ = '$'; u8hex(arg);
		*d++ = ',', *d++ = 'S', *d++ = ')', *d++ = ',', *d++ = 'Y';
		break;
	case OPM_DL: /* "[$xx]"                - **(u16*)u8 */
		*d++ = '[', *d++ = '$'; u8hex(arg); *d++ = ']';
		break;
	case OPM_DI: /* "[$xx],Y"              - ((u8*)*(u16*)u8)[Y] */
		*d++ = '[', *d++ = '$'; u8hex(arg); *d++ = ']', *d++ = ',', *d++ = 'Y';
		break;
	case OPM_IL: /* "[$xxxx]"              - **(u16*)u16 */
		*d++ = '[', *d++ = '$'; u16hex(arg); *d++ = ']';
		break;
	case OPM_DK: /* src/dst banks */
		*d++ = '$'; u8hex(arg); *d++ = ',';
		*d++ = '$'; u8hex(arg>>8);
		break;

	default: /* erase needless whitespace */
		if (d[-1]==sp)
			--d;
		if (d[-1]==sp)
			--d;
	}
	*d = 0; /* terminate */

	return A;
}
