/*
 interpreter.h
  v1.9.3

  Created: 2010.03.21
  Revised: 2014.08.01

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __INTERPRETER_H
#include "../enviro64.h"

#include <stdlib.h>

/* ---- constants */

#define INTERP_VERS			0x192

enum
{ /* error codes */
	_e_param_separation	= 1,	/* white space */
	_e_return		= 0x7FFFFFFF,
	_e_repeat		= 0x7FFFFFFE, /* loop current line */
	_e_repeat_last		= 0x7FFFFFFD, /* loop last line(s) */
	_e_repeat_end		= 0x7FFFFFFC, /* loop ended (not an expr line) */

	_e_general_error	=-1,	/* syntax/param error */
	_e_memory_error		=-2,	/* alloc failed */
	_e_read_error		=-3,	/* I/O read error */
	_e_write_error		=-4,	/* I/O write error */
	_e_incomplete_error	=-5,	/* expression missing symbols */
	_e_open_interrogator	=-6,	/* incomplete error for '?' */
	_e_no_identifier	=-7,	/* bad/missing ID */
	_e_handler_error	=-8,	/* handler not supplied for op */
	_e_stack_error		=-9,	/* too many or missing parentheses */
	_e_range_error		=-10,	/* displacement too large for size */
	_e_open_comment		=-11,	/* comment still open at end */
	_e_open_string		=-12,	/* unterminated string */
	_e_identifier_conflict	=-13,	/* ID already added */
	_e_op_error		=-14,	/* can't encode that! */
	_e_label_error		=-15,	/* can't add this as a label here */
	_e_language_error	=-16,	/* not equipped to process this */
	_e_macro_error		=-17,	/* no such macro */
	_e_cannot_optimize	=-18,	/* can't remove/reduce this code */
	_e_value_set		=-19,	/* already set, different value */
	_e_lvalue		=-20,	/* destination store is a number */
	_e_overlong_data	=-21,	/* data (e.g., string) too large */
	_e_no_field		=-22,	/* no such struct/list field */
	_e_no_expr		=-23	/* no expression for loop */
};

/* ---- types */

/* multi-parameter element types */
/* IMPORTANT: these do not fit into 16 bits (use LONG_TREC for 32 bits) */
#define TYP_PAIR(a,b)			(((a)&511)|(b)<<9)
#define TYP_TRIPLET(a,b,c)		(((a)&511)|((b)&31)<<9|(c)<<14)
#define TYP_QUADRUPLET(a,b,c,d)		(((a)&511)|((b)&31)<<9|((c)&31)<<14|(d)<<19)

enum
{ /* types - values -1 to -255 are pascal string sizes */
	_int8		=-256,
	_int16		=-257,
	_int32		=-258,
	_int64		=-259,

	_uint8		=-260,
	_uint16		=-261,
	_uint32		=-262,
	_uint64		=-263,
	/* printed in hexadecimal, unclipped */
	_hint8		=-264,
	_hint16		=-265,
	_hint32		=-266,
	_hint64		=-267,

	_bool		=-268,	/* non-zero byte */
	_ifrc16		=-269,	/* i6:10 - all 1 bits = -1u */
	_float		=-270,	/* IEEE 32-bit (s1.e8.m23) */
	_double		=-271,	/* IEEE 64-bit (s1.e11.m52) */

	_uflt8		=-272,	/* Lua 8-bit (e5.m3) */
	_flt16		=-273,	/* f16:0 */
	_dbl32		=-274,	/* d32:0 */
	_ifix16		=-275,	/* i8:8 */

	_ifixed		=-276,	/* i16:16 */
	_ufixed		=-277,	/* u16:16 */
	_ifract		=-278,	/* i2:30 */
	_ufract		=-279,	/* u2:30 */

	_alpha8		=-280,	/* 255=1.0 */
	_ufrc8		=-281,	/* u0:8 */
	_ufrc16		=-282,	/* u0:16 */
	_ufrc32		=-283,	/* u0:32 */

	_styl8		=-284,	/* [style.8: 1=bold,2=italic] */
	/* 1/30 sec. tick counts */
	_itck16		=-285,
	_utck16		=-286,
	_itck32		=-287,
 /* some multi-element types */
	_rgb24		= TYP_TRIPLET(_uint8,_uint8,_uint8),
	_argb32		= TYP_QUADRUPLET(_alpha8,_uint8,_uint8,_uint8),
	_rgb48		= TYP_TRIPLET(_uint16,_uint16,_uint16),
	_pt2d16		= TYP_PAIR(_int16,_int16),
	_pt3d16		= TYP_TRIPLET(_int16,_int16,_int16),
	_pt2dq		= TYP_PAIR(_ifixed,_ifixed),
	_pt3dq		= TYP_TRIPLET(_ifixed,_ifixed,_ifixed),
	_rect16		= TYP_QUADRUPLET(_int16,_int16,_int16,_int16),
	_fnts80		= TYP_QUADRUPLET(-65,_int16,_styl8,_uint8)
};

/* unsigned long type */
#if SIZEOF_LONGINT>4
 #define _ulong				_uint64
#else
 #define _ulong				_uint32
#endif

#ifdef mac
 #define CHR_NE				'\255' /* inequal */
 #define CHR_LE				'\262' /* lesser or equal */
 #define CHR_GE				'\263' /* greater or equal */
 #define CHR_RA				'\303' /* radical (square root) */
 #define CHR_NB				'\312' /* non-breaking space */
 #define CHR_MN				'\320' /* minus "en-dash" */
 #define CHR_DV				'\326' /* divide */
 #define CHR_NL				'\015' /* new line: CR */
#else /* ISO-8895-1-compatible */
/* Windows-1252 */
 //no NE (U+2260)
 //no official LE (U+2264)
 //no official GE (U+2265)
 //no RA (U+221A)
 #define CHR_NB				'\240' /* non-breaking space */
 #define CHR_DV				'\367' /* divide */
 //en-dash (U+2013) not in ISO-8859-1
 #define CHR_MN				'\226' /* minus "en-dash" */
 #define CHR_NL				'\012' /* new line: LF */
#endif

#ifndef CHR_MN
 #define CHR_MN				'-'
#endif

/* ---- macros */

#define CASE_NUMERIC			case'0':case'1':case'2':case'3':case'4': \
	case'5':case'6':case'7':case'8':case'9':
#define CASE_ALPHABETIC			case'A':case'B':case'C':case'D':case'E': \
	case'F':case'G':case'H':case'I':case'J':case'K':case'L':case'M':case'N': \
	case'O':case'P':case'Q':case'R':case'S':case'T':case'U':case'V':case'W': \
	case'X':case'Y':case'Z':	case'a':case'b':case'c':case'd':case'e': \
	case'f':case'g':case'h':case'i':case'j':case'k':case'l':case'm':case'n': \
	case'o':case'p':case'q':case'r':case's':case't':case'u':case'v':case'w': \
	case'x':case'y':case'z':
#define CASE_BINARY_OPERATOR		case'%':case'&':case'*':case'/':case'<': \
	case'=':case'>':case'^':case'|':

#define IS_EOL(c)			((c)-10u<4u)
#define IS_DIG(c)			((c)-'0'<10u)
#define IS_ALA(c)			(((c)|32)-'a'<26u)

/* efficient, in-function copy routine that's fast for small cases */
/* this advances two int8* streams */
#if env68k || envppc /* M68020+ */
 #define _amcpy(d,s,z)			do{ \
	if (z&1) *d++=*s++; \
	if (z&2) *((int16*)d)++=*((int16*)s)++; \
	if (z>>=2) do *((int32*)d)++=*((int32*)s)++; while(--z); \
 }while(0)
#elif (envx64 || envx86) && defined(__POCC__)
 extern void __cdecl __movsb(uint8*, const uint8*, size_t); /*v6.5*/
 #define _amcpy(d,s,z)			do{ \
	__movsb((uint8*)d,(const uint8*)s,z), d+=z, s+=z; \
 }while(0)
#else
 #define _amcpy(d,s,z)			do{ \
	if (z&1) *d++=*s++; \
	if (z&2) *d++=*s++, *d++=*s++; \
	if (z>>=2) do *d++=*s++,*d++=*s++,*d++=*s++,*d++=*s++; while(--z); \
 }while(0)
#endif

/* ---- typedefs/structures */

#ifdef env68k
 #define double_t			long double
#else
 #define double_t			double
#endif

#if TARGET_OS_MAC
 #define INFINITY			__inf()
 double_t scalb(double_t,
	long);
#elif __POCC__
 #define scalb				scalbn
#endif

#if EVAL_FLOATING
typedef double		evalvalue;
typedef double		evalparse;
 #define sizeof_evalvalue		8
 #define EVAL_HUGEVAL			(evalvalue)((1L<<22)*(evalvalue)(1L<<30)-1)
 #define parse_eval(s,e,i)		parse_number(s,e,i,1)
#else
typedef int32		evalvalue;
typedef long int	evalparse;
 #define sizeof_evalvalue		4
 #define EVAL_HUGEVAL			((evalvalue)1<<(sizeof(evalvalue)<<3)-1)
 #define parse_eval(s,e,i)		readanyinteger(s,e-s,i)
#endif

/* text/data object */
typedef struct tobj
{
	int32		z; /* size - bit #31 = not allocated */
	char		*s;
} tobj;

/* reference object */
typedef struct tref
{
	int32		o; /* offset */
	char		*p;
} tref;

/* still a work in progress */
/*
#define IS_STR(l)			((l)->typ>=0)
#define IS_ITM(l)			((l)->typ>=-0x40000000)
#define IS_LST(l)			((l)->typ<-0x40000000)
#define IS_PTR(l)			((l)->idx<0)
*/
typedef struct list_item/*(16)*/
{
	int32		typ; /* var type; pos.=string size; lists are <-0x40000000 */
	int32		idx; /* non-list case: [alloc.1] [count.31] */
	union
	{
		void		*ptr;
		char		buf[8]; /* empty is NULL malloc */

		struct list_item	*ref;
	} data;
} list_item;

/* ---- globals */

#ifndef NO_TABLES
 #define IUCT_ISWID(X)			(interpUCType[(uint8)(x)]<(char)0xA5)
 #define IUCT_ISHEX(X)			(interpUCType[(uint8)(x)]<(char)0x90)
 #define IUCT_ISWSP(x)			((uint8)interpUCType[(uint8)(x)]<2u)
 #define IUCT_ISBLK(x)			((uint8)interpUCType[(uint8)(x)]<7u)
extern const char interpUCType[256];
#endif

/* ---- prototypes */

/* system standard */

#ifndef NO_TABLES
int strcasecmp(const char *a,
	const char *b);
int strncasecmp(const char *a,
	const char *b,
	size_t z);
#endif

/* extended standard */

/* parses ASCII-compatible chars to return length of an entire line; 2011.08.13 */
/* escape char can be anything, such as '\', or 0 (null) for no escape char */
long sreadline(const char *p,
	long length,
	uint32 *lecount,
	char escape);
/* same as sreadline, but uses fgetc, allocates string w. reallocfunc; */
/* output is capped to 0x7FFFFFF0 bytes */
char *freadline(struct FILE*,
	uint32 *length, /* in/out; can be NULL (C-string result) */
	char *store, /* in-place realloc'd data; can be NULL */
	void *(reallocfunc)(void*, /* can be NULL (default: realloc) */
		size_t),
	uint32 *lecount, /* line-ending count, in/out; can be NULL */
	int escape); /* can be 0 (no escape char) */

/* reads and buffers the string for an entire file */
char *freadfile(struct FILE *f,
	int32 *z, /* out length; can be NULL */
	int (delim)(const char *s,
		const char *e));

int makebasestring(void *outstring,
	const char *basechars, /* can be NULL (use default chars) */
	long value,
	unsigned base,
	int minlength,
	char sign); /* pass zero to interpret unsigned */

/* returns ptr to next character */
const char *readanyinteger(const char *str,
	long length,
	long int *outvalue);
/* added 2013.04.01 */
const char *readanyint64(const char *s,
	const char *e,
	int64 *v);

const char *parse_number(const char *s,
	const char *e,
	double *v,
	uint8 r); /* real (take the fractional point) */
double_t scal10(double_t v,
	int e);
/* converts a fixed-point number to a decimal-adjusted double_t */
double_t fix2double(uint32 v,
	int m, /* binary scale - negate for 'signed' */
	int c); /* maximum fraction digit count */
/* Lua-style 8-bit ufloat conversion */
uint8 dbl2uflt(double v);
double uflt2dbl(uint8 v);

/* this requires that the characters have already been processed */
int parse_literal(const char *s,
	int64 *r,
	long z);

const char *parse_eval_literal(const char *str,
	const char *e,
	char endquote,
	evalvalue *v);

/* returns next position (after closing quote) */
const char *parse_textloop(const char *s,
	const char *e,
	char endquote,
	int32 *z, /* in max./out length; negative on memory error */
	uint8 *isPascal, /* can be NULL */
	char **r, /* will use realloc to buffer this; can be NULL */
	void *u, /* ptr passed to b func */
	void (*b)(void *u,
		char v));

/* prints an escaped string to file output */
/* the bounding quote chars are not printed by this function */
/* returns total printed char count */
uint32 print_tchars(struct FILE *f,
	const char *s,
	long z,
	char q);

#ifndef NO_TABLES
const char *parse_wordbreak(const char *s, /* does nothing if either are NULL */
	const char *e,
	uint8 *h); /* non-zero first char hash (can be NULL); 2011.09.15 */
const char *parse_rwordbreak(const char *s,
	const char *f);
#endif

const char *parse_skipwsp(const char *s,
	const char *e);
const char *parse_rskipwsp(const char *s,
	const char *f);

/* each division by zero will yield an infinite (or largest possible value) */
const char *parse_parameter(const char *s,
	long length,
	char *outstring, /* can be NULL; otherwise, expects at least 24 chars allocated */
	int *outerr,
	const char *(*getv)(const char *s,
		const char *e,
		evalvalue *i, /* may be NULL */
		void *u),
	const char *(*setv)(const char *s,
		const char *e,
		evalvalue i,
		int *err,
		void *u),
	void *userdata, /* for getv */
	evalvalue (*evalindir)(evalvalue)); /* indirection; (`@`); can be NULL */

/* returns parsed length */
long evaluate_line(const char *s,
	const char *e,
	evalvalue *v,
	int *err,
#ifdef ALLOW_LONG_COMMENTS
	uint8 *com, /* within a long-comment (` #[..]# `)*/
#endif
	const char *(*getv)( /* can be NULL - return NULL on error */
		const char *s,
		const char *e,
		evalvalue*, /* may be NULL */
		void *u),
	const char *(*setv)( /* can be NULL - return NULL on error */
		const char *s,
		const char *e,
		evalvalue,
		int *err,
		void *u),
	void *u,
	evalvalue (*evalindir)(evalvalue)); /* can be NULL */

/* ---- prototypes/var.c */

const char *elem_tname(int32 t);

int32 elem_tsize(int32 typ);

/* basic string storage */

/* tobj delete (free and zero) */
/* returns 1 if deallocation occured, or 0 */
int deltobj(tobj *o);

/* tobj unreference (allocate if a ref ptr) */
/* returns 1 if allocation occured, _e_memory_error on failure, or 0 */
int unrtobj(tobj *o);

/* conversion */

double cvt_tdata(int32 z,
	const void *p);
int set_tdata(int32 *type,
	const tobj *s,
	int8 *p,
	int8 lit,
	int8 first);

const int8 *read_typ_be(void *p, /* must be valid */
	const int8 *s, /* must be valid */
	int32 typ);

/* printing */

int print_string(struct FILE *f,
	const char *s,
	int32 z,
	int t,
	int idbreak);
int print_tdata(struct FILE *f,
	int32 z,
	const void *p,
	int idbreak);
/* prints the text equivalent of an item */
#if 0
int eval_print(struct FILE *f,
	const list_item*);
#endif

/* scanning */

/* single-allocation data case */
/* returns offset to data, or ~length if not in list */
/* should never return zero */
int32 get_hbuf_entry(const char *name,
	int nameLength,
	uint8 hash, /* as returned by parse_wordbreak */
	const uint8 *u, /* zero-hash terminated list */
	uint32 dz); /* data size - must be valid */
/* no in-data case (no null termination needed) */
/* returns index or ~length */
int32 get_hname_index(const char *name,
	int nameLength,
	uint8 hash,
	const uint8 *u,
	uint32 tz); /* total size */

const char *parse_entrybreak(const char *str,
	const char *e,
	int32 context);

/* basic variable storage */

const char *eval_getv(const char *s,
	const char *e,
	evalvalue *i,
	void *u);

const char *eval_setv(const char *s,
	const char *e,
	evalvalue i,
	int *err,
	void *u);

/* call this when done with calls to eval_getv/eval_setv */
int eval_free(list_item*);

#define __INTERPRETER_H
#endif
