/*
 var.c
  interpreter 1.9.1: variable storage

  Created: 2013.11.23
  Revised: 2014.03.20

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#if TARGET_OS_MAC
 #include <fp.h>	/* for nan */
#else
 #include <math.h>
#endif

#include "interpreter.h"

/* ---- globals */

int8		_typSzArray[32] =
{ /* prev. used the `[..] = ,` method, not supported by SC/MrC */
  /* because the enum runs negative, the order here is in reverse (vs. interpreter.h) */
	4, 2, 2, 1, /* _itck32, _utck16, _itck16, _styl8 */
	4, 2, 1, 1, /* _ufrc32, _ufrc16, _ufrc8, _alpha8 */
	4, 4, 4, 4, /* _ufract, _ifract, _ufixed, _ifixed */
	2, 4, 2, 1, /* _ifix16, _dbl32, _flt16, _uflt8 */
	8, 4, 2, 1, /* _double, _float, _ifrc16, _bool */
	8, 4, 2, 1, /* _hint64, _hint32, _hint16, _hint8 */
	8, 4, 2, 1, /* _uint64, _uint32, _uint16, _uint8 */
	8, 4, 2, 1  /* _int64, _int32, _int16, _int8 */
};

/* ---- basic string storage */

int deltobj(tobj *o)
{
	if (o->z<=0 || !o->s)
		{o->s = NULL, o->z = 0; return 0;}
	free(o->s);
	o->s = NULL, o->z = 0;
	return 1;
}

int unrtobj(tobj *o)
{
	char		*d;
	int32		z;

	if (!o->s)
		{o->z = 0; return 0;}
	if ((z=o->z)>=0)
		return 0;
	if ((d=malloc(z&=0x7FFFFFFF))==NULL)
		return _e_memory_error;
	memcpy(d,o->s,z);
	o->s = d;
	return 1;
}

/* ---- scanning */

int32 get_hbuf_entry(const char *n,
	int z,
	uint8 h,
	const uint8 *u,
	uint32 dz)
{
	const uint8	*l;

	if ((l=u)==NULL)
		return ~0;
	for(++dz; *l++; l+=*l,l+=dz)
		if (*l==z && h==l[-1] && !memcmp(l+1,n,z))
			return 1+z+l-u;
	return u-l; /* becomes ~total */
}

int32 get_hname_index(const char *n,
	int z,
	uint8 h,
	const uint8 *u,
	uint32 tz)
{
	const uint8	*l, *e;
	int32		c;

	if ((l=u)==NULL)
		return ~0;
	for(e=l+tz,c=0; l<e; l+=*l,++c)
		if (*++l==z && h==l[-1] && !memcmp(l+1,n,z))
			return c;
	return u-l; /* becomes ~total */
}

/* ---- calculation */

int32 elem_tsize(int32 typ)
{
	int32		t = 0, z;
	int		h = 9;

	if (typ>=0)
		return typ;
	z = typ|~511;
	goto start;
	do
	{
		h = 5, z = (typ|(~287|~511))&~256;
	start:	if (--z>-257)
			t-= z;
		else if ((z+=288)<0)
			return _e_op_error;
		else
			t+= _typSzArray[z];
	} while((typ>>=h)<-31u);
	return t;
}

double cvt_tdata(int32 z,
	const void *p)
{
	char		b[8];

	if (z<-255)
		switch(z)
		{
		case _hint8:
		case _uint8:
		case _styl8: return *(const uint8*)p;
		case _hint16:
		case _uint16: return *(const uint16*)p;
		case _hint32:
		case _uint32: return *(const uint32*)p;
		case _hint64:
		case _uint64: return int64tox(*(const uint64*)p);
		case _int8: return *(const int8*)p;
		case _int16: return *(const int16*)p;
		case _int32: return *(const int32*)p;
		case _int64: *(int64*)b = *(const int64*)p;
			return int64tox(*(int64*)b);
		case _uflt8: return uflt2dbl(*(const uint8*)p);
		case _bool: return *(const int8*)p!=0;
		case _ufrc8: return fix2double(*(const uint8*)p,8,3);
		case _ufrc16: return *(const uint16*)p/65536.;
		case _ufrc32: return *(const uint32*)p/4294967296.;
		case _ifix16: return fix2double(*(const int16*)p,-8,3);
		case _ifrc16: return fix2double(*(const int16*)p,-10,4);
		case _itck16: z = *(const int16*)p; goto ftck;
		case _utck16: z = *(const uint16*)p; goto ftck;
		case _itck32: z = *(const int32*)p;
		ftck:	return z/30.;
		case _ifixed: return *(const int32*)p/65536.;
		case _ufixed: return *(const uint32*)p/65536.;
		case _ifract: return *(const int32*)p/1073741824.;
		case _ufract: return *(const uint32*)p/1073741824.;
		case _flt16:
			((int16*)b)[BIG_ENDIAN] = 0;
			((int16*)b)[!BIG_ENDIAN] = *(const int16*)p;
			return *(float*)b;
		case _float: return *(const float*)p;
		case _dbl32:
			((int32*)b)[BIG_ENDIAN] = 0;
			((int32*)b)[!BIG_ENDIAN] = *(const int32*)p;
			return *(double*)b;
		case _double: return *(const double*)p;

		default: goto end;
		}
	if (z<0)
		z =-z, p = 1+(const int8*)p;
	if (parse_number(p,z+(const char*)p,(double*)b,1))
		return *(double*)b;
end:	return nan("255");
}

/* ---- parsing */

static int parsehex(const char *s,
	const char *e,
	int8 *d,
	int z)
{
	int		t = 0;
	uint16		v = 0x80, c;

	if (s<e && *s=='#')
		++s;
	while(z>0)
	{
		if (s>=e)
			return _e_op_error;
		if ((c=*s++)=='\t' || c==' ')
			continue;
		if ((c-='0')<=9u)
			v = v<<4|c;
		else if ((c-='A'-'0')<=25u || (c-='a'-'A')<=25u)
			c+= 10, v = v<<4|c;
		else
			return _e_op_error;
		if (v&0x8000)
			d[t++] = v, v = 0x80, --z;
	}
	return parse_skipwsp(s,e)!=e ? _e_op_error:t;
}

static int equalstr(const char *d,
	const char *s)
{
	while(*s++)
		if (s[-1]!=tolower((uint8)*d++))
			return 0;
	return 1;
}

int set_tdata(int32 *type,
	const tobj *s,
	int8 *p,
	int8 lit,
	int8 first)
{
	const char	*u, *e, *n;
	double		v;
	int8		*d;
	int32		z, t, typ;

	if (!type)
		return _e_read_error;
	if ((d=p)==NULL)
		return _e_write_error;
	if (s)
		u = s->s, z = u ? s->z&0x7FFFFFFF:0;
	else
		u = 0, z = 0;
	if ((typ=*type)<0)
	{
		typ|=~511;
		if (!first)
			typ|=~287, typ&=~256;
	}
	t = ABS(typ), e = u+z;
	if (typ<-255)
	{
		int8		uns;

		if (typ>=_hint64)
			if (lit)
			{
				if (z>8)
					goto bdat;
				if (parse_literal(u,(int64*)&v,z)<0)
					{t = _e_range_error; goto end;}
				goto t5;
			} else if ((typ<=_uint64 || typ==_int64)
					&& (n=readanyint64(u,e,(int64*)&v))!=NULL
					&& (n>=e || *n!='.'))
			{
			t5:	switch(typ&3)
				{
				default: *d = int64lo(*(int64*)&v), ++d; break;
				case 3: *(int16*)d = int64lo(*(int64*)&v), d+= 2; break;
				case 2: *(int32*)d = int64lo(*(int64*)&v), d+= 4; break;
				case 1: *(int64*)d = *(int64*)&v, d+= 8;
				}
				goto end;
			}
		uns = 0;
		if (*type==_rgb24 && (*u=='#' || memchr(u,' ',z)))
		{
			if ((t=parsehex(u,e,d,3))>=0)
				d+= t;
			*type =~0;
			goto end;
		}
		if ((n=parse_number(u,e,&v,1))!=NULL)
		{
			if (*(u=n)=='u' || *u=='U')
				++u, uns = 1;
		} else if (z==4)
			if (equalstr(u,"true"))
				v = 1;
			else if (equalstr(u,"-inf"))
				v =-INFINITY;
			else if (typ==_styl8 && equalstr(u,"bold")) //(typ|1)==_styl16
				v = 1;
			else if (typ==_ifrc16 && equalstr(u,"none"))
				v =-.001;
			else
				goto tbool;
		else if (z==5 && equalstr(u,"false"))
			v = 0;
		else if (z==3 && equalstr(u,"inf"))
			v = INFINITY;
		else if (z==6 && typ==_styl8 && equalstr(u,"italic")) //(typ|1)==_styl16
			v = 2; /* italic */
		else
		tbool:	if (typ==_bool)
				v = s && s->z ? *u:0;
			else
				goto bdat;
		switch(typ)
		{
		case _itck32:
			if (uns)
				goto r16;
			v*= 30./1073741824;
		case _ifract: v*= 16384;
		case _ifixed: v*= 65536, v+= v<0 ? -.5:.5;
		case _int32:
			if (v>IMAX(int32))
				*(int32*)d = IMAX(int32);
			else
				*(int32*)d = v;
			break;
		case _utck16: 
			if (uns)
				goto r16;
			v*= 30./65536;
		case _ufrc16: v*= 65536, v+= .5;
		case _uint16:
			if (v<0 && uns)
				v+= 65536;
			if (v>65535)
				*(uint16*)d = 65535;
			else
				*(uint16*)d = v;
			break;
		case _int8:
			if (v>IMAX(int8))
				*(int8*)d = IMAX(int8);
			else
				*(int8*)d = v;
			break;
		case _itck16:
			if (uns)
				goto r16;
			v*= 30./1024;
		case _ifrc16:
			if (uns)
				goto r16;
			v*= 4;
		case _ifix16: v*= 256;
			v+= v<0 ? -.5:.5;
		case _int16:
		r16:	if (v>IMAX(int16))
				*(int16*)d = IMAX(int16);
			else
				*(int16*)d = v;
			break;
		case _int64:
			if (v>=-IMIN64)
				int64sethl(*(int64*)d,IMAX(int32),-1ul);
			else
				int64xset(*(int64*)d,v);
			break;
		case _uflt8:
			if (v<0 && uns)
				v+= 4294967296.;
			v = dbl2uflt(v);
			goto u8;
		case _alpha8:
			v*= 255./256;
		case _ufrc8:
			v*= 256, v+= .5;
		case _uint8:
		u8:	if (v<0 && uns)
				v+= 256;
			if (v>255)
				*(uint8*)d = 255;
			else
				*(uint8*)d = v;
			break;
		case _hint8: case _styl8:
			*(uint8*)d = (int32)v;
			break;
		case _hint16: *(uint16*)d = (int32)v; break;
		case _ufrc32: v*= 4;
		case _ufract: v*= 16384;
		case _ufixed: v*= 65536, v+= .5;
		case _uint32:
			if (v<0 && uns)
				v+= 4294967296.;
			if (v>-1ul)
				*(uint32*)d =-1ul;
			else
				*(uint32*)d = v;
			break;
		case _hint32: /* fmod method is too expensive */
			int64xset(*(int64*)&v,v);
			*(uint32*)d = int64lo(*(int64*)&v);
			break;
		case _uint64:
			if (v<0 && uns)
				v+=-2*IMIN64;
			if (v>=-2*IMIN64)
				int64sethl(*(int64*)d,-1ul,-1ul);
			else
		case _hint64: int64xset(*(uint64*)d,v); break;
		case _bool: *(int8*)d = v!=0; break;
		case _float: *(float*)d = v; break;
		case _double: *(double*)d = v; break;
		case _flt16: *(float*)&v = v;
			if (~*(int32*)&v&0x7F800000)
				*(int32*)&v+= 0x8000;
			*(int16*)d = ((int16*)&v)[!BIG_ENDIAN];
			break;
		case _dbl32: if (~((int32*)&v)[!BIG_ENDIAN]&0x7FF00000)
				((int32*)&v)[!BIG_ENDIAN]+= ((uint32*)&v)[BIG_ENDIAN]>>31;
			*(int32*)d = ((int32*)&v)[!BIG_ENDIAN];
			break;
		bdat:
		default: t = _e_op_error;
			goto end;
		}
		d+= _typSzArray[typ+287];
	} else
	{
		if (typ<0)
			++d;
		memset(d,0,ABS(typ)); /* zero out undefined chars - C term. auto */
		if (z>t)
			z = t;
		if (typ<0)
			d[-1] = z;
		memcpy(d,u,z);
		d+= t;
	}
end:	if ((*type>>=first ? 9:5)>~31u)
		*type = 0;
	return t<0 ? t:d-p;
}

const int8 *read_typ_be(void *p,
	const int8 *s,
	int32 typ)
{
	int32		z;
	int8		*d = p;
	int		h = 9;

	if (typ>=0)
		{_amcpy(d,s,typ); goto end;}
	z = typ|~511;
	goto start;
	do
	{
		h = 5, z = (typ|(~287|~511))&~256;
	start:	switch(z)
		{
		case _int8:
		case _uint8:
		case _hint8:
		case _bool:
		case _ufrc8:
		case _uflt8:
		case _alpha8:
		case _styl8:
			*d++ = *s++;
			break;
		case _int16:
		case _uint16:
		case _hint16:
		case _ufrc16:
		case _flt16:
		case _ifix16:
		case _ifrc16:
		case _itck16:
		case _utck16:
			*(uint16*)d = va_be16(s), d+= 2;
			break;
		case _int32:
		case _uint32:
		case _hint32:
		case _ifixed:
		case _ufixed:
		case _ifract:
		case _ufract:
		case _ufrc32:
		case _dbl32:
		case _float:
		case _itck32:
			*(uint32*)d = va_be32(s), d+= 4;
			break;
		case _int64:
		case _uint64:
		case _hint64:
		case _double:
			((uint32*)d)[!BIG_ENDIAN] = va_be32(s);
			((uint32*)d)[BIG_ENDIAN] = va_be32(s);
			d+= 8;
			break;

		default: if (--z>-257)
			{
				z =-z;
				_amcpy(d,s,z);
			}
		}
	} while((typ>>=h)<-31u);
end:	return s;
}

/* ---- basic eval */

const char *eval_getv(const char *s,
	const char *e,
	evalvalue *i,
	void *u)
{
	const uint8	*d;
	int32		x;
	uint8		h;

	if (!u)
		return NULL;
	d = *(uint8**)u;
	e = parse_wordbreak(s,e,&h), ++s;
	if ((x=e-s)>255u)
		return NULL; /* bad length */
	if ((x=get_hbuf_entry(s,x,h,d,sizeof(evalvalue)+1))<=0)
		return NULL;
	if (i)
		memcpy(i,d+x+1,sizeof(evalvalue));
	return e;
}

const char *eval_setv(const char *s,
	const char *e,
	evalvalue i,
	int *err,
	void *u)
{
	uint8		*d, h;
	int32		x, z;

	if (!u)
	{
		if (err)
			*err = _e_write_error;
		return NULL;
	}
	d = *(uint8**)u;
	e = parse_wordbreak(s,e,&h), ++s;
	if ((z=e-s)>255u)
		return NULL; /* bad length */
	if ((x=get_hbuf_entry(s,z,h,d,sizeof(evalvalue)+1))>=0)
	{
		if (!x)
			return NULL;
		d+= x;
	} else
	{ /* (-x) abs. end offset; automagic +1 includes null term. */
		if ((d=realloc(d,2+1+sizeof(evalvalue)+z-x))==NULL)
		{
			if (err)
				*err = _e_memory_error;
			return NULL;
		}
		*(uint8**)u = d, d-= x;
		d[-1] = h, *d++ = z;
		memcpy(d,s,z), d+= z, d[sizeof(evalvalue)+1] = 0;
	}
	memcpy(d+1,&i,sizeof(evalvalue));
	return e;
}

int eval_free(list_item *p)
{
	if (!p)
		return 0;
	//if (p->data.ptr)
	//	{free(p->data.ptr); p->data.ptr = NULL;}
	free(p);
	return 1;
}

#define tcase(t)		case _##t:return#t;

const char *elem_tname(int32 t)
{
	switch(t)
	{
	tcase(int8)
	tcase(int16)
	tcase(int32)
	tcase(int64)

	tcase(uint8)
	tcase(uint16)
	tcase(uint32)
	tcase(uint64)

	tcase(hint8)
	tcase(hint16)
	tcase(hint32)
	tcase(hint64)

	tcase(bool)
	tcase(ifrc16)
	tcase(float)
	tcase(double)

	tcase(uflt8)
	tcase(flt16)
	tcase(dbl32)
	tcase(ifix16)

	tcase(ifixed)
	tcase(ufixed)
	tcase(ifract)
	tcase(ufract)

	tcase(alpha8)
	tcase(ufrc8)
	tcase(ufrc16)
	tcase(ufrc32)

	tcase(styl8)
	tcase(itck16)
	tcase(utck16)
	tcase(itck32)

	default: return "";
	}
}

/* ---- printing */

int print_string(FILE *f,
	const char *s,
	int32 z,
	int t,
	int w)
{
	if (z<0)
	{
		z =-z;
		if (*(const uint8*)s<z)
			z = *(const uint8*)s;
		++s;
	} else /* don't print trailing nulls (C-string default) */
		while(z)
			if (s[--z])
				{++z; break;}
	if (z && w && parse_entrybreak(s,s+z,t)==s+z)
		return fwrite(s,z,1,f);
	return (fputc('"',f)>0)+print_tchars(f,s,z,'"')
		+(fputc('"',f)>0);
}

int print_tdata(FILE *f,
	int32 typ,
	const void *p,
	int w)
{
	char		b[32];
	double		v;
	int32		r, z, m, t = 0;
	int		h = 9;

	goto start;
	do
	{
		h = 5, t+= !fputc(',',f);
	start:	if ((z=typ)<0)
		{
			z|=~511;
			if (h&4)
				z|=~287, z&=~256;
		}
		if (z>-256)
			{t+= print_string(f,p,z,1,w); continue;}
		switch(z)
		{
		case _hint8: r = *(const uint8*)p; goto h10;
		case _hint16: r = *(const uint16*)p; goto h10;
		case _hint32: r = *(const uint32*)p;
		h10:	t+= fprintf(f,"0x%X",
				r);
			continue;
		case _hint64:
			t+= fprintf(f,"0x%llX",
				*(const uint64*)p);
			continue;
		case _int8: r = *(const int8*)p; goto i10;
		case _int16: r = *(const int16*)p; goto i10;
		case _int32: r = *(const int32*)p;
		i10:	t+= fprintf(f,"%d",
				r);
			continue;
		case _int64:
			*(int64*)b = *(const int64*)p;
			t+= fprintf(f,"%lld",
				*(int64*)b);
			continue;
		case _uint8:
		case _styl8: r = *(const uint8*)p; goto u10;
		case _uflt8:
			t+= fprintf(f,"%.3g",
				uflt2dbl(*(const uint8*)p));
			continue;
		case _uint16: r = *(const uint16*)p; goto u10;
		case _uint32: r = *(const uint32*)p;
		u10:	t+= fprintf(f,"%u",
				r);
			continue;
		case _uint64:
			t+= fprintf(f,"%llu",
				*(const uint64*)p);
			continue;
		case _bool: r = *(const int8*)p!=0; goto u10;
		case _ufrc8: v = fix2double(*(const uint8*)p,8,3);
			goto xfs;
		case _ufrc16: r = *(const uint16*)p, m = 16; goto frc;
		case _ufrc32: m = 32; goto x30;
		case _ifix16: v = fix2double(*(const int16*)p,-8,3);
			goto xfs;
		case _ifrc16:
			if (*(const int16*)p==-1)
				{t+= fputs("-1u",f) ? 0:3; continue;}
			v = fix2double(*(const int16*)p,-10,4);
			goto xfs;
		case _itck16: r = *(const int16*)p;
			goto ftck;
		case _utck16: r = *(const uint16*)p;
			goto ftck;
		case _itck32: r = *(const int32*)p;
		ftck:	m = 0;
			if (r<0)
				r =-r, m = 1, *b = '-';
			m+= makebasestring(b+m,NULL,r/30,10,0,0);
			if (r%=30)
			{
				b[m++] = '.';
				m+= makebasestring(b+m,NULL,(r*20+3)/6,10,2,0);
				if (b[m-1]=='0')
					--m;
			}
			b[m] = 0;
			break;
		case _ifixed: m =-16; goto ifrc;
		case _ufixed: m = 16;
		ifrc:	r = *(const int32*)p;
		frc:	v = fix2double(r,m,5);
			goto xfs;
		case _ifract: m =-30; goto x30;
		case _ufract: m = 30;
		x30:	v = fix2double(*(const int32*)p,m,10);
		xfs:	m = sprintf(b,"%.10g",
				v);
			break;
		case _flt16:
			((int16*)b)[BIG_ENDIAN] = 0;
			((int16*)b)[!BIG_ENDIAN] = *(const int16*)p;
			m = sprintf(b,"%.4g",
				*(float*)b);
			break;
		case _float:
			m = sprintf(b,"%.8g",
				*(const float*)p);
			break;
		case _dbl32:
			((int32*)b)[BIG_ENDIAN] = 0;
			((int32*)b)[!BIG_ENDIAN] = *(const int32*)p;
			m = sprintf(b,"%.8g",
				*(double*)b);
			break;
		case _double: v = *(const double*)p;
			m = sprintf(b,"%.16g",
				v);
			break;
		case _alpha8:
			m = sprintf(b,"%.3g",
				fix2double(*(const uint8*)p*8421505u>>16,15,3));
			break;

		default: return 0;
		}
		/* contains self-modifying expression to remove unnecessary zero */
		t+= fputs((*b=='-' && b[1]=='0' && b[2]=='.' && (b[1]='-',1))
			|| (*b=='0' && b[1]=='.') ? b+1:b,f)<0 ? 0:m;
	} while(p=elem_tsize(z)+(int8*)p,(typ>>=h)-1<-32u);
	return t;
}

#if 0
int eval_print(FILE *f,
	const list_item *l)
{
	int		i, t, y;

	if (!f || !l)
		return _e_general_error;
	i = t = 0;
	if (!(l->typ&1))
	{
		i = l->idx;
		if ((l=l->data.ref)==NULL)
			goto null;
	}
	y = l->typ>>1;
	if (y>0)
	{
		if (l->idx&0x80000000u)
		{
			const char	*e, *s = IS_PTR(l) ? l->data.ptr:l->data.buf;

			e = s+(l->idx&0x7FFFFFFF);
			if (fputc('(',f))
				y = 1;
			if (--y)
				for(;;)
				{
					if (s<e)
					{
						char	c = *s++;
						long	z = (uint8)*s++;

						t+= fprintf(f,"%c%.*s:",
							(c&0x3F)|0x40,z,s);
						s+= z;
					}
					y+= eval_print(f,++l);
					if (!--y)
						break;
					if (fputc(',',f))
						++t;
				}
			if (fputc(')',f))
				++t;
		}
		goto end;
	}
	if ((t=print_tdata(f,t,IS_PTR(l) ? l->data.ptr:l->data.buf,0))!=0)
		goto end;
null:	t+= fprintf(f,l->typ&~1 ? "%s<?>":"%s<null>",
		elem_tname(l->typ>>1));
end:	if (i)
		t+= fprintf(f,"+%d",
			i);
	return t;
}
#endif
