/*
 macos7-9.c
  cross-platform routines for MacOS 7.x-9.x

  Created: 2008.07.04 18:28 EST
  Revised: 2014.04.02

Saturday, July 11, 2009 3:02 PM EDST
  some stuff changed to work with CALL_NOT_IN_CARBON=0.
Friday, July 4, 2008 7:58 PM EDST
  finished implementation for wide positioning in iocopydata.
Monday, July 21, 2008 8:52 AM
  fixed spec2path.  faster, smaller code, less stack usage, simpler.  no more memset.
Monday, July 21, 2008 9:36 AM
  all FSMakeFSSpec -> PBMakeFSSpecSync.
Friday, August 1, 2008 2:59 AM
  hopefully fixed spec2path (MakeFSSpec may overwrite early).
Tuesday, August 18, 2009 10:59 PM
  fixed applyappledouble bugs (sizeof(adheader),(UInt8*)&adheader,...).
Tuesday, August 18, 2009 11:48 PM
  removing defer stuff for auto-deletion (can be slower, but less RAM used).
*/

#include <errno.h>
#include <stdlib.h>

#include <Resources.h>
#include <Devices.h>
#include <Gestalt.h>
#include <MacErrors.h>

#include "macos7-9.h"

#ifdef NOGLOBALS
 #define global
#else
 #define global static
#endif

/* ---- code */

/* something like this should already exist in MacOS standard library code
  -- the one found in IntEnv.o (MPW) is _mapOSerr */
static int translatemacoserror(int error)
{ /* alternative case-style (the plain switch generated more code and was slower) */
	/*if (error==opWrErr) // file already open w/ write perm. (returned the one open)
	  return 0;*/
	if (error==noErr)
		return 0;
	if ((error+=-fnfErr)==0) /* no such file or directory */
		return ENOENT; /* EISDIR = can't open directory for write */
	if ((error+=fnfErr-controlErr)==0 /* something broke in _Control */
			|| (error+=controlErr-statusErr)==0 /* ..in _Status */
			|| (error+=statusErr-openErr)==0 /* ..in _Open */
			|| (error+=openErr-closErr)==0 /* ..in _Close */
			|| (error+=closErr-ioErr)==0) /* I/O error (process level) */
		return EIO; /* EAGAIN = resource temporarily unavailable */
	if ((error+=ioErr-fLckdErr)==0 /* file is locked */
			|| (error+=fLckdErr-permErr)==0) /* file access permission denied */
		return EACCES;
	if ((error+=permErr-fBsyErr)==0) /* file is in use */
		return EBUSY;
	if ((error+=fBsyErr-dupFNErr)==0) /* file already exists */
		return EEXIST;
	if ((error+=dupFNErr-dskFulErr)==0) /* no space left on device */
		return ENOSPC;
	if ((error+=dskFulErr-wrPermErr)==0) /* attempt to modify a read-only file */
		return EROFS; /* EPERM = operation not permitted */
	if ((error+=wrPermErr-memFullErr)==0) /* out of memory */
		return ENOMEM;
	if ((error+=memFullErr-paramErr)==0) /* invalid argument */
		return EINVAL; /* ERANGE = result too large, E2BIG = arg list too long */
	if ((error+=paramErr-tmfoErr)==0) /* too many files open */
		return ENFILE; /* EMFILE = process, ENFILE = system */
	if ((error+=tmfoErr-rfNumErr)==0) /* bad file descriptor (or number) */
		return EBADF;
	if ((error+=rfNumErr-posErr)==0) /* negative position in file */
		return ESPIPE;
	if ((error+=posErr-noDriveErr)==0 /* drive isn't installed */
			|| (error+=noDriveErr-nsDrvErr)==0 /* no such drive */
			|| (error+=nsDrvErr-nsvErr)==0 /* no such volume */
			|| (error+=nsvErr-volGoneErr)==0) /* volume is unavailable */
		return ENODEV;
	if ((error+=volGoneErr-badUnitErr)==0) /* I/O system error */
		return ENXIO; /* no such device or address */
	if ((error+=badUnitErr-readErr)==0 /* something broke in _Read */
			|| (error+=readErr-writErr)==0) /* ..in _Write */
		return EPIPE;
	/*case eofErr: // EOF would be checked (fgetx() returns EOF constant, feof, etc.) */
	/* return ENOTBLK; // ??? */

	/* EINTR = interrupted function call */
	/* ECHILD = no child process */
	/* EXDEV = improper link attempted */
	/* EMLINK = too many links on a single file */
		/* (on MacOS, overflow of read-only files -> tmfoErr) */

	if ((error+=writErr-errFSNotAFolder)==0)
		return ENOTDIR;
	if ((error+=errFSNotAFolder-fileBoundsErr)==0) /* file is too large */
		return EFBIG;
	if ((error+=fileBoundsErr-procNotFound)==0) /* no process found (check for existance first) */
		return ENOEXEC; /* ENORSRC = No such process */
	if ((error+=procNotFound-protocolErr)==0) /* bad process protocol */
		return ENOTTY; /* inappropriate I/O control operation */
	return -1;
}

OSErr iocopydata(short dstref,
	short srcref,
	long int dstpos,
	long int srcpos,
	long length,
	void *buffer)
{
	XIOParam		file;
	char			resort[512];
	OSErr			rerr = noErr, werr = noErr;
	long			transfersize = length;

	file.ioCompletion = NULL;
	if ((file.ioBuffer=buffer)==NULL)
	{
		file.ioBuffer = resort;
		if ((unsigned)transfersize>sizeof(resort)
				&& (file.ioBuffer=NewPtr(transfersize=(unsigned)length>65536 ? 65536:length))==NULL
				&& (file.ioBuffer=NewPtr(transfersize>>=4))==NULL)
			file.ioBuffer = resort, transfersize = sizeof(resort); /* frustrated enough? */
	}

	if (srcpos>0x7FFFFFFFL || dstpos>0x7FFFFFFFL)
	{ /* extended format (MacOS 9.x+ only) */
		global long		sysversion = 0;

		if (sysversion==0 && Gestalt(gestaltSystemVersion,&sysversion)!=noErr)
			sysversion = 0x600; /* assume system 6.x */
		if (sysversion<0x900)
			return posErr;
		file.ioPosMode = fsFromStart|kUseWidePositioning;
	} else
		file.ioPosMode = fsFromStart;

	while(length>0)
	{
		file.ioReqCount = length>transfersize ? transfersize:length;

		if (file.ioPosMode&kUseWidePositioning)
			file.ioWPosOffset.hi = srcpos>>32, file.ioWPosOffset.lo = srcpos;
		else
			file.ioWPosOffset.hi = srcpos;

		file.ioRefNum = srcref;
		if ((rerr=PBReadSync((ParmBlkPtr)&file))!=noErr)
		{
			length = 0;
			if (rerr!=eofErr)
				break;
		}

		length -= (file.ioReqCount=file.ioActCount);

		if (file.ioPosMode&kUseWidePositioning)
		{
			srcpos = (file.ioWPosOffset.hi<<32)|file.ioWPosOffset.lo;
			file.ioWPosOffset.hi = dstpos>>32, file.ioWPosOffset.lo = dstpos;
		} else
			srcpos = file.ioWPosOffset.hi, file.ioWPosOffset.hi = dstpos;

		file.ioRefNum = dstref;
		if ((werr=PBWriteSync((ParmBlkPtr)&file))!=noErr)
			break;

		if (file.ioPosMode&kUseWidePositioning)
			dstpos = (file.ioWPosOffset.hi<<32)|file.ioWPosOffset.lo;
		else
			dstpos = file.ioWPosOffset.hi;
	}

	if (buffer==NULL && file.ioBuffer!=resort)
		DisposePtr(file.ioBuffer);

	return werr!=noErr ? werr:rerr;
}

/* yeah, i stole these from myself */
OSErr path2spec(const void *path, int length, FSSpec *file)
{
	HParamBlockRec		params;

	const char		*s, *p, *q = NULL;
	int				retries = 999; /* to avoid hang time, period. */
	OSErr			err = 0;

	if ((params.ioParam.ioMisc=(Ptr)file)==NULL)
		return -1;
	file->vRefNum = vRefNumLocal, file->parID = 0;
	if ((s=path)==NULL || length<=0)
		{file->name[0] = 0; return 0;}
	p = s;
	do
	{
		UInt8		buf[256], *d, *e;
		int			z;
		OSErr		perr = noErr;

		/* copy chars; FSSpec will point to parent, each next string offset = next to previous colon */
		d = 1+buf;
		if (q==NULL)
			s = path, z = length;
		else
			s = q-1, z = length-(s-(const char*)path);

		e = (z<255 ? z:255)+d;
		do
		{
			if ((UInt8)(*d++=*s++)==(UInt8)':')
				q = p, p = s;
		} while(d<e);

		buf[0] = d-(1+buf);
		if (s-(const char*)path>=length)
			p = s;
		else
			buf[0] -= s-p;
		params.fileParam.ioVRefNum = file->vRefNum;
		params.fileParam.ioDirID = file->parID;
		params.fileParam.ioNamePtr = buf;
		if ((perr=PBMakeFSSpecSync(&params))!=noErr)
			if (perr!=fnfErr)
				return perr;
			else
				err = perr; /* caller should worry about object existance */

		if (--retries==0)
			return bdNamErr;
	} while(s!=p);

	return err;
}

int spec2path(const FSSpec *file, void *path, int remaining)
{
	char			*d = path;
	int				c = 0, total = 0;

	if (path==NULL || remaining<=0)
		return 0;

	d[0] = 0;
	if (file->vRefNum<=0) /* only if the volume reference is valid */
	{
		HParamBlockRec		params;

		FSSpec		current;
		OSErr		err = noErr;

		params.ioParam.ioMisc = (Ptr)&current;
		params.fileParam.ioVRefNum = file->vRefNum;
		params.fileParam.ioDirID = file->parID;
		params.fileParam.ioNamePtr = (UInt8*)file->name; /* shouldn't be altered */
		if ((err=PBMakeFSSpecSync(&params))!=noErr
				&& err!=fnfErr)
			return err;
		c = current.name[0];
		if (current.parID==fsRtParID) /* force colon for root; 2008.07.21 14:40 GMT */
			goto copy;
		else
		{
			if (c>=remaining)
				return 0; /* may have colon at start */
			memcpy(d,1+current.name,c);
			goto next;
		}

		/* insert each directory name until we hit the root parent of the volume */
		do
		{
			params.fileParam.ioVRefNum = current.vRefNum;
			params.fileParam.ioDirID = current.parID;
			params.fileParam.ioNamePtr = "\p"; /* blank for parent */
			if ((err=PBMakeFSSpecSync(&params))!=noErr
					&& err!=fnfErr) /* late fix; 2008.07.20 08:11 GMT */
				return err;
			c = current.name[0];
		copy:
			if (++c>=remaining)
				return 0; /* may have colon at start */
			memmove(c+d,d,total);
			d[c-1] = ':'; /* and sandwich in the colon */
			memcpy(d,1+current.name,c-1);
		next:
			total += c, remaining -= c;
		} while(current.parID!=fsRtParID);
	}

	/* and terminate */
	if (remaining>0)
		d[total] = 0;

	return total;
}

/* MacOS substitute for (truncated) UNIX API */
/* since this cannot set the errno (Carbon only allows for a function call), */
/*  use the value returned as a negative errno */
int mkdir(const char *path)
{
	FSSpec			dirspec;
	long			dirid;
	OSErr			err;

	if ((err=path2spec(path,strlen(path),&dirspec))==noErr)
		return EEXIST;
	if (err!=fnfErr || (err=FSpDirCreate(&dirspec,0,&dirid))!=noErr)
		return -translatemacoserror(err);
	return dirid>=0 ? dirid:0;
}

typedef struct suffixtc/*(16)*/
{ /* important note: extensionToFileAndCreator requires all non-name chars to be nulls */
	UInt8			suffix[8]; /* lowercase, can do full eight (no termination) */
	OSType			type, creator;
} suffixtc;

/* returns -1 if the table isn't installed, 0 if an entry is not found; 1 if it is */
int extensionToFileAndCreator(const void *name, int length, OSType *filecreator)
{
	global Handle	tclistresource = NULL;
	global int		tclistcount = 0;

	const suffixtc	*list;
	UInt32			matcha, matchb; /* first and second 32 bits of suffix chardata */
	int				x;

	if (tclistresource==NULL)
	{
		if ((tclistresource=GetResource('sfTC',0))==NULL) /* was Get1Resource */
			return -1;
		tclistcount = GetHandleSize(tclistresource)>>3;
	}

	if (length<=0 || name==NULL)
		matcha = matchb = 0;
	else
	{ /* scan backwards until we reach a period, seperator or surpass eight chars */
		const UInt8		*n = name;
		UInt8			c;

		/* "nullify" matching data */
		//((UInt32*)match)[0] = ((UInt32*)match)[1] = 0;
		n += length, x = 8, matcha = matchb = 0;
		do
		{
			c = *--n;
			/* period is included */
			if ((c+=-'/')==0 || (c+='/'-':')==0 || (c+=':'-'\\')==0)
				break;
			if (--x<0)
				return 0;
			if ((UInt8)((c+='\\')-'A')<='Z'-'A')
				c += 'a'-'A'; /* make lowercase */
			matchb = (matchb>>8)|(matcha<<24);
			matcha = (matcha>>8)|(c<<24);
		} while(--length!=0 && c!='.');
	}

	if ((x=tclistcount)<=0)
		return 0;

	/* no need to lock the handle, we call nothing from here on */
	list = (const suffixtc*)*tclistresource;
	do
	{
		if (((UInt32*)list->suffix)[0]!=matcha || ((UInt32*)list->suffix)[1]!=matchb)
			continue;
		if (filecreator!=NULL)
			*filecreator++ = list->type, *filecreator++ = list->creator;
		return 1;
	} while(++list,--x!=0);

	return 0;
}

/* the following structure needs byte packed alignment */

#if PRAGMA_STRUCT_ALIGN
# pragma options align=packed
#elif PRAGMA_STRUCT_PACKPUSH
# pragma pack(push, 1)
#elif PRAGMA_STRUCT_PACK
# pragma pack(1)
#endif

typedef struct AppleDoubleHeader/*(589)*/
{ /* any number of these fields may be wrong (don't study to be an astronaut) */
	UInt32			appleDoubleSignature; /* should be 0x00051607 */
	SInt16			type; /* should be (1) - single target file or directory */
	SInt16	unused[8]; /* zeros */
	SInt32			atomCount; /* should be (5) */
/*0x1A*/
	SInt32			atomicKey1; /* should be (2) */
	SInt32			resourceDataOffset, resourceDataLength;

	SInt32			atomicKey2; /* should be (3) */
	SInt32			nameOffset, nameLength;

	SInt32			atomicKey3; /* should be (4) */
	UInt32			commentOffset, commentLength;

	SInt32			atomicKey4; /* should be (7) */
	SInt32			checksumOffset, checksumLength;

	SInt32			atomicKey5; /* should be (9) */
	SInt32			xInfoOffset, xInfoLength;
/*0x56*/
	UInt8			data[503]; /* constant positioning is unreliable; use offsets */
} AppleDoubleHeader;

#if PRAGMA_STRUCT_ALIGN
# pragma options align=reset
#elif PRAGMA_STRUCT_PACKPUSH
# pragma pack(pop)
#elif PRAGMA_STRUCT_PACK
# pragma pack()
#endif

static OSErr applyappledouble(
	AppleDoubleHeader *adheader,/* not on stack here so it'd last in renaming */
	CInfoPBRec *cinfo,			/* maliable out data - even the name ptr is changed */
	const FSSpec *spec,			/* destin filespec */
	const UInt8 *adname)		/* AppleDouble file name (cinfo stores vRefNum/parID) */
{
	CInfoPBRec		adcinfo;
	OSErr			err;
	long			count;
	short			adrefnum, rfrefnum;

	adcinfo.dirInfo.ioFDirIndex = 0;
	adcinfo.dirInfo.ioFlAttrib = 0;
	adcinfo.dirInfo.ioACUser = 0;
	adcinfo.dirInfo.ioNamePtr = "\p:.AppleDouble:";
	adcinfo.dirInfo.ioVRefNum = spec->vRefNum;
	adcinfo.dirInfo.ioDrDirID = adcinfo.dirInfo.ioDrParID =
		(cinfo->hFileInfo.ioFlAttrib&kioFlAttribDirMask) ?
		cinfo->dirInfo.ioDrDirID:cinfo->dirInfo.ioDrParID;
	if ((err=PBGetCatInfoSync(&adcinfo))!=noErr
			|| (err=HOpen(adcinfo.dirInfo.ioVRefNum,adcinfo.dirInfo.ioDrDirID,
			adname,fsRdPerm,&adrefnum))!=noErr
			|| adheader==NULL)
		return err;

	count = sizeof(*adheader);
	if (FSRead(adrefnum,&count,adheader)!=noErr
			|| adheader->appleDoubleSignature!=0x00051607
			|| adheader->atomCount!=5
			|| adheader->atomicKey1!=2
			|| adheader->atomicKey2!=3
			|| adheader->atomicKey3!=4
			|| adheader->atomicKey4!=7
			|| adheader->atomicKey5!=9
			|| adheader->resourceDataOffset<sizeof(*adheader)
			|| (unsigned)adheader->resourceDataLength>1UL<<24
			|| (unsigned)(adheader->nameOffset-0x56)>sizeof(*adheader)-0x56
			|| (unsigned)adheader->xInfoLength>32
			|| adheader->xInfoOffset+adheader->xInfoLength>sizeof(*adheader))
		{err = -1; goto endo;}

	memcpy(&cinfo->hFileInfo.ioFlFndrInfo,adheader->xInfoOffset+(UInt8*)adheader,
		adheader->xInfoLength);
	/* sorry, no seperation between creation and modification dates */

	if (adheader->resourceDataLength>0)
		if ((err=FSpOpenRF(spec,fsRdWrPerm,&rfrefnum))==noErr)/* || err==opWrErr)*/
		{
			if ((err=iocopydata(rfrefnum,adrefnum,0,adheader->resourceDataOffset,
					adheader->resourceDataLength,NULL))!=noErr)
				; /*fprintf(stderr,
					" error writing resource fork for '%P' (#%d)\n",
					prog,
					spec->name,
					err);*/
			FSClose(rfrefnum);
		}/* else
			fprintf(stderr,
				" error opening resource fork for '%P' (#%d)\n",
				prog,
				spec->name,
				err);*/

	FSClose(adrefnum); adrefnum = -1;
	/* ..and remove the piece of.. */
	HDelete(adcinfo.dirInfo.ioVRefNum,adcinfo.dirInfo.ioDrDirID,adname);
	/* instead of deference... */
	HDelete(adcinfo.dirInfo.ioVRefNum,adcinfo.dirInfo.ioDrDirID,"\p");

	{ /* overwriting other data to form pascal string - but what do i care? */
		UInt8		*name = adheader->nameOffset-1+(UInt8*)adheader;

		name[0] = adheader->nameLength;
		if (HRename(cinfo->hFileInfo.ioVRefNum,cinfo->hFileInfo.ioFlParID,
				spec->name,name)==noErr)
			cinfo->hFileInfo.ioNamePtr = name;
	}
endo:
	if (adrefnum>0)
		FSClose(adrefnum);
	return err;
}

int fsapplyinfo(short vrefnum,
	long dirid,
	const UInt8 *fname,
	time_t settime)
{
	HParamBlockRec		params;
	AppleDoubleHeader	adheader;

	CInfoPBRec		cinfo;
	FSSpec			spec;
	int				err = 0;

	/* we need the vRefNum and parent directory ID to find the AppleDouble info */
	params.ioParam.ioMisc = (Ptr)&spec;
	params.fileParam.ioVRefNum = vrefnum;
	params.fileParam.ioDirID = dirid;
	params.fileParam.ioNamePtr = (UInt8*)fname; /* contents will not be altered */
	if (PBMakeFSSpecSync(&params)!=noErr)
		return 0;

	/* get info for target */
	cinfo.hFileInfo.ioCompletion = NULL;
	cinfo.hFileInfo.ioFDirIndex = 0;
	cinfo.hFileInfo.ioFlAttrib = 0;
	cinfo.hFileInfo.ioACUser = 0;
	cinfo.hFileInfo.ioNamePtr = spec.name;
	cinfo.hFileInfo.ioVRefNum = spec.vRefNum;
	cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = spec.parID;
	if (PBGetCatInfoSync(&cinfo)!=noErr) /* should not err */
		return 0;

	/* set info (spec may be used for getting further data - set up the long way) */
	cinfo.hFileInfo.ioFDirIndex = 0;
	cinfo.hFileInfo.ioACUser = 0;
	cinfo.hFileInfo.ioNamePtr = spec.name;

	if (settime!=0)
		cinfo.hFileInfo.ioFlCrDat = cinfo.hFileInfo.ioFlMdDat = settime;

	/* filename for parent info is always this constant */
	if (cinfo.hFileInfo.ioFlAttrib&kioFlAttribDirMask)
	{
		if (memcmp(spec.name,"\p.AppleDouble",13)==0)
		{ /* we have the directory at end in our grasp, try to remove it */
			FSpDelete(&spec);
			return noErr;
		}
		fname = "\p.Parent";
		applyappledouble(&adheader,&cinfo,&spec,fname);
	} else if (applyappledouble(&adheader,&cinfo,&spec,spec.name)!=noErr)
		extensionToFileAndCreator(1+fname,fname[0],
			&cinfo.hFileInfo.ioFlFndrInfo.fdType);

	cinfo.hFileInfo.ioVRefNum = spec.vRefNum;
	cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = spec.parID;

	return PBSetCatInfoSync(&cinfo)==noErr ? (err!=0 ? err:1):0;
}
