/*
 * untgz.c -- Display contents and extract files from a gzip'd TAR file
 *  version 0.2.1
 *
 * written by Pedro A. Aranda Gutierrez <paag@tid.es>
 * adaptation to Unix by Jean-loup Gailly <jloup@gzip.org>
 * various fixes by Cosmin Truta <cosmint@cs.ubbcluj.ro>
 * (glue) adaptation to Mac OS Classic by Adam P. <deadbeat@home>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#ifdef mac
# include <Files.h>
# include "::macos:macos7-9.h"
# include ":::zlib.h"
# define TPATH_SEP      ':' /* after translation only */
# define TPATH_EXTRA    1
#else
# include "zlib.h"
# define TPATH_SEP      '/'
# define TPATH_EXTRA    0
#endif

#ifdef unix
#  include <unistd.h>
#elif !defined(mac)
#  include <direct.h>
#  include <io.h>
#endif

#ifdef WIN32
#include <windows.h>
#  ifndef F_OK
#    define F_OK  0
#  endif
#  define mymkdir(dirname,mode) _mkdir(dirname)
#  ifdef _MSC_VER
#    define access(path,mode)   _access(path,mode)
#    define chmod(path,mode)    _chmod(path,mode)
#    define strdup(str)         _strdup(str)
#  endif
#elif defined(mac)
/* yes, I implemented it - AP; 2009.07.09 09:57 EST */
static char *strdup(char *s)
{
  char  *d;
  int   z;

  if (s==NULL || (z=strlen(s))<0 || (d=malloc(1+z))==NULL)
      return NULL;
  memcpy(d,s,z);
  d[z] = 0;

  return d;
}
#  define mymkdir(n,m)          mkdir(n)
#else
#  include <utime.h>
#  define mymkdir               mkdir
#endif


/* values used in typeflag field */

#define REGTYPE  '0'            /* regular file */
#define AREGTYPE '\0'           /* regular file */
#define LNKTYPE  '1'            /* link */
#define SYMTYPE  '2'            /* reserved */
#define CHRTYPE  '3'            /* character special */
#define BLKTYPE  '4'            /* block special */
#define DIRTYPE  '5'            /* directory */
#define FIFOTYPE '6'            /* FIFO special */
#define CONTTYPE '7'            /* reserved */

/* GNU tar extensions */

#define GNUTYPE_DUMPDIR  'D'    /* file names from dumped directory */
#define GNUTYPE_LONGLINK 'K'    /* long link name */
#define GNUTYPE_LONGNAME 'L'    /* long file name */
#define GNUTYPE_MULTIVOL 'M'    /* continuation of file from another volume */
#define GNUTYPE_NAMES    'N'    /* file name that does not fit into main hdr */
#define GNUTYPE_SPARSE   'S'    /* sparse file */
#define GNUTYPE_VOLHDR   'V'    /* tape/volume header */


/* tar header */

#define BLOCKSIZE     512
#define SHORTNAMESIZE 100

struct tar_header
{                               /* byte offset */
  char name[100];               /*   0 */
  char mode[8];                 /* 100 */
  char uid[8];                  /* 108 */
  char gid[8];                  /* 116 */
  char size[12];                /* 124 */
  char mtime[12];               /* 136 */
  char chksum[8];               /* 148 */
  unsigned char typeflag;       /* 156 */
  char linkname[100];           /* 157 */
  char magic[6];                /* 257 */
  char version[2];              /* 263 */
  char uname[32];               /* 265 */
  char gname[32];               /* 297 */
  char devmajor[8];             /* 329 */
  char devminor[8];             /* 337 */
  char prefix[155];             /* 345 */
                                /* 500 */
};

union tar_buffer
{
  char               buffer[BLOCKSIZE];
  struct tar_header  header;
};

struct attr_item
{
  struct attr_item  *next;
  char              *fname;
  unsigned int       mode;
  time_t             time;
};

enum { TGZ_EXTRACT, TGZ_EXTRACTVERBOSE, TGZ_LIST, TGZ_INVALID };

char *TGZfname          OF((const char *));
void TGZnotfound        OF((const char *));

unsigned int getoct     OF((char *, int));
char *strtime           OF((time_t *));
int setfiletime         OF((char *, time_t));
void push_attr          OF((struct attr_item **, char *, int, time_t));
void restore_attr       OF((struct attr_item **));

int ExprMatch           OF((char *, char *));

char *printablefpath    OF((char *));
int makedir             OF((char *));
int matchname           OF((int, int, char **, char *));

void error              OF((const char *));
int tar                 OF((gzFile, int, int, int, char **));

void help               OF((int));
int main                OF((int, char **));

char *prog;

const char *TGZsuffix[] = { "\0", ".tar", ".tar.gz", ".taz", ".tgz", NULL };

/* return the file name of the TGZ archive */
/* or NULL if it does not exist */

char *TGZfname (const char *arcname)
{
  static char buffer[1024];
  const char *s;
  int origlen,i;

  strcpy(buffer,arcname);
  origlen = strlen(buffer);

  for (i=0; (s=TGZsuffix[i]) != NULL; i++)
    {
       strcpy(buffer+origlen,s);
#ifdef mac
       { /* stupid alternative; 2008.07.02 11:27 GMT */
         FILE *file;

         if ((file=fopen(buffer,"rb")) == NULL)
           continue;
         fclose(file);
       }
#else
       if (access(buffer,F_OK) == 0)
#endif
         return buffer;
    }
  return NULL;
}


/* error message for the filename */

void TGZnotfound (const char *arcname)
{
  int i;

  fprintf(stderr,"%s: Couldn't find ", prog);
  for (i=0;TGZsuffix[i];i++)
    fprintf(stderr,(TGZsuffix[i+1]) ? "%s%s, " : "or %s%s\n",
            arcname,
            TGZsuffix[i]);
  exit(1);
}


/* convert octal digits to int */
/* on error return -1 */

unsigned int getoct (char *s,int width)
{ /* AP: optimized (again); 2009.07.10 15:04 EST */
  unsigned int result = 0;
  unsigned char c, *p = (unsigned char*)s;

  if (width>0)
    do
      {
        if ((c=*p++) == 0)
          break;
        if (c == ' ')
          continue;
        if ((c-='0') > 7) /* also accounts for c < '0' */
          return -1;
        result = result * 8 + c;
      } while(--width != 0);

  return result;
}


/* convert time_t to string */
/* use the "YYYY/MM/DD hh:mm:ss" format */

char *strtime (time_t *t)
{
  struct tm   *local;
  static char result[32];

  local = localtime(t);
  sprintf(result,"%4d/%02d/%02d %02d:%02d:%02d",
          local->tm_year+1900, local->tm_mon+1, local->tm_mday,
          local->tm_hour, local->tm_min, local->tm_sec);
  return result;
}


/* set file time */

#ifndef mac
int setfiletime (char *fname,time_t ftime)
{
#ifdef WIN32
  static int isWinNT = -1;
  SYSTEMTIME st;
  FILETIME locft, modft;
  struct tm *loctm;
  HANDLE hFile;
  int result;

  loctm = localtime(&ftime);
  if (loctm == NULL)
    return -1;

  st.wYear         = (WORD)loctm->tm_year + 1900;
  st.wMonth        = (WORD)loctm->tm_mon + 1;
  st.wDayOfWeek    = (WORD)loctm->tm_wday;
  st.wDay          = (WORD)loctm->tm_mday;
  st.wHour         = (WORD)loctm->tm_hour;
  st.wMinute       = (WORD)loctm->tm_min;
  st.wSecond       = (WORD)loctm->tm_sec;
  st.wMilliseconds = 0;
  if (!SystemTimeToFileTime(&st, &locft) ||
      !LocalFileTimeToFileTime(&locft, &modft))
    return -1;

  if (isWinNT < 0)
    isWinNT = (GetVersion() < 0x80000000) ? 1 : 0;
  hFile = CreateFile(fname, GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
                     (isWinNT ? FILE_FLAG_BACKUP_SEMANTICS : 0),
                     NULL);
  if (hFile == INVALID_HANDLE_VALUE)
    return -1;
  result = SetFileTime(hFile, NULL, NULL, &modft) ? 0 : -1;
  CloseHandle(hFile);
  return result;
#else
  struct utimbuf settime;

  settime.actime = settime.modtime = ftime;
  return utime(fname,&settime);
#endif
}
#endif


/* push file attributes */

void push_attr(struct attr_item **list,char *fname,int mode,time_t time)
{
  struct attr_item *item;

  item = (struct attr_item *)malloc(sizeof(struct attr_item));
  if (item == NULL)
    error("Out of memory");
  item->fname = strdup(fname);
  item->mode  = mode;
  item->time  = time;
  item->next  = *list;
  *list       = item;
}


/* restore file attributes */

void restore_attr(struct attr_item **list)
{
  struct attr_item *item, *prev;

  for (item = *list; item != NULL; )
    {
#ifdef mac
     int        len;

     if ((len=strlen(item->fname))<=255)
       {
         unsigned char   fname[256];

         fname[0] = len;
         memcpy(1+fname,item->fname,len);
         if (fsapplyinfo(vRefNumLocal,0,fname,item->time)<0)
           fprintf(stderr,
             " error applying AppleDouble data to '%s'",
             prog,
             1+fname);
       }
#else
      setfiletime(item->fname,item->time);
      chmod(item->fname,item->mode);
#endif
      prev = item;
      item = item->next;
      free(prev);
    }
  *list = NULL;
}


/* match regular expression */

/*#define ISSPECIAL(c) (((c) == '*') || ((c) == '/'))*/

int ExprMatch (char *string,char *expr)
{
  while (1)
    {
      if (*expr == TPATH_SEP)
        {
          if (*string != '\\' && *string != TPATH_SEP)
            return 0;
          string++; expr++;
        }
      else if (*expr == '*')
        {
          if (*expr++ == 0)
            return 1;
          while (*++string != *expr)
            if (*string == 0)
              return 0;
        }
      else
        {
          if (*string++ != *expr)
            return 0;
          if (*expr++ == 0)
            return 1;
        }
    }
}

/* any char outside [0x20,0xFF) -> space */
char *printablefpath(char *name)
{
    unsigned char   c, *d;

    if ((d=(unsigned char*)name)==NULL)
      return NULL;
    if ((c=*d++)!=0)
      do
        {
          if ((c+=-0x20)>=0xFF-0x20)
          d[-1] = ' ';
        } while((c=*d++)!=0);

    return name;
}

/* recursive mkdir */
/* abort on ENOENT; ignore other errors like "directory already exists" */
/* return 1 if OK */
/*        0 on error */

int makedir (char *newdir)
{
  char              *buffer;
  unsigned char     hold, *p;
  int               len;

  if (((len=strlen(newdir)) <= 0) || (buffer=malloc(1+len)) == NULL)
    return 0;
  memcpy(buffer,newdir,len);
  buffer[len] = 0;
  if (buffer[len-1] == TPATH_SEP) {
    buffer[len-1] = '\0';
  }

  if (mymkdir(buffer, 0755) == 0)
    goto end;

  p = 1+(unsigned char*)buffer;
  do
    {
      do{if ((hold=*p++)==0) break;} while(hold != '\\' && hold != TPATH_SEP);
      *--p = 0;
      if ((mymkdir(buffer, 0755) == -1) && (errno == ENOENT))
        {
          fprintf(stderr,"%s: Couldn't create directory %s\n", prog,buffer);
          free(buffer);
          return 0;
        }
    } while((*p++=hold)!=0);
end:
  free(buffer);
  return 1;
}


int matchname (int arg,int argc,char **argv,char *fname)
{
  if (arg == argc)      /* no arguments given (untgz tgzarchive) */
    return 1;

  while (arg < argc)
    if (ExprMatch(fname,argv[arg++]))
      return 1;

  return 0; /* ignore this for the moment being */
}


/* tar file list or extract */

int tar (gzFile in,int action,int arg,int argc,char **argv)
{
  union  tar_buffer buffer;
  int               len;
  int               err;
  int               getheader = 1;
  int               remaining = 0;
  FILE             *outfile = NULL;
  char              fname[TPATH_EXTRA+BLOCKSIZE];
  unsigned int      tarmode;
  time_t            tartime;
  struct attr_item *attributes = NULL;

  if (action == TGZ_LIST)
    printf("    date      time     size                       file\n"
           " ---------- -------- --------- -------------------------------------\n");
  while (1)
    {
      tarmode = 0; /* initialize to eliminate 'unset' warning; 2009.07.10 15:41 EST */
      len = gzread(in, &buffer, BLOCKSIZE);
      if (len < 0)
        error(gzerror(in, &err));
      /*
       * Always expect complete blocks to process
       * the tar information.
       */
      if (len != BLOCKSIZE)
        {
          action = TGZ_INVALID; /* force error exit */
          remaining = 0;        /* force I/O cleanup */
        }

      /*
       * If we have to get a tar header
       */
      if (getheader >= 1)
        {
          /*
           * if we met the end of the tar
           * or the end-of-tar block,
           * we are done
           */
          if (len == 0 || buffer.header.name[0] == 0)
            break;

          tarmode = getoct(buffer.header.mode,8);
          tartime = (time_t)getoct(buffer.header.mtime,12);
          if ((tarmode == -1) || (tartime == (time_t)-1))
            {
              buffer.header.name[0] = 0;
              action = TGZ_INVALID;
            }
#ifdef mac
          else if (tartime != (time_t)-1)
            tartime += 2082844800;
#endif

          if (getheader == 1)
            {
              strncpy(TPATH_EXTRA+fname,buffer.header.name,SHORTNAMESIZE);
              if (fname[TPATH_EXTRA+SHORTNAMESIZE-1] != 0)
                  fname[TPATH_EXTRA+SHORTNAMESIZE] = 0;

#if TPATH_EXTRA
              { /* translate to native path (w/o an additional function) */
                unsigned char   c, *s = (unsigned char*)fname;

                *s++ = TPATH_SEP;
                if ((c=*s++) != 0)
                  do
                    {
                      if ((c+=-'/') == 0 || (c+='/'-'\\') == 0)
                          s[-1] = TPATH_SEP;
                    } while((c=*s++) != 0);
              }
#endif
            }
          else
            {
              /*
               * The file name is longer than SHORTNAMESIZE
               */
              if (strncmp(fname,buffer.header.name,SHORTNAMESIZE-1) != 0)
                  error("bad long name");
              getheader = 1;
            }

          /*
           * Act according to the type flag
           */
          switch (buffer.header.typeflag)
            {
            case DIRTYPE:
              if (action == TGZ_LIST)
                {
                  printablefpath(fname);
                  printf(" %s     <dir> %s\n", strtime(&tartime),fname);
                }
              else if ((action <= TGZ_EXTRACTVERBOSE) && (matchname(arg,argc,argv,fname) != 0))
                {
                  if (action==TGZ_EXTRACTVERBOSE)
                    printf(" creating directory '%s'\n", fname);
                  makedir(fname);
                  push_attr(&attributes,fname,tarmode,tartime);
                }
              break;
            case REGTYPE:
            case AREGTYPE:
              remaining = getoct(buffer.header.size,12);
              if (remaining == -1)
                {
                  action = TGZ_INVALID;
                  break;
                }
              if (action == TGZ_LIST)
                {
                  printablefpath(fname);
                  printf(" %s %9d %s\n", strtime(&tartime),remaining,fname);
                }
              else if (action <= TGZ_EXTRACTVERBOSE)
                {
                  if (matchname(arg,argc,argv,fname))
                    {
                      if (action == TGZ_EXTRACTVERBOSE)
                        printf("  extracting file '%s'\n", fname);
                      outfile = fopen(fname,"wb");
                      if (outfile == NULL) {
                        /* try creating directory */
                        char *p = strrchr(fname, TPATH_SEP);
                        if (p != NULL) {
                          *p = '\0';
                          makedir(fname);
                          *p = TPATH_SEP;
                          outfile = fopen(fname,"wb");
                        }
                      }
                      if (outfile != NULL)
                        printf("Extracting %s\n", fname);
                      else
                        fprintf(stderr,"%s: Couldn't create %s", prog,fname);
                    }
                  else
                    outfile = NULL;
                }
              getheader = 0;
              break;
            case GNUTYPE_LONGLINK:
            case GNUTYPE_LONGNAME:
              remaining = getoct(buffer.header.size,12);
              if (remaining < 0 || remaining >= BLOCKSIZE)
                {
                  action = TGZ_INVALID;
                  break;
                }
              len = gzread(in, fname, BLOCKSIZE);
              if (len < 0)
                error(gzerror(in, &err));
              if (fname[BLOCKSIZE-1] != 0 || (int)strlen(fname) > remaining)
                {
                  action = TGZ_INVALID;
                  break;
                }
              getheader = 2;
              break;
            default:
              if (action == TGZ_LIST) {
                printablefpath(fname);
                printf(" %s     <---> %s\n", strtime(&tartime),fname);
              }
              break;
            }
        }
      else
        {
          unsigned int bytes = (remaining > BLOCKSIZE) ? BLOCKSIZE : remaining;

          if (outfile != NULL)
            {
              if (fwrite(&buffer,sizeof(char),bytes,outfile) != bytes)
                {
                  fprintf(stderr,"%s: Error writing %s -- skipping\n",
                    prog,fname);
                  fclose(outfile);
                  outfile = NULL;
                  remove(fname);
                }
            }
          remaining -= bytes;
        }

      if (remaining == 0)
        {
          getheader = 1;
          if (outfile != NULL)
            {
              fclose(outfile);
              outfile = NULL;
              if (action != TGZ_INVALID)
                push_attr(&attributes,fname,tarmode,tartime);
            }
        }

      /*
       * Abandon if errors are found
       */
      if (action == TGZ_INVALID)
        {
          error("broken archive");
          break;
        }
    }

  /*
   * Restore file modes and time stamps
   */
  printf("restoring attributes...");
  fflush(stdout);
  restore_attr(&attributes);

  if (gzclose(in) != Z_OK)
    error("failed gzclose");

  return 0;
}


/* ============================================================ */

void help(int exitval)
{
  printf("untgz version 0.2.1"
#ifdef mac
         " (MacOS Classic port by Adam P., update 2009.08.19)"
#endif
         "\n  using zlib version %s\n\n",
         zlibVersion());
  printf("Usage: untgz file.tgz            extract all files\n"
         "       untgz file.tgz fname ...  extract selected files\n"
         "       untgz -v file.tgz [...]   extract verbose\n"
         "       untgz -l file.tgz         list archive contents\n"
         "       untgz -h                  display this help\n");
  exit(exitval);
}

void error(const char *msg)
{
  fprintf(stderr,"%s: %s\n", prog, msg);
  exit(1);
}


/* ============================================================ */

#if defined(WIN32) && defined(__GNUC__)
int _CRT_glob = 0;      /* disable argument globbing in MinGW */
#endif

int main(int argc,char **argv)
{
    int         action = TGZ_EXTRACT;
    int         arg = 1;
    char        *TGZfile;
    gzFile      *f;

    prog = strrchr(argv[0],'\\');
    if (prog == NULL)
      {
        prog = strrchr(argv[0],'/');
        if (prog == NULL)
          {
            prog = strrchr(argv[0],':');
            if (prog == NULL)
              prog = argv[0];
            else
              prog++;
          }
        else
          prog++;
      }
    else
      prog++;

    if (argc == 1)
      help(0);

    if (argv[arg][0] == '-')
      if (argv[arg][1] == 'v')
        {
          action = TGZ_EXTRACTVERBOSE;
          if (argc == ++arg)
            help(0);
        }
      else if (argv[arg][1] == 'l')
        {
          action = TGZ_LIST;
          if (argc == ++arg)
            help(0);
        }
      else if (argv[arg][1] == '-h')
        {
          help(0);
        }

    if ((TGZfile = TGZfname(argv[arg])) == NULL)
      TGZnotfound(argv[arg]);

    ++arg;
    if ((action == TGZ_LIST) && (arg != argc))
      help(1);

/*
 *  Process the TGZ file
 */
    switch(action)
      {
      case TGZ_EXTRACT:
      case TGZ_EXTRACTVERBOSE:
      case TGZ_LIST:
        f = gzopen(TGZfile,"rb");
        if (f == NULL)
          {
            fprintf(stderr,"%s: Couldn't gzopen %s\n", prog,TGZfile);
            return 1;
          }

        exit(tar(f, action, arg, argc, argv));
        break;

      default:
        error("Unknown option");
        exit(1);
      }

    return 0;
}
