/* minigzip.c -- simulate gzip using the zlib compression library
 * Copyright (C) 1995-2005 Jean-loup Gailly.
 * For conditions of distribution and use, see copyright notice in zlib.h
 */

/*
 * minigzip is a minimal implementation of the gzip utility. This is
 * only an example of using zlib and isn't meant to replace the
 * full-featured gzip. No attempt is made to deal with file systems
 * limiting names to 14 or 8+3 characters, etc... Error checking is
 * very limited. So use minigzip only for testing; use gzip for the
 * real thing. On MSDOS, use only on file names without extension
 * or in pipe mode.
 */

/* @(#) $Id$ */

#include <stdio.h>
#include "zlib.h"

#ifdef STDC
#  include <string.h>
#  include <stdlib.h>
#endif

#ifdef USE_MMAP
#  include <sys/types.h>
#  include <sys/mman.h>
#  include <sys/stat.h>
#endif

#if defined(MSDOS) || defined(OS2) || defined(WIN32) || defined(__CYGWIN__)
#  include <fcntl.h>
#  include <io.h>
#  define SET_BINARY_MODE(file) setmode(fileno(file), O_BINARY)
#else
#  define SET_BINARY_MODE(file)
#endif

#ifdef mac
# include <fcntl.h>
 /* MPW works by this... */
# define stream_gzdopen(s,m)    gzopen("Dev:" #s,m)

# include ":contrib:macos:macos7-9.h"
# include <Debugging.h>

# define signalbreak()          ((*((volatile UInt8*)0x179)&*((volatile UInt8*)0x17A))!=0)
# define DOS_TO_NATIVE_TIME     2082844800
# define TPATH_EXTRA            1
# define TPATH_SEP              ':'
#else
# define stream_gzdopen(s,m)    gzdopen(fileno(s),m)

# define signalbreak()          0
# define DOS_TO_NATIVE_TIME     0
# define TPATH_SEP              '/'
# define TPATH_EXTRA            0
#endif

#ifdef VMS
#  define unlink delete
#  define GZ_SUFFIX "-gz"
#endif
#ifdef RISCOS
#  define unlink remove
#  define GZ_SUFFIX "-gz"
#  define fileno(file) file->__file
#endif
#if defined(__MWERKS__) && __dest_os != __be_os && __dest_os != __win32_os
#  include <unix.h> /* for fileno */
#endif

#ifndef WIN32 /* unlink already in stdio.h for WIN32 */
  extern int unlink OF((const char *));
#endif

#ifndef GZ_SUFFIX
#  define GZ_SUFFIX     ".gz"
#  define GZ_SUFFIX_UC  ".GZ"
#endif
#define SUFFIX_LEN (sizeof(GZ_SUFFIX)-1)

#define BUFLEN         16384
#define MAX_NAME_LEN    1024

enum /* options (2010.12.28) */
{
    _unlink             = 0x1,
    _original_name      = 0x2
};

char *prog;
char pathbuf[MAX_NAME_LEN];
char databuf[BUFLEN];

void warn             OF((const char *msg));
void error            OF((const char *msg));

char *getpathname     OF((char *path));

void setfileinfo      OF((const char *filepath, uLong dostime));
uLong getfileinfo     OF((const char *f, uLong *dt));

int  gz_compress      OF((FILE   *in, gzFile out));
#ifdef USE_MMAP
int  gz_compress_mmap OF((FILE   *in, gzFile out));
#endif
int  gz_uncompress    OF((gzFile in, FILE   *out));
int  file_compress    OF((char  *file, char *mode, char *comment, int opt));
int  file_uncompress  OF((char  *file, int opt));
int  file_list        OF((char  *file, int header));
int  main             OF((int argc, char *argv[]));

/* ===========================================================================
 * Display error message and exit
 */
void warn(msg)
    const char *msg;
{
    fprintf(stderr, "%s: %s\n", prog, msg);
}

void error(msg)
    const char *msg;
{
    warn(msg);
    exit(1);
}

/* ===========================================================================
 * Find the object name in the path, or the path string (no other names).
 */
char *getpathname(path)
    char *path;
{
    char *name = path, *n0;

    if (TPATH_SEP!=':' && (n0=strrchr(name,'\\')) == 0)
        name = 1+n0;
    if ((n0=strrchr(name,TPATH_SEP)) != 0)
        name = 1+n0;

    return name;
}

/* ===========================================================================
 * Get/set file attributes (only partly implemented)
 */

void setfileinfo(filepath, dostime)
    const char *filepath;
    uLong dostime;
{
#ifdef WIN32
    HANDLE      hFile;
    FILETIME    ftm,ftLocal,ftCreate,ftLastAcc,ftLastWrite;

    hFile = CreateFile(filepath,GENERIC_READ | GENERIC_WRITE,
        0,NULL,OPEN_EXISTING,0,NULL);

    if (dostime != 0 && dostime != 1UL) {
        GetFileTime(hFile,&ftCreate,&ftLastAcc,&ftLastWrite);
        DosDateTimeToFileTime((WORD)(dostime>>16),(WORD)dostime,&ftLocal);
        LocalFileTimeToFileTime(&ftLocal,&ftm);
        SetFileTime(hFile,&ftm,&ftLastAcc,&ftm);
    }
    CloseHandle(hFile);
#else
    {
#ifdef mac
        CInfoPBRec      cinfo;
        unsigned char   name[256];
        int             len = strlen(filepath);

        if (len>255)
            return;
        memcpy(1+name,filepath,len);
        name[0] = len;
        cinfo.hFileInfo.ioCompletion = NULL;
        cinfo.hFileInfo.ioFDirIndex = 0;
        cinfo.hFileInfo.ioFlAttrib = 0;
        cinfo.hFileInfo.ioACUser = 0;
        cinfo.hFileInfo.ioNamePtr = name;
        cinfo.hFileInfo.ioVRefNum = vRefNumLocal;
        cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = 0;
        if (PBGetCatInfoSync(&cinfo)!=noErr)
            return;
        if (dostime != 0 && dostime != 1UL) {
            cinfo.hFileInfo.ioFlMdDat = dostime+DOS_TO_NATIVE_TIME;
        /* only override creation date if new */
            if (cinfo.hFileInfo.ioFlCrDat==0
                    || cinfo.hFileInfo.ioFlCrDat>cinfo.hFileInfo.ioFlMdDat)
                cinfo.hFileInfo.ioFlCrDat = cinfo.hFileInfo.ioFlMdDat;
        }
        extensionToFileAndCreator(1+name,len,&cinfo.hFileInfo.ioFlFndrInfo.fdType);
        cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = 0;
        PBSetCatInfoSync(&cinfo);
#elif defined(unix)
        struct utimbuf ut;

        if ((ut.actime=ut.modtime=dostime) != 0 && dostime != 1UL)
            utime(filepath,&ut);
#endif
    }
#endif
}

uLong getfileinfo(f, dt)
    const char *f;         /* name of file to get info of */
    uLong *dt;             /* dostime */
{
  int ret = 0;
#ifdef WIN32
  FILETIME ftLocal;
  HANDLE hFind;
  WIN32_FIND_DATA  ff32;

  hFind = FindFirstFile(f,&ff32);
  if (hFind != INVALID_HANDLE_VALUE)
  {
    FileTimeToLocalFileTime(&(ff32.ftLastWriteTime),&ftLocal);
    FileTimeToDosDateTime(&ftLocal,((LPWORD)dt)+1,((LPWORD)dt)+0);
    FindClose(hFind);
    ret = 1;
  }
#elif defined(unix) || defined(mac)
# ifdef unix
  struct stat s;        /* results of stat() */
# endif
  time_t tm_t=0;

  if (strcmp(f,"-")!=0)
  {
    static char name[MAX_NAME_LEN + 1 + TPATH_EXTRA];
    int len;

    if ((len=strlen(f)) > MAX_NAME_LEN)
      len = MAX_NAME_LEN;

    strncpy(TPATH_EXTRA+name, f, MAX_NAME_LEN-1);
    /* strncpy doesn't append the trailing null if the string is too long. */
    name[ TPATH_EXTRA + MAX_NAME_LEN ] = '\0';

    if (name[len + (TPATH_EXTRA-1)] == TPATH_SEP)
      name[len + (TPATH_EXTRA-1)] = '\0';
#ifdef mac
    {
      CInfoPBRec    cinfo;
      char          *p;
      long          parid = 0;

      cinfo.hFileInfo.ioCompletion = NULL;
      cinfo.hFileInfo.ioVRefNum = 0x8003; /*vRefNumLocal*/
      cinfo.hFileInfo.ioFDirIndex = 0;
      cinfo.hFileInfo.ioACUser = 0;

      if (len>=255) {
        if ((p=strrchr(1+name,TPATH_SEP))==NULL || (name[0]=p-name)<=1)
          goto skip;
        len -= name[0], ++len;
        cinfo.hFileInfo.ioFlAttrib = 0;
        cinfo.hFileInfo.ioNamePtr = (unsigned char*)p;
        cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = 0;
        if (PBGetCatInfoSync(&cinfo)!=noErr)
          goto skip;
        parid = cinfo.hFileInfo.ioFlParID;
      } else
        p = name;

      name[0] = len;
      cinfo.hFileInfo.ioFlAttrib = 0;
      cinfo.hFileInfo.ioNamePtr = (unsigned char*)p;
      cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = parid;
      if (PBGetCatInfoSync(&cinfo)==noErr)
      {
        tm_t = cinfo.hFileInfo.ioFlMdDat-DOS_TO_NATIVE_TIME;
        ret = (cinfo.hFileInfo.ioFlAttrib&kioFlAttribDirMask) ? (-1):(1);
      }
    }
#else
    /* not all systems allow stat'ing a file with / appended */
    if (stat(name,&s)==0)
    {
      tm_t = s.st_mtime;
      ret = 1;
    }
#endif
    *dt = tm_t;
  }
skip:
#endif
  return ret;
}

/* ===========================================================================
 * Compress input to output then close both files.
 */

int gz_compress(in, out)
    FILE   *in;
    gzFile out;
{
    int len;
    int err = 0;
#ifdef USE_MMAP
    /* Try first compressing with mmap. If mmap fails (minigzip used in a
     * pipe), use the normal fread loop.
     */
    if (gz_compress_mmap(in, out) == Z_OK) return;
#endif
    for (;;) {
        len = (int)fread(databuf, 1, sizeof(databuf), in);
        if (ferror(in)) {
            perror("fread");
            exit(1);
        }
        if (len == 0) break;

        if (gzwrite(out, databuf, (unsigned)len) != len) {
            warn(gzerror(out, &err));
            break;
        }
        if (signalbreak())
            {err = -128; break;}
    }
    fclose(in);
    if (gzclose(out) != Z_OK) error("failed gzclose");

    return err;
}

#ifdef USE_MMAP /* MMAP version, Miguel Albrecht <malbrech@eso.org> */

/* Try compressing the input file at once using mmap. Return Z_OK if
 * if success, Z_ERRNO otherwise.
 */
int gz_compress_mmap(in, out)
    FILE   *in;
    gzFile out;
{
    int len;
    int err;
    int ifd = fileno(in);
    caddr_t buf;    /* mmap'ed buffer for the entire input file */
    off_t buf_len;  /* length of the input file */
    struct stat sb;

    /* Determine the size of the file, needed for mmap: */
    if (fstat(ifd, &sb) < 0 || (buf_len=sb.st_size) <= 0)
        return Z_ERRNO;

    /* Now do the actual mmap: */
    buf = mmap((caddr_t) 0, buf_len, PROT_READ, MAP_SHARED, ifd, (off_t)0);
    if (buf == (caddr_t)(-1))
        return Z_ERRNO;

    /* Compress the whole file at once: */
    len = gzwrite(out, (char *)buf, (unsigned)buf_len);

    if (len != (int)buf_len)
        warn(gzerror(out, &err));
    else
        munmap(buf, buf_len);
    fclose(in);
    if (gzclose(out) != Z_OK) error("failed gzclose");
    return Z_OK;
}
#endif /* USE_MMAP */

/* ===========================================================================
 * Uncompress input to output then close both files.
 */
int gz_uncompress(in, out)
    gzFile in;
    FILE   *out;
{
    int len;
    int err = 0;

    while ( (len = gzread(in, databuf, sizeof(databuf))) != 0 ) {
        if (len < 0) {
            warn (gzerror(in, &err));
            break;
        }

        if ((int)fwrite(databuf, 1, (unsigned)len, out) != len) {
            warn("failed fwrite");
            break;
        }
        if (signalbreak())
            {err = -128; break;}
    }
    if (fclose(out)) warn("failed fclose");

    if (gzclose(in) != Z_OK) error("failed gzclose");

    return err;
}


/* ===========================================================================
 * Compress the given file: create a corresponding .gz file and remove the
 * original.
 * -- AJP: removal now optional.
 */
int file_compress(file, mode, comment, opt)
    char  *file, *mode, *comment;
    int   opt;
{
    FILE  *in;
    gzFile out;
    uLong  dt = 0;
    int    err = 0;

    strcpy(pathbuf, file);
    strcat(pathbuf, GZ_SUFFIX);

    in = fopen(file, "rb");
    if (in == NULL) {
        perror(file);
        exit(1);
    }

    getfileinfo(file,&dt);

    if (opt&_original_name)
        out = gzopen2(pathbuf, mode,
            getpathname(file), comment, -1UL>>1, -1UL>>1, &dt);
    else
        out = gzopen2(pathbuf, mode, NULL, comment,
            -1UL>>1, -1UL>>1, &dt);
    if (out == NULL) {
        fprintf(stderr, "%s: can't gzopen %s\n", prog, pathbuf);
        exit(1);
    }
    err = gz_compress(in, out);

    if (opt&_unlink)
        unlink(file);

    return err;
}


/* ===========================================================================
 * Uncompress the given file and remove the original.
 * -- AJP: removal now optional.
 */
int file_uncompress(file, opt)
    char  *file;
    int   opt;
{
    char  origname[260];
    char  *infile, *outfile;
    FILE  *out;
    gzFile in;
    uLong  dt = 0;
    uInt   len = (uInt)strlen(file);
    int    err = 0;

    strcpy(pathbuf, file);

    if (len > SUFFIX_LEN && (strcmp(file+len-SUFFIX_LEN, GZ_SUFFIX) == 0
            || strcmp(file+len-SUFFIX_LEN, GZ_SUFFIX_UC) == 0)) {
        outfile = pathbuf;
        outfile[len-3] = '\0';
        infile = file;
    } else {
        outfile = file;
        infile = pathbuf;
        strcat(infile, GZ_SUFFIX);
    }
    in = gzopen2(infile, "rb", (opt&_original_name) ? origname:NULL, NULL,
        sizeof(origname)-1, 0, &dt);
    if (in == NULL) {
        fprintf(stderr, "%s: can't gzopen %s\n", prog, infile);
        exit(1);
    }
    if (opt&_original_name) {
        strcpy(pathbuf,outfile);
        outfile = getpathname(pathbuf);
        strcpy(outfile,origname);
        outfile = pathbuf;
    }
    out = fopen(pathbuf, "wb");
    if (out == NULL) {
        perror(file);
        exit(1);
    }

    err = gz_uncompress(in, out);

    if (opt&_unlink)
        unlink(infile);

    setfileinfo(pathbuf,dt);

    return err;
}

/* ===========================================================================
 * List information contained in the gz file.
 */
int file_list(file, header)
    char  *file;
    int    header;
{
    char        namebuf[300], commbuf[300], datebuf[66];
    char       *infile = file;
    gzFile      gzin = NULL;
    long int    total, csize = 0;
    uLong       dt = 0;
    uInt        len = (uInt)strlen(file);
    int         err = 0, fd;

    strcpy(pathbuf, file);

    if (len <= SUFFIX_LEN || (strcmp(file+len-SUFFIX_LEN, GZ_SUFFIX) != 0
            && strcmp(file+len-SUFFIX_LEN, GZ_SUFFIX_UC) != 0)) {
        infile = pathbuf;
        strcat(infile, GZ_SUFFIX);
    }

    *commbuf = 0;
    namebuf[1] = *namebuf = 0;

    if ((fd=open(infile,O_BINARY)) >= 0)
    { /* first get compressed size by opening fd */
        csize = lseek(fd,0,SEEK_END);
        lseek(fd,0,SEEK_SET);
        gzin = gzdopen2(fd, "rb", 1+namebuf, commbuf,
            sizeof(namebuf)-3, sizeof(commbuf)-1, &dt);
    }
    if (gzin == NULL) {
        fprintf(stderr, "%s: can't open %s\n", prog, file);
        exit(1);
    }
    if (header)
        puts(
"   Size    Ratio   Date      Time    Orig. Name, Comment\n"
"---------- ---- ---------- -------- ---------------------");

    /* the crude way of calculating file size */
    for(total=0; (len=gzread(gzin, databuf, sizeof(databuf)))!=0; total+=len) {
        if (len < 0) {
            warn (gzerror(gzin, &err));
            break;
        }
        if (total > 99999 && total <= 99999 + sizeof(databuf)) /* delay */
            fflush(stdout);
    }

    if (gzclose(gzin) != Z_OK) error("failed gzclose");

    if (dt != 0 && dt != -1UL) {
        time_t tm_t = dt+DOS_TO_NATIVE_TIME;
        struct tm* date = localtime(&tm_t);
        int y = date->tm_year>999 ? date->tm_year:date->tm_year+1900;

        sprintf(datebuf,"%4u.%02u.%02u %02u:%02u:%02u",
            y,date->tm_mon+1,date->tm_mday,
            date->tm_hour,date->tm_min,date->tm_sec);
    } else
        datebuf[0] = 0;

    if (namebuf[1] == 0) {
        strncpy(1+namebuf,getpathname(pathbuf),sizeof(namebuf)-3);
        namebuf[0] = '(', strcat(namebuf,")");
    } else
        namebuf[0] = '\'', strcat(namebuf,"'");

    printf(*commbuf ?
"% 10d % 3d%% % 19s %s  %s\n":
"% 10d % 3d%% % 19s %s\n",
        total, (unsigned)((csize*100.0)/total), datebuf, namebuf, commbuf);

    return err;
}

int main(argc, argv)
    int argc;
    char *argv[];
{
    int mode = 0, opt = 0, err = 0;
    char outmode[20], **av = argv, *comment = 0;

    strcpy(outmode, "wb6 ");

    prog = av[0];
    while (++av,--argc > 0)
        if (**av != '-')
            break;
        else if ((*av)[1] == 'c')
            comment = 2+*av;
        else if ((*av)[2] != 0)
            break;
        else if ((*av)[1] == 'd')
            mode|= 1;
        else if ((*av)[1] == 'l')
            mode|= 2;
        else if ((*av)[1] == 'u')
            opt|= _unlink;
        else if ((*av)[1] == 'n')
            opt|= _original_name;
        else if ((*av)[1] == 'f')
            outmode[3] = 'f';
        else if ((*av)[1] == 'h')
            outmode[3] = 'h';
        else if ((*av)[1] == 'r')
            outmode[3] = 'R';
        else if ((*av)[1] >= '1' && (*av)[1] <= '9')
            outmode[2] = (*av)[1];
        else if ((*av)[1] == '?')
        {
            printf(
"# Usage:  %s [-d] [options] [files...]\n"
"#   -d : decompress\n"
"#   -f : compress with Z_FILTERED\n"
"#   -h : compress with Z_HUFFMAN_ONLY\n"
"#   -r : compress with Z_RLE\n"
"#   -n : store/use original names\n"
"#   -u : unlink (remove) passed names when done\n"
"#   -c... : add comment (compress only)\n"
"#   -l : list header info only\n"
"#   -1 to -9 : compression level\n", argv[0]);
            return 0;
        } else
            break;

    if (mode > 2)
        error("can't decompress *and* list!");

    if (outmode[3] == ' ')
        outmode[3] = 0;
#ifdef mac
    if (((volatile char*)0x174)[60>>3]&(0x80>>(60&7))) /* keymap - control */
        SafeDebugger();
#endif
    if (argc == 0) {
        gzFile file;

        SET_BINARY_MODE(stdin);
        SET_BINARY_MODE(stdout);

        if (mode == 1) {
            if ((file=stream_gzdopen(stdin,"rb")) == NULL)
                error("can't gzdopen stdin");
            gz_uncompress(file, stdout);
        } else if (mode == 0) {
            if ((file=stream_gzdopen(stdout,outmode)) == NULL)
                error("can't gzdopen stdout");
            gz_compress(stdin, file);
        } else
            error("can't get info for stdin!");
    } else {
        int first = 1;

        do
            if (mode == 1)
                err = file_uncompress(*av, opt);
            else if (mode == 0)
                err = file_compress(*av, outmode, comment, opt);
            else
                err = file_list(*av,first);
        while (first=0,++av,err==0 && --argc!=0);
    }

    return err;
}
